/******************************************************************************/
/*                                                                            */
/*                              DIPSWITCH SUPPORT                             */
/*                                                                            */
/******************************************************************************/

#include "deftypes.h"

#define MAX_DIPSWITCHES		3

typedef struct DSW
{
   char *DSWName;		// Switch or setting name
   UBYTE bits;			// Switch bitmask or setting bitpattern
   UBYTE values;		// Number of settings defined for this switch
} DSW;

typedef struct DSTAT
{
   UBYTE pos;			//
   UBYTE offset;		//
} DSTAT;

typedef struct DIPSW
{
   UBYTE value;			// Current value
   UBYTE def;			// Default value (factory setting)
   ULONG count;			// Items in diplist
   ULONG address;		// RAM[] address to place dsw
   DSW   diplist[64];
   ULONG statcount;		// Items in statlist
   DSTAT statlist[16];
} DIPSW;

struct DIPSW dipswitch[MAX_DIPSWITCHES];

// Fill DSW bytes from stat list

void make_dipswitch_bytes(void);

// Fill stat list from DSW bytes

void make_dipswitch_statlist(void);

void init_dsw(void);

extern struct DSW_DATA dsw_data_default_0[];
extern struct DSW_DATA dsw_data_default_1[];
extern struct DSW_DATA dsw_data_default_2[];

void RestoreDSWDefault(void);

UBYTE get_dsw(int i);

void load_dipswitches(char *section);

void save_dipswitches(char *section);

/******************************************************************************/
/*                                                                            */
/*         ROM SWITCH SUPPORT (consider them unofficial dip switches)         */
/*                                                                            */
/******************************************************************************/

typedef struct ROMSW
{
   char *Name;			// ROMSwitch Name
   ULONG Address;		// Address of Switch
   int Count;			// Number of Settings(modes) for this switch
   char *Mode[16];		// Mode Name x16
   UBYTE Data[16];		// Mode Byte x16
   UBYTE def;
} ROMSW;

struct ROMSW LanguageSw;	// ROMSwitch for Language Selection (Taito roms)

void init_romsw(void);

void SetLanguageSwitch(int number);

int GetLanguageSwitch(void);

void load_romswitches(char *section);

void save_romswitches(char *section);

/******************************************************************************/
/*                                                                            */
/*                ROM PATCHING SUPPORT (something like an .ips)               */
/*                                                                            */
/******************************************************************************/

typedef struct ROM_PATCH
{
   ULONG offset;		// offset to patch
   ULONG data_0;		// first longword
   ULONG data_1;		// second longword
} ROM_PATCH;

void patch_rom(UBYTE *src, ROM_PATCH *patch);

/******************************************************************************/
