/******************************************************************************/
/*                                                                            */
/*                BUBBLE SYMPHONY (C) 1994 TAITO CORPORATION                  */
/*                                                                            */
/******************************************************************************/

#include "gameinc.h"
#include "bubsymph.h"
#include "f3system.h"
#include "tc003vcu.h"
#include "tc200obj.h"

static struct DIR_INFO bubble_symphony_dirs[] =
{
   { "bubble_symphony", },
   { "bubsymph", },
   { NULL, },
};

static struct ROM_INFO bubble_symphony_roms[] =
{
   {       "d90.01", 0x00100000, 0x8AEDB9E5, },
   {       "d90.02", 0x00100000, 0x5AB04CA2, },
   {       "d90.03", 0x00100000, 0x6FA894A1, },
   {       "d90.04", 0x00200000, 0xFEEE5FDA, },
   {       "d90.05", 0x00200000, 0xC192331F, },
   {       "d90.06", 0x00100000, 0x166A72B8, },
   {       "d90.07", 0x00100000, 0xB436B42D, },
   {       "d90.08", 0x00100000, 0x25A4FB2C, },
   {       "d90.09", 0x00040000, 0x3F2090B7, },
   {       "d90.10", 0x00040000, 0x8E957D3D, },
   {       "d90.11", 0x00040000, 0xEDFDBB7F, },
   {       "d90.12", 0x00040000, 0x9E523996, },
   {       "d90.13", 0x00040000, 0x6762BD90, },
   {       "d90.14", 0x00040000, 0x8E33357E, },
   {           NULL,          0,          0, },
};

static struct ROMSW_DATA romsw_data_bubble_symphony_0[] =
{
   { "Taito Japan (Japanese)", 0x01 },
   { "Taito America",          0x02 },
   { "Taito 3",                0x03 },
   { "Taito 4",                0x04 },
   { "Taito 5",                0x05 },
   { "Taito 6",                0x06 },
   { NULL,                     0    },
};

static struct ROMSW_INFO bubble_symphony_romsw[] =
{
   { 0x0FFFFF, 0x01, romsw_data_bubble_symphony_0 },
   { 0,        0,    NULL },
};

static struct VIDEO_INFO bubble_symphony_video =
{
   DrawBubbleSymphony,
   320,
   224,
   64,
   VIDEO_ROTATE_NORMAL,
};

struct GAME_MAIN game_bubble_symphony =
{
   bubble_symphony_dirs,
   bubble_symphony_roms,
   f3_system_inputs,
   NULL,
   bubble_symphony_romsw,

   LoadBubbleSymphony,
   ClearBubbleSymphony,
   &bubble_symphony_video,
   ExecuteBubbleSymphonyFrame,
   "bubsymph",
   "Bubble Symphony",
   "ouVtHj[",
   COMPANY_ID_TAITO,
   "D90",
   1994,
   NULL,
};

static struct DIR_INFO bubble_bobble_2_dirs[] =
{
   { "bubble_bobble_2", },
   { "bublbob2", },
   { ROMOF("bubsymph"), },
   { CLONEOF("bubsymph"), },
   { NULL, },
};

static struct ROM_INFO bubble_bobble_2_roms[] =
{
   {       "d90.01", 0x00100000, 0x8AEDB9E5, },
   {       "d90.02", 0x00100000, 0x5AB04CA2, },
   {       "d90.03", 0x00100000, 0x6FA894A1, },
   {       "d90.04", 0x00200000, 0xFEEE5FDA, },
   {       "d90.05", 0x00200000, 0xC192331F, },
   {       "d90.06", 0x00100000, 0x166A72B8, },
   {       "d90.07", 0x00100000, 0xB436B42D, },
   {       "d90.08", 0x00100000, 0x25A4FB2C, },
   {       "d90.17", 0x00040000, 0x711f1894, },
   {       "d90.10", 0x00040000, 0x8E957D3D, },
   {       "d90.11", 0x00040000, 0xEDFDBB7F, },
   {       "d90.12", 0x00040000, 0x9E523996, },
   {       "d90.13", 0x00040000, 0x6762BD90, },
   {       "d90.14", 0x00040000, 0x8E33357E, },
   {           NULL,          0,          0, },
};


static struct ROMSW_DATA romsw_data_bubble_bobble_2[] =
{
   { "Taito World (Bubble Bobble 2)", 0x03 },
   { NULL,                     0    },
};

static struct ROMSW_INFO bubble_bobble_2_romsw[] =
{
   { 0x0FFFFF, 0x03, romsw_data_bubble_bobble_2 },
   { 0,        0,    NULL },
};

struct GAME_MAIN game_bubble_bobble_2 =
{
   bubble_bobble_2_dirs,
   bubble_bobble_2_roms,
   f3_system_inputs,
   NULL,
   bubble_bobble_2_romsw, 

   LoadBubbleSymphony,
   ClearBubbleSymphony,
   &bubble_symphony_video,
   ExecuteBubbleSymphonyFrame,
   "bublbob2",
   "Bubble Bobble 2",
   "ouVtHj[",
   COMPANY_ID_TAITO,
   "D90",
   1994,
   NULL,
};


static UBYTE *RAM_BG0;
static UBYTE *RAM_BG1;
static UBYTE *RAM_BG2;
static UBYTE *RAM_BG3;
static UBYTE *RAM_BG4;		// PIXEL

static UBYTE *RAM_SCR0;
static UBYTE *RAM_SCR1;
static UBYTE *RAM_SCR2;
static UBYTE *RAM_SCR3;
static UBYTE *RAM_SCR4;		// PIXEL

static ULONG SCR0_XOFS;
static ULONG SCR1_XOFS;
static ULONG SCR2_XOFS;
static ULONG SCR3_XOFS;
static ULONG SCR4_XOFS;		// PIXEL

static ULONG SCR0_YOFS;
static ULONG SCR1_YOFS;
static ULONG SCR2_YOFS;
static ULONG SCR3_YOFS;
static ULONG SCR4_YOFS;		// PIXEL

static UBYTE *GFX_BG0;
static UBYTE *GFX_BG0_SOLID;

static UBYTE *GFX_SPR;
static UBYTE *GFX_SPR_SOLID;

void LoadBubbleSymphony(void)
{
   int ta,tb,tc;

   RAMSize=0x80000;

   if(!(RAM=AllocateMem(0x80000))) return;
   if(!(ROM=AllocateMem(0x100000))) return;
   if(!(GFX=AllocateMem(0x3E8D00+0x336E00))) return;

   GFX_BG0 = GFX+0x3E8D00;
   GFX_SPR = GFX+0x000000;

   tb=0;
   if(!load_rom("d90.03", ROM, 0xFA340)) return;	// 16x16 SPRITES ($3E8D)
   for(ta=0;ta<0xFA340;ta++){
      GFX[tb++]=ROM[ta]&15;
      GFX[tb++]=ROM[ta]>>4;
      tb+=2;
   }
   tb=2;
   if(!load_rom("d90.02", ROM, 0xFA340)) return;	// 16x16 SPRITES
   for(ta=0;ta<0xFA340;ta++){
      GFX[tb++]=ROM[ta]&15;
      GFX[tb++]=ROM[ta]>>4;
      tb+=2;
   }
   tb=0;
   if(!load_rom("d90.01", ROM, 0xFA340)) return;	// 16x16 SPRITES (MASK)
   for(ta=0;ta<0xFA340;ta++){
      tc=ROM[ta];
      GFX[tb+3]|=((tc&0xC0)>>6)<<4;
      GFX[tb+2]|=((tc&0x30)>>4)<<4;
      GFX[tb+1]|=((tc&0x0C)>>2)<<4;
      GFX[tb+0]|=((tc&0x03)>>0)<<4;
      tb+=4;
   }

   memset(GFX+0x1C8900,0x00,0x100);

   tb=0;
   if(!load_rom("d90.08", ROM, 0xCDB80)) return;	// 16x16 TILES ($336E)
   for(ta=0;ta<0xCDB80;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=4;
   if(!load_rom("d90.07", ROM, 0xCDB80)) return;	// 16x16 TILES
   for(ta=0;ta<0xCDB80;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=0;
   if(!load_rom("d90.06", ROM, 0xCDB80)) return;	// 16x16 TILES (MASK)
   for(ta=0;ta<0xCDB80;ta+=2){
      tc=ROM[ta];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<4;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<4;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<4;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<4;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<4;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<4;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<4;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<4;
      tb+=8;
   }

   GFX_BG0_SOLID = MakeSolidTileMap16x16(GFX_BG0, 0x336E);
   GFX_SPR_SOLID = make_solid_mask_16x16(GFX_SPR, 0x3E8D);

   // Setup 68020 Memory Map
   // ----------------------

   AddF3MemoryMap(0x100000);

   RAM_BG3=RAM+0x30000;
   RAM_BG2=RAM+0x32000;
   RAM_BG0=RAM+0x34000;
   RAM_BG1=RAM+0x36000;
   RAM_BG4=RAM+0x50000;

   RAM_SCR3=RAM+0x6A000;
   RAM_SCR2=RAM+0x6A002;
   RAM_SCR0=RAM+0x6A004;
   RAM_SCR1=RAM+0x6A006;
   RAM_SCR4=RAM+0x6A018;

   SCR3_XOFS=(0xFFD8)<<6;
   SCR2_XOFS=(0xFFDC)<<6;
   SCR0_XOFS=(0xFFE0)<<6;
   SCR1_XOFS=(0xFFE4)<<6;
   SCR4_XOFS=(0x0029-1)<<0;

   SCR3_YOFS=(0xFFFF)<<7;
   SCR2_YOFS=(0xFFFF)<<7;
   SCR0_YOFS=(0xFFFF)<<7;
   SCR1_YOFS=(0xFFFF)<<7;
   SCR4_YOFS=(0x001F-1)<<0;

   if(!load_rom("d90.12", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+0]=RAM[ta];
   }
   if(!load_rom("d90.11", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+1]=RAM[ta];
   }
   if(!load_rom("d90.10", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+2]=RAM[ta];
   }

   /* This one is shared between the two sets */
   if(!load_rom_index(8, RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+3]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x80000);
   memset(RAM+0x69000,0xFF,0x01000);

   // EEPROM HACKS
   // ------------

   WriteWord68k(&ROM[0x0ECD6A],0x7F00);		//	raine	#$00 <read eeprom>
   WriteWord68k(&ROM[0x0ECD6C],0x4E75);		//	rts

   WriteWord68k(&ROM[0x0ECDE6],0x7F01);		//	raine	#$01 <write eeprom>
   WriteWord68k(&ROM[0x0ECDE8],0x4E75);		//	rts

   // SPEED HACKS
   // -----------

   WriteWord68k(&ROM[0x0E9A3A],0x7F02);		//	raine	#$02 <stop cpu>
   WriteWord68k(&ROM[0x0E9A3C],0x4E71);		//	nop
   WriteWord68k(&ROM[0x0E9A3E],0x4E71);		//	nop

   WriteWord68k(&ROM[0x0E9902],0x4EF9);		//	jmp	$FFA00
   WriteLong68k(&ROM[0x0E9904],0x000FFA00);	//

   WriteLong68k(&ROM[0x0FFA00],0x526D8124);	//	addq	#1,(-32476,a5)
   WriteWord68k(&ROM[0x0FFA04],0x7F02);		//	raine	#$02 <stop cpu>

   WriteWord68k(&ROM[0x0FFA06],0x4EF9);		//	jmp	$E98F8
   WriteLong68k(&ROM[0x0FFA08],0x000E98F8);	//

   F3SystemEEPROMAccess=&F3SysEEPROMAccessMode1;


   InitPaletteMap(RAM+0x60000, 0x200, 0x40, 0x1000);

   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu.RAM	= RAM+0x20000;
   tc0003vcu.mapper	= &Map_24bit_xRGB;
   tc0003vcu.bmp_x	= 64;
   tc0003vcu.bmp_y	= 64;
   tc0003vcu.bmp_w	= 320;
   tc0003vcu.bmp_h	= 224;
   tc0003vcu.scr_x	= 0;
   tc0003vcu.scr_y	= 0;

   // Init tc0200obj emulation
   // ------------------------

   tc0200obj.RAM	= RAM+0x20000;
   tc0200obj.RAM_B	= RAM+0x28000;
   tc0200obj.GFX	= GFX_SPR;
   tc0200obj.MASK	= GFX_SPR_SOLID;
   tc0200obj.bmp_x	= 64;
   tc0200obj.bmp_y	= 64;
   tc0200obj.bmp_w	= 320;
   tc0200obj.bmp_h	= 224;
   tc0200obj.mapper	= &Map_24bit_xRGB;
   tc0200obj.tile_mask	= 0x3FFF;
   tc0200obj.ofs_x	= 0-0x2E;
   tc0200obj.ofs_y	= 0-0x1F;

   tc0200obj.cols	= 32;

   init_tc0200obj();

   init_f3_system_ioc(F3_IOC_2P_3BUTTON);

   init_m68k();
}

void ClearBubbleSymphony(void)
{
   save_eeprom();

   #ifdef HACKS
      save_debug("ROM.bin",ROM,0x100000,0);
      save_debug("RAM.bin",RAM,0x080000,0);
      //save_debug("GFX.bin",GFX,0x64AB00+0x3C2300,0);
   #endif
}

void ExecuteBubbleSymphonyFrame(void)
{
   Execute68020(400000);
   /*#ifdef HACKS
      print_debug("[BEFORE INT2] PC:%06x SR:%04x\n",regs.pc,regs.sr);
   #endif*/
   Interrupt68020(2);

   Execute68020(400000);
   /*#ifdef HACKS
      print_debug("[BEFORE INT3] PC:%06x SR:%04x\n",regs.pc,regs.sr);
   #endif*/
   Interrupt68020(3);

   IntF3System();
}

void DrawBubbleSymphony(void)
{
   int x,y,ta,zz,zzz,zzzz,x16,y16;
   UBYTE *MAP;

   ClearPaletteMap();

   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu_layer_count = 0;

   if(check_layer_enabled(f3_bg0_id)){
   MAKE_SCROLL_1024x512_4_16(
      (ReadWord68k(&RAM_SCR0[0])-SCR0_XOFS)>>6,
      (ReadWord68k(&RAM_SCR0[8])-SCR0_YOFS)>>7
   );

   START_SCROLL_1024x512_4_16(64,64,320,224);

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG0[zz])&0x1FF,
               32
            );

      switch(RAM_BG0[zz]&0xC0){
      case 0x00: Draw16x16_Mapped(&GFX_BG0[(ReadWord68k(&RAM_BG0[zz+2])&0x3FFF)<<8],x,y,MAP);        break;
      case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[(ReadWord68k(&RAM_BG0[zz+2])&0x3FFF)<<8],x,y,MAP);  break;
      case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[(ReadWord68k(&RAM_BG0[zz+2])&0x3FFF)<<8],x,y,MAP);  break;
      case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[(ReadWord68k(&RAM_BG0[zz+2])&0x3FFF)<<8],x,y,MAP); break;
      }

   END_SCROLL_1024x512_4_16();
   }

   if(check_layer_enabled(f3_bg1_id)){
   MAKE_SCROLL_1024x512_4_16(
      (ReadWord68k(&RAM_SCR1[0])-SCR1_XOFS)>>6,
      (ReadWord68k(&RAM_SCR1[8])-SCR1_YOFS)>>7
   );

   START_SCROLL_1024x512_4_16(64,64,320,224);

      ta=ReadWord68k(&RAM_BG1[zz+2])&0x3FFF;
      if(ta!=0){

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG1[zz])&0x1FF,
               32
            );

      if(GFX_BG0_SOLID[ta]==0){
         switch(RAM_BG1[zz]&0xC0){
         case 0x00: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      else{
         switch(RAM_BG1[zz]&0xC0){
         case 0x00: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      }

   END_SCROLL_1024x512_4_16();
   }

   if(check_layer_enabled(f3_bg2_id)){
   MAKE_SCROLL_1024x512_4_16(
      (ReadWord68k(&RAM_SCR2[0])-SCR2_XOFS)>>6,
      (ReadWord68k(&RAM_SCR2[8])-SCR2_YOFS)>>7
   );

   START_SCROLL_1024x512_4_16(64,64,320,224);

      ta=ReadWord68k(&RAM_BG2[zz+2])&0x3FFF;
      if(ta!=0){

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG2[zz])&0x1FF,
               32
            );

      if(GFX_BG0_SOLID[ta]==0){
         switch(RAM_BG2[zz]&0xC0){
         case 0x00: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      else{
         switch(RAM_BG2[zz]&0xC0){
         case 0x00: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      }

   END_SCROLL_1024x512_4_16();
   }

   if(check_layer_enabled(f3_bg3_id)){
   MAKE_SCROLL_1024x512_4_16(
      (ReadWord68k(&RAM_SCR3[0])-SCR3_XOFS)>>6,
      (ReadWord68k(&RAM_SCR3[8])-SCR3_YOFS)>>7
   );

   START_SCROLL_1024x512_4_16(64,64,320,224);

      ta=ReadWord68k(&RAM_BG3[zz+2])&0x3FFF;
      if(ta!=0){

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG3[zz])&0x1FF,
               32
            );

      if(GFX_BG0_SOLID[ta]==0){
         switch(RAM_BG3[zz]&0xC0){
         case 0x00: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      else{
         switch(RAM_BG3[zz]&0xC0){
         case 0x00: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      }

   END_SCROLL_1024x512_4_16();
   }

   if(check_layer_enabled(f3_bg6_id)){
   zzz=0-((ReadWord68k(&RAM_SCR4[0])-SCR4_XOFS)>>0);
   zzzz=((zzz&0x1F8)>>3)<<10;			// X Offset (16-1024)
   x16=7-(zzz&7);				// X Offset (0-15)
   zz=17-((ReadWord68k(&RAM_SCR4[2])-SCR4_YOFS)>>0);
   zz&=0xFF;					// Y Offset (0-255)
   zzzz+=zz<<2;					// Y Offset (0-255)

   zzzz&=0xFFFF;

   MAP = MAP_PALETTE(
            Map_24bit_xRGB,
            16,
            16
         );

   if(zz<=48){

   for(x=56+x16;x<(320+64);x+=8){
      Draw8xH_Trans_Packed_Mapped_FlipY_Column(&RAM_BG4[zzzz],x,64+16,208,MAP);
      zzzz=(zzzz+0x400)&0xFFFF;
   }

   }
   else{

   for(x=56+x16;x<(320+64);x+=8){
      Draw8xH_Trans_Packed_Mapped_FlipY_Column(&RAM_BG4[zzzz],x,64+16,256-zz,MAP);
      Draw8xH_Trans_Packed_Mapped_FlipY_Column(&RAM_BG4[zzzz+((256-zz)<<2)-0x400],x,64+16+(256-zz),208-(256-zz),MAP);
      zzzz=(zzzz+0x400)&0xFFFF;
   }

   }

   }

   render_tc0200obj_mapped_f3system();

   if(check_layer_enabled(f3_bg5_id)){
      f3video_render_fg0();
   }
}
