/******************************************************************************/
/*                                                                            */
/*                        CAVE / ATLUS GAMES (C) 1998                         */
/*                                                                            */
/******************************************************************************/

#include "gameinc.h"
#include "cave.h"
#include "cat93c46.h"

static struct DIR_INFO esprade_dirs[] =
{
   { "esprade", },
   { NULL, },
};

static struct ROM_INFO esprade_roms[] =
{
   {      "u19.bin", 0x00400000, 0xf54b1cab, },
   {      "u41.bin", 0x00080000, 0xdef30539, },
   {      "u42.bin", 0x00080000, 0x0718c7e5, },
   {      "u51.bin", 0x00400000, 0x0b9b875c, },
   {      "u52.bin", 0x00400000, 0xe7abe7b4, },
   {      "u53.bin", 0x00400000, 0x51a0f391, },
   {      "u54.bin", 0x00400000, 0xe7ca6936, },
   {      "u55.bin", 0x00400000, 0xf53bd94f, },
   {      "u63.bin", 0x00400000, 0x2f2fe92c, },
   {      "u64.bin", 0x00400000, 0x491a3da4, },
   {      "u65.bin", 0x00400000, 0x06563efe, },
   {      "u66.bin", 0x00400000, 0x7bbe4cfc, },
   {           NULL,          0,          0, },
};

static struct INPUT_INFO cave_68k_inputs[] =
{
   { KB_DEF_COIN1,        MSG_COIN1,               0x000000, 0x01, BIT_ACTIVE_0 },
   { KB_DEF_COIN2,        MSG_COIN2,               0x000002, 0x01, BIT_ACTIVE_0 },
   { KB_DEF_TEST,         MSG_TEST,                0x000000, 0x02, BIT_ACTIVE_0 },
   { KB_DEF_SERVICE,      MSG_SERVICE,             0x000002, 0x02, BIT_ACTIVE_0 },

   { KB_DEF_P1_START,     MSG_P1_START,            0x000001, 0x80, BIT_ACTIVE_0 },
   { KB_DEF_P1_UP,        MSG_P1_UP,               0x000001, 0x01, BIT_ACTIVE_0 },
   { KB_DEF_P1_DOWN,      MSG_P1_DOWN,             0x000001, 0x02, BIT_ACTIVE_0 },
   { KB_DEF_P1_LEFT,      MSG_P1_LEFT,             0x000001, 0x04, BIT_ACTIVE_0 },
   { KB_DEF_P1_RIGHT,     MSG_P1_RIGHT,            0x000001, 0x08, BIT_ACTIVE_0 },
   { KB_DEF_P1_B1,        MSG_P1_B1,               0x000001, 0x10, BIT_ACTIVE_0 },
   { KB_DEF_P1_B2,        MSG_P1_B2,               0x000001, 0x20, BIT_ACTIVE_0 },
   { KB_DEF_P1_B3,        MSG_P1_B3,               0x000001, 0x40, BIT_ACTIVE_0 },

   { KB_DEF_P2_START,     MSG_P2_START,            0x000003, 0x80, BIT_ACTIVE_0 },
   { KB_DEF_P2_UP,        MSG_P2_UP,               0x000003, 0x01, BIT_ACTIVE_0 },
   { KB_DEF_P2_DOWN,      MSG_P2_DOWN,             0x000003, 0x02, BIT_ACTIVE_0 },
   { KB_DEF_P2_LEFT,      MSG_P2_LEFT,             0x000003, 0x04, BIT_ACTIVE_0 },
   { KB_DEF_P2_RIGHT,     MSG_P2_RIGHT,            0x000003, 0x08, BIT_ACTIVE_0 },
   { KB_DEF_P2_B1,        MSG_P2_B1,               0x000003, 0x10, BIT_ACTIVE_0 },
   { KB_DEF_P2_B2,        MSG_P2_B2,               0x000003, 0x20, BIT_ACTIVE_0 },
   { KB_DEF_P2_B3,        MSG_P2_B3,               0x000003, 0x40, BIT_ACTIVE_0 },

   { 0,                   NULL,                    0,        0,    0            },
};

static struct VIDEO_INFO cave_68k_video =
{
   draw_cave_68k,
   320,
   240,
   32,
   VIDEO_ROTATE_NORMAL |
   VIDEO_ROTATABLE,
};

static struct VIDEO_INFO cave_68k_video_r270 =
{
   draw_cave_68k,
   320,
   240,
   32,
   VIDEO_ROTATE_270 |
   VIDEO_ROTATABLE,
};

struct GAME_MAIN game_esprade =
{
   esprade_dirs,
   esprade_roms,
   cave_68k_inputs,
   NULL,
   NULL,

   load_esprade,
   clear_cave_68k,
   &cave_68k_video_r270,
   execute_cave_68k_frame,
   "esprade",
   "Esprade",
   "CV-04",
   COMPANY_ID_CAVE,
   NULL,
   1998,
   NULL,
};

static struct DIR_INFO uo_poko_dirs[] =
{
   { "uo_poko", },
   { "uopoko", },
   { NULL, },
};

static struct ROM_INFO uo_poko_roms[] =
{
   {       "u4.bin", 0x00200000, 0xa2d0d755, },
   {     "u25j.bin", 0x00080000, 0x68cb6211, },
   {     "u26j.bin", 0x00080000, 0xe7eec050, },
   {      "u33.bin", 0x00400000, 0x5d142ad2, },
   {      "u49.bin", 0x00400000, 0x12fb11bb, },
   {           NULL,          0,          0, },
};

struct GAME_MAIN game_uo_poko =
{
   uo_poko_dirs,
   uo_poko_roms,
   cave_68k_inputs,
   NULL,
   NULL,

   load_uo_poko,
   clear_cave_68k,
   &cave_68k_video,
   execute_cave_68k_frame,
   "uopoko",
   "Uo Poko",
   NULL,
   COMPANY_ID_CAVE,
   "CV-02",
   1998,
   NULL,
};

static struct DIR_INFO feveron_dirs[] =
{
   { "feveron", },
   { "dfeveron", },
   { "dangun_feveron", },
   { NULL, },
};

static struct ROM_INFO feveron_roms[] =
{
   {      "u19.bin", 0x00400000, 0x5f5514da, },
   {      "u25.bin", 0x00400000, 0xa6f6a95d, },
   {      "u26.bin", 0x00400000, 0x32edb62a, },
   {      "u33.bin", 0x00080000, 0xe53a7db3, },
   {      "u34.bin", 0x00080000, 0xbe87f19d, },
   {      "u49.bin", 0x00200000, 0xd21cdda7, },
   {      "u50.bin", 0x00200000, 0x7a344417, },
   {           NULL,          0,          0, },
};

struct GAME_MAIN game_feveron =
{
   feveron_dirs,
   feveron_roms,
   cave_68k_inputs,
   NULL,
   NULL,

   load_feveron,
   clear_cave_68k,
   &cave_68k_video_r270,
   execute_cave_68k_frame,
   "dfeveron",
   "Dangun Feveron",
   NULL,
   COMPANY_ID_CAVE,
   "CV-05",
   1998,
   NULL,
};

static UBYTE *GFX_BG0;
static UBYTE *GFX_BG1;
static UBYTE *GFX_BG2;

static UBYTE *GFX_OBJ_A;
static UBYTE *GFX_OBJ_B;
static UBYTE *GFX_OBJ_SIZE;

static UBYTE *GFX_BG0_SOLID;
static UBYTE *GFX_BG1_SOLID;
static UBYTE *GFX_BG2_SOLID;

static UBYTE *GFX_OBJ_SOLID;

static UBYTE *RAM_BG0;
static UBYTE *RAM_BG1;
static UBYTE *RAM_BG2;

static UBYTE *RAM_BG0_CTRL;
static UBYTE *RAM_BG1_CTRL;
static UBYTE *RAM_BG2_CTRL;

static UBYTE *RAM_CTRL;

static ULONG gfx_colours;

static UBYTE esp_input_rb(ULONG offset)
{
   switch(offset & 3){
      case 0x00:
         return (input_buffer[0] & 0x03) | 0xFC;
      break;
      case 0x01:
         return input_buffer[1];
      break;
      case 0x02:
         return (input_buffer[2] & 0x03) | 0xF4 | (eeprom_93c46_rb_cave() << 3);
      break;
      case 0x03:
         return input_buffer[3];
      break;
      default:
         return 0;
      break;
   }
}

static UWORD esp_input_rw(ULONG offset)
{
   return (esp_input_rb(offset+0) << 8) |
          (esp_input_rb(offset+1) << 0);
}

static void esp_input_wb(ULONG offset, UBYTE data)
{
   #ifdef HACKS
      print_debug("esp_input_wb(%02x,%02x)\n", offset & 3, data);
   #endif
}

static void esp_input_ww(ULONG offset, UWORD data)
{
   esp_input_wb(offset+0, data >> 8);
   esp_input_wb(offset+1, data >> 0);
}

static UBYTE esp_800000_rb(ULONG offset)
{
   offset &= 0x7F;

   switch(offset){
      case 0x01:
         return 0xFF ^ (0x04 | 0x01);
      break;
      default:
         #ifdef HACKS
            print_debug("esp_800000_rb(%02x)\n", offset);
         #endif
         return 0xFF;
      break;
   }
}

static UWORD esp_800000_rw(ULONG offset)
{
   return (esp_800000_rb(offset+0) << 8) |
          (esp_800000_rb(offset+1) << 0);
}

static void esp_800000_wb(ULONG offset, UBYTE data)
{
   offset &= 0x7F;

   switch(offset){
      case 0x08:
      case 0x09:
         RAM_CTRL[offset^1] = data;
      break;
      case 0x68:
      case 0x69:
      break;
      default:
         RAM_CTRL[offset^1] = data;
         #ifdef HACKS
            print_debug("esp_800000_wb(%02x,%02x)\n", offset, data);
         #endif
      break;
   }
}

static void esp_800000_ww(ULONG offset, UWORD data)
{
   esp_800000_wb(offset+0, data >> 8);
   esp_800000_wb(offset+1, data >> 0);
}


static void esp_eeprom_ww(ULONG offset, UWORD data)
{
   if(data&0x0400)

      eeprom_93c46_wb((data & 0x0800) >> 11);

}

static int layer_id_data[4];

static char *layer_id_name[4] =
{
   "BG0", "BG1", "BG2", "OBJECT",
};

void load_esprade(void)
{
   int ta,tb;
   UBYTE *TMP;

   if(!(GFX_OBJ_A   =AllocateMem(0x1000000))) return;
   if(!(GFX_OBJ_B   =AllocateMem(0x1000000))) return;
   if(!(GFX_OBJ_SIZE=AllocateMem(0x1000000/0x100))) return;

   if(!(GFX_BG0=AllocateMem(0x0800000))) return;
   if(!(GFX_BG1=AllocateMem(0x0800000))) return;
   if(!(GFX_BG2=AllocateMem(0x0400000))) return;
   if(!(TMP=    AllocateMem(0x0400000))) return;

   if(!load_rom("u54.bin", TMP+0x000000, 0x400000)) return;		// GFX BG0
   tb=0;
   for(ta = 0; ta < 0x400000; ta += 8, tb += 8){
      GFX_BG0[tb+0]  = (TMP[ta+0] >> 4) & 0x0F;
      GFX_BG0[tb+1]  = (TMP[ta+0] >> 0) & 0x0F;
      GFX_BG0[tb+2]  = (TMP[ta+2] >> 4) & 0x0F;
      GFX_BG0[tb+3]  = (TMP[ta+2] >> 0) & 0x0F;
      GFX_BG0[tb+4]  = (TMP[ta+4] >> 4) & 0x0F;
      GFX_BG0[tb+5]  = (TMP[ta+4] >> 0) & 0x0F;
      GFX_BG0[tb+6]  = (TMP[ta+6] >> 4) & 0x0F;
      GFX_BG0[tb+7]  = (TMP[ta+6] >> 0) & 0x0F;
      GFX_BG0[tb+0] |= (TMP[ta+1] << 0) & 0xF0;
      GFX_BG0[tb+1] |= (TMP[ta+1] << 4) & 0xF0;
      GFX_BG0[tb+2] |= (TMP[ta+3] << 0) & 0xF0;
      GFX_BG0[tb+3] |= (TMP[ta+3] << 4) & 0xF0;
      GFX_BG0[tb+4] |= (TMP[ta+5] << 0) & 0xF0;
      GFX_BG0[tb+5] |= (TMP[ta+5] << 4) & 0xF0;
      GFX_BG0[tb+6] |= (TMP[ta+7] << 0) & 0xF0;
      GFX_BG0[tb+7] |= (TMP[ta+7] << 4) & 0xF0;
   }
   if(!load_rom("u55.bin", TMP+0x000000, 0x400000)) return;		// GFX BG0
   for(ta = 0; ta < 0x400000; ta += 8, tb += 8){
      GFX_BG0[tb+0]  = (TMP[ta+0] >> 4) & 0x0F;
      GFX_BG0[tb+1]  = (TMP[ta+0] >> 0) & 0x0F;
      GFX_BG0[tb+2]  = (TMP[ta+2] >> 4) & 0x0F;
      GFX_BG0[tb+3]  = (TMP[ta+2] >> 0) & 0x0F;
      GFX_BG0[tb+4]  = (TMP[ta+4] >> 4) & 0x0F;
      GFX_BG0[tb+5]  = (TMP[ta+4] >> 0) & 0x0F;
      GFX_BG0[tb+6]  = (TMP[ta+6] >> 4) & 0x0F;
      GFX_BG0[tb+7]  = (TMP[ta+6] >> 0) & 0x0F;
      GFX_BG0[tb+0] |= (TMP[ta+1] << 0) & 0xF0;
      GFX_BG0[tb+1] |= (TMP[ta+1] << 4) & 0xF0;
      GFX_BG0[tb+2] |= (TMP[ta+3] << 0) & 0xF0;
      GFX_BG0[tb+3] |= (TMP[ta+3] << 4) & 0xF0;
      GFX_BG0[tb+4] |= (TMP[ta+5] << 0) & 0xF0;
      GFX_BG0[tb+5] |= (TMP[ta+5] << 4) & 0xF0;
      GFX_BG0[tb+6] |= (TMP[ta+7] << 0) & 0xF0;
      GFX_BG0[tb+7] |= (TMP[ta+7] << 4) & 0xF0;
   }

   if(!load_rom("u52.bin", TMP+0x000000, 0x400000)) return;		// GFX BG1
   tb=0;
   for(ta = 0; ta < 0x400000; ta += 8, tb += 8){
      GFX_BG1[tb+0]  = (TMP[ta+0] >> 4) & 0x0F;
      GFX_BG1[tb+1]  = (TMP[ta+0] >> 0) & 0x0F;
      GFX_BG1[tb+2]  = (TMP[ta+2] >> 4) & 0x0F;
      GFX_BG1[tb+3]  = (TMP[ta+2] >> 0) & 0x0F;
      GFX_BG1[tb+4]  = (TMP[ta+4] >> 4) & 0x0F;
      GFX_BG1[tb+5]  = (TMP[ta+4] >> 0) & 0x0F;
      GFX_BG1[tb+6]  = (TMP[ta+6] >> 4) & 0x0F;
      GFX_BG1[tb+7]  = (TMP[ta+6] >> 0) & 0x0F;
      GFX_BG1[tb+0] |= (TMP[ta+1] << 0) & 0xF0;
      GFX_BG1[tb+1] |= (TMP[ta+1] << 4) & 0xF0;
      GFX_BG1[tb+2] |= (TMP[ta+3] << 0) & 0xF0;
      GFX_BG1[tb+3] |= (TMP[ta+3] << 4) & 0xF0;
      GFX_BG1[tb+4] |= (TMP[ta+5] << 0) & 0xF0;
      GFX_BG1[tb+5] |= (TMP[ta+5] << 4) & 0xF0;
      GFX_BG1[tb+6] |= (TMP[ta+7] << 0) & 0xF0;
      GFX_BG1[tb+7] |= (TMP[ta+7] << 4) & 0xF0;
   }
   if(!load_rom("u53.bin", TMP+0x000000, 0x400000)) return;		// GFX BG1
   for(ta = 0; ta < 0x400000; ta += 8, tb += 8){
      GFX_BG1[tb+0]  = (TMP[ta+0] >> 4) & 0x0F;
      GFX_BG1[tb+1]  = (TMP[ta+0] >> 0) & 0x0F;
      GFX_BG1[tb+2]  = (TMP[ta+2] >> 4) & 0x0F;
      GFX_BG1[tb+3]  = (TMP[ta+2] >> 0) & 0x0F;
      GFX_BG1[tb+4]  = (TMP[ta+4] >> 4) & 0x0F;
      GFX_BG1[tb+5]  = (TMP[ta+4] >> 0) & 0x0F;
      GFX_BG1[tb+6]  = (TMP[ta+6] >> 4) & 0x0F;
      GFX_BG1[tb+7]  = (TMP[ta+6] >> 0) & 0x0F;
      GFX_BG1[tb+0] |= (TMP[ta+1] << 0) & 0xF0;
      GFX_BG1[tb+1] |= (TMP[ta+1] << 4) & 0xF0;
      GFX_BG1[tb+2] |= (TMP[ta+3] << 0) & 0xF0;
      GFX_BG1[tb+3] |= (TMP[ta+3] << 4) & 0xF0;
      GFX_BG1[tb+4] |= (TMP[ta+5] << 0) & 0xF0;
      GFX_BG1[tb+5] |= (TMP[ta+5] << 4) & 0xF0;
      GFX_BG1[tb+6] |= (TMP[ta+7] << 0) & 0xF0;
      GFX_BG1[tb+7] |= (TMP[ta+7] << 4) & 0xF0;
   }

   if(!load_rom("u51.bin", TMP+0x000000, 0x400000)) return;		// GFX BG2
   tb=0;
   for(ta = 0; ta < 0x400000; ta += 8, tb += 8){
      GFX_BG2[tb+0]  = (TMP[ta+0] >> 4) & 0x0F;
      GFX_BG2[tb+1]  = (TMP[ta+0] >> 0) & 0x0F;
      GFX_BG2[tb+2]  = (TMP[ta+2] >> 4) & 0x0F;
      GFX_BG2[tb+3]  = (TMP[ta+2] >> 0) & 0x0F;
      GFX_BG2[tb+4]  = (TMP[ta+4] >> 4) & 0x0F;
      GFX_BG2[tb+5]  = (TMP[ta+4] >> 0) & 0x0F;
      GFX_BG2[tb+6]  = (TMP[ta+6] >> 4) & 0x0F;
      GFX_BG2[tb+7]  = (TMP[ta+6] >> 0) & 0x0F;
      GFX_BG2[tb+0] |= (TMP[ta+1] << 0) & 0xF0;
      GFX_BG2[tb+1] |= (TMP[ta+1] << 4) & 0xF0;
      GFX_BG2[tb+2] |= (TMP[ta+3] << 0) & 0xF0;
      GFX_BG2[tb+3] |= (TMP[ta+3] << 4) & 0xF0;
      GFX_BG2[tb+4] |= (TMP[ta+5] << 0) & 0xF0;
      GFX_BG2[tb+5] |= (TMP[ta+5] << 4) & 0xF0;
      GFX_BG2[tb+6] |= (TMP[ta+7] << 0) & 0xF0;
      GFX_BG2[tb+7] |= (TMP[ta+7] << 4) & 0xF0;
   }

   if(!load_rom("u64.bin", TMP, 0x400000)) return;		// OBJECT
   tb=0;
   for(ta = 0; ta < 0x400000; ta ++, tb += 2){
      GFX_OBJ_A[tb+1] = (TMP[ta] >> 4) & 0x0F;
      GFX_OBJ_A[tb+0] = (TMP[ta] >> 0) & 0x0F;
   }
   if(!load_rom("u66.bin", TMP, 0x400000)) return;		// OBJECT
   for(ta = 0; ta < 0x400000; ta ++, tb += 2){
      GFX_OBJ_A[tb+1] = (TMP[ta] >> 4) & 0x0F;
      GFX_OBJ_A[tb+0] = (TMP[ta] >> 0) & 0x0F;
   }
   if(!load_rom("u63.bin", TMP, 0x400000)) return;		// OBJECT
   tb=0;
   for(ta = 0; ta < 0x400000; ta ++, tb += 2){
      GFX_OBJ_A[tb+1] |= (TMP[ta] << 0) & 0xF0;
      GFX_OBJ_A[tb+0] |= (TMP[ta] << 4) & 0xF0;
   }
   if(!load_rom("u65.bin", TMP, 0x400000)) return;		// OBJECT
   for(ta = 0; ta < 0x400000; ta ++, tb += 2){
      GFX_OBJ_A[tb+1] |= (TMP[ta] << 0) & 0xF0;
      GFX_OBJ_A[tb+0] |= (TMP[ta] << 4) & 0xF0;
   }

   FreeMem(TMP);

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x0800000/0x40);
   GFX_BG1_SOLID = make_solid_mask_8x8(GFX_BG1, 0x0800000/0x40);
   GFX_BG2_SOLID = make_solid_mask_8x8(GFX_BG2, 0x0400000/0x40);

   memcpy(GFX_OBJ_B,GFX_OBJ_A,0x1000000);
   memset(GFX_OBJ_SIZE,0x00,0x1000000/0x100);
   GFX_OBJ_SOLID = make_solid_mask_16x16(GFX_OBJ_B, 0x1000000/0x100);

   gfx_colours = 256;

   RAMSize=0x80000;

   if(!(RAM=AllocateMem(RAMSize))) return;
   if(!(ROM=AllocateMem(0x100000))) return;

   if(!load_rom("u42.bin", RAM, 0x80000)) return;		// 68000 ROM
   for(ta=0;ta<0x80000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   if(!load_rom("u41.bin", RAM, 0x80000)) return;		// 68000 ROM
   for(ta=0;ta<0x80000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x80000);

   RAM_BG0 = RAM + 0x020000;
   RAM_BG1 = RAM + 0x028000;
   RAM_BG2 = RAM + 0x030000;

   RAM_BG0_CTRL = RAM + 0x048080;
   RAM_BG1_CTRL = RAM + 0x048088;
   RAM_BG2_CTRL = RAM + 0x048090;

   RAM_CTRL = RAM + 0x48100;

/*
   WriteWord68k(&ROM[0x02D4A],0x4E71);
*/
   InitPaletteMap(RAM+0x38000, 0x80, 0x100, 0x8000);

   set_colour_mapper(&col_map_xggg_ggrr_rrrb_bbbb);

   EEPROM = RAM + 0x48000;

   eeprom_data = (UWORD *) EEPROM;

   add_eeprom(EEPROM, 0x80, EPR_INVALIDATE_ON_ROM_CHANGE);
   load_eeprom();

   layer_id_data[0] = add_layer_info(layer_id_name[0]);
   layer_id_data[1] = add_layer_info(layer_id_name[1]);
   layer_id_data[2] = add_layer_info(layer_id_name[2]);
   layer_id_data[3] = add_layer_info(layer_id_name[3]);

/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x100000);
   ByteSwap(RAM,0x048000);

   AddMemFetch(0x000000, 0x0FFFFF, ROM+0x000000-0x000000);		// 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x0FFFFF, NULL, ROM+0x000000);			// 68000 ROM
   AddReadByte(0x100000, 0x10FFFF, NULL, RAM+0x000000);			// 68000 RAM
   AddReadByte(0x400000, 0x40FFFF, NULL, RAM+0x010000);			// OBJECT RAM
   AddReadByte(0x500000, 0x507FFF, NULL, RAM+0x020000);			// BG0 RAM
   AddReadByte(0x600000, 0x607FFF, NULL, RAM+0x028000);			// BG1 RAM
   AddReadByte(0x700000, 0x707FFF, NULL, RAM+0x030000);			// BG2 RAM
   AddReadByte(0x800000, 0x80007F, esp_800000_rb, NULL);		// ?
   AddReadByte(0xC00000, 0xC0FFFF, NULL, RAM+0x038000);			// COLOR RAM
   AddReadByte(0xD00000, 0xD00003, esp_input_rb, NULL);			// INPUT
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);		// <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x0FFFFF, NULL, ROM+0x000000);			// 68000 ROM
   AddReadWord(0x100000, 0x10FFFF, NULL, RAM+0x000000);			// 68000 RAM
   AddReadWord(0x400000, 0x40FFFF, NULL, RAM+0x010000);			// OBJECT RAM
   AddReadWord(0x500000, 0x507FFF, NULL, RAM+0x020000);			// BG0 RAM
   AddReadWord(0x600000, 0x607FFF, NULL, RAM+0x028000);			// BG1 RAM
   AddReadWord(0x700000, 0x707FFF, NULL, RAM+0x030000);			// BG2 RAM
   AddReadWord(0x800000, 0x80007F, esp_800000_rw, NULL);		// ?
   AddReadWord(0xC00000, 0xC0FFFF, NULL, RAM+0x038000);			// COLOR RAM
   AddReadWord(0xD00000, 0xD00003, esp_input_rw, NULL);			// INPUT
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);		// <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x100000, 0x10FFFF, NULL, RAM+0x000000);		// 68000 RAM
   AddWriteByte(0x400000, 0x40FFFF, NULL, RAM+0x010000);		// OBJECT RAM
   AddWriteByte(0x500000, 0x507FFF, NULL, RAM+0x020000);		// BG0 RAM
   AddWriteByte(0x600000, 0x607FFF, NULL, RAM+0x028000);		// BG1 RAM
   AddWriteByte(0x700000, 0x707FFF, NULL, RAM+0x030000);		// BG2 RAM
   AddWriteByte(0x800000, 0x80007F, esp_800000_wb, NULL);		// ?
   AddWriteByte(0xC00000, 0xC0FFFF, NULL, RAM+0x038000);		// COLOR RAM
   AddWriteByte(0xD00000, 0xD00003, esp_input_wb, NULL);		// INPUT
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);			// Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);		// <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x100000, 0x10FFFF, NULL, RAM+0x000000);		// 68000 RAM
   AddWriteWord(0x400000, 0x40FFFF, NULL, RAM+0x010000);		// OBJECT RAM
   AddWriteWord(0x500000, 0x507FFF, NULL, RAM+0x020000);		// BG0 RAM
   AddWriteWord(0x600000, 0x607FFF, NULL, RAM+0x028000);		// BG1 RAM
   AddWriteWord(0x700000, 0x707FFF, NULL, RAM+0x030000);		// BG2 RAM
   AddWriteWord(0xC00000, 0xC0FFFF, NULL, RAM+0x038000);		// COLOR RAM
   AddWriteWord(0xD00000, 0xD00003, esp_input_ww, NULL);		// INPUT
   AddWriteWord(0xE00000, 0xE00001, esp_eeprom_ww, NULL);		// EEPROM
   AddWriteWord(0x800000, 0x80007F, esp_800000_ww, NULL);		// ?
   AddWriteWord(0x900000, 0x900007, NULL, RAM+0x048080);		// BG0 CTRL RAM
   AddWriteWord(0xA00000, 0xA00007, NULL, RAM+0x048088);		// BG1 CTRL RAM
   AddWriteWord(0xB00000, 0xB00007, NULL, RAM+0x048090);		// BG2 CTRL RAM
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);		// <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();	// Set Starscream mem pointers... 
}

void load_feveron(void)
{
   int ta,tb;
   UBYTE *TMP;

   if(!(GFX_OBJ_A   =AllocateMem(0x1000000))) return;
   if(!(GFX_OBJ_B   =AllocateMem(0x1000000))) return;
   if(!(GFX_OBJ_SIZE=AllocateMem(0x1000000/0x100))) return;

   if(!(GFX_BG0=AllocateMem(0x0800000))) return;
   if(!(GFX_BG1=AllocateMem(0x0800000))) return;
   if(!(TMP=    AllocateMem(0x0400000))) return;

   GFX_BG2 = NULL;

   if(!load_rom("u50.bin", TMP+0x000000, 0x200000)) return;		// GFX BG0
   tb=0;
   for(ta = 0; ta < 0x200000; ta ++, tb += 2){
      GFX_BG0[tb+0] = (TMP[ta] >> 4) & 0x0F;
      GFX_BG0[tb+1] = (TMP[ta] >> 0) & 0x0F;
   }

   if(!load_rom("u49.bin", TMP+0x000000, 0x200000)) return;		// GFX BG1
   tb=0;
   for(ta = 0; ta < 0x200000; ta ++, tb += 2){
      GFX_BG1[tb+0] = (TMP[ta] >> 4) & 0x0F;
      GFX_BG1[tb+1] = (TMP[ta] >> 0) & 0x0F;
   }

   if(!load_rom("u25.bin", TMP, 0x400000)) return;		// OBJECT
   tb=0;
   for(ta = 0; ta < 0x400000; ta ++, tb += 2){
      GFX_OBJ_A[tb+1] = (TMP[ta] >> 4) & 0x0F;
      GFX_OBJ_A[tb+0] = (TMP[ta] >> 0) & 0x0F;
   }
   if(!load_rom("u26.bin", TMP, 0x400000)) return;		// OBJECT
   for(ta = 0; ta < 0x400000; ta ++, tb += 2){
      GFX_OBJ_A[tb+1] = (TMP[ta] >> 4) & 0x0F;
      GFX_OBJ_A[tb+0] = (TMP[ta] >> 0) & 0x0F;
   }

   FreeMem(TMP);

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x0800000/0x40);
   GFX_BG1_SOLID = make_solid_mask_8x8(GFX_BG1, 0x0800000/0x40);
   GFX_BG2_SOLID = NULL;

   memcpy(GFX_OBJ_B,GFX_OBJ_A,0x1000000);
   memset(GFX_OBJ_SIZE,0x00,0x1000000/0x100);
   GFX_OBJ_SOLID = make_solid_mask_16x16(GFX_OBJ_B, 0x1000000/0x100);

   gfx_colours = 16;

   RAMSize=0x80000;

   if(!(RAM=AllocateMem(RAMSize))) return;
   if(!(ROM=AllocateMem(0x100000))) return;

   if(!load_rom("u34.bin", RAM, 0x80000)) return;		// 68000 ROM
   for(ta=0;ta<0x80000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   if(!load_rom("u33.bin", RAM, 0x80000)) return;		// 68000 ROM
   for(ta=0;ta<0x80000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x80000);

   RAM_BG0 = RAM + 0x020000;
   RAM_BG1 = RAM + 0x028000;
   RAM_BG2 = NULL;

   RAM_BG0_CTRL = RAM + 0x048080;
   RAM_BG1_CTRL = RAM + 0x048088;
   RAM_BG2_CTRL = NULL;

   RAM_CTRL = RAM + 0x48100;

/*
   WriteWord68k(&ROM[0x02D4A],0x4E71);
*/

   InitPaletteMap(RAM+0x38000, 0x80, 0x10, 0x8000);

   set_colour_mapper(&col_map_xggg_ggrr_rrrb_bbbb);

   EEPROM = RAM + 0x48000;

   eeprom_data = (UWORD *) EEPROM;

   add_eeprom(EEPROM, 0x80, EPR_INVALIDATE_ON_ROM_CHANGE);
   load_eeprom();

   layer_id_data[0] = add_layer_info(layer_id_name[0]);
   layer_id_data[1] = add_layer_info(layer_id_name[1]);
   layer_id_data[2] = 0;
   layer_id_data[3] = add_layer_info(layer_id_name[3]);

/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x100000);
   ByteSwap(RAM,0x048000);

   AddMemFetch(0x000000, 0x0FFFFF, ROM+0x000000-0x000000);		// 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x0FFFFF, NULL, ROM+0x000000);			// 68000 ROM
   AddReadByte(0x100000, 0x10FFFF, NULL, RAM+0x000000);			// 68000 RAM
   AddReadByte(0x400000, 0x40FFFF, NULL, RAM+0x010000);			// OBJECT RAM
   AddReadByte(0x500000, 0x507FFF, NULL, RAM+0x020000);			// BG0 RAM
   AddReadByte(0x600000, 0x607FFF, NULL, RAM+0x028000);			// BG1 RAM
   AddReadByte(0x800000, 0x80007F, esp_800000_rb, NULL);		// ?
   AddReadByte(0x708000, 0x708FFF, NULL, RAM+0x038000);			// COLOR RAM
   AddReadByte(0xB00000, 0xB00003, esp_input_rb, NULL);			// INPUT
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);		// <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x0FFFFF, NULL, ROM+0x000000);			// 68000 ROM
   AddReadWord(0x100000, 0x10FFFF, NULL, RAM+0x000000);			// 68000 RAM
   AddReadWord(0x400000, 0x40FFFF, NULL, RAM+0x010000);			// OBJECT RAM
   AddReadWord(0x500000, 0x507FFF, NULL, RAM+0x020000);			// BG0 RAM
   AddReadWord(0x600000, 0x607FFF, NULL, RAM+0x028000);			// BG1 RAM
   AddReadWord(0x800000, 0x80007F, esp_800000_rw, NULL);		// ?
   AddReadWord(0x708000, 0x708FFF, NULL, RAM+0x038000);			// COLOR RAM
   AddReadWord(0xB00000, 0xB00003, esp_input_rw, NULL);			// INPUT
   AddReadWord(0x710000, 0x710FFF, NULL, RAM+0x038000);			// COLOR RAM [MIRROR]
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);		// <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x100000, 0x10FFFF, NULL, RAM+0x000000);		// 68000 RAM
   AddWriteByte(0x400000, 0x40FFFF, NULL, RAM+0x010000);		// OBJECT RAM
   AddWriteByte(0x500000, 0x507FFF, NULL, RAM+0x020000);		// BG0 RAM
   AddWriteByte(0x600000, 0x607FFF, NULL, RAM+0x028000);		// BG1 RAM
   AddWriteByte(0x800000, 0x80007F, esp_800000_wb, NULL);		// ?
   AddWriteByte(0x708000, 0x708FFF, NULL, RAM+0x038000);		// COLOR RAM
   AddWriteByte(0xB00000, 0xB00003, esp_input_wb, NULL);		// INPUT
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);			// Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);		// <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x100000, 0x10FFFF, NULL, RAM+0x000000);		// 68000 RAM
   AddWriteWord(0x400000, 0x40FFFF, NULL, RAM+0x010000);		// OBJECT RAM
   AddWriteWord(0x500000, 0x507FFF, NULL, RAM+0x020000);		// BG0 RAM
   AddWriteWord(0x600000, 0x607FFF, NULL, RAM+0x028000);		// BG1 RAM
   AddWriteWord(0x708000, 0x708FFF, NULL, RAM+0x038000);		// COLOR RAM
   AddWriteWord(0xB00000, 0xB00003, esp_input_ww, NULL);		// INPUT
   AddWriteWord(0xC00000, 0xC00001, esp_eeprom_ww, NULL);		// EEPROM
   AddWriteWord(0x800000, 0x80007F, esp_800000_ww, NULL);		// ?
   AddWriteWord(0x900000, 0x900007, NULL, RAM+0x048080);		// BG0 CTRL RAM
   AddWriteWord(0xA00000, 0xA00007, NULL, RAM+0x048088);		// BG1 CTRL RAM
   AddWriteWord(0x710000, 0x710FFF, NULL, RAM+0x038000);		// COLOR RAM [MIRROR]
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);		// <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();	// Set Starscream mem pointers... 
}

void load_uo_poko(void)
{
   int ta,tb;
   UBYTE *TMP;

   if(!(GFX_OBJ_A   =AllocateMem(0x1000000))) return;
   if(!(GFX_OBJ_B   =AllocateMem(0x1000000))) return;
   if(!(GFX_OBJ_SIZE=AllocateMem(0x1000000/0x100))) return;

   if(!(GFX_BG0  =AllocateMem(0x0800000))) return;
   if(!(TMP      =AllocateMem(0x0400000))) return;

   GFX_BG1 = NULL;
   GFX_BG2 = NULL;

   if(!load_rom("u49.bin", TMP, 0x400000)) return;		// GFX BG0
   tb=0;
   for(ta = 0; ta < 0x400000; ta += 8, tb += 8){
      GFX_BG0[tb+0]  = (TMP[ta+0] >> 4) & 0x0F;
      GFX_BG0[tb+1]  = (TMP[ta+0] >> 0) & 0x0F;
      GFX_BG0[tb+2]  = (TMP[ta+2] >> 4) & 0x0F;
      GFX_BG0[tb+3]  = (TMP[ta+2] >> 0) & 0x0F;
      GFX_BG0[tb+4]  = (TMP[ta+4] >> 4) & 0x0F;
      GFX_BG0[tb+5]  = (TMP[ta+4] >> 0) & 0x0F;
      GFX_BG0[tb+6]  = (TMP[ta+6] >> 4) & 0x0F;
      GFX_BG0[tb+7]  = (TMP[ta+6] >> 0) & 0x0F;
      GFX_BG0[tb+0] |= (TMP[ta+1] << 0) & 0xF0;
      GFX_BG0[tb+1] |= (TMP[ta+1] << 4) & 0xF0;
      GFX_BG0[tb+2] |= (TMP[ta+3] << 0) & 0xF0;
      GFX_BG0[tb+3] |= (TMP[ta+3] << 4) & 0xF0;
      GFX_BG0[tb+4] |= (TMP[ta+5] << 0) & 0xF0;
      GFX_BG0[tb+5] |= (TMP[ta+5] << 4) & 0xF0;
      GFX_BG0[tb+6] |= (TMP[ta+7] << 0) & 0xF0;
      GFX_BG0[tb+7] |= (TMP[ta+7] << 4) & 0xF0;
   }

   if(!load_rom("u33.bin", TMP, 0x400000)) return;		// OBJECT
   tb=0;
   for(ta = 0; ta < 0x400000; ta ++, tb += 2){
      GFX_OBJ_A[tb+1]  = (TMP[ta+0] >> 4) & 0x0F;
      GFX_OBJ_A[tb+0]  = (TMP[ta+0] >> 0) & 0x0F;
   }

   FreeMem(TMP);

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x0800000/0x40);
   GFX_BG1_SOLID = NULL;
   GFX_BG2_SOLID = NULL;

   memcpy(GFX_OBJ_B,GFX_OBJ_A,0x1000000);
   memset(GFX_OBJ_SIZE,0x00,0x1000000/0x100);
   GFX_OBJ_SOLID = make_solid_mask_16x16(GFX_OBJ_B, 0x1000000/0x100);

   gfx_colours = 256;

   RAMSize=0x80000;

   if(!(RAM=AllocateMem(RAMSize))) return;
   if(!(ROM=AllocateMem(0x100000))) return;

   if(!load_rom("u26j.bin", RAM, 0x80000)) return;		// 68000 ROM
   for(ta=0;ta<0x80000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   if(!load_rom("u25j.bin", RAM, 0x80000)) return;		// 68000 ROM
   for(ta=0;ta<0x80000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x80000);

   RAM_BG0 = RAM + 0x020000;
   RAM_BG1 = NULL;
   RAM_BG2 = NULL;

   RAM_BG0_CTRL = RAM + 0x048080;
   RAM_BG1_CTRL = NULL;
   RAM_BG2_CTRL = NULL;

   RAM_CTRL = RAM + 0x48100;

/*
   WriteWord68k(&ROM[0x02D4A],0x4E71);
*/
   InitPaletteMap(RAM+0x38000, 0x80, 0x100, 0x8000);

   set_colour_mapper(&col_map_xggg_ggrr_rrrb_bbbb);

   EEPROM = RAM + 0x48000;

   eeprom_data = (UWORD *) EEPROM;

   add_eeprom(EEPROM, 0x80, EPR_INVALIDATE_ON_ROM_CHANGE);
   load_eeprom();

   layer_id_data[0] = add_layer_info(layer_id_name[0]);
   layer_id_data[1] = 0;
   layer_id_data[2] = 0;
   layer_id_data[3] = add_layer_info(layer_id_name[3]);

/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x100000);
   ByteSwap(RAM,0x048000);

   AddMemFetch(0x000000, 0x0FFFFF, ROM+0x000000-0x000000);		// 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x0FFFFF, NULL, ROM+0x000000);			// 68000 ROM
   AddReadByte(0x100000, 0x10FFFF, NULL, RAM+0x000000);			// 68000 RAM
   AddReadByte(0x400000, 0x40FFFF, NULL, RAM+0x010000);			// OBJECT RAM
   AddReadByte(0x500000, 0x507FFF, NULL, RAM+0x020000);			// BG0 RAM
   AddReadByte(0x600000, 0x60007F, esp_800000_rb, NULL);		// ?
   AddReadByte(0x800000, 0x80FFFF, NULL, RAM+0x038000);			// COLOR RAM
   AddReadByte(0x900000, 0x900003, esp_input_rb, NULL);			// INPUT
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);		// <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x0FFFFF, NULL, ROM+0x000000);			// 68000 ROM
   AddReadWord(0x100000, 0x10FFFF, NULL, RAM+0x000000);			// 68000 RAM
   AddReadWord(0x400000, 0x40FFFF, NULL, RAM+0x010000);			// OBJECT RAM
   AddReadWord(0x500000, 0x507FFF, NULL, RAM+0x020000);			// BG0 RAM
   AddReadWord(0x600000, 0x60007F, esp_800000_rw, NULL);		// ?
   AddReadWord(0x800000, 0x80FFFF, NULL, RAM+0x038000);			// COLOR RAM
   AddReadWord(0x900000, 0x900003, esp_input_rw, NULL);			// INPUT
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);		// <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x100000, 0x10FFFF, NULL, RAM+0x000000);		// 68000 RAM
   AddWriteByte(0x400000, 0x40FFFF, NULL, RAM+0x010000);		// OBJECT RAM
   AddWriteByte(0x500000, 0x507FFF, NULL, RAM+0x020000);		// BG0 RAM
   AddWriteByte(0x600000, 0x60007F, esp_800000_wb, NULL);		// ?
   AddWriteByte(0x800000, 0x80FFFF, NULL, RAM+0x038000);		// COLOR RAM
   AddWriteByte(0x900000, 0x900003, esp_input_wb, NULL);		// INPUT
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);			// Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);		// <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x100000, 0x10FFFF, NULL, RAM+0x000000);		// 68000 RAM
   AddWriteWord(0x400000, 0x40FFFF, NULL, RAM+0x010000);		// OBJECT RAM
   AddWriteWord(0x500000, 0x507FFF, NULL, RAM+0x020000);		// BG0 RAM
   AddWriteWord(0x600000, 0x60007F, esp_800000_ww, NULL);		// ?
   AddWriteWord(0x800000, 0x80FFFF, NULL, RAM+0x038000);		// COLOR RAM
   AddWriteWord(0x900000, 0x900003, esp_input_ww, NULL);		// INPUT
   AddWriteWord(0xA00000, 0xA00001, esp_eeprom_ww, NULL);		// EEPROM
   AddWriteWord(0x700000, 0x700007, NULL, RAM+0x048080);		// BG0 CTRL RAM
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);		// <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();	// Set Starscream mem pointers... 
}

void clear_cave_68k(void)
{
   save_eeprom();

   #ifdef HACKS
      save_debug("RAM.bin",RAM,0x080000,1);
   #endif
}

void debug_cave_vcu(void)
{
   #ifdef HACKS

   clear_ingame_message_list();

   if(RAM_BG0_CTRL)

      print_ingame(60,"BG0: (%04x,%04x) %04x", ReadWord(RAM_BG0_CTRL+0), ReadWord(RAM_BG0_CTRL+2), ReadWord(RAM_BG0_CTRL+4));

   if(RAM_BG1_CTRL)

      print_ingame(60,"BG1: (%04x,%04x) %04x", ReadWord(RAM_BG1_CTRL+0), ReadWord(RAM_BG1_CTRL+2), ReadWord(RAM_BG1_CTRL+4));

   if(RAM_BG2_CTRL)

      print_ingame(60,"BG2: (%04x,%04x) %04x", ReadWord(RAM_BG2_CTRL+0), ReadWord(RAM_BG2_CTRL+2), ReadWord(RAM_BG2_CTRL+4));


   print_ingame(60,"CTRL: %04x %04x %04x %04x %04x %04x %04x %04x", ReadWord(RAM_CTRL+0x00), ReadWord(RAM_CTRL+0x02), ReadWord(RAM_CTRL+0x04), ReadWord(RAM_CTRL+0x06), ReadWord(RAM_CTRL+0x08), ReadWord(RAM_CTRL+0x0A), ReadWord(RAM_CTRL+0x0C), ReadWord(RAM_CTRL+0x0E));

   #endif
}

void execute_cave_68k_frame(void)
{
   cpu_execute_cycles(CPU_68K_0, CPU_FRAME_MHz(28,60));	// M68000 28MHz (60fps)

   #ifdef HACKS
      print_debug("PC:%06x SR:%04x\n",s68000context.pc,s68000context.sr);
   #endif

   cpu_interrupt(CPU_68K_0, 1);

   debug_cave_vcu();
}

static void make_tile_new(ULONG num, ULONG width)
{
   if(GFX_OBJ_SIZE[num] != width){

   ULONG x,y;
   UBYTE *src,*dst;

   src = GFX_OBJ_A + (num << 8);
   dst = GFX_OBJ_B + (num << 8);

   for(y = 0; y < (0x010 * 0x010); y += 0x010){
   for(x = 0; x < (width * 0x100); x += 0x100){

      dst[x + tile_16x16_map[y +  0]] = src[ 0];
      dst[x + tile_16x16_map[y +  1]] = src[ 1];
      dst[x + tile_16x16_map[y +  2]] = src[ 2];
      dst[x + tile_16x16_map[y +  3]] = src[ 3];
      dst[x + tile_16x16_map[y +  4]] = src[ 4];
      dst[x + tile_16x16_map[y +  5]] = src[ 5];
      dst[x + tile_16x16_map[y +  6]] = src[ 6];
      dst[x + tile_16x16_map[y +  7]] = src[ 7];
      dst[x + tile_16x16_map[y +  8]] = src[ 8];
      dst[x + tile_16x16_map[y +  9]] = src[ 9];
      dst[x + tile_16x16_map[y + 10]] = src[10];
      dst[x + tile_16x16_map[y + 11]] = src[11];
      dst[x + tile_16x16_map[y + 12]] = src[12];
      dst[x + tile_16x16_map[y + 13]] = src[13];
      dst[x + tile_16x16_map[y + 14]] = src[14];
      dst[x + tile_16x16_map[y + 15]] = src[15];

      src += 16;

   }
   }

   for(x = 0; x < width; x ++){
      GFX_OBJ_SIZE[num + x] = width;
      GFX_OBJ_SOLID[num + x] = check_tile_solid(GFX_OBJ_B + ((num + x) << 8), 0x100);
   }

   }

}


static layer_count;

void draw_cave_layer(int num, int pri)
{
   int x,y,ta;
   int z,zz;
   int x16,y16,zzz,zzzz;
   UBYTE *map;

   UBYTE *RAM_BG;
   UBYTE *SCR_BG;
   UBYTE *GFX_BG;
   UBYTE *MSK_BG;

   if(! check_layer_enabled(layer_id_data[num]))

      return;

   switch(num){
      case 0:
         RAM_BG = RAM_BG0;
         SCR_BG = RAM_BG0_CTRL;
         GFX_BG = GFX_BG0;
         MSK_BG = GFX_BG0_SOLID;
      break;
      case 1:
         RAM_BG = RAM_BG1;
         SCR_BG = RAM_BG1_CTRL;
         GFX_BG = GFX_BG1;
         MSK_BG = GFX_BG1_SOLID;
      break;
      case 2:
         RAM_BG = RAM_BG2;
         SCR_BG = RAM_BG2_CTRL;
         GFX_BG = GFX_BG2;
         MSK_BG = GFX_BG2_SOLID;
      break;
   }

   if(! RAM_BG)

      return;

   if(! pri){

     if(ReadWord(SCR_BG+2) & 0x200)

        return;

   }
   else{

     if(! (ReadWord(SCR_BG+2) & 0x200))

        return;

   }

   if(! layer_count){

   MAKE_SCROLL_512x512_4_16(
      ReadWord(SCR_BG+0) - 0x1D3 + ((ReadWord(SCR_BG+0)&0xE000)>>9),
      ReadWord(SCR_BG+2) - 0x03F + ((ReadWord(SCR_BG+2)&0xE000)>>9)
   );

   START_SCROLL_512x512_4_16(32,32,320,240);

      ta = (ReadWord(&RAM_BG[zz+2])&0x7FFF) << 2;

      MAP_PALETTE_MAPPED_NEW(
         ((ReadWord(&RAM_BG[zz])>>8) | 0x40) & 0x7F,
         gfx_colours,
         map
      );

      Draw8x8_Mapped_Rot(&GFX_BG[(ta+0)<<6],x+0,y+0,map);
      Draw8x8_Mapped_Rot(&GFX_BG[(ta+1)<<6],x+8,y+0,map);
      Draw8x8_Mapped_Rot(&GFX_BG[(ta+2)<<6],x+0,y+8,map);
      Draw8x8_Mapped_Rot(&GFX_BG[(ta+3)<<6],x+8,y+8,map);

   END_SCROLL_512x512_4_16();

   }
   else{

   MAKE_SCROLL_512x512_4_16(
      ReadWord(SCR_BG+0) - 0x1D3 + ((ReadWord(SCR_BG+0)&0xE000)>>9),
      ReadWord(SCR_BG+2) - 0x03F + ((ReadWord(SCR_BG+2)&0xE000)>>9)
   );

   START_SCROLL_512x512_4_16(32,32,320,240);

      ta = (ReadWord(&RAM_BG[zz+2])&0x7FFF) << 2;

      MAP_PALETTE_MAPPED_NEW(
         ((ReadWord(&RAM_BG[zz])>>8) | 0x40) & 0x7F,
         gfx_colours,
         map
      );

      Draw8x8_Trans_Mapped_Rot(&GFX_BG[(ta+0)<<6],x+0,y+0,map);
      Draw8x8_Trans_Mapped_Rot(&GFX_BG[(ta+1)<<6],x+8,y+0,map);
      Draw8x8_Trans_Mapped_Rot(&GFX_BG[(ta+2)<<6],x+0,y+8,map);
      Draw8x8_Trans_Mapped_Rot(&GFX_BG[(ta+3)<<6],x+8,y+8,map);

   END_SCROLL_512x512_4_16();

   }

   layer_count ++;
}

static UBYTE flip_map[4] =
{
   0, 2, 1, 3
};

void draw_cave_68k(void)
{
   int x,y,ta,xx;
   int z,w,h,ww,hh,flip;
   UBYTE *map, *RAM_OBJ;

   ClearPaletteMap();

   layer_count = 0;

   draw_cave_layer(0,0);
   draw_cave_layer(1,0);
   draw_cave_layer(2,0);

   draw_cave_layer(0,1);
   draw_cave_layer(1,1);
   draw_cave_layer(2,1);

   if(check_layer_enabled(layer_id_data[3])){

   if(! ReadWord(RAM_CTRL+8))

      RAM_OBJ = RAM + 0x10000;

   else

      RAM_OBJ = RAM + 0x14000;


   for(ta=0; ta < 0x4000; ta+=16){

      ww = (ReadWord(&RAM_OBJ[12+ta]) >> 8) & 0x0F;
      hh = (ReadWord(&RAM_OBJ[12+ta]) >> 0) & 0x0F;

      if((ww) && (hh)){

      x = ((ReadWord(&RAM_OBJ[0+ta]) >> 6) + 32) & 0x3FF;
      y = ((ReadWord(&RAM_OBJ[2+ta]) >> 6) + 32) & 0x3FF;

      z = ReadWord(&RAM_OBJ[6+ta]) & 0xFFFF;

      flip = flip_map[(ReadWord(&RAM_OBJ[4+ta])>>2) & 0x03];

      MAP_PALETTE_MAPPED_NEW(
         (ReadWord(&RAM_OBJ[4+ta])>>8) & 0x3F,
         gfx_colours,
         map
      );

      switch(flip){
      case 0:
      for(h = 0; h < hh; h++){

         make_tile_new(z, ww);

         xx = x;

         for(w = 0; w < ww; w++){

            if((xx > 16) && (xx < 320+32) && (y > 16) && (y < 240+32)){

               if(GFX_OBJ_SOLID[z+w]){

                  if(GFX_OBJ_SOLID[z+w]==1)
                     Draw16x16_Trans_Mapped_flip_Rot(&GFX_OBJ_B[(z+w)<<8], xx, y, map, flip);
                  else
                     Draw16x16_Mapped_flip_Rot(&GFX_OBJ_B[(z+w)<<8], xx, y, map, flip);

               }

            }

            xx = (xx + 16) & 0x3FF;

         }

         y = (y + 16) & 0x3FF;
         z = (z + ww) & 0xFFFF;

      }
      break;
      case 1:
      x = (x + ((ww - 1) * 16)) & 0x3FF;
      for(h = 0; h < hh; h++){

         make_tile_new(z, ww);

         xx = x;

         for(w = 0; w < ww; w++){

            if((xx > 16) && (xx < 320+32) && (y > 16) && (y < 240+32)){

               if(GFX_OBJ_SOLID[z+w]){

                  if(GFX_OBJ_SOLID[z+w]==1)
                     Draw16x16_Trans_Mapped_flip_Rot(&GFX_OBJ_B[(z+w)<<8], xx, y, map, flip);
                  else
                     Draw16x16_Mapped_flip_Rot(&GFX_OBJ_B[(z+w)<<8], xx, y, map, flip);

               }

            }

            xx = (xx - 16) & 0x3FF;

         }

         y = (y + 16) & 0x3FF;
         z = (z + ww) & 0xFFFF;

      }
      break;
      case 2:
      y = (y + ((hh - 1) * 16)) & 0x3FF;
      for(h = 0; h < hh; h++){

         make_tile_new(z, ww);

         xx = x;

         for(w = 0; w < ww; w++){

            if((xx > 16) && (xx < 320+32) && (y > 16) && (y < 240+32)){

               if(GFX_OBJ_SOLID[z+w]){

                  if(GFX_OBJ_SOLID[z+w]==1)
                     Draw16x16_Trans_Mapped_flip_Rot(&GFX_OBJ_B[(z+w)<<8], xx, y, map, flip);
                  else
                     Draw16x16_Mapped_flip_Rot(&GFX_OBJ_B[(z+w)<<8], xx, y, map, flip);

               }

            }

            xx = (xx + 16) & 0x3FF;

         }

         y = (y - 16) & 0x3FF;
         z = (z + ww) & 0xFFFF;

      }
      break;
      case 3:
      x = (x + ((ww - 1) * 16)) & 0x3FF;
      y = (y + ((hh - 1) * 16)) & 0x3FF;
      for(h = 0; h < hh; h++){

         make_tile_new(z, ww);

         xx = x;

         for(w = 0; w < ww; w++){

            if((xx > 16) && (xx < 320+32) && (y > 16) && (y < 240+32)){

               if(GFX_OBJ_SOLID[z+w]){

                  if(GFX_OBJ_SOLID[z+w]==1)
                     Draw16x16_Trans_Mapped_flip_Rot(&GFX_OBJ_B[(z+w)<<8], xx, y, map, flip);
                  else
                     Draw16x16_Mapped_flip_Rot(&GFX_OBJ_B[(z+w)<<8], xx, y, map, flip);

               }

            }

            xx = (xx - 16) & 0x3FF;

         }

         y = (y - 16) & 0x3FF;
         z = (z + ww) & 0xFFFF;

      }
      break;
      }

      }
   }

   }

}

/*

bg ram 0x0000-0x0FFF:

byte |     bit(s)     | use
-----+FEDCBA9876543210+-----------
 0-1 |..xxxxxx........| colour
 2-3 |.xxxxxxxxxxxxxxx| tile

bg ram 0x1000-0x1FFF:

byte |     bit(s)     | use
-----+FEDCBA9876543210+-----------
 0-1 |..........xxxxxx| ?
 2-3 |................|

bg ctrl ram:

byte |     bit(s)     | use
-----+FEDCBA9876543210+-----------
 0-1 |xxx.............| x offset
 0-1 |.......xxxxxxxxx| x pos
 2-3 |xxx.............| y offset
 2-3 |......x.........| priority ?
 2-3 |.......xxxxxxxxx| y pos

object ram:

byte |     bit(s)     | use
-----+FEDCBA9876543210+-----------
 0-1 |xxxxxxxxxx......| x pos
 2-3 |xxxxxxxxxx......| y pos
 4-5 |..xxxxxx........| colour
 4-5 |............x...| flip x axis
 4-5 |.............x..| flip y axis
 6-7 |xxxxxxxxxxxxxxxx| tile
 8-9 |................|
 A-B |................|
 C-D |....xxxx........| x chain
 C-D |............xxxx| y chain
 E-F |................|

ctrl ram:

byte |     bit(s)     | use
-----+FEDCBA9876543210+-----------
08-09|...............x| object double buffer

*/









