/******************************************************************************/
/*                                                                            */
/*              CLEOPATRA'S FORTUNE (C) 1996 TAITO CORPORATION                */
/*                                                                            */
/******************************************************************************/

#include "gameinc.h"
#include "cleofort.h"
#include "f3system.h"
#include "tc003vcu.h"
#include "tc200obj.h"

static struct DIR_INFO cleopatras_fortune_dirs[] =
{
   { "cleopatras_fortune", },
   { "cleofort", },
   { NULL, },
};

static struct ROM_INFO cleopatras_fortune_roms[] =
{
   {   "e28-01.bin", 0x00080000, 0x4440E659, },
   {   "e28-02.bin", 0x00080000, 0xB20D47CB, },
   {   "e28-03.bin", 0x00200000, 0x15C7989D, },
   {   "e28-04.bin", 0x00100000, 0x57AEF029, },
   {   "e28-05.bin", 0x00100000, 0x2870DBBC, },
   {   "e28-06.bin", 0x00100000, 0x21D0C454, },
   {   "e28-07.bin", 0x00080000, 0xA507797B, },
   {   "e28-08.bin", 0x00080000, 0x7564F199, },
   {   "e28-09.bin", 0x00080000, 0x1C48A1F9, },
   {   "e28-10.bin", 0x00080000, 0x013FBC39, },
   {   "e28-11.bin", 0x00020000, 0x01A06950, },
   {   "e28-12.bin", 0x00020000, 0xDC19260F, },
   {           NULL,          0,          0, },
};

static struct ROMSW_DATA romsw_data_cleopatras_fortune_0[] =
{
   { "Taito Japan",           0x01 },
   { "Taito America",         0x02 },
   { "Taito Worldwide",       0x03 },
   { NULL,                    0    },
};

static struct ROMSW_INFO cleopatras_fortune_romsw[] =
{
   { 0x07FFFF, 0x01, romsw_data_cleopatras_fortune_0 },
   { 0,        0,    NULL },
};

static struct VIDEO_INFO cleopatras_fortune_video =
{
   DrawCleoFortune,
   320,
   224,
   64,
   VIDEO_ROTATE_NORMAL,
};

struct GAME_MAIN game_cleopatras_fortune =
{
   cleopatras_fortune_dirs,
   cleopatras_fortune_roms,
   f3_system_inputs,
   NULL,
   cleopatras_fortune_romsw,

   LoadCleoFortune,
   ClearCleoFortune,
   &cleopatras_fortune_video,
   ExecuteCleoFortuneFrame,
   "cleofort",
   "Cleopatra's Fortune",
   "NIpgEtH[`",
   COMPANY_ID_TAITO,
   "E28",
   1996,
   NULL,
};

static UBYTE *RAM_BG0;
static UBYTE *RAM_BG1;
static UBYTE *RAM_BG2;
static UBYTE *RAM_BG3;

static UBYTE *RAM_SCR0;
static UBYTE *RAM_SCR1;
static UBYTE *RAM_SCR2;
static UBYTE *RAM_SCR3;

static ULONG SCR0_XOFS;
static ULONG SCR1_XOFS;
static ULONG SCR2_XOFS;
static ULONG SCR3_XOFS;

static ULONG SCR0_YOFS;
static ULONG SCR1_YOFS;
static ULONG SCR2_YOFS;
static ULONG SCR3_YOFS;

static UBYTE *GFX_BG0;
static UBYTE *GFX_BG0_SOLID;

static UBYTE *GFX_SPR;
static UBYTE *GFX_SPR_SOLID;

void LoadCleoFortune(void)
{
   int ta,tb,tc;

   RAMSize=0x80000;

   if(!(GFX=AllocateMem(0x2DBA00+0x0FC000))) return;
   if(!(ROM=AllocateMem(0xB6E80))) return;

   GFX_BG0 = GFX+0x000000;
   GFX_SPR = GFX+0x2DBA00;

   tb=0;
   if(!load_rom("e28-02.bin", ROM, 0x3F000)) return;	// 16x16 SPRITES ($0FC0)
   for(ta=0;ta<0x3F000;ta++){
      GFX_SPR[tb++]=ROM[ta]&15;
      GFX_SPR[tb++]=ROM[ta]>>4;
      tb+=2;
   }
   tb=2;
   if(!load_rom("e28-01.bin", ROM, 0x3F000)) return;	// 16x16 SPRITES
   for(ta=0;ta<0x3F000;ta++){
      GFX_SPR[tb++]=ROM[ta]&15;
      GFX_SPR[tb++]=ROM[ta]>>4;
      tb+=2;
   }

   memset(GFX_SPR+0x25200,0x00,0x100);			// Transparent Green Layer

   tb=0;
   if(!load_rom("e28-06.bin", ROM, 0xB6E80)) return;	// 16x16 TILES ($2DBA)
   for(ta=0;ta<0xB6E80;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=4;
   if(!load_rom("e28-05.bin", ROM, 0xB6E80)) return;	// 16x16 TILES
   for(ta=0;ta<0xB6E80;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=0;
   if(!load_rom("e28-04.bin", ROM, 0xB6E80)) return;	// 16x16 TILES (MASK)
   for(ta=0;ta<0xB6E80;ta+=2){
      tc=ROM[ta];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<4;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<4;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<4;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<4;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<4;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<4;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<4;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<4;
      tc=ROM[ta+1];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<5;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<5;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<5;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<5;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<5;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<5;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<5;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<5;
      tb+=8;
   }

   FreeMem(ROM);
   if(!(ROM=AllocateMem(0x80000))) return;
   if(!(RAM=AllocateMem(0x80000))) return;

   GFX_BG0_SOLID = MakeSolidTileMap16x16(GFX_BG0, 0x2DBA);
   GFX_SPR_SOLID = make_solid_mask_16x16(GFX_SPR, 0x0FC0);

   // Setup 68020 Memory Map
   // ----------------------

   AddF3MemoryMap(0x80000);

   RAM_BG0=RAM+0x30000;
   RAM_BG1=RAM+0x31000;
   RAM_BG2=RAM+0x32000;
   RAM_BG3=RAM+0x33000;

   RAM_SCR0=RAM+0x6A000;
   RAM_SCR1=RAM+0x6A002;
   RAM_SCR2=RAM+0x6A004;
   RAM_SCR3=RAM+0x6A006;

   SCR0_XOFS=0xF600;
   SCR1_XOFS=0xF700;
   SCR2_XOFS=0xF800;
   SCR3_XOFS=0xF900;

   SCR0_YOFS=0x0780-(16<<7);
   SCR1_YOFS=0x0780-(16<<7);
   SCR2_YOFS=0x0780-(16<<7);
   SCR3_YOFS=0x0780-(16<<7);

   if(!load_rom("e28-10.bin", RAM, 0x20000)) return;
   for(ta=0;ta<0x20000;ta++){
      ROM[(ta<<2)+0]=RAM[ta];
   }
   if(!load_rom("e28-09.bin", RAM, 0x20000)) return;
   for(ta=0;ta<0x20000;ta++){
      ROM[(ta<<2)+1]=RAM[ta];
   }
   if(!load_rom("e28-08.bin", RAM, 0x20000)) return;
   for(ta=0;ta<0x20000;ta++){
      ROM[(ta<<2)+2]=RAM[ta];
   }
   if(!load_rom("e28-07.bin", RAM, 0x20000)) return;
   for(ta=0;ta<0x20000;ta++){
      ROM[(ta<<2)+3]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x80000);
   memset(RAM+0x69000,0xFF,0x01000);

   // FIX ROM ERROR
   // -------------

   WriteWord68k(&ROM[0x000D7E],0x4E71);		//	nop

   // EEPROM HACKS
   // ------------

   WriteWord68k(&ROM[0x00425E],0x7F00);		//	raine	#$00 <read eeprom>
   WriteWord68k(&ROM[0x004260],0x4E75);		//	rts

   WriteWord68k(&ROM[0x0042DC],0x7F01);		//	raine	#$01 <write eeprom>
   WriteWord68k(&ROM[0x0042DE],0x4E75);		//	rts

   // SPEED HACK#1
   // ------------

   WriteLong68k(&ROM[0x0053A],0x4E714E71);

   WriteWord68k(&ROM[0x0025A],0x7F02);		//	raine	#$02 <stop cpu>
   WriteWord68k(&ROM[0x0025C],0x60EC+6);

   F3SystemEEPROMAccess=&F3SysEEPROMAccessMode1;

   InitPaletteMap(RAM+0x60000, 0x200, 0x40, 0x1000);

   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu.RAM	= RAM+0x20000;
   tc0003vcu.mapper	= &Map_21bit_xRGB;
   tc0003vcu.bmp_x	= 64;
   tc0003vcu.bmp_y	= 64;
   tc0003vcu.bmp_w	= 320;
   tc0003vcu.bmp_h	= 224;
   tc0003vcu.scr_x	= 0;
   tc0003vcu.scr_y	= 16;

   // Init tc0200obj emulation
   // ------------------------

   tc0200obj.RAM	= RAM+0x20000;
   tc0200obj.RAM_B	= RAM+0x28000;
   tc0200obj.GFX	= GFX_SPR;
   tc0200obj.MASK	= GFX_SPR_SOLID;
   tc0200obj.bmp_x	= 64;
   tc0200obj.bmp_y	= 64;
   tc0200obj.bmp_w	= 320;
   tc0200obj.bmp_h	= 224;
   tc0200obj.mapper	= &Map_21bit_xRGB;
   tc0200obj.tile_mask	= 0x0FFF;
   tc0200obj.ofs_x	= 0;	//-0x2E;
   tc0200obj.ofs_y	= -16;	//-0x18;

   tc0200obj.cols	= 16;

   init_tc0200obj();

   init_f3_system_ioc(F3_IOC_2P_3BUTTON);

   init_m68k();
}

void ClearCleoFortune(void)
{
   save_eeprom();

   #ifdef HACKS
      save_debug("ROM.bin",ROM,0x080000,0);
      save_debug("RAM.bin",RAM,0x080000,0);
      //save_debug("GFX.bin",GFX,0x64AB00+0x3C2300,0);
   #endif
}

void ExecuteCleoFortuneFrame(void)
{
   Execute68020(300000);
   /*#ifdef HACKS
      print_debug("PC:%06x SR:%04x\n",regs.pc,regs.sr);
   #endif*/
   //Interrupt68020(3);
   Interrupt68020(2);

   IntF3System();
}

void DrawCleoFortune(void)
{
   int x16,y16,zz,zzz,zzzz;
   int ta,x,y;
   UBYTE *MAP;

   ClearPaletteMap();

   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu_layer_count = 0;

   if(check_layer_enabled(f3_bg0_id)){
   MAKE_SCROLL_512x512_4_16(
      (0-320)-((ReadWord68k(&RAM_SCR0[0])-SCR0_XOFS)>>6),
      (0-224)-((ReadWord68k(&RAM_SCR0[8])-SCR0_YOFS)>>7)
   );

   START_SCROLL_512x512_4_16_R180(64,64,320,224);

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG0[zz])&0x1FF,
               64
            );

      (*Draw16x16_Mapped_JumpListR[(RAM_BG0[zz]&0xC0)>>6])(&GFX_BG0[ReadWord68k(&RAM_BG0[zz+2])<<8],x,y,MAP);

   END_SCROLL_512x512_4_16();
   }


   if(check_layer_enabled(f3_bg1_id)){
   MAKE_SCROLL_512x512_4_16(
      (0-320)-((ReadWord68k(&RAM_SCR1[0])-SCR1_XOFS)>>6),
      (0-224)-((ReadWord68k(&RAM_SCR1[8])-SCR1_YOFS)>>7)
   );

   START_SCROLL_512x512_4_16_R180(64,64,320,224);

   ta=ReadWord68k(&RAM_BG1[zz+2]);
   if(ta!=0){

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG1[zz])&0x1FF,
               64
            );

      if(GFX_BG0_SOLID[ta]==0){
         (*Draw16x16_Trans_Mapped_JumpListR[(RAM_BG1[zz]&0xC0)>>6])(&GFX_BG0[ta<<8],x,y,MAP);
      }
      else{
         (*Draw16x16_Mapped_JumpListR[(RAM_BG1[zz]&0xC0)>>6])(&GFX_BG0[ta<<8],x,y,MAP);
      }
   }

   END_SCROLL_512x512_4_16();
   }

   render_tc0200obj_mapped_f3system_r180();

   if(check_layer_enabled(f3_bg2_id)){
   MAKE_SCROLL_512x512_4_16(
      (0-320)-((ReadWord68k(&RAM_SCR2[0])-SCR2_XOFS)>>6),
      (0-224)-((ReadWord68k(&RAM_SCR2[8])-SCR2_YOFS)>>7)
   );

   START_SCROLL_512x512_4_16_R180(64,64,320,224);

   ta=ReadWord68k(&RAM_BG2[zz+2]);
   if(ta!=0){

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG2[zz])&0x1FF,
               64
            );

      if(GFX_BG0_SOLID[ta]==0){
         (*Draw16x16_Trans_Mapped_JumpListR[(RAM_BG2[zz]&0xC0)>>6])(&GFX_BG0[ta<<8],x,y,MAP);
      }
      else{
         (*Draw16x16_Mapped_JumpListR[(RAM_BG2[zz]&0xC0)>>6])(&GFX_BG0[ta<<8],x,y,MAP);
      }
   }

   END_SCROLL_512x512_4_16();
   }

   if(check_layer_enabled(f3_bg3_id)){
   MAKE_SCROLL_512x512_4_16(
      (0-320)-((ReadWord68k(&RAM_SCR3[0])-SCR3_XOFS)>>6),
      (0-224)-((ReadWord68k(&RAM_SCR3[8])-SCR3_YOFS)>>7)
   );

   START_SCROLL_512x512_4_16_R180(64,64,320,224);

   ta=ReadWord68k(&RAM_BG3[zz+2]);
   if(ta!=0){

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG3[zz])&0x1FF,
               64
            );

      if(GFX_BG0_SOLID[ta]==0){
         (*Draw16x16_Trans_Mapped_JumpListR[(RAM_BG3[zz]&0xC0)>>6])(&GFX_BG0[ta<<8],x,y,MAP);
      }
      else{
         (*Draw16x16_Mapped_JumpListR[(RAM_BG3[zz]&0xC0)>>6])(&GFX_BG0[ta<<8],x,y,MAP);
      }
   }

   END_SCROLL_512x512_4_16();
   }

   if(check_layer_enabled(f3_bg5_id)){
      f3video_render_fg0_r180();
   }
}
