/******************************************************************************/
/*                                                                            */
/*                DYNAMITE LEAGUE (C) 1990 TAITO CORPORATION                  */
/*                                                                            */
/******************************************************************************/

#include "gameinc.h"
#include "dleague.h"
#include "tc004vcu.h"
#include "tc140syt.h"
#include "tc220ioc.h"
#include "taitosnd.h"

static struct DIR_INFO dynamite_league_dirs[] =
{
   { "dynamite_league", },
   { "dleague", },
   { NULL, },
};

static struct ROM_INFO dynamite_league_roms[] =
{
   {  "dl_ic06.bin", 0x00010000, 0x20D3732C, },
   {  "dl_ic07.bin", 0x00010000, 0x70D2BDDF, },
   {  "dl_ic08.bin", 0x00010000, 0x1E99E3E5, },
   {  "dl_ic09.bin", 0x00010000, 0x9E06F787, },
   {  "dl_ic14.bin", 0x00010000, 0xC7AF1571, },
   {  "dl_ic15.bin", 0x00080000, 0xB273F854, },
   {  "dl_ic16.bin", 0x00010000, 0xF436DEDF, },
   {  "dl_ic17.bin", 0x00080000, 0xC3FD0DCD, },
   {  "dl_ic18.bin", 0x00010000, 0x7E9B7EAC, },
   {  "dl_ic19.bin", 0x00080000, 0x18EF740A, },
   {  "dl_ic20.bin", 0x00010000, 0x6AA2EFB1, },
   {  "dl_ic21.bin", 0x00080000, 0xFE3A5179, },
   {  "dl_ic33.bin", 0x00020000, 0x7E904E45, },
   {  "dl_ic34.bin", 0x00010000, 0xCDF593F3, },
   {  "dl_ic36.bin", 0x00020000, 0x18C8A32B, },
   {  "dl_ic37.bin", 0x00010000, 0xF50DB2D7, },
   {  "dl_ic40.bin", 0x00010000, 0x5632EE49, },
   {   "c02-01.rom", 0x00080000, 0xD5A3D1AA, },
   {           NULL,          0,          0, },
};

static struct INPUT_INFO dynamite_league_inputs[] =
{
   { KB_DEF_COIN1,        MSG_COIN1,               0x032004, 0x04, BIT_ACTIVE_0 },
   { KB_DEF_COIN2,        MSG_COIN2,               0x032004, 0x08, BIT_ACTIVE_0 },
   { KB_DEF_TILT,         MSG_TILT,                0x032004, 0x02, BIT_ACTIVE_0 },
   { KB_DEF_SERVICE,      MSG_SERVICE,             0x032004, 0x01, BIT_ACTIVE_0 },

   { KB_DEF_P1_START,     MSG_P1_START,            0x032004, 0x10, BIT_ACTIVE_0 },
   { KB_DEF_P1_UP,        MSG_P1_UP,               0x032006, 0x01, BIT_ACTIVE_0 },
   { KB_DEF_P1_DOWN,      MSG_P1_DOWN,             0x032006, 0x02, BIT_ACTIVE_0 },
   { KB_DEF_P1_LEFT,      MSG_P1_LEFT,             0x032006, 0x04, BIT_ACTIVE_0 },
   { KB_DEF_P1_RIGHT,     MSG_P1_RIGHT,            0x032006, 0x08, BIT_ACTIVE_0 },
   { KB_DEF_P1_B1,        MSG_P1_B1,               0x03200E, 0x01, BIT_ACTIVE_0 },
   { KB_DEF_P1_B2,        MSG_P1_B2,               0x03200E, 0x02, BIT_ACTIVE_0 },
   { KB_DEF_P1_B3,        MSG_P1_B3,               0x03200E, 0x04, BIT_ACTIVE_0 },

   { KB_DEF_P2_START,     MSG_P2_START,            0x032004, 0x20, BIT_ACTIVE_0 },
   { KB_DEF_P2_UP,        MSG_P2_UP,               0x032006, 0x10, BIT_ACTIVE_0 },
   { KB_DEF_P2_DOWN,      MSG_P2_DOWN,             0x032006, 0x20, BIT_ACTIVE_0 },
   { KB_DEF_P2_LEFT,      MSG_P2_LEFT,             0x032006, 0x40, BIT_ACTIVE_0 },
   { KB_DEF_P2_RIGHT,     MSG_P2_RIGHT,            0x032006, 0x80, BIT_ACTIVE_0 },
   { KB_DEF_P2_B1,        MSG_P2_B1,               0x03200E, 0x10, BIT_ACTIVE_0 },
   { KB_DEF_P2_B2,        MSG_P2_B2,               0x03200E, 0x20, BIT_ACTIVE_0 },
   { KB_DEF_P2_B3,        MSG_P2_B3,               0x03200E, 0x40, BIT_ACTIVE_0 },

   { 0,                   NULL,                    0,        0,    0            },
};

static struct DSW_DATA dsw_data_dynamite_league_0[] =
{
   { MSG_DSWA_BIT1,           0x01, 0x02 },
   { MSG_OFF,                 0x01, 0x00 },
   { MSG_ON,                  0x00, 0x00 },
   { MSG_DSWA_BIT2,           0x02, 0x02 },
   { MSG_OFF,                 0x02, 0x00 },
   { MSG_ON,                  0x00, 0x00 },
   { MSG_TEST_MODE,           0x04, 0x02 },
   { MSG_OFF,                 0x04, 0x00 },
   { MSG_ON,                  0x00, 0x00 },
   { MSG_DEMO_SOUND,          0x08, 0x02 },
   { MSG_ON,                  0x08, 0x00 },
   { MSG_OFF,                 0x00, 0x00 },
   { MSG_DSWA_BIT5,           0x10, 0x02 },
   { MSG_OFF,                 0x10, 0x00 },
   { MSG_ON,                  0x00, 0x00 },
   { MSG_DSWA_BIT6,           0x20, 0x02 },
   { MSG_OFF,                 0x20, 0x00 },
   { MSG_ON,                  0x00, 0x00 },
   { MSG_DSWA_BIT7,           0x40, 0x02 },
   { MSG_OFF,                 0x40, 0x00 },
   { MSG_ON,                  0x00, 0x00 },
   { MSG_DSWA_BIT8,           0x80, 0x02 },
   { MSG_OFF,                 0x80, 0x00 },
   { MSG_ON,                  0x00, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO dynamite_league_dsw[] =
{
   { 0x032000, 0xFF, dsw_data_dynamite_league_0 },
   { 0x032002, 0xFF, dsw_data_default_1 },
   { 0,        0,    NULL,      },
};

static struct ROMSW_DATA romsw_data_dynamite_league_0[] =
{
   { "Taito Japan (Japanese)", 0x00 },
   { "Taito America",          0x01 },
   { "Taito Japan",            0x02 },
   { NULL,                     0    },
};

static struct ROMSW_INFO dynamite_league_romsw[] =
{
   { 0x05FFFF, 0x00, romsw_data_dynamite_league_0 },
   { 0,        0,    NULL },
};

static struct VIDEO_INFO dynamite_league_video =
{
   DrawDLeague,
   320,
   240,
   64,
   VIDEO_ROTATE_NORMAL,
};

struct GAME_MAIN game_dynamite_league =
{
   dynamite_league_dirs,
   dynamite_league_roms,
   dynamite_league_inputs,
   dynamite_league_dsw,
   dynamite_league_romsw,

   LoadDLeague,
   ClearDLeague,
   &dynamite_league_video,
   ExecuteDLeagueFrame,
   "dleague",
   "Dynamite League",
   NULL,
   COMPANY_ID_TAITO,
   "C02",
   1990,
   taito_ym2610_sound,
};

static UBYTE *RAM_VIDEO;
static UBYTE *RAM_COLOUR;
static UBYTE *RAM_INPUT;

static UBYTE *GFX_BG0;
static UBYTE *GFX_BG0_SOLID;

static UBYTE *GFX_FG0;

void LoadDLeague(void)
{
   int ta,tb,tc;

   if(!(RAM=AllocateMem(0x80000))) return;
   if(!(ROM=AllocateMem(0x60000))) return;
   if(!(GFX=AllocateMem(0x400000+0x100000))) return;

   GFX_BG0 = GFX+0x000000;

   if(!load_rom("dl_ic06.bin", RAM+0x00000, 0x10000)) return;	// GFX
   tb=0x400000;
   for(ta=0;ta<0x10000;ta+=16){
      for(tc=0;tc<16;tc++){
         GFX[tb+(tc<<4)]=RAM[ta+tc]&15;
         GFX[tb+(tc<<4)+1]=RAM[ta+tc]>>4;
      }
      tb+=256;
   }
   if(!load_rom("dl_ic07.bin", RAM+0x00000, 0x10000)) return;	// GFX
   tb=0x400004;
   for(ta=0;ta<0x10000;ta+=16){
      for(tc=0;tc<16;tc++){
         GFX[tb+(tc<<4)]=RAM[ta+tc]&15;
         GFX[tb+(tc<<4)+1]=RAM[ta+tc]>>4;
      }
      tb+=256;
   }
   if(!load_rom("dl_ic08.bin", RAM+0x00000, 0x10000)) return;	// GFX
   tb=0x400008;
   for(ta=0;ta<0x10000;ta+=16){
      for(tc=0;tc<16;tc++){
         GFX[tb+(tc<<4)]=RAM[ta+tc]&15;
         GFX[tb+(tc<<4)+1]=RAM[ta+tc]>>4;
      }
      tb+=256;
   }
   if(!load_rom("dl_ic09.bin", RAM+0x00000, 0x10000)) return;	// GFX
   tb=0x40000C;
   for(ta=0;ta<0x10000;ta+=16){
      for(tc=0;tc<16;tc++){
         GFX[tb+(tc<<4)]=RAM[ta+tc]&15;
         GFX[tb+(tc<<4)+1]=RAM[ta+tc]>>4;
      }
      tb+=256;
   }
   if(!load_rom("dl_ic14.bin", RAM+0x00000, 0x10000)) return;	// GFX
   tb=0x400002;
   for(ta=0;ta<0x10000;ta+=16){
      for(tc=0;tc<16;tc++){
         GFX[tb+(tc<<4)]=RAM[ta+tc]&15;
         GFX[tb+(tc<<4)+1]=RAM[ta+tc]>>4;
      }
      tb+=256;
   }
   if(!load_rom("dl_ic16.bin", RAM+0x00000, 0x10000)) return;	// GFX
   tb=0x400006;
   for(ta=0;ta<0x10000;ta+=16){
      for(tc=0;tc<16;tc++){
         GFX[tb+(tc<<4)]=RAM[ta+tc]&15;
         GFX[tb+(tc<<4)+1]=RAM[ta+tc]>>4;
      }
      tb+=256;
   }
   if(!load_rom("dl_ic18.bin", RAM+0x00000, 0x10000)) return;	// GFX
   tb=0x40000A;
   for(ta=0;ta<0x10000;ta+=16){
      for(tc=0;tc<16;tc++){
         GFX[tb+(tc<<4)]=RAM[ta+tc]&15;
         GFX[tb+(tc<<4)+1]=RAM[ta+tc]>>4;
      }
      tb+=256;
   }
   if(!load_rom("dl_ic20.bin", RAM+0x00000, 0x10000)) return;	// GFX
   tb=0x40000E;
   for(ta=0;ta<0x10000;ta+=16){
      for(tc=0;tc<16;tc++){
         GFX[tb+(tc<<4)]=RAM[ta+tc]&15;
         GFX[tb+(tc<<4)+1]=RAM[ta+tc]>>4;
      }
      tb+=256;
   }
   if(!load_rom("dl_ic15.bin", RAM+0x00000, 0x80000)) return;	// GFX
   tb=0;
   for(ta=0;ta<0x80000;ta+=2){
      GFX[tb+0]=RAM[ta]&15;
      GFX[tb+1]=RAM[ta]>>4;
      GFX[tb+2]=RAM[ta+1]&15;
      GFX[tb+3]=RAM[ta+1]>>4;
      tb+=16;
   }
   if(!load_rom("dl_ic17.bin", RAM+0x00000, 0x80000)) return;	// GFX
   tb=4;
   for(ta=0;ta<0x80000;ta+=2){
      GFX[tb+0]=RAM[ta]&15;
      GFX[tb+1]=RAM[ta]>>4;
      GFX[tb+2]=RAM[ta+1]&15;
      GFX[tb+3]=RAM[ta+1]>>4;
      tb+=16;
   }
   if(!load_rom("dl_ic19.bin", RAM+0x00000, 0x80000)) return;	// GFX
   tb=8;
   for(ta=0;ta<0x80000;ta+=2){
      GFX[tb+0]=RAM[ta]&15;
      GFX[tb+1]=RAM[ta]>>4;
      GFX[tb+2]=RAM[ta+1]&15;
      GFX[tb+3]=RAM[ta+1]>>4;
      tb+=16;
   }
   if(!load_rom("dl_ic21.bin", RAM+0x00000, 0x80000)) return;	// GFX
   tb=12;
   for(ta=0;ta<0x80000;ta+=2){
      GFX[tb+0]=RAM[ta]&15;
      GFX[tb+1]=RAM[ta]>>4;
      GFX[tb+2]=RAM[ta+1]&15;
      GFX[tb+3]=RAM[ta+1]>>4;
      tb+=16;
   }

   if(!load_rom("dl_ic33.bin", RAM, 0x20000)) return;		// 68000 ROM
   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   if(!load_rom("dl_ic36.bin", RAM, 0x20000)) return;
   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }
   if(!load_rom("dl_ic34.bin", RAM, 0x10000)) return;
   for(ta=0;ta<0x10000;ta++){
      ROM[ta+ta+0x40000]=RAM[ta];
   }
   if(!load_rom("dl_ic37.bin", RAM, 0x10000)) return;
   for(ta=0;ta<0x10000;ta++){
      ROM[ta+ta+0x40001]=RAM[ta];
   }

   RAMSize=0x48000+0x10000;

   /*-----[Sound Setup]-----*/

   Z80ROM=RAM+0x48000;
   if(!load_rom("dl_ic40.bin", Z80ROM, 0x10000)) return;	// Z80 SOUND ROM

   if(!(PCMROM=AllocateMem(0x80000))) return;
   if(!load_rom("c02-01.rom",PCMROM,0x80000)) return;		// ADPCM A rom
   //memset(PCMROM,0x80,0x80000);
   YM2610SetBuffers(PCMROM, PCMROM, 0x80000, 0x80000);

   AddTaitoYM2610(0x033A, 0x02A7, 0x10000);

   /*-----------------------*/

   memset(RAM+0x00000,0x00,0x40000);

   GFX_FG0   = RAM+0x40000;
   RAM_VIDEO = RAM+0x10000;
   RAM_COLOUR= RAM+0x31000+0x800;
   RAM_INPUT = RAM+0x32000;

   GFX_BG0_SOLID = make_solid_mask_16x16(GFX_BG0, 0x5000);

   InitPaletteMap(RAM_COLOUR, 0x40, 0x10, 0x8000);

   // 68000 Speed Hack
   // ----------------

   WriteWord68k(&ROM[0x0464],0x4EF9);		//	jmp	$300
   WriteLong68k(&ROM[0x0466],0x00000300);

   WriteLong68k(&ROM[0x0300],0x46FC2000);	//	move	#$2000,SR
   WriteLong68k(&ROM[0x0304],0x13FC0000);	// 	move.b	#$00,$AA0000
   WriteLong68k(&ROM[0x0308],0x00AA0000);
   WriteWord68k(&ROM[0x030C],0x6100-10);	//	bra.s	<loop>

   // Init tc0004vcu emulation
   // ------------------------

   tc0004vcu.RAM	= RAM_VIDEO;
   tc0004vcu.GFX_BG0	= GFX_BG0;
   tc0004vcu.GFX_BG0_MSK= GFX_BG0_SOLID;
   tc0004vcu.GFX_FG0	= GFX_FG0;
   tc0004vcu.tile_mask	= 0x7FFF;
   tc0004vcu.mapper	= &Map_15bit_xBGR;
   tc0004vcu.bmp_x	= 64;
   tc0004vcu.bmp_y	= 64;
   tc0004vcu.bmp_w	= 320;
   tc0004vcu.bmp_h	= 240;
   tc0004vcu.scr_x	= 16-1;
   tc0004vcu.scr_y	= 32-1;

   tc0004vcu_init();

   // Init tc0220ioc emulation
   // ------------------------

   tc0220ioc.RAM  = RAM_INPUT;
   tc0220ioc.ctrl = 0;		//TC0220_STOPCPU;
   reset_tc0220ioc();

/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x60000);
   ByteSwap(RAM,0x40000);

   AddMemFetch(0x000000, 0x05FFFF, ROM+0x000000-0x000000);	// 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x05FFFF, NULL, ROM+0x000000);			// 68000 ROM
   AddReadByte(0x100000, 0x10FFFF, NULL, RAM+0x000000);			// 68000 RAM
   AddReadByte(0x400000, 0x420FFF, NULL, RAM+0x010000);			// SCREEN RAM
   AddReadByte(0x500000, 0x500FFF, NULL, RAM+0x031000);			// COLOR RAM
   AddReadByte(0x200000, 0x20000F, NULL, RAM_INPUT);			// INPUT
   AddReadByte(0x300000, 0x300003, tc0140syt_read_main_68k, NULL);	// SOUND COMM
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);		// <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x05FFFF, NULL, ROM+0x000000);			// 68000 ROM
   AddReadWord(0x100000, 0x10FFFF, NULL, RAM+0x000000);			// 68000 RAM
   AddReadWord(0x400000, 0x420FFF, NULL, RAM+0x010000);			// SCREEN RAM
   AddReadWord(0x500000, 0x500FFF, NULL, RAM+0x031000);			// COLOR RAM
   AddReadWord(0x200000, 0x20000F, NULL, RAM_INPUT);			// INPUT
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);		// <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x100000, 0x10FFFF, NULL, RAM+0x000000);		// 68000 RAM
   AddWriteByte(0x400000, 0x420FFF, NULL, RAM+0x010000);		// SCREEN RAM
   AddWriteByte(0x500000, 0x500FFF, NULL, RAM+0x031000);		// COLOR RAM
   AddWriteByte(0x200000, 0x20000F, tc0220ioc_wb, NULL);		// INPUT
   AddWriteByte(0x300000, 0x300003, tc0140syt_write_main_68k, NULL);	// SOUND COMM
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);			// Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);		// <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x100000, 0x10FFFF, NULL, RAM+0x000000);		// 68000 RAM
   AddWriteWord(0x400000, 0x420FFF, NULL, RAM+0x010000);		// SCREEN RAM
   AddWriteWord(0x500000, 0x500FFF, NULL, RAM+0x031000);		// COLOR RAM
   AddWriteWord(0x200000, 0x20000F, tc0220ioc_ww, NULL);		// INPUT
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);		// <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();	// Set Starscream mem pointers... 
}

void ClearDLeague(void)
{
   RemoveTaitoYM2610();

   #ifdef HACKS
      save_debug("ROM.bin",ROM,0x060000,1);
      save_debug("RAM.bin",RAM,0x040000,1);
      //save_debug("GFX.bin",GFX,0x500000,0);
   #endif
}

void ExecuteDLeagueFrame(void)
{
   //print_ingame(60,"%04x",ReadWord(&RAM[0x30800]));

   cpu_execute_cycles(CPU_68K_0, CPU_FRAME_MHz(12,60));	// M68000 12MHz (60fps)
   #ifdef HACKS
      print_debug("PC:%06x SR:%04x\n",s68000context.pc,s68000context.sr);
   #endif
   cpu_interrupt(CPU_68K_0, 1);
   cpu_interrupt(CPU_68K_0, 3);

   Taito2610_Frame();			// Z80 and YM2610
}

void DrawDLeague(void)
{
   ClearPaletteMap();

   // Init tc0180vcu emulation
   // ------------------------

   tc0004vcu_layer_count = 0;

   // BG0
   // ---

   tc0004vcu_render_bg0();

   // BG1
   // ---

   tc0004vcu_render_bg1();

   // OBJECT
   // ------

   tc0004vcu_render_obj();

   // FG0
   // ---

   //tc0004vcu_render_fg0();
}
