/******************************************************************************/
/*                                                                            */
/*                               RAINE GAME LISTS                             */
/*                                                                            */
/******************************************************************************/

#ifndef __raine_games_h_
#define __raine_games_h_

#include "deftypes.h"

enum company_num
{
   COMPANY_ID_UNKNOWN = 0,
   COMPANY_ID_BOOTLEG,
   COMPANY_ID_CAPCOM,
   COMPANY_ID_DECO,
   COMPANY_ID_IREM,
   COMPANY_ID_JALECO,
   COMPANY_ID_KONAMI,
   COMPANY_ID_NICHIBUTSU,
   COMPANY_ID_SEGA,
   COMPANY_ID_SNK,
   COMPANY_ID_TAITO,
   COMPANY_ID_TECMO,
   COMPANY_ID_TOAPLAN,
   COMPANY_ID_BANPREST,
   COMPANY_ID_UPL,
   COMPANY_ID_NAMCO,
   COMPANY_ID_NTC,
   COMPANY_ID_VISCO,
   COMPANY_ID_FACE,
   COMPANY_ID_SEIBU,
   COMPANY_ID_SAMMY,
   COMPANY_ID_VIDEOSYSTEM,
   COMPANY_ID_ALPHA,
   COMPANY_ID_NAZCA,
   COMPANY_ID_TECHNOS,
   COMPANY_ID_NMK,
   COMPANY_ID_SUNSOFT,
   COMPANY_ID_AICOM,
   COMPANY_ID_YUMEKOBO,
   COMPANY_ID_HUDSON,
   COMPANY_ID_TAD,
   COMPANY_ID_COMAD,
   COMPANY_ID_SETA,
   COMPANY_ID_EX_SYSTEM,
   COMPANY_ID_KANEKO,
   COMPANY_ID_RAIZING,
   COMPANY_ID_CAVE,
};

#define ROMOF(game) \
"#" game            \

#define CLONEOF(game) \
"$" game              \

#define IS_ROMOF(string) \
(string##[0] == '#')     \

#define IS_CLONEOF(string) \
(string##[0] == '$')       \

typedef struct DIR_INFO
{
   UBYTE *maindir;              // directory name
} DIR_INFO;

typedef struct ROM_INFO
{
   UBYTE *name;                 // file name
   ULONG  size;                 // file size
   ULONG  crc32;                // file crc32
} ROM_INFO;

typedef struct INPUT_INFO
{
   UWORD  default_key;          // default input mapping
   UBYTE *name;                 // input name
   ULONG  offset;               // offset in (input) ram
   UBYTE  bit_mask;             // bit mask
   UBYTE  flags;                // type flags
} INPUT_INFO;

typedef struct DSW_DATA
{
   UBYTE *name;                 // dsw name          / dsw setting name
   UBYTE  bit_mask;             // dsw bit mask      / dsw setting bit pattern
   UBYTE  count;                // dsw setting count / zero
} DSW_DATA;

typedef struct DSW_INFO
{
   ULONG     offset;            // offset in (dsw) ram
   UBYTE     factory_setting;   // dsw default data
   DSW_DATA *data;              // dsw data list
} DSW_INFO;

typedef struct ROMSW_DATA
{
   UBYTE *name;                 // romsw setting name
   UBYTE  data;                 // romsw setting byte
} ROMSW_DATA;

typedef struct ROMSW_INFO
{
   ULONG       offset;          // offset in rom
   UBYTE       factory_setting; // romsw default data
   ROMSW_DATA *data;            // romsw data list
} ROMSW_INFO;

/*

flags for VIDEO_INFO

*/

#define VIDEO_ROTATE_NORMAL	(0x00000000)
#define VIDEO_ROTATE_90		(0x00000001)
#define VIDEO_ROTATE_180	(0x00000002)
#define VIDEO_ROTATE_270	(0x00000003)

#define VIDEO_ROTATE(flags)	((flags >> 0) & 3)

#define VIDEO_ROTATABLE		(0x00000004)

#define VIDEO_FLIP_NORMAL	(0x00000000)
#define VIDEO_FLIP_X		(0x00000008)
#define VIDEO_FLIP_Y		(0x00000010)
#define VIDEO_FLIP_XY		(0x00000018)

#define VIDEO_FLIP(flags)	((flags >> 3) & 3)

typedef struct VIDEO_INFO
{
   void       (*draw_game)();   // pointer to screen update function
   int        screen_x;         // screen width
   int        screen_y;         // screen height
   int        border_size;      // clipping border size
   ULONG      flags;		// extra info
} VIDEO_INFO;

typedef struct SOUND_INFO
{
   ULONG      type;             // chip type
   void      *interface;        // chip specific inteface
} SOUND_INFO;

/*

main game structure

*/

typedef struct GAME_MAIN
{
   DIR_INFO    *dir_list;       // dir list
   ROM_INFO    *rom_list;       // rom list
   INPUT_INFO  *input_list;     // input list
   DSW_INFO    *dsw_list;       // dsw list
   ROMSW_INFO  *romsw_list;     // romsw list

   void       (*load_game)();   // Pointer to load/init game function
   void       (*clear_game)();  // Pointer to clear game function
   VIDEO_INFO  *video_info;     // video hardware information
   void       (*exec_frame)();  // Pointer to exec game function

   /*

   game 'database' information

   */

   UBYTE       *main_name;      // id game name (8 char, lower case)
   UBYTE       *long_name;      // full game name
   UBYTE       *long_name_jpn;  // full game name
   UBYTE        company_id;     // company who produced the game
   UBYTE       *board;          // board number (company specific)
   ULONG        year;           // year of release

   /*

   sound cpu information

   */

   SOUND_INFO  *sound_list;     // sound list

} GAME_MAIN;

/*

the list of all games

*/

extern struct GAME_MAIN *game_list[];

/*

the number of games in game_list

*/

int game_count;

/*

the currently loaded game (or NULL)

*/

struct GAME_MAIN *current_game;

/*

initialize game_list (and other things)

*/

void init_game_list(void);

/*

return string for company_id

*/

char *game_company_name(UBYTE company_id);

#endif // __raine_games_h_
