/******************************************************************************/
/*                                                                            */
/*                   KAISER KNUCKLE (C) 1994 TAITO CORPORATION                */
/*                                                                            */
/******************************************************************************/

#include "gameinc.h"
#include "f3system.h"
#include "kaiserkn.h"
#include "tc003vcu.h"
#include "tc200obj.h"

static struct DIR_INFO kaiser_knuckle_dirs[] =
{
   { "kaiser_knuckle", },
   { "kaiserkn", },
   { NULL, },
};

static struct ROM_INFO kaiser_knuckle_roms[] =
{
   {   "d84-01.rom", 0x00200000, 0x9AD22149, },	// 0
   {   "d84-02.rom", 0x00200000, 0x9E1827E4, },	// 1
   {   "d84-03.rom", 0x00200000, 0xD786F552, },	// 2
   {   "d84-04.rom", 0x00200000, 0xD1F32B5D, },	// 3
   {   "d84-05.rom", 0x00200000, 0x31A3C75D, },	// 4
   {   "d84-06.rom", 0x00200000, 0xFA924DAB, },	// 5
   {   "d84-07.rom", 0x00200000, 0x54517A6B, },	// 6
   {   "d84-08.rom", 0x00200000, 0x07347BF1, },	// 7
   {   "d84-09.rom", 0x00200000, 0xFAA78D98, },	// 8
   {   "d84-10.rom", 0x00200000, 0xB84B7320, },	// 9
   {   "d84-11.rom", 0x00200000, 0xA062C1D4, },	// 10
   {   "d84-12.rom", 0x00200000, 0x66A7A9AA, },	// 11
   {   "d84-13.rom", 0x00200000, 0xAE125516, },	// 12
   {   "d84-14.rom", 0x00200000, 0x2B2E693E, },	// 13
   {   "d84-15.rom", 0x00100000, 0x31CEB152, },	// 14
   {   "d84-16.rom", 0x00100000, 0xBCFF9B2D, },	// 15
   {   "d84-17.rom", 0x00100000, 0x0BE37CC3, },	// 16
   {   "d84-18.rom", 0x00100000, 0xE812BCC5, },	// 17
   {   "d84-19.rom", 0x00080000, 0x6DDF77E5, },	// 18
   {   "d84-20.rom", 0x00080000, 0xF85041E5, },	// 19
   {   "d84-21.rom", 0x00080000, 0x89F68B66, },	// 20
   {   "d84-29.rom", 0x00080000, 0x9821F17A, },	// 21
   {   "d84-23.rom", 0x00080000, 0x39F12A9B, },	// 22
   {   "d84-24.rom", 0x00080000, 0xBF20C755, },	// 23
   {   "d84-25.rom", 0x00080000, 0x2840893F, },	// 24
   {   "d84-26.rom", 0x00040000, 0x4F5B8563, },	// 25
   {   "d84-27.rom", 0x00040000, 0xFB0CB1BA, },	// 26
   {           NULL,          0,          0, },
};

static struct ROMSW_DATA romsw_data_kaiser_knuckle_0[] =
{
   { "Taito Japan",                     0x01 },
   { "Taito America (Global Champion)", 0x02 },
   { "Taito",                           0x03 },
   { NULL,                              0    },
};

static struct ROMSW_INFO kaiser_knuckle_romsw[] =
{
   { 0x1FFFFF, 0x03, romsw_data_kaiser_knuckle_0 },
   { 0,        0,    NULL },
};

static struct VIDEO_INFO kaiser_knuckle_video =
{
   DrawKaiserKnuckle,
   320,
   224,
   64,
   VIDEO_ROTATE_NORMAL,
};

struct GAME_MAIN game_kaiser_knuckle =
{
   kaiser_knuckle_dirs,
   kaiser_knuckle_roms,
   f3_system_inputs_6_button,
   NULL,
   kaiser_knuckle_romsw,

   LoadKaiserKnuckle,
   ClearKaiserKnuckle,
   &kaiser_knuckle_video,
   ExecuteF3SystemFrame,
   "kaiserkn",
   "Kaiser Knuckle",
   "JCU[ibN",
   COMPANY_ID_TAITO,
   "D84",
   1994,
   NULL,
};

static struct DIR_INFO kaiser_knuckle_alt_dirs[] =
{
   { "dankuga", },
   { ROMOF("kaiserkn"), },
   { CLONEOF("kaiserkn"), },
   { NULL, },
};

static struct ROM_INFO kaiser_knuckle_alt_roms[] =
{
   {   "d84-01.rom", 0x00200000, 0x9AD22149, },
   {   "d84-02.rom", 0x00200000, 0x9E1827E4, },
   {   "d84-03.rom", 0x00200000, 0xD786F552, },
   {   "d84-04.rom", 0x00200000, 0xD1F32B5D, },
   {   "d84-05.rom", 0x00200000, 0x31A3C75D, },
   {   "d84-06.rom", 0x00200000, 0xFA924DAB, },
   {   "d84-07.rom", 0x00200000, 0x54517A6B, },
   {   "d84-08.rom", 0x00200000, 0x07347BF1, },
   {   "d84-09.rom", 0x00200000, 0xFAA78D98, },
   {   "d84-10.rom", 0x00200000, 0xB84B7320, },
   {   "d84-11.rom", 0x00200000, 0xA062C1D4, },
   {   "d84-12.rom", 0x00200000, 0x66A7A9AA, },
   {   "d84-13.rom", 0x00200000, 0xAE125516, },
   {   "d84-14.rom", 0x00200000, 0x2B2E693E, },
   {   "d84-15.rom", 0x00100000, 0x31CEB152, },
   {   "d84-16.rom", 0x00100000, 0xBCFF9B2D, },
   {   "d84-17.rom", 0x00100000, 0x0BE37CC3, },
   {   "d84-18.rom", 0x00100000, 0xE812BCC5, },
   {   "d84-19.rom", 0x00080000, 0x6DDF77E5, },
   {   "d84-20.rom", 0x00080000, 0xF85041E5, },
   {   "d84-21.rom", 0x00080000, 0x89F68B66, },
   { "dkg_mpr0.bin", 0x00080000, 0xad6ada07, },
   { "dkg_mpr1.bin", 0x00080000, 0x97566f69, },
   { "dkg_mpr2.bin", 0x00080000, 0x18a4748b, },
   { "dkg_mpr3.bin", 0x00080000, 0xee1531ca, },
   {   "d84-26.rom", 0x00040000, 0x4F5B8563, },
   {   "d84-27.rom", 0x00040000, 0xFB0CB1BA, },
   {           NULL,          0,          0, },
};

static struct ROMSW_DATA romsw_data_kaiser_knuckle_alt_0[] =
{
   { "Taito Japan",                     0x01 },
   { "Taito America (Global Champion)", 0x02 },
   { "Taito",                           0x03 },
   { NULL,                              0    },
};

static struct ROMSW_INFO kaiser_knuckle_alt_romsw[] =
{
   { 0x1FFFFF, 0x01, romsw_data_kaiser_knuckle_alt_0 },
   { 0,        0,    NULL },
};

struct GAME_MAIN game_kaiser_knuckle_alt =
{
   kaiser_knuckle_alt_dirs,
   kaiser_knuckle_alt_roms,
   f3_system_inputs_6_button,
   NULL,
   kaiser_knuckle_alt_romsw,

   LoadKaiserKnuckle,
   ClearKaiserKnuckle,
   &kaiser_knuckle_video,
   ExecuteF3SystemFrame,
   "dankuga",
   "Dankuga",
   "JCU[ibN (alternate)",
   COMPANY_ID_TAITO,
   "D84",
   1994,
   NULL,
};

static UBYTE *RAM_BG0;
static UBYTE *RAM_BG1;
static UBYTE *RAM_BG2;
static UBYTE *RAM_BG3;

static UBYTE *RAM_SCR0;
static UBYTE *RAM_SCR1;
static UBYTE *RAM_SCR2;
static UBYTE *RAM_SCR3;

static ULONG SCR0_XOFS;
static ULONG SCR1_XOFS;
static ULONG SCR2_XOFS;
static ULONG SCR3_XOFS;

static ULONG SCR0_YOFS;
static ULONG SCR1_YOFS;
static ULONG SCR2_YOFS;
static ULONG SCR3_YOFS;

static UBYTE *GFX2;
static UBYTE *GFX_BG0;
static UBYTE *GFX_BG0_SOLID;

static UBYTE *GFX_SPR;
static UBYTE *GFX_SPR_SOLID;

void LoadKaiserKnuckle(void)
{
   int ta,tb,tc;

   RAMSize=0x80000;

   if(!(RAM=AllocateMem(0x80000))) return;
   if(!(ROM=AllocateMem(0x200000))) return;
   if(!(GFX2=AllocateMem(0x9BAE00))) return;
   if(!(GFX=AllocateMem(0x1900400))) return;	// ouch!!

   GFX_BG0=GFX2;
   GFX_SPR=GFX;

   if(!load_rom("d84-03.rom", ROM, 0x200000)) return;	// 16x16 SPRITES ($8000)
   tb=0;
   for(ta=0;ta<0x200000;ta++,tb+=4){
      WriteWord(&GFX_SPR[tb],((ROM[ta])|(ROM[ta]<<4))&0x0F0F);
   }
   if(!load_rom("d84-04.rom", ROM, 0x200000)) return;	// 16x16 SPRITES
   tb=2;
   for(ta=0;ta<0x200000;ta++,tb+=4){
      WriteWord(&GFX_SPR[tb],((ROM[ta])|(ROM[ta]<<4))&0x0F0F);
   }

   tb=0;
   if(!load_rom("d84-05.rom", ROM, 0x200000)) return;	// 16x16 SPRITES (MASK)
   for(ta=0;ta<0x200000;ta++){
      tc=ROM[ta];
      GFX[tb+3]|=((tc&0x40)>>6)<<4;
      GFX[tb+2]|=((tc&0x10)>>4)<<4;
      GFX[tb+1]|=((tc&0x04)>>2)<<4;
      GFX[tb+0]|=((tc&0x01)>>0)<<4;
      tb+=4;
   }

   if(!load_rom("d84-06.rom", ROM, 0x200000)) return;	// 16x16 SPRITES ($8000)
   tb=0x800000;
   for(ta=0;ta<0x200000;ta++,tb+=4){
      WriteWord(&GFX_SPR[tb],((ROM[ta])|(ROM[ta]<<4))&0x0F0F);
   }
   if(!load_rom("d84-07.rom", ROM, 0x200000)) return;	// 16x16 SPRITES
   tb=0x800002;
   for(ta=0;ta<0x200000;ta++,tb+=4){
      WriteWord(&GFX_SPR[tb],((ROM[ta])|(ROM[ta]<<4))&0x0F0F);
   }

   tb=0x800000;
   if(!load_rom("d84-08.rom", ROM, 0x200000)) return;	// 16x16 SPRITES (MASK)
   for(ta=0;ta<0x200000;ta++){
      tc=ROM[ta];
      GFX[tb+3]|=((tc&0x40)>>6)<<4;
      GFX[tb+2]|=((tc&0x10)>>4)<<4;
      GFX[tb+1]|=((tc&0x04)>>2)<<4;
      GFX[tb+0]|=((tc&0x01)>>0)<<4;
      tb+=4;
   }

   if(!load_rom("d84-09.rom", ROM, 0x200000)) return;	// 16x16 SPRITES ($8000)
   tb=0x1000000;
   for(ta=0;ta<0x200000;ta++,tb+=4){
      WriteWord(&GFX_SPR[tb],((ROM[ta])|(ROM[ta]<<4))&0x0F0F);
   }
   if(!load_rom("d84-10.rom", ROM, 0x200000)) return;	// 16x16 SPRITES
   tb=0x1000002;
   for(ta=0;ta<0x200000;ta++,tb+=4){
      WriteWord(&GFX_SPR[tb],((ROM[ta])|(ROM[ta]<<4))&0x0F0F);
   }

   tb=0x1000000;
   if(!load_rom("d84-11.rom", ROM, 0x200000)) return;	// 16x16 SPRITES (MASK)
   for(ta=0;ta<0x200000;ta++){
      tc=ROM[ta];
      GFX[tb+3]|=((tc&0x40)>>6)<<4;
      GFX[tb+2]|=((tc&0x10)>>4)<<4;
      GFX[tb+1]|=((tc&0x04)>>2)<<4;
      GFX[tb+0]|=((tc&0x01)>>0)<<4;
      tb+=4;
   }

   if(!load_rom("d84-19.rom", ROM, 0x40100)) return;	// 16x16 SPRITES ($1004)
   tb=0x1800000;
   for(ta=0;ta<0x40100;ta++,tb+=4){
      WriteWord(&GFX_SPR[tb],((ROM[ta])|(ROM[ta]<<4))&0x0F0F);
   }
   if(!load_rom("d84-20.rom", ROM, 0x40100)) return;	// 16x16 SPRITES
   tb=0x1800002;
   for(ta=0;ta<0x40100;ta++,tb+=4){
      WriteWord(&GFX_SPR[tb],((ROM[ta])|(ROM[ta]<<4))&0x0F0F);
   }

   tb=0x1800000;
   if(!load_rom("d84-21.rom", ROM, 0x40100)) return;	// 16x16 SPRITES (MASK)
   for(ta=0;ta<0x40100;ta++){
      tc=ROM[ta];
      GFX[tb+3]|=((tc&0x40)>>6)<<4;
      GFX[tb+2]|=((tc&0x10)>>4)<<4;
      GFX[tb+1]|=((tc&0x04)>>2)<<4;
      GFX[tb+0]|=((tc&0x01)>>0)<<4;
      tb+=4;
   }

   tb=0;
   if(!load_rom("d84-12.rom", ROM, 0x200000)) return;	// 16x16 TILES ($8000)
   for(ta=0;ta<0x200000;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=4;
   if(!load_rom("d84-13.rom", ROM, 0x200000)) return;	// 16x16 TILES
   for(ta=0;ta<0x200000;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=0;
   if(!load_rom("d84-14.rom", ROM, 0x200000)) return;	// 16x16 TILES (MASK)
   for(ta=0;ta<0x200000;ta+=2){
      tc=ROM[ta];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<4;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<4;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<4;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<4;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<4;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<4;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<4;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<4;
      tb+=8;
   }
   tb=0x800000;
   if(!load_rom("d84-16.rom", ROM, 0x6EB80)) return;	// 16x16 TILES ($1BAE)
   for(ta=0;ta<0x6EB80;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=0x800004;
   if(!load_rom("d84-17.rom", ROM, 0x6EB80)) return;	// 16x16 TILES
   for(ta=0;ta<0x6EB80;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=0x800000;
   if(!load_rom("d84-18.rom", ROM, 0x6EB80)) return;	// 16x16 TILES (MASK)
   for(ta=0;ta<0x6EB80;ta+=2){
      tc=ROM[ta];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<4;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<4;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<4;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<4;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<4;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<4;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<4;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<4;
      tb+=8;
   }

   GFX_BG0_SOLID = MakeSolidTileMap16x16(GFX_BG0, 0x09BAE);
   GFX_SPR_SOLID = make_solid_mask_16x16(GFX_SPR, 0x19004);

   // Setup 68020 Memory Map
   // ----------------------

   AddF3MemoryMap(0x200000);

   RAM_BG1=RAM+0x30000;
   RAM_BG2=RAM+0x31000;
   RAM_BG3=RAM+0x32000;
   RAM_BG0=RAM+0x33000;

   RAM_SCR1=RAM+0x6A000;
   RAM_SCR2=RAM+0x6A002;
   RAM_SCR3=RAM+0x6A004;
   RAM_SCR0=RAM+0x6A006;

   SCR1_XOFS=(0xF600)+(1<<6);
   SCR2_XOFS=(0xF700)+(1<<6);
   SCR3_XOFS=(0xF800)+(1<<6);
   SCR0_XOFS=(0xF900)+(1<<6);

   SCR1_YOFS=(0x0780-(16<<7));
   SCR2_YOFS=(0x0780-(16<<7));
   SCR3_YOFS=(0x0780-(16<<7));
   SCR0_YOFS=(0x0780-(16<<7));

   if(!load_rom_index(24, RAM, 0x80000)) return;
   for(ta=0;ta<0x80000;ta++){
      ROM[(ta<<2)+0]=RAM[ta];
   }
   if(!load_rom_index(23, RAM, 0x80000)) return;
   for(ta=0;ta<0x80000;ta++){
      ROM[(ta<<2)+1]=RAM[ta];
   }
   if(!load_rom_index(22, RAM, 0x80000)) return;
   for(ta=0;ta<0x80000;ta++){
      ROM[(ta<<2)+2]=RAM[ta];
   }
   if(!load_rom_index(21, RAM, 0x80000)) return;
   for(ta=0;ta<0x80000;ta++){
      ROM[(ta<<2)+3]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x80000);
   memset(RAM+0x69000,0xFF,0x01000);

   if(! stricmp(current_game->main_name,"kaiserkn")){

   // FIX ROM CHECKSUM

   WriteWord68k(&ROM[0x11CE],0x4E75);		//	rts

   // EEPROM HACKS

   WriteWord68k(&ROM[0xC30C],0x7F00);		//	raine	#$00	<eeprom read>
   WriteWord68k(&ROM[0xC30E],0x4E75);		//	rts

   WriteWord68k(&ROM[0xC388],0x7F01);		//	raine	#$01	<eeprom write>
   WriteWord68k(&ROM[0xC38A],0x4E75);		//	rts

   // SPEED HACK#1

   WriteLong68k(&ROM[0x0534],0x4E714E71);

   WriteLong68k(&ROM[0x025C],0x7F024E71);	// 	raine	#$02 <stop cpu>
   WriteLong68k(&ROM[0x0260],0x4E714E71);	//	nop
   WriteWord68k(&ROM[0x0264],0x60EC);

   }
   else{

   // FIX ROM CHECKSUM

   WriteWord68k(&ROM[0x11F4],0x4E75);		//	rts

   // EEPROM HACKS

   WriteWord68k(&ROM[0xD93A],0x7F00);		//	raine	#$00	<eeprom read>
   WriteWord68k(&ROM[0xD93C],0x4E75);		//	rts

   WriteWord68k(&ROM[0xD9B6],0x7F01);		//	raine	#$01	<eeprom write>
   WriteWord68k(&ROM[0xD9B8],0x4E75);		//	rts

   // SPEED HACK#1

   WriteLong68k(&ROM[0x052E],0x4E714E71);

   WriteLong68k(&ROM[0x025C],0x7F024E71);	// 	raine	#$02 <stop cpu>
   WriteLong68k(&ROM[0x0260],0x4E714E71);	//	nop
   WriteWord68k(&ROM[0x0264],0x60EC);

   }

   F3SystemEEPROMAccess=&F3SysEEPROMAccessMode1;


   InitPaletteMap(RAM+0x60000, 0x200, 0x40, 0x1000);


   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu.RAM	= RAM+0x20000;
   tc0003vcu.mapper	= &Map_24bit_xRGB;
   tc0003vcu.bmp_x	= 64;
   tc0003vcu.bmp_y	= 64;
   tc0003vcu.bmp_w	= 320;
   tc0003vcu.bmp_h	= 224;
   tc0003vcu.scr_x	= 0;
   tc0003vcu.scr_y	= 16;

   // Init tc0200obj emulation
   // ------------------------

   tc0200obj.RAM	= RAM+0x20000;
   tc0200obj.RAM_B	= RAM+0x28000;
   tc0200obj.GFX	= GFX_SPR;
   tc0200obj.MASK	= GFX_SPR_SOLID;
   tc0200obj.bmp_x	= 64;
   tc0200obj.bmp_y	= 64;
   tc0200obj.bmp_w	= 320;
   tc0200obj.bmp_h	= 224;
   tc0200obj.mapper	= &Map_24bit_xRGB;
   tc0200obj.tile_mask	= 0x1FFFF;
   tc0200obj.ofs_x	= 0;	//-0x2E;
   tc0200obj.ofs_y	= -16;	//-0x18;

   tc0200obj.cols	= 32;

   init_tc0200obj();

   init_f3_system_ioc(F3_IOC_2P_6BUTTON);

   init_m68k();
}

void ClearKaiserKnuckle(void)
{
   save_eeprom();

   #ifdef HACKS
      save_debug("ROM.bin",ROM,0x200000,0);
   #endif
}

void DrawKaiserKnuckle(void)
{
   int x16,y16,zz,zzz,zzzz;
   int ta,x,y;
   UBYTE *MAP;

   ClearPaletteMap();

   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu_layer_count = 0;

   if(check_layer_enabled(f3_bg0_id)){
   MAKE_SCROLL_512x512_4_16(
      (0-320)-((ReadWord68k(&RAM_SCR0[0])-SCR0_XOFS)>>6),
      (0-224)-((ReadWord68k(&RAM_SCR0[8])-SCR0_YOFS)>>7)
   );

   START_SCROLL_512x512_4_16_R180(64,64,320,224);

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG0[zz])&0x1FF,
               32
            );

      (*Draw16x16_Mapped_JumpListR[(RAM_BG0[zz]&0xC0)>>6])(&GFX_BG0[ReadWord68k(&RAM_BG0[zz+2])<<8],x,y,MAP);

   END_SCROLL_512x512_4_16();
   }


   if(check_layer_enabled(f3_bg1_id)){
   MAKE_SCROLL_512x512_4_16(
      (0-320)-((ReadWord68k(&RAM_SCR1[0])-SCR1_XOFS)>>6),
      (0-224)-((ReadWord68k(&RAM_SCR1[8])-SCR1_YOFS)>>7)
   );

   START_SCROLL_512x512_4_16_R180(64,64,320,224);

   ta=ReadWord68k(&RAM_BG1[zz+2]);
   if(ta!=0){

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG1[zz])&0x1FF,
               32
            );

      if(GFX_BG0_SOLID[ta]==0){
         switch(RAM_BG1[zz]&0xC0){
         case 0xC0: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x80: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x40: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x00: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      else{
         switch(RAM_BG1[zz]&0xC0){
         case 0xC0: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x80: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x40: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x00: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
   }

   END_SCROLL_512x512_4_16();
   }

   if(check_layer_enabled(f3_bg2_id)){
   MAKE_SCROLL_512x512_4_16(
      (0-320)-((ReadWord68k(&RAM_SCR2[0])-SCR2_XOFS)>>6),
      (0-224)-((ReadWord68k(&RAM_SCR2[8])-SCR2_YOFS)>>7)
   );

   START_SCROLL_512x512_4_16_R180(64,64,320,224);

   ta=ReadWord68k(&RAM_BG2[zz+2]);
   if(ta!=0){

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG2[zz])&0x1FF,
               32
            );

      if(GFX_BG0_SOLID[ta]==0){
         switch(RAM_BG2[zz]&0xC0){
         case 0xC0: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x80: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x40: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x00: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      else{
         switch(RAM_BG2[zz]&0xC0){
         case 0xC0: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x80: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x40: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x00: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
   }

   END_SCROLL_512x512_4_16();
   }

   if(check_layer_enabled(f3_bg3_id)){
   MAKE_SCROLL_512x512_4_16(
      (0-320)-((ReadWord68k(&RAM_SCR3[0])-SCR3_XOFS)>>6),
      (0-224)-((ReadWord68k(&RAM_SCR3[8])-SCR3_YOFS)>>7)
   );

   START_SCROLL_512x512_4_16_R180(64,64,320,224);

   ta=ReadWord68k(&RAM_BG3[zz+2]);
   if(ta!=0){

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG3[zz])&0x1FF,
               32
            );

      if(GFX_BG0_SOLID[ta]==0){
         switch(RAM_BG3[zz]&0xC0){
         case 0xC0: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x80: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x40: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x00: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      else{
         switch(RAM_BG3[zz]&0xC0){
         case 0xC0: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x80: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x40: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x00: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
   }

   END_SCROLL_512x512_4_16();
   }

   render_tc0200obj_mapped_f3system_r180();

   if(check_layer_enabled(f3_bg5_id)){
      f3video_render_fg0_r180();
   }
}
