/******************************************************************************/
/*                                                                            */
/*               PUZZLE BOBBLE 2/2x (C) 1994 TAITO CORPORATION                */
/*                                                                            */
/******************************************************************************/

#include "gameinc.h"
#include "f3system.h"
#include "pbobble2.h"
#include "tc003vcu.h"
#include "tc200obj.h"

static struct DIR_INFO ensoniq_test_dirs[] =
{
   { "ensoniq", },
   { ROMOF("popnpop"), },
   { NULL, },
};

static struct ROM_INFO ensoniq_test_roms[] =
{
   {    "e51-04.38", 0x00200000, 0x66790F55, },
   {    "e51-05.41", 0x00200000, 0x4D08B26D, },
   {    "e51-13.32", 0x00040000, 0x3B9E3986, },
   {    "e51-14.33", 0x00040000, 0x1F9A5015, },
   {           NULL,          0,          0, },
};

static struct VIDEO_INFO ensoniq_test_video =
{
   DrawPuzzleBobble2Sound,
   320,
   240,
   0,
   VIDEO_ROTATE_NORMAL,
};

struct GAME_MAIN game_ensoniq_test =
{
   ensoniq_test_dirs,
   ensoniq_test_roms,
   NULL,
   NULL,
   NULL,

   LoadPBobble2Sound,
   ClearPBobble2Sound,
   &ensoniq_test_video,
   ExecutePBobble2SoundFrame,
   "ensoniq",
   "Ensoniq Test",
   NULL,
   COMPANY_ID_BOOTLEG,
   NULL,
   1999,
   NULL,
};

static struct DIR_INFO puzzle_bobble_2_dirs[] =
{
   { "puzzle_bobble_2", },
   { "pbobble2", },
   { NULL, },
};

static struct ROM_INFO puzzle_bobble_2_roms[] =
{
   {   "e10-04.rom", 0x00200000, 0x5C0862A6, },
   {   "e10-01.rom", 0x00100000, 0x8C26FF49, },
   {   "e10-02.rom", 0x00100000, 0xC0564490, },
   {   "e10-03.rom", 0x00200000, 0x46D68AC8, },
   {   "e10-05.rom", 0x00100000, 0x81266151, },
   {   "e10-06.rom", 0x00100000, 0x1B0F20E2, },
   {   "e10-07.rom", 0x00100000, 0xDCB3C29B, },
   {   "e10-09.rom", 0x00040000, 0xE0B1B599, },
   {   "e10-10.rom", 0x00040000, 0xF432267A, },
   {   "e10-11.rom", 0x00040000, 0xB82F81DA, },
   {   "e10-12.rom", 0x00040000, 0xB92DC8AD, },
   {   "e10-13.rom", 0x00040000, 0x87842C13, },
   {   "e10-15.rom", 0x00040000, 0xA2C0A268, },
   {           NULL,          0,          0, },
};

static struct ROMSW_DATA romsw_data_puzzle_bobble_2_0[] =
{
   { "Taito Japan (Japanese)",            0x01 },
   { "Taito America (Bust-a-move again)", 0x02 },
   { "Taito Japan",                       0x03 },
   { NULL,                                0    },
};

static struct ROMSW_INFO puzzle_bobble_2_romsw[] =
{
   { 0x0FFFFF, 0x03, romsw_data_puzzle_bobble_2_0 },
   { 0,        0,    NULL },
};

static struct VIDEO_INFO puzzle_bobble_2_video =
{
   DrawPuzzleBobble2,
   320,
   232,
   64,
   VIDEO_ROTATE_NORMAL,
};

struct GAME_MAIN game_puzzle_bobble_2 =
{
   puzzle_bobble_2_dirs,
   puzzle_bobble_2_roms,
   f3_system_inputs,
   NULL,
   puzzle_bobble_2_romsw,

   LoadPuzzleBobble2,
   ClearPuzzleBobble2,
   &puzzle_bobble_2_video,
   ExecuteF3SystemFrame,
   "pbobble2",
   "Puzzle Bobble 2",
   "pY{uQ",
   COMPANY_ID_TAITO,
   "E10",
   1995,
   NULL,
};

static struct DIR_INFO puzzle_bobble_2x_dirs[] =
{
   { "puzzle_bobble_2x", },
   { "puzbob2x", },
   { ROMOF("pbobble2"), },
   { CLONEOF("pbobble2"), },
   { NULL, },
};

static struct ROM_INFO puzzle_bobble_2x_roms[] =
{
   {       "e10.01", 0x00100000, 0x8C26FF49, },
   {       "e10.02", 0x00100000, 0xC0564490, },
   {       "e10.03", 0x00200000, 0x46D68AC8, },
   {       "e10.04", 0x00200000, 0x5C0862A6, },
   {       "e10.05", 0x00100000, 0x81266151, },
   {       "e10.06", 0x00100000, 0x1B0F20E2, },
   {       "e10.07", 0x00100000, 0xDCB3C29B, },
   {       "e10.26", 0x00040000, 0xA5C24047, },
   {       "e10.27", 0x00040000, 0x88CC0B5C, },
   {       "e10.28", 0x00040000, 0x412A3602, },
   {       "e10.29", 0x00040000, 0xF1E9AD3F, },
   {       "e10.30", 0x00040000, 0xBB090C1E, },
   {       "e10.31", 0x00040000, 0xF4B88D65, },
   {           NULL,          0,          0, },
};

static struct ROMSW_DATA romsw_data_puzzle_bobble_2x_0[] =
{
   { "Taito Japan (Japanese)",               0x01 },
   { "Taito America (Bust-a-move again EX)", 0x02 },
   { "Taito Japan",                          0x03 },
   { NULL,                                   0    },
};

static struct ROMSW_INFO puzzle_bobble_2x_romsw[] =
{
   { 0x0FFFFF, 0x01, romsw_data_puzzle_bobble_2x_0 },
   { 0,        0,    NULL },
};

struct GAME_MAIN game_puzzle_bobble_2x =
{
   puzzle_bobble_2x_dirs,
   puzzle_bobble_2x_roms,
   f3_system_inputs,
   NULL,
   puzzle_bobble_2x_romsw,

   LoadPuzzleBobble2x,
   ClearPuzzleBobble2x,
   &puzzle_bobble_2_video,
   ExecuteF3SystemFrame,
   "puzbob2x",
   "Puzzle Bobble 2 (extra)",
   "pY{uQw",
   COMPANY_ID_TAITO,
   "E10",
   1995,
   NULL,
};

static UBYTE *RAM_BG0;
static UBYTE *RAM_BG1;
static UBYTE *RAM_BG2;
static UBYTE *RAM_BG3;

static UBYTE *RAM_SCR0;
static UBYTE *RAM_SCR1;
static UBYTE *RAM_SCR2;
static UBYTE *RAM_SCR3;

static ULONG SCR0_XOFS;
static ULONG SCR1_XOFS;
static ULONG SCR2_XOFS;
static ULONG SCR3_XOFS;

static ULONG SCR0_YOFS;
static ULONG SCR1_YOFS;
static ULONG SCR2_YOFS;
static ULONG SCR3_YOFS;

static UBYTE *GFX_BG0;
static UBYTE *GFX_BG0_SOLID;
static UBYTE *GFX_BG0_PENS;

static UBYTE *GFX_SPR;
static UBYTE *GFX_SPR_SOLID;

void LoadPuzzleBobble2(void)
{
   int ta,tb,tc;

   RAMSize=0x80000;

   if(!(RAM=AllocateMem(0x80000))) return;
   if(!(ROM=AllocateMem(0x100000))) return;
   if(!(GFX=AllocateMem(0x3F5D00+0x3EC700))) return;

   GFX_BG0 = GFX+0x3F5D00;
   GFX_SPR = GFX+0x000000;

   tb=0;
   if(!load_rom("e10-02.rom", ROM, 0xFD740)) return;	// 16x16 SPRITES ($3F5D)
   for(ta=0;ta<0xFD740;ta++){
      GFX[tb++]=ROM[ta]&15;
      GFX[tb++]=ROM[ta]>>4;
      tb+=2;
   }
   tb=2;
   if(!load_rom("e10-01.rom", ROM, 0xFD740)) return;	// 16x16 SPRITES
   for(ta=0;ta<0xFD740;ta++){
      GFX[tb++]=ROM[ta]&15;
      GFX[tb++]=ROM[ta]>>4;
      tb+=2;
   }
   tb=0;
   if(!load_rom("e10-07.rom", ROM, 0xFB1C0)) return;	// 16x16 TILES ($3EC7)
   for(ta=0;ta<0xFB1C0;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=4;
   if(!load_rom("e10-06.rom", ROM, 0xFB1C0)) return;	// 16x16 TILES
   for(ta=0;ta<0xFB1C0;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=0;
   if(!load_rom("e10-05.rom", ROM, 0xFB1C0)) return;	// 16x16 TILES (MASK)
   for(ta=0;ta<0xFB1C0;ta+=2){
      tc=ROM[ta];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<4;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<4;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<4;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<4;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<4;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<4;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<4;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<4;
      tc=ROM[ta+1];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<5;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<5;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<5;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<5;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<5;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<5;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<5;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<5;
      tb+=8;
   }

   GFX_BG0_SOLID = make_solid_mask_16x16(GFX_BG0, 0x3EC7);
   GFX_SPR_SOLID = make_solid_mask_16x16(GFX_SPR, 0x3F5D);

   GFX_BG0_PENS = make_colour_count_16x16(GFX_BG0, 0x3EC7);

   // Setup 68020 Memory Map
   // ----------------------

   AddF3MemoryMap(0x100000);

   RAM_BG0=RAM+0x33000;
   RAM_BG1=RAM+0x32000;
   RAM_BG2=RAM+0x31000;
   RAM_BG3=RAM+0x30000;

   RAM_SCR0=RAM+0x6A006;
   RAM_SCR1=RAM+0x6A004;
   RAM_SCR2=RAM+0x6A002;
   RAM_SCR3=RAM+0x6A000;

   SCR3_XOFS=0xF640;
   SCR2_XOFS=0xF740;
   SCR1_XOFS=0xF840;
   SCR0_XOFS=0xF940;

   SCR3_YOFS=0xF400;
   SCR2_YOFS=0xF400;
   SCR1_YOFS=0xF400;
   SCR0_YOFS=0xF400;

   if(!load_rom("e10-11.rom", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+0]=RAM[ta];
   }
   if(!load_rom("e10-10.rom", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+1]=RAM[ta];
   }
   if(!load_rom("e10-09.rom", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+2]=RAM[ta];
   }
   if(!load_rom("e10-15.rom", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+3]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x80000);
   memset(RAM+0x69000,0xFF,0x01000);

   // EEPROM ACCESS
   // -------------

   WriteWord68k(&ROM[0x0044B2],0x7F00);		//	raine	#$00 <eeprom access>
   WriteWord68k(&ROM[0x0044B4],0x4ED6);		//	jmp	(a6)

   // SPEED HACK
   // ----------

   WriteWord68k(&ROM[0x31B4],0x4EF9);		//	jmp	$F8000
   WriteLong68k(&ROM[0x31B6],0x000F8000);

   WriteWord68k(&ROM[0xF8000],0x4EB9);		//	jsr	$028FC <random gen>
   WriteLong68k(&ROM[0xF8002],0x000028FC);
   WriteWord68k(&ROM[0xF8006],0x7F02);		//	raine	#$02 <stop cpu>
   WriteWord68k(&ROM[0xF8008],0x6100-10);	//	bra.s	<loop>

   F3SystemEEPROMAccess=&F3SysEEPROMAccessMode2;


   InitPaletteMap(RAM+0x60000, 0x200, 0x40, 0x1000);


   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu.RAM	= RAM+0x20000;
   tc0003vcu.mapper	= &Map_24bit_xRGB;
   tc0003vcu.bmp_x	= 64;
   tc0003vcu.bmp_y	= 64;
   tc0003vcu.bmp_w	= 320;
   tc0003vcu.bmp_h	= 232;
   tc0003vcu.scr_x	= 0;
   tc0003vcu.scr_y	= 0;

   // Init tc0200obj emulation
   // ------------------------

   tc0200obj.RAM	= RAM+0x20000;
   tc0200obj.RAM_B	= RAM+0x28000;
   tc0200obj.GFX	= GFX_SPR;
   tc0200obj.MASK	= GFX_SPR_SOLID;
   tc0200obj.bmp_x	= 64;
   tc0200obj.bmp_y	= 64;
   tc0200obj.bmp_w	= 320;
   tc0200obj.bmp_h	= 232;
   tc0200obj.mapper	= &Map_24bit_xRGB;
   tc0200obj.tile_mask	= 0x3FFF;
   tc0200obj.ofs_x	= 0-0x2E;
   tc0200obj.ofs_y	= 0-0x18;

   tc0200obj.cols	= 16;

   init_tc0200obj();

   init_f3_system_ioc(F3_IOC_2P_3BUTTON);

   init_m68k();
}

void LoadPuzzleBobble2x(void)
{
   int ta,tb,tc;

   RAMSize=0x80000;

   if(!(RAM=AllocateMem(0x80000))) return;
   if(!(ROM=AllocateMem(0x100000))) return;
   if(!(GFX=AllocateMem(0x3F5D00+0x3EC700))) return;

   GFX_BG0 = GFX+0x3F5D00;
   GFX_SPR = GFX+0x000000;

   tb=0;
   if(!load_rom("e10.02", ROM, 0xFD740)) return;	// 16x16 SPRITES ($3F5D)
   for(ta=0;ta<0xFD740;ta++){
      GFX[tb++]=ROM[ta]&15;
      GFX[tb++]=ROM[ta]>>4;
      tb+=2;
   }
   tb=2;
   if(!load_rom("e10.01", ROM, 0xFD740)) return;	// 16x16 SPRITES
   for(ta=0;ta<0xFD740;ta++){
      GFX[tb++]=ROM[ta]&15;
      GFX[tb++]=ROM[ta]>>4;
      tb+=2;
   }
   tb=0;
   if(!load_rom("e10.07", ROM, 0xFB1C0)) return;	// 16x16 TILES ($3EC7)
   for(ta=0;ta<0xFB1C0;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=4;
   if(!load_rom("e10.06", ROM, 0xFB1C0)) return;	// 16x16 TILES
   for(ta=0;ta<0xFB1C0;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=0;
   if(!load_rom("e10.05", ROM, 0xFB1C0)) return;	// 16x16 TILES (MASK)
   for(ta=0;ta<0xFB1C0;ta+=2){
      tc=ROM[ta];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<4;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<4;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<4;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<4;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<4;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<4;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<4;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<4;
      tc=ROM[ta+1];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<5;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<5;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<5;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<5;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<5;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<5;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<5;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<5;
      tb+=8;
   }

   GFX_BG0_SOLID = make_solid_mask_16x16(GFX_BG0, 0x3EC7);
   GFX_SPR_SOLID = make_solid_mask_16x16(GFX_SPR, 0x3F5D);

   GFX_BG0_PENS = make_colour_count_16x16(GFX_BG0, 0x3EC7);

   // Setup 68020 Memory Map
   // ----------------------

   AddF3MemoryMap(0x100000);

   RAM_BG0=RAM+0x33000;
   RAM_BG1=RAM+0x32000;
   RAM_BG2=RAM+0x31000;
   RAM_BG3=RAM+0x30000;

   RAM_SCR0=RAM+0x6A006;
   RAM_SCR1=RAM+0x6A004;
   RAM_SCR2=RAM+0x6A002;
   RAM_SCR3=RAM+0x6A000;

   SCR3_XOFS=0xF640;
   SCR2_XOFS=0xF740;
   SCR1_XOFS=0xF840;
   SCR0_XOFS=0xF940;

   SCR3_YOFS=0xF400;
   SCR2_YOFS=0xF400;
   SCR1_YOFS=0xF400;
   SCR0_YOFS=0xF400;

   if(!load_rom("e10.29", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+0]=RAM[ta];
   }
   if(!load_rom("e10.28", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+1]=RAM[ta];
   }
   if(!load_rom("e10.27", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+2]=RAM[ta];
   }
   if(!load_rom("e10.26", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[(ta<<2)+3]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x80000);
   memset(RAM+0x69000,0xFF,0x01000);

   // EEPROM ACCESS
   // -------------

   WriteWord68k(&ROM[0x0044D2],0x7F00);		//	raine	#$00 <eeprom access>
   WriteWord68k(&ROM[0x0044D4],0x4ED6);		//	jmp	(a6)

   // SPEED HACK
   // ----------

   WriteWord68k(&ROM[0x031D4],0x4EF9);		//	jmp	$F8000
   WriteLong68k(&ROM[0x031D6],0x000F8000);

   WriteWord68k(&ROM[0xF8000],0x4EB9);		//	jsr	$0291C <random gen>
   WriteLong68k(&ROM[0xF8002],0x0000291C);
   WriteWord68k(&ROM[0xF8006],0x7F02);		//	raine	#$02 <stop cpu>
   WriteWord68k(&ROM[0xF8008],0x6100-10);	//	bra.s	<loop>

   F3SystemEEPROMAccess=&F3SysEEPROMAccessMode2;


   InitPaletteMap(RAM+0x60000, 0x200, 0x40, 0x1000);


   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu.RAM	= RAM+0x20000;
   tc0003vcu.mapper	= &Map_24bit_xRGB;
   tc0003vcu.bmp_x	= 64;
   tc0003vcu.bmp_y	= 64;
   tc0003vcu.bmp_w	= 320;
   tc0003vcu.bmp_h	= 232;
   tc0003vcu.scr_x	= 0;
   tc0003vcu.scr_y	= 0;

   // Init tc0200obj emulation
   // ------------------------

   tc0200obj.RAM	= RAM+0x20000;
   tc0200obj.RAM_B	= RAM+0x28000;
   tc0200obj.GFX	= GFX_SPR;
   tc0200obj.MASK	= GFX_SPR_SOLID;
   tc0200obj.bmp_x	= 64;
   tc0200obj.bmp_y	= 64;
   tc0200obj.bmp_w	= 320;
   tc0200obj.bmp_h	= 232;
   tc0200obj.mapper	= &Map_24bit_xRGB;
   tc0200obj.tile_mask	= 0x3FFF;
   tc0200obj.ofs_x	= 0-0x2E;
   tc0200obj.ofs_y	= 0-0x18;

   tc0200obj.cols	= 16;

   init_tc0200obj();

   init_f3_system_ioc(F3_IOC_2P_3BUTTON);

   init_m68k();
}

void ClearPuzzleBobble2(void)
{
   save_eeprom();

   #ifdef HACKS
      //save_debug("ROM.bin",ROM,0x100000,0);
      //save_debug("RAM.bin",RAM,0x080000,0);
      //save_debug("GFX.bin",GFX,0x400000,0);
   #endif
}

void ClearPuzzleBobble2x(void)
{
   save_eeprom();

   #ifdef HACKS
      //save_debug("ROM.bin",ROM,0x100000,0);
      //save_debug("RAM.bin",RAM,0x080000,0);
      //save_debug("GFX.bin",GFX,0x400000,0);
   #endif
}

void DrawPuzzleBobble2(void)
{
   int x16,y16,zz,zzz,zzzz;
   int ta,x,y,z;
   UBYTE *map;

   ClearPaletteMap();

   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu_layer_count = 0;

   if(check_layer_enabled(f3_bg0_id)){
   MAKE_SCROLL_512x512_4_16(
      (ReadWord68k(&RAM_SCR0[0])-SCR0_XOFS)>>6,
      (ReadWord68k(&RAM_SCR0[8])-SCR0_YOFS)>>7
   );

   START_SCROLL_512x512_4_16(64,64,320,232);

      ta=ReadWord68k(&RAM_BG0[zz+2])&0x3FFF;

      MAP_PALETTE_MULTI_MAPPED(
         Map_24bit_xRGB,
         ReadWord68k(&RAM_BG0[zz])&0x1FF,
         GFX_BG0_PENS[ta],
         map
      );

      switch(RAM_BG0[zz]&0xC0){
         case 0x00: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,map);        break;
         case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,map);  break;
         case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,map);  break;
         case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,map); break;
      }

   END_SCROLL_512x512_4_16();
   }

   if(check_layer_enabled(f3_bg1_id)){
   MAKE_SCROLL_512x512_4_16(
      (ReadWord68k(&RAM_SCR1[0])-SCR1_XOFS)>>6,
      (ReadWord68k(&RAM_SCR1[8])-SCR1_YOFS)>>7
   );

   START_SCROLL_512x512_4_16(64,64,320,232);

      ta=ReadWord68k(&RAM_BG1[zz+2])&0x3FFF;
      if(GFX_BG0_SOLID[ta]!=0){			// No pixels; skip

         MAP_PALETTE_MULTI_MAPPED(
            Map_24bit_xRGB,
            ReadWord68k(&RAM_BG1[zz])&0x1FF,
            GFX_BG0_PENS[ta],
            map
         );

         if(GFX_BG0_SOLID[ta]==1){		// Some pixels; trans
            switch(RAM_BG1[zz]&0xC0){
               case 0x00: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,map);        break;
               case 0x40: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0x80: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0xC0: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,map); break;
            }
         }
         else{					// all pixels; solid
            switch(RAM_BG1[zz]&0xC0){
               case 0x00: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,map);        break;
               case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,map); break;
            }
         }

      }

   END_SCROLL_512x512_4_16();
   }

   if(check_layer_enabled(f3_bg2_id)){
   MAKE_SCROLL_512x512_4_16(
      (ReadWord68k(&RAM_SCR2[0])-SCR2_XOFS)>>6,
      (ReadWord68k(&RAM_SCR2[8])-SCR2_YOFS)>>7
   );

   START_SCROLL_512x512_4_16(64,64,320,232);

      ta=ReadWord68k(&RAM_BG2[zz+2])&0x3FFF;
      if(GFX_BG0_SOLID[ta]!=0){			// No pixels; skip

         MAP_PALETTE_MULTI_MAPPED(
            Map_24bit_xRGB,
            ReadWord68k(&RAM_BG2[zz])&0x1FF,
            GFX_BG0_PENS[ta],
            map
         );

         if(GFX_BG0_SOLID[ta]==1){		// Some pixels; trans
            switch(RAM_BG2[zz]&0xC0){
               case 0x00: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,map);        break;
               case 0x40: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0x80: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0xC0: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,map); break;
            }
         }
         else{					// all pixels; solid
            switch(RAM_BG2[zz]&0xC0){
               case 0x00: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,map);        break;
               case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,map); break;
            }
         }

      }

   END_SCROLL_512x512_4_16();
   }

   if(check_layer_enabled(f3_bg3_id)){
   MAKE_SCROLL_512x512_4_16(
      (ReadWord68k(&RAM_SCR3[0])-SCR3_XOFS)>>6,
      (ReadWord68k(&RAM_SCR3[8])-SCR3_YOFS)>>7
   );

   START_SCROLL_512x512_4_16(64,64,320,232);

      ta=ReadWord68k(&RAM_BG3[zz+2])&0x3FFF;
      if(GFX_BG0_SOLID[ta]!=0){			// No pixels; skip

         MAP_PALETTE_MULTI_MAPPED(
            Map_24bit_xRGB,
            ReadWord68k(&RAM_BG3[zz])&0x1FF,
            GFX_BG0_PENS[ta],
            map
         );

         if(GFX_BG0_SOLID[ta]==1){		// Some pixels; trans
            switch(RAM_BG3[zz]&0xC0){
               case 0x00: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,map);        break;
               case 0x40: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0x80: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0xC0: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,map); break;
            }
         }
         else{					// all pixels; solid
            switch(RAM_BG3[zz]&0xC0){
               case 0x00: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,map);        break;
               case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,map);  break;
               case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,map); break;
            }
         }

      }

   END_SCROLL_512x512_4_16();
   }

   // object

   render_tc0200obj_mapped_f3system_b();

   if(check_layer_enabled(f3_bg5_id)){
   zz=0;
   for(y=64;y<232+64;y+=8){
   for(x=64;x<320+64;x+=8){
      z=RAM[0x3C001+zz];
      if((z!=0)&&(z!=0x10)){		// 0x10='brown' rectangle

         MAP_PALETTE_MULTI_MAPPED(
            Map_24bit_xRGB,
            (RAM[0x3C000+zz]>>1)&0x3F,
            16,
            map
         );

         if((RAM[0x3C000+zz]&0x01)!=0){
            Draw8x8_Trans_Packed_Mapped(&RAM[0x3E000+(z<<5)],x,y,map);
         }
         else{
            Draw8x8_Trans_Packed_Mapped_FlipY(&RAM[0x3E000+(z<<5)],x,y,map);
         }

      }
      zz+=2;
   }
   zz+=(128-80);
   }
   }
}


/*************************************************/
/* ENSONIQ OTIS SOUND TEST                       */
/*************************************************/

static int BadReadByte(ULONG address)
{
   static int ta=0;

   #ifdef HACKS
      print_debug("Rb(%06x) [%06x]\n",address,s68000context.pc);
   #endif
   ta^=0xFF;
   return(ta);
}

static int BadReadWord(ULONG address)
{
   static int ta=0;

   #ifdef HACKS
      print_debug("Rw(%06x) [%06x]\n",address,s68000context.pc);
   #endif
   ta^=0xFFFF;
   return(ta);
}

static void BadWriteByte(ULONG address, UBYTE data)
{
   #ifdef HACKS
      print_debug("Wb(%06x,%02x) [%06x]\n",address,data,s68000context.pc);
   #endif
}

static void BadWriteWord(ULONG address, UWORD data)
{
   #ifdef HACKS
      print_debug("Ww(%06x,%04x) [%06x]\n",address,data,s68000context.pc);
   #endif
}

/******************************************************************************/
/* DEBUG COMMUNICATION                                                        */
/******************************************************************************/

static UBYTE EnsoniqScr[0x400];
static int EnsoniqCount=0;

static void DebugWriteByte(ULONG address, UBYTE data)
{
   address &= 0x1F;
   address |= 0x01;

   #ifdef HACKS
   print_debug("DEBUG_WRITE: %02x:%02x [PC:%06x]\n",address,data,s68000context.pc);
   #endif

   switch(address){
      case 0x15:				// 0x15: Debug Mode?
         #ifdef HACKS
         print_debug("Debug Mode: %02x\n",data);
         #endif
      break;
      case 0x17:				// 0x17: Write Debug Data
         #ifdef HACKS
         print_debug("Debug Data: %02x\n",data);
         #endif
         RAM[0x60000+EnsoniqCount]=data;	// Readme.txt
         EnsoniqScr[EnsoniqCount++]=data;
         EnsoniqCount&=0x3FF;
      break;
      default:
      break;
   }
}

static int DebugReadByte(ULONG address)
{
   static int ta=0;

   address &= 0x1F;
   address |= 0x01;

   #ifdef HACKS
   print_debug("DEBUG_READ: %02x:xx [PC:%06x]\n",address,s68000context.pc);
   #endif

   switch(address){
      case 0x0B:				// 0x0B: ?
         //return 0x0F;
         return 0x28;				// bit#3 should be set (bit#0 = ?/bit#1-2 = ?/bit#5 = ?)
      break;
      case 0x13:				// 0x13: ?
         return 0x08;				// bit#3 should be set (bit#4 / bit#6 = ?)
      break;
      case 0x17:				// 0x17: ?
         return 0x00;				// Data seems to be discared (sync read?)
      break;
      default:
         ta ^= 0xFF;
         return ta;
      break;
   }
}

/******************************************************************************/
/* ENSONIQ ES-5505 [OTIS]                                                     */
/******************************************************************************/

/*

1E - Page Select

Page	Address	Function
00 - 1F	00 - 1D	Voice Specific Control
40	00 - ?? 

*/

#define ES_5505_PAGE	(0x1E)

static ULONG es_5505_page;

static void es_5505_ww(ULONG address, UWORD data)
{
   switch(address&0x1E){

      case ES_5505_PAGE:
         es_5505_page = data;
         #ifdef HACKS
         print_debug("es_5505_page_write[%04x]\n", es_5505_page);
         #endif
      break;

      default:
         #ifdef HACKS
         print_debug("es_5505_write[%04x:%04x] [PC:%06x]\n", address&0x1F, data, s68000context.pc);
         #endif
      break;

   }
}

static int es_5505_rw(ULONG address)
{
   static int ta=0;

   #ifdef HACKS
      print_debug("es_5505_read[%04x] [PC:%06x]\n", address&0x1F, s68000context.pc);
   #endif

   ta+=0x5555;
   return ta;
}

/******************************************************************************/
/* ENSONIQ ES-5510 [Ensoniq Signal Processor / DSP]                           */
/******************************************************************************/

/*

Address	Register Name	Bits	Description / Function	
---------------------------------------------------------------
00	GPR        	23-16	GPR / SPR Data MS-byte	
01	GPR 		15-8	GPR / SPR  Data	
02	GPR   		7-0	GPR / SPR  Data LS-byte	
03	INSTR           47-40	Mult D Operand Address	
04	INSTR   	39-32	Mult C Operand Address	
05	INSTR 		31-24 	ALU B Operand Address	
06	INSTR    	23-16	ALU A Operand Address	
07	INSTR    	15-8	ALU Opcode / Operand Select	
08	INSTR    	7-3	Skip Flag +  MAC + RAM Control	
09	DIL 		23-16	DRAM Input Latch MS-byte	
0A	DIL 		15-8	DRAM Input Latch 	
0B	DIL		7-0	DRAM Input Latch LS-byte	 (always read as 00 )
0C	DOL 		23-16	DRAM Ouput Latch MS-byte	
0D	DOL 		15-8	DRAM Ouput Latch 	
0E	DOL		7-0	DRAM Ouput Latch LS-byte	 (always read as FF )
0F	DADR		23-16	DRAM Address MS-byte	 (write last to transfer data)
10	DADR		15-8	DRAM Address 	
11	DADR  		7-4	DRAM Address LS-byte	
12	Host Control 	2-0	ESP State / Control Flags	
14	RAM Access Control 	7-6	RAM,I/O, Read/Write Select	
16	Program Counter            		For Test Purposes Only	
17	Internal Refresh counter        		For Test Purposes Only	
18	Host serial control	7-0	Serial I/O Format and Control	
1F	Halt enable       (write only) Frame Counter (read only)		Stop ESP Execution For Test Purposes Only	

80	Read  Select:  GPR + INSTR  	7-0	Read  Address	Address Ranges:	
A0	Write Select:  GPR      	7-0	Write Address	GPR       $00 - $BF	
C0	Write Select:  INSTR      	7-0	Write Address	SPR       $EA - $FF	
E0	Write Select:  GPR+INSTR	7-0	Write Address	INSTR   $00 - $9F	

*/

static UBYTE es_5510_data[0x100];

#define ES_5510_GPR		(0x00)
#define ES_5510_INSTR		(0x03)
#define ES_5510_DIL		(0x09)
#define ES_5510_DOL		(0x0C)
#define ES_5510_DADR		(0x0F)
#define ES_5510_HOST_CTRL	(0x12)
#define ES_5510_RAM_CTRL	(0x14)
#define ES_5510_PC		(0x16)
#define ES_5510_IRC		(0x17)
#define ES_5510_HOST_SER_CTRL	(0x18)
#define ES_5510_HALT		(0x1F)
#define ES_5510_FRAME_CNT	(0x1F)

/*

Host Control
------------

bit#2 = Host Access OK (this should be zero)
bit#1 = RAM Clear (this is set to clear esp ram)
bit#0 = Refresh Disable (this should be set 0 when setting ram clear)
					1 	          0	
*/

static void es_5510_wb(ULONG address, UBYTE data)
{
   // convert to 256 byte address space

   address = (address>>1)&0xFF;

   // store the data

   es_5510_data[address] = data;

   // process the data

   switch(address){
      default:
         #ifdef HACKS
         print_debug("es_5510_write[%02x:%02x] [PC:%06x]\n", address, data, s68000context.pc);
         #endif
      break;
   }
}

static int es_5510_rb(ULONG address)
{
   static int ta=0;

   // convert to 256 byte address space

   address = (address>>1)&0xFF;

   #ifdef HACKS
   print_debug("es_5510_read[%02x] [PC:%06x]\n", address, s68000context.pc);
   #endif

   switch(address){

      case ES_5510_GPR+0:
      case ES_5510_GPR+1:
      case ES_5510_GPR+2:
         return es_5510_data[address];
      break;

      case ES_5510_DIL+0:
      case ES_5510_DIL+1:
         return es_5510_data[address];
      break;
      case ES_5510_DIL+2:
         return 0x00;
      break;

      case ES_5510_DOL+0:
      case ES_5510_DOL+1:
         return es_5510_data[address];
      break;
      case ES_5510_DOL+2:		// 0x00 or 0xFF? Manual contradicts itself!
         return 0xFF;
      break;

      case ES_5510_HOST_CTRL:
         return es_5510_data[address]&0x01;
      break;

      case ES_5510_PC:			// It is checked against $28 = 40
         return 0x00;
      break;

      default:
         ta^=0xFF;
         return ta;
      break;
   }

}

/******************************************************************************/

void GenF3Samples(UBYTE *TMP)
{
   int ta,tb,tc,samstart,samend;
   char s[32];
   UBYTE *SCB;

   #ifdef HACKS
   print_debug(
      "F3 Sample Convert\n"
      "-----------------\n"
      "\n"
   );
   #endif

   ta=0;
   s[12]=0;
   SCB=TMP+1;
   while((SCB[ta]!=0)||(SCB[ta+1]!=0)||(SCB[ta+2]!=0)){
      tc=ReadWord68k(&SCB[ta]);
      for(tb=0;tb<12;tb++){
         s[tb]=SCB[ta+tb+2];
      }
      #ifdef HACKS
      print_debug("Instrument: %s\n",s);
      #endif
      samstart=ReadLong68k(&TMP[tc+5])&0xFFFFFF;
      samend=ReadLong68k(&TMP[tc+9])&0xFFFFFF;
      #ifdef HACKS
      print_debug("Sample: %06x-%06x\n",samstart,samend);
      #endif
      ta+=14;
   }
}

void LoadPBobble2Sound(void)
{
   int ta;
   UBYTE *TMP;

   RAMSize=0x80000;

   if(!(RAM=AllocateMem(0x80000))) return;
   if(!(ROM=AllocateMem(0x80000))) return;

   if(!(TMP=AllocateMem(0x200000))) return;
 
   //if(!load_rom("e10-03.rom", TMP, 0x200000)) return;	// SAMPLES#1
   if(!load_rom("e51-04.38", TMP, 0x200000)) return;	// SAMPLES#1

   GenF3Samples(&TMP[0x000000]);
   GenF3Samples(&TMP[0x080000]);
   GenF3Samples(&TMP[0x100000]);
   GenF3Samples(&TMP[0x180000]);

   //if(!load_rom("e10-04.rom", TMP, 0x200000)) return;	// SAMPLES#2
   if(!load_rom("e51-05.41", TMP, 0x200000)) return;	// SAMPLES#2

   GenF3Samples(&TMP[0x000000]);
   GenF3Samples(&TMP[0x080000]);
   GenF3Samples(&TMP[0x100000]);
   GenF3Samples(&TMP[0x180000]);

   //if(!load_rom("e10-12.rom", RAM, 0x40000)) return;	// 68000 ROM
   if(!load_rom("e51-13.32", RAM, 0x40000)) return;	// 68000 ROM
   for(ta=0;ta<0x40000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   //if(!load_rom("e10-13.rom", RAM, 0x40000)) return;
   if(!load_rom("e51-14.33", RAM, 0x40000)) return;
   for(ta=0;ta<0x40000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x80000);

   // Starscream takes PC+SR from 0x000000 (no MMU emulation)

   memcpy(RAM+0x8000,ROM,0x08);


/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x80000);
   ByteSwap(RAM,0x80000);

   AddMemFetch(0xC00000, 0xC7FFFF, ROM+0x000000-0xC00000);	// 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0xC00000, 0xC7FFFF, NULL, ROM+0x000000);			// 68000 ROM
   AddReadByte(0x000000, 0x03FFFF, NULL, RAM+0x008000);			// 68000 RAM
   AddReadByte(0xFF8000, 0xFFFFFF, NULL, RAM+0x000000);			// 68000 RAM
   AddReadByte(0x260000, 0x2601FF, es_5510_rb, NULL);			// ES-5510
   AddReadByte(0x280000, 0x28001F, DebugReadByte, NULL);		// DEBUG COMM
   AddReadByte(0x000000, 0xFFFFFF, BadReadByte, NULL);			// <bad reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0xC00000, 0xC7FFFF, NULL, ROM+0x000000);			// 68000 ROM
   AddReadWord(0x000000, 0x03FFFF, NULL, RAM+0x008000);			// 68000 RAM
   AddReadWord(0xFF8000, 0xFFFFFF, NULL, RAM+0x000000);			// 68000 RAM
   AddReadWord(0x200000, 0x20001F, es_5505_rw, NULL);			// ES-5505 OTIS
   AddReadWord(0x000000, 0xFFFFFF, BadReadWord, NULL);			// <bad reads>
   AddReadWord(-1, -1, NULL, NULL);

   AddWriteByte(0x000000, 0x03FFFF, NULL, RAM+0x008000);		// 68000 RAM
   AddWriteByte(0xFF8000, 0xFFFFFF, NULL, RAM+0x000000);		// 68000 RAM
   AddWriteByte(0x260000, 0x2601FF, es_5510_wb, NULL);			// ES-5510
   AddWriteByte(0x280000, 0x28001F, DebugWriteByte, NULL);		// DEBUG COMM
   AddWriteByte(0x000000, 0xFFFFFF, BadWriteByte, NULL);		// <bad writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x000000, 0x03FFFF, NULL, RAM+0x008000);		// 68000 RAM
   AddWriteWord(0xFF8000, 0xFFFFFF, NULL, RAM+0x000000);		// 68000 RAM
   AddWriteWord(0x200000, 0x20001F, es_5505_ww, NULL);			// ES-5505 OTIS
   AddWriteWord(0x000000, 0xFFFFFF, BadWriteWord, NULL);		// <bad writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();	// Set Starscream mem pointers... 
}

void ClearPBobble2Sound(void)
{
   #ifdef HACKS
      save_debug("ROM.bin",ROM,0x080000,1);
      save_debug("RAM.bin",RAM,0x080000,1);
   #endif
}

void ExecutePBobble2SoundFrame(void)
{
   ULONG ta;

   cpu_execute_cycles(CPU_68K_0, CPU_FRAME_MHz(16,60));	// M68000 16MHz (60fps)

   #ifdef HACKS
      print_debug("[68000] PC:%06x SR:%04x V:%04x%04x\n",s68000context.pc,s68000context.sr,ReadWord(&RAM[0x8100]),ReadWord(&RAM[0x8102]));
   #endif

   // Hack to execute user defined vector
   // -----------------------------------

   ta = ReadLong(&RAM[0x8064]);

   WriteLong(&RAM[0x8064],ReadLong(&RAM[0x8100]));

   cpu_interrupt(CPU_68K_0, 1);

   WriteLong(&RAM[0x8064],ta);
}

void DrawPuzzleBobble2Sound(void)
{
   int ta,tx,ty;
   UBYTE tb[2];

   clear_game_screen(0);		// Game has no solid BG0

   pal[0].r=0;
   pal[0].g=0;
   pal[0].b=0;
   pal[1].r+=1;
   pal[1].g-=2;
   pal[1].b+=4;

   ta=0;
   for(ty=0;ty<240;ty+=8){
   for(tx=0;tx<320;tx+=8){
   tb[0]=EnsoniqScr[ta];
   tb[1]=0;
   textout(GameBitmap,font,tb,tx,ty,1);
   ta++;
   if(ta==0x400){tx=320;ty=140;}
   }
   }
}

/*

Elevator Action 2
-----------------

Ensoniq
OTISR2
609-0381303
FP05164
K9442

Ensoniq
5701000101
9416MVA82

Ensoniq
609-0390750
FP03457 K9420
ES5510000106
ESPR6

Motorola
MC68681P

Twin Qix
--------

Ensoniq
OTISR2
609-0381303
FP05164
K9442

Ensoniq
5701000101
9416MVA82

Ensoniq
609-0390750
FP03457 K9420
ES5510000106
ESPR6

Motorola
MC68681P

Super Chase
-----------

68000
68681
MB8421 (x2)
MB87078
Ensoniq 5510
Ensoniq 5505

*/
