/******************************************************************************/
/*                                                                            */
/*                 PUZZLE BOBBLE 4 (C) 1997 TAITO CORPORATION                 */
/*                                                                            */
/******************************************************************************/

#include "gameinc.h"
#include "f3system.h"
#include "pbobble4.h"
#include "tc003vcu.h"
#include "tc200obj.h"

static struct DIR_INFO puzzle_bobble_4_dirs[] =
{
   { "puzzle_bobble_4", },
   { "puzbob4", },
   { NULL, },
};

static struct ROM_INFO puzzle_bobble_4_roms[] =
{
   {       "e49.01", 0x00100000, 0xA3DD5F85, },
   {       "e49.02", 0x00100000, 0xC7D2F40B, },
   {       "e49.03", 0x00200000, 0xF64303E0, },
   {       "e49.04", 0x00200000, 0x09BE229C, },
   {       "e49.05", 0x00200000, 0x5CE90EE2, },
   {       "e49.06", 0x00100000, 0xEC85F7CE, },
   {       "e49.07", 0x00100000, 0x1E1E8E1C, },
   {       "e49.08", 0x00100000, 0x87408106, },
   {       "e49.10", 0x00080000, 0x0307460B, },
   {       "e49.11", 0x00080000, 0xBF69A087, },
   {       "e49.12", 0x00080000, 0xFFFEA203, },
   {       "e49.13", 0x00040000, 0x83536F7F, },
   {       "e49.14", 0x00040000, 0x19815BDB, },
   {       "e49.16", 0x00080000, 0x0A021624, },
   {           NULL,          0,          0, },
};

static struct ROMSW_DATA romsw_data_puzzle_bobble_4_0[] =
{
   { "Taito Japan (Japanese)", 0x01 },
   { "Taito America",          0x02 },
   { "Taito Worldwide",        0x03 },
   { NULL,                     0    },
};

static struct ROMSW_INFO puzzle_bobble_4_romsw[] =
{
   { 0x1FFFFF, 0x03, romsw_data_puzzle_bobble_4_0 },
   { 0,        0,    NULL },
};

static struct VIDEO_INFO puzzle_bobble_4_video =
{
   DrawPuzzleBobble4,
   320,
   232,
   64,
   VIDEO_ROTATE_NORMAL,
};

struct GAME_MAIN game_puzzle_bobble_4 =
{
   puzzle_bobble_4_dirs,
   puzzle_bobble_4_roms,
   f3_system_inputs,
   NULL,
   puzzle_bobble_4_romsw,

   LoadPuzzleBobble4,
   ClearPuzzleBobble4,
   &puzzle_bobble_4_video,
   ExecutePuzzleBobble4Frame,
   "puzbob4",
   "Puzzle Bobble 4",
   NULL,
   COMPANY_ID_TAITO,
   "E49",
   1997,
   NULL,
};

static UBYTE *RAM_BG0;
static UBYTE *RAM_BG1;
static UBYTE *RAM_BG2;
static UBYTE *RAM_BG3;

static UBYTE *RAM_SCR0;
static UBYTE *RAM_SCR1;
static UBYTE *RAM_SCR2;
static UBYTE *RAM_SCR3;

static ULONG SCR0_XOFS;
static ULONG SCR1_XOFS;
static ULONG SCR2_XOFS;
static ULONG SCR3_XOFS;

static ULONG SCR0_YOFS;
static ULONG SCR1_YOFS;
static ULONG SCR2_YOFS;
static ULONG SCR3_YOFS;

static UBYTE *GFX_BG0;
static UBYTE *GFX_BG0_SOLID;

static UBYTE *GFX_SPR;
static UBYTE *GFX_SPR_SOLID;

void LoadPuzzleBobble4(void)
{
   int ta,tb,tc;

   RAMSize=0x80000;

   if(!(RAM=AllocateMem(0x80000))) return;
   if(!(ROM=AllocateMem(0x200000))) return;
   if(!(GFX=AllocateMem(0x3FA100+0x3FA100))) return;

   GFX_BG0 = GFX+0x3FA100;
   GFX_SPR = GFX+0x000000;

   tb=0;
   if(!load_rom("E49.02", ROM, 0xFE840)) return;		// 16x16 SPRITES ($3FA1)
   for(ta=0;ta<0xFE840;ta++){
      GFX[tb++]=ROM[ta]&15;
      GFX[tb++]=ROM[ta]>>4;
      tb+=2;
   }
   tb=2;
   if(!load_rom("E49.01", ROM, 0xFE840)) return;		// 16x16 SPRITES
   for(ta=0;ta<0xFE840;ta++){
      GFX[tb++]=ROM[ta]&15;
      GFX[tb++]=ROM[ta]>>4;
      tb+=2;
   }

   tb=0;
   if(!load_rom("E49.08", ROM, 0xFE840)) return;		// 16x16 TILES ($3FA1)
   for(ta=0;ta<0xFE840;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }
   tb=4;
   if(!load_rom("E49.07", ROM, 0xFE840)) return;		// 16x16 TILES
   for(ta=0;ta<0xFE840;ta+=2){
      GFX_BG0[tb++]=ROM[ta]&15;
      GFX_BG0[tb++]=ROM[ta]>>4;
      GFX_BG0[tb++]=ROM[ta+1]&15;
      GFX_BG0[tb++]=ROM[ta+1]>>4;
      tb+=4;
   }

   tb=0;
   if(!load_rom("E49.06", ROM, 0xFE840)) return;		// 16x16 TILES (MASK)
   for(ta=0;ta<0xFE840;ta+=2){
      tc=ROM[ta];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<4;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<4;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<4;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<4;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<4;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<4;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<4;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<4;
      tc=ROM[ta+1];
      GFX_BG0[tb+7]|=((tc&0x80)>>7)<<5;
      GFX_BG0[tb+6]|=((tc&0x40)>>6)<<5;
      GFX_BG0[tb+5]|=((tc&0x20)>>5)<<5;
      GFX_BG0[tb+4]|=((tc&0x10)>>4)<<5;
      GFX_BG0[tb+3]|=((tc&0x08)>>3)<<5;
      GFX_BG0[tb+2]|=((tc&0x04)>>2)<<5;
      GFX_BG0[tb+1]|=((tc&0x02)>>1)<<5;
      GFX_BG0[tb+0]|=((tc&0x01)>>0)<<5;
      tb+=8;
   }

   GFX_BG0_SOLID = MakeSolidTileMap16x16(GFX_BG0, 0x3FA1);
   GFX_SPR_SOLID = make_solid_mask_16x16(GFX_SPR, 0x3FA1);

   // Setup 68020 Memory Map
   // ----------------------

   AddF3MemoryMap(0x200000);

   RAM_BG0=RAM+0x30000;
   RAM_BG3=RAM+0x31000;
   RAM_BG1=RAM+0x32000;
   RAM_BG2=RAM+0x33000;

   RAM_SCR0=RAM+0x6A000;
   RAM_SCR3=RAM+0x6A002;
   RAM_SCR1=RAM+0x6A004;
   RAM_SCR2=RAM+0x6A006;

   SCR0_XOFS=0xF640;
   SCR3_XOFS=0xF740;
   SCR1_XOFS=0xF840;
   SCR2_XOFS=0xF940;

   SCR0_YOFS=0xF400;
   SCR3_YOFS=0xF400;
   SCR1_YOFS=0xF400;
   SCR2_YOFS=0xF400;

   if(!load_rom("E49.12", RAM, 0x80000)) return;
   for(ta=0;ta<0x80000;ta++){
      ROM[(ta<<2)+0]=RAM[ta];
   }
   if(!load_rom("E49.11", RAM, 0x80000)) return;
   for(ta=0;ta<0x80000;ta++){
      ROM[(ta<<2)+1]=RAM[ta];
   }
   if(!load_rom("E49.10", RAM, 0x80000)) return;
   for(ta=0;ta<0x80000;ta++){
      ROM[(ta<<2)+2]=RAM[ta];
   }
   if(!load_rom("E49.16", RAM, 0x80000)) return;
   for(ta=0;ta<0x80000;ta++){
      ROM[(ta<<2)+3]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x80000);
   memset(RAM+0x69000,0xFF,0x01000);

   // PROTECTION HACK
   // ---------------

   WriteLong68k(&ROM[0x00BA96],0x001E8000);
   W24[0x1E] = ROM+0x1E0000;

   // EEPROM HACKS
   // ------------

   WriteWord68k(&ROM[0x0047AE],0x7F00);		//	raine	#$00 <read/write eeprom>
   WriteWord68k(&ROM[0x0047B0],0x4ED6);		//	jmp	(a6)


   // SPEED HACK#1
   // ------------

   WriteWord68k(&ROM[0x001500],0x4EF9);		//	jmp	$1E8100
   WriteLong68k(&ROM[0x001502],0x001E8100);	//

   WriteWord68k(&ROM[0x1E8100],0x4EB9);		//	jsr	$15EA
   WriteLong68k(&ROM[0x1E8102],0x000015EA);	//

   WriteWord68k(&ROM[0x1E8106],0x7F02);		//	raine	#$02 <stop cpu>

   WriteWord68k(&ROM[0x1E8108],0x6100-10);	//	bra.s	<loop>

   F3SystemEEPROMAccess=&F3SysEEPROMAccessMode2;

   InitPaletteMap(RAM+0x60000, 0x200, 0x40, 0x1000);

   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu.RAM	= RAM+0x20000;
   tc0003vcu.mapper	= &Map_24bit_xRGB;
   tc0003vcu.bmp_x	= 64;
   tc0003vcu.bmp_y	= 64;
   tc0003vcu.bmp_w	= 320;
   tc0003vcu.bmp_h	= 232;
   tc0003vcu.scr_x	= 0;
   tc0003vcu.scr_y	= 0;

   // Init tc0200obj emulation
   // ------------------------

   tc0200obj.RAM	= RAM+0x20000;
   tc0200obj.RAM_B	= RAM+0x28000;
   tc0200obj.GFX	= GFX_SPR;
   tc0200obj.MASK	= GFX_SPR_SOLID;
   tc0200obj.bmp_x	= 64;
   tc0200obj.bmp_y	= 64;
   tc0200obj.bmp_w	= 320;
   tc0200obj.bmp_h	= 232;
   tc0200obj.mapper	= &Map_24bit_xRGB;
   tc0200obj.tile_mask	= 0x3FFF;
   tc0200obj.ofs_x	= 0-0x2E;
   tc0200obj.ofs_y	= 0-0x18;

   tc0200obj.cols	= 16;

   init_tc0200obj();


   init_f3_system_ioc(F3_IOC_2P_3BUTTON);

   init_m68k();
}

void ClearPuzzleBobble4(void)
{
   save_eeprom();

   #ifdef HACKS
      save_debug("ROM.bin",ROM,0x200000,0);
      save_debug("RAM.bin",RAM,0x080000,0);
      //save_debug("GFX.bin",GFX,0x64AB00+0x3C2300,0);
   #endif
}

void ExecutePuzzleBobble4Frame(void)
{
/*
   print_ingame(240,"0X:%04x 1X:%04x 2X:%04x 3X:%04x",ReadWord68k(&RAM[0x6A000]),ReadWord68k(&RAM[0x6A002]),ReadWord68k(&RAM[0x6A004]),ReadWord68k(&RAM[0x6A006]));
   print_ingame(240,"0Y:%04x 1Y:%04x 2Y:%04x 3Y:%04x",ReadWord68k(&RAM[0x6A008]),ReadWord68k(&RAM[0x6A00A]),ReadWord68k(&RAM[0x6A00C]),ReadWord68k(&RAM[0x6A00E]));
   print_ingame(240,"00:%04x 00:%04x PX:%04x PY:%04x",ReadWord68k(&RAM[0x6A010]),ReadWord68k(&RAM[0x6A012]),ReadWord68k(&RAM[0x6A014]),ReadWord68k(&RAM[0x6A016]));
   print_ingame(240,"00:%04x 00:%04x 00:%04x ??:%04x",ReadWord68k(&RAM[0x6A018]),ReadWord68k(&RAM[0x6A01A]),ReadWord68k(&RAM[0x6A01C]),ReadWord68k(&RAM[0x6A01E]));

   print_ingame(240,"-------------------------------");

   print_ingame(240,"%04x %04x %04x %04x",ReadWord68k(&RAM[0x687F8]),ReadWord68k(&RAM[0x687FA]),ReadWord68k(&RAM[0x687FC]),ReadWord68k(&RAM[0x687FE]));
   print_ingame(240,"%04x %04x %04x %04x",ReadWord68k(&RAM[0x40000]),ReadWord68k(&RAM[0x40200]),ReadWord68k(&RAM[0x40400]),ReadWord68k(&RAM[0x40600]));
   print_ingame(240,"%04x %04x %04x %04x",ReadWord68k(&RAM[0x40800]),ReadWord68k(&RAM[0x40A00]),ReadWord68k(&RAM[0x40C00]),ReadWord68k(&RAM[0x40E00]));
   print_ingame(240,"%04x",ReadWord68k(&RAM[0x41000]));
*/
   Execute68020(800000);
   /*#ifdef HACKS
      print_debug("PC:%06x SR:%04x A7:%06x\n",regs.pc,regs.sr,m68k_areg(regs,7));
   #endif*/
   Interrupt68020(2);
   Interrupt68020(5);
   Interrupt68020(3);

   IntF3System();
}

static int BGPRI[3];

void DrawPuzzleBobble4(void)
{
   UBYTE *MAP;
   int x,y,z,ta,zz,zzz,zzzz,x16,y16;
   UWORD d0;

   ClearPaletteMap();

   // Init tc0003vcu emulation
   // ------------------------

   tc0003vcu_layer_count = 0;

   d0=0;

   if(check_layer_enabled(f3_bg0_id)){
   MAKE_SCROLL_512x512_4_16(
      (ReadWord68k(&RAM_SCR0[0])-SCR0_XOFS)>>6,
      (ReadWord68k(&RAM_SCR0[8])-SCR0_YOFS)>>7
   );

   START_SCROLL_512x512_4_16(64,64,320,232);

      if(d0==0) d0=ReadWord68k(&RAM_BG0[zz])&0x3E00;

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG0[zz])&0x1FF,
               64
            );

      switch(RAM_BG0[zz]&0xC0){
      case 0x00: Draw16x16_Mapped(&GFX_BG0[(ReadWord68k(&RAM_BG0[zz+2])&0x3FFF)<<8],x,y,MAP);        break;
      case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[(ReadWord68k(&RAM_BG0[zz+2])&0x3FFF)<<8],x,y,MAP);  break;
      case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[(ReadWord68k(&RAM_BG0[zz+2])&0x3FFF)<<8],x,y,MAP);  break;
      case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[(ReadWord68k(&RAM_BG0[zz+2])&0x3FFF)<<8],x,y,MAP); break;
      }

   END_SCROLL_512x512_4_16();
   }

   BGPRI[0]=d0;
   d0=0;

   if(check_layer_enabled(f3_bg1_id)){
   MAKE_SCROLL_512x512_4_16(
      (ReadWord68k(&RAM_SCR1[0])-SCR1_XOFS)>>6,
      (ReadWord68k(&RAM_SCR1[8])-SCR1_YOFS)>>7
   );

   START_SCROLL_512x512_4_16(64,64,320,232);

   ta=ReadWord68k(&RAM_BG1[zz+2])&0x3FFF;
   if(ta!=0){

      if(d0==0) d0=ReadWord68k(&RAM_BG1[zz])&0x3E00;

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG1[zz])&0x1FF,
               64
            );

      if(GFX_BG0_SOLID[ta]==0){
         switch(RAM_BG1[zz]&0xC0){
         case 0x00: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      else{
         switch(RAM_BG1[zz]&0xC0){
         case 0x00: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
   }

   END_SCROLL_512x512_4_16();
   }

   BGPRI[1]=d0;
   d0=0;

   if(check_layer_enabled(f3_bg2_id)){
   MAKE_SCROLL_512x512_4_16(
      (ReadWord68k(&RAM_SCR2[0])-SCR2_XOFS)>>6,
      (ReadWord68k(&RAM_SCR2[8])-SCR2_YOFS)>>7
   );

   START_SCROLL_512x512_4_16(64,64,320,232);

   ta=ReadWord68k(&RAM_BG2[zz+2])&0x3FFF;
   if(ta!=0){

      if(d0==0) d0=ReadWord68k(&RAM_BG2[zz])&0x3E00;

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG2[zz])&0x1FF,
               64
            );

      if(GFX_BG0_SOLID[ta]==0){
         switch(RAM_BG2[zz]&0xC0){
         case 0x00: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      else{
         switch(RAM_BG2[zz]&0xC0){
         case 0x00: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
   }

   END_SCROLL_512x512_4_16();
   }

   BGPRI[2]=d0;
   d0=0;

   if(check_layer_enabled(f3_bg3_id)){
   MAKE_SCROLL_512x512_4_16(
      (ReadWord68k(&RAM_SCR3[0])-SCR3_XOFS)>>6,
      (ReadWord68k(&RAM_SCR3[8])-SCR3_YOFS)>>7
   );

   START_SCROLL_512x512_4_16(64,64,320,232);

   ta=ReadWord68k(&RAM_BG3[zz+2])&0x3FFF;
   if(ta!=0){

      if(d0==0) d0=ReadWord68k(&RAM_BG3[zz])&0x3E00;

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               ReadWord68k(&RAM_BG3[zz])&0x1FF,
               64
            );

      if(GFX_BG0_SOLID[ta]==0){
         switch(RAM_BG3[zz]&0xC0){
         case 0x00: Draw16x16_Trans_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Trans_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Trans_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Trans_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
      else{
         switch(RAM_BG3[zz]&0xC0){
         case 0x00: Draw16x16_Mapped(&GFX_BG0[ta<<8],x,y,MAP);        break;
         case 0x40: Draw16x16_Mapped_FlipY(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0x80: Draw16x16_Mapped_FlipX(&GFX_BG0[ta<<8],x,y,MAP);  break;
         case 0xC0: Draw16x16_Mapped_FlipXY(&GFX_BG0[ta<<8],x,y,MAP); break;
         }
      }
   }

   END_SCROLL_512x512_4_16();
   }

   BGPRI[3]=d0;

   render_tc0200obj_mapped_f3system_b();

   if(check_layer_enabled(f3_bg5_id)){
      //memset(RAM+0x3E000+(5<<5),0x20,0x00);
      //f3video_render_fg0();

   zz=0;
   for(y=64;y<232+64;y+=8){
   for(x=64;x<320+64;x+=8){
      z=RAM[0x3C001+zz];
      if((z!=0)&&(z!=5)){		// 5='black' transparent shade

      MAP = MAP_PALETTE(
               Map_24bit_xRGB,
               (RAM[0x3C000+zz]>>1)&0x3F,
               16
            );

         if((RAM[0x3C000+zz]&0x01)!=0){
            Draw8x8_Trans_Packed_Mapped(&RAM[0x3E000+(z<<5)],x,y,MAP);
         }
         else{
            Draw8x8_Trans_Packed_Mapped_FlipY(&RAM[0x3E000+(z<<5)],x,y,MAP);
         }

      }
      zz+=2;
   }
   zz+=(128-80);
   }
   }

//     print_ingame(1,"$$$: %04x %04x %04x %04x",BGPRI[0],BGPRI[1],BGPRI[2],BGPRI[3]);

     if((BGPRI[0]!=0)||(BGPRI[1]!=0)||(BGPRI[2]!=0)||(BGPRI[3]!=0))
     {
       RAM_BG0=RAM+0x33000;
       RAM_BG2=RAM+0x31000;
       RAM_BG3=RAM+0x30000;

       RAM_SCR0=RAM+0x6A006;
       RAM_SCR2=RAM+0x6A002;
       RAM_SCR3=RAM+0x6A000;

       SCR3_XOFS=0xF640;
       SCR2_XOFS=0xF740;
       SCR0_XOFS=0xF940;

     }
     else
     {
       RAM_BG0=RAM+0x30000;
       RAM_BG3=RAM+0x31000;
       RAM_BG2=RAM+0x33000;

       RAM_SCR0=RAM+0x6A000;
       RAM_SCR3=RAM+0x6A002;
       RAM_SCR2=RAM+0x6A006;

       SCR0_XOFS=0xF640;
       SCR3_XOFS=0xF740;
       SCR2_XOFS=0xF940;

     }

}

/*

'Raven' Romlist
---------------

File   | Contents
-------+----------------------------------
pb4.17 | 68020 3
pb4.18 | 68020 2
pb4.19 | 68020 1
pb4.20 | 68020 0
pb4.32 | 68000 1
pb4.33 | 68000 0
pb4.45 | Tiles odd
pb4.47 | Tiles even
pb4.38 | Samples 0 (missing odd bytes)
pb4.39 | Samples 1 (missing odd bytes)
pb4.40 | Samples 2 (missing odd bytes)
pb4.8  | Sprites odd  (missing odd bytes)
pb4.12 | Sprites even (missing odd bytes)

What's missing:

- Odd bytes on 5 of the roms
- 1 Mask rom for tiles

*/
