/******************************************************************************/
/*                                                                            */
/*                  SNOW BROS/WINTER BOBBLE (C) 1990 TOAPLAN                  */
/*                                                                            */
/******************************************************************************/

#include "gameinc.h"
#include "snowbros.h"
#include "3812intf.h"
#include "ym3812x.h"

static struct DIR_INFO snow_bros_dirs[] =
{
   { "snow_bros", },
   { "snowbros", },
   { NULL, },
};

static struct ROM_INFO snow_bros_roms[] =
{
   {      "sn5.bin", 0x00020000, 0xad310d3f, }, // 0
   {      "sn6.bin", 0x00020000, 0x4899ddcf, }, // 1
   {   "snowbros.4", 0x00008000, 0xE6EAB4E4, }, // 2
   {          "ch0", 0x00020000, 0x36D84DFE, }, // 3
   {          "ch1", 0x00020000, 0x76347256, }, // 4
   {          "ch2", 0x00020000, 0xFDAA634C, }, // 5
   {          "ch3", 0x00020000, 0x34024AEF, }, // 6
   {           NULL,          0,          0, },
};

static struct INPUT_INFO snow_bros_inputs[] =
{
   { KB_DEF_COIN1,        MSG_COIN1,               0x006405, 0x04, BIT_ACTIVE_0 },
   { KB_DEF_COIN2,        MSG_COIN2,               0x006405, 0x08, BIT_ACTIVE_0 },
   { KB_DEF_TILT,         MSG_TILT,                0x006405, 0x20, BIT_ACTIVE_0 },
   { KB_DEF_P1_START,     MSG_P1_START,            0x006405, 0x01, BIT_ACTIVE_0 },
   { KB_DEF_P1_UP,        MSG_P1_UP,               0x006401, 0x01, BIT_ACTIVE_0 },
   { KB_DEF_P1_DOWN,      MSG_P1_DOWN,             0x006401, 0x02, BIT_ACTIVE_0 },
   { KB_DEF_P1_LEFT,      MSG_P1_LEFT,             0x006401, 0x04, BIT_ACTIVE_0 },
   { KB_DEF_P1_RIGHT,     MSG_P1_RIGHT,            0x006401, 0x08, BIT_ACTIVE_0 },
   { KB_DEF_P1_B1,        MSG_P1_B1,               0x006401, 0x10, BIT_ACTIVE_0 },
   { KB_DEF_P1_B2,        MSG_P1_B2,               0x006401, 0x20, BIT_ACTIVE_0 },
   { KB_DEF_P2_START,     MSG_P2_START,            0x006405, 0x02, BIT_ACTIVE_0 },
   { KB_DEF_P2_UP,        MSG_P2_UP,               0x006403, 0x01, BIT_ACTIVE_0 },
   { KB_DEF_P2_DOWN,      MSG_P2_DOWN,             0x006403, 0x02, BIT_ACTIVE_0 },
   { KB_DEF_P2_LEFT,      MSG_P2_LEFT,             0x006403, 0x04, BIT_ACTIVE_0 },
   { KB_DEF_P2_RIGHT,     MSG_P2_RIGHT,            0x006403, 0x08, BIT_ACTIVE_0 },
   { KB_DEF_P2_B1,        MSG_P2_B1,               0x006403, 0x10, BIT_ACTIVE_0 },
   { KB_DEF_P2_B2,        MSG_P2_B2,               0x006403, 0x20, BIT_ACTIVE_0 },
   { 0,                   NULL,                    0,        0,    0            },
};

static struct DSW_DATA dsw_data_snow_bros_0[] =
{
   { "Country",               0x01, 0x02 },
   { "America",               0x01, 0x00 },
   { "Europe",                0x00, 0x00 },
   { MSG_SCREEN,              0x02, 0x02 },
   { MSG_NORMAL,              0x02, 0x00 },
   { MSG_INVERT,              0x00, 0x00 },
   { MSG_TEST_MODE,           0x04, 0x02 },
   { MSG_OFF,                 0x04, 0x00 },
   { MSG_ON,                  0x00, 0x00 },
   { MSG_DEMO_SOUND,          0x08, 0x02 },
   { MSG_ON,                  0x08, 0x00 },
   { MSG_OFF,                 0x00, 0x00 },
   { MSG_COIN1,               0x30, 0x04 },
   { "1/1",                   0x30, 0x00 },
   { "1/2; 2/1",              0x20, 0x00 },
   { "2/1; 3/1",              0x10, 0x00 },
   { "2/3; 4/1",              0x00, 0x00 },
   { MSG_COIN2,               0xC0, 0x04 },
   { "1/1",                   0xC0, 0x00 },
   { "1/2; 1/3",              0x80, 0x00 },
   { "2/1; 1/4",              0x40, 0x00 },
   { "2/3; 1/6",              0x00, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_DATA dsw_data_snow_bros_1[] =
{
   { MSG_DIFFICULTY,          0x03, 0x04 },
   { MSG_NORMAL,              0x03, 0x00 },
   { MSG_EASY,                0x02, 0x00 },
   { MSG_HARD,                0x01, 0x00 },
   { MSG_HARDEST,             0x00, 0x00 },
   { "Extra Life at",         0x0C, 0x04 },
   { "100k",                  0x0C, 0x00 },
   { "200k",                  0x08, 0x00 },
   { "100k 200k",             0x04, 0x00 },
   { "Nothing",               0x00, 0x00 },
   { "Lives",                 0x30, 0x04 },
   { "3",                     0x30, 0x00 },
   { "1",                     0x20, 0x00 },
   { "4",                     0x10, 0x00 },
   { "2",                     0x00, 0x00 },
   { "Cheat",                 0x40, 0x02 },
   { MSG_OFF,                 0x40, 0x00 },
   { MSG_ON,                  0x00, 0x00 },
   { "Continue",              0x80, 0x02 },
   { MSG_ON,                  0x80, 0x00 },
   { MSG_OFF,                 0x00, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO snow_bros_dsw[] =
{
   { 0x006400, 0xFF, dsw_data_snow_bros_0 },
   { 0x006402, 0xFF, dsw_data_snow_bros_1 },
   { 0,        0,    NULL,      },
};

static struct VIDEO_INFO snow_bros_video =
{
   DrawSnowBros,
   256,
   224,
   32,
   VIDEO_ROTATE_NORMAL |
   VIDEO_ROTATABLE,
};

static struct YM3812interface ym3812_interface =
{
   1,              // 1 chip
   3600000,        // 3.6 MHz
   { 255 },        // Volume (emu only)
    NULL
};

static struct SOUND_INFO snow_bros_sound[] =
{
   { SOUND_YM3812,  &ym3812_interface,    },
   { 0,             NULL,                 },
};

struct GAME_MAIN game_snow_bros =
{
   snow_bros_dirs,
   snow_bros_roms,
   snow_bros_inputs,
   snow_bros_dsw,
   NULL,

   LoadSnowBros,
   ClearSnowBros,
   &snow_bros_video,
   ExecuteSnowBrosFrame,
   "snowbros",
   "Snow Bros",
   "Xm[uU[Y American",
   COMPANY_ID_TOAPLAN,
   NULL,
   1990,
   snow_bros_sound,
};

static struct DIR_INFO snow_bros_alt_dirs[] =
{
   { "snow_bros_alt", },
   { "snowbroa", },
   { ROMOF("snowbros"), },
   { CLONEOF("snowbros"), },
   { NULL, },
};

static struct ROM_INFO snow_bros_alt_roms[] =
{
   {  "snowbros.2a", 0x00020000, 0xAB91CC1E, },
   {  "snowbros.3a", 0x00020000, 0x10CB37E1, },
   {   "snowbros.4", 0x00008000, 0xE6EAB4E4, },
   {          "ch0", 0x00020000, 0x36D84DFE, },
   {          "ch1", 0x00020000, 0x76347256, },
   {          "ch2", 0x00020000, 0xFDAA634C, },
   {          "ch3", 0x00020000, 0x34024AEF, },
   {           NULL,          0,          0, },
};

struct GAME_MAIN game_snow_bros_alt =
{
   snow_bros_alt_dirs,
   snow_bros_alt_roms,
   snow_bros_inputs,
   snow_bros_dsw,
   NULL,

   LoadSnowBrosAlt,
   ClearSnowBros,
   &snow_bros_video,
   ExecuteSnowBrosFrame,
   "snowbroa",
   "Snow Bros (alternate)",
   "Xm[uU[Y (alternate)",
   COMPANY_ID_TOAPLAN,
   NULL,
   1990,
   snow_bros_sound,
};

static struct DIR_INFO snow_bros_japanese_dirs[] =
{
   { "snow_bros_japanese", },
   { "snowbroj", },
   { ROMOF("snowbros"), },
   { CLONEOF("snowbros"), },
   { NULL, },
};

static struct ROM_INFO snow_bros_japanese_roms[] =
{
   {   "snowbros.2", 0x00020000, 0x854b02bc, },
   {   "snowbros.3", 0x00020000, 0x3f504f9e, },
   {   "snowbros.4", 0x00008000, 0xE6EAB4E4, },
   {          "ch0", 0x00020000, 0x36D84DFE, },
   {          "ch1", 0x00020000, 0x76347256, },
   {          "ch2", 0x00020000, 0xFDAA634C, },
   {          "ch3", 0x00020000, 0x34024AEF, },
   {           NULL,          0,          0, },
};

struct GAME_MAIN game_snow_bros_japanese =
{
   snow_bros_japanese_dirs,
   snow_bros_japanese_roms,
   snow_bros_inputs,
   snow_bros_dsw,
   NULL,

   LoadSnowBrosJ,
   ClearSnowBros,
   &snow_bros_video,
   ExecuteSnowBrosFrame,
   "snowbroj",
   "Snow Bros (Japanese)",
   "Xm[uU[Y",
   COMPANY_ID_TOAPLAN,
   NULL,
   1990,
   snow_bros_sound,
};

static struct DIR_INFO snow_bros_alt_2_dirs[] =
{
   { "snow_bros_alt_2", },
   { "snowbra2", },
   { ROMOF("snowbros"), },
   { CLONEOF("snowbros"), },
   { NULL, },
};

static struct ROM_INFO snow_bros_alt_2_roms[] =
{
   {     "sbros2-a", 0x00020000, 0xf6689f41, },
   {     "sbros3-a", 0x00020000, 0x301627d6, },
   {       "sbros4", 0x00008000, 0xE6EAB4E4, },
   {       "sbros1", 0x00080000, 0x16f06b3a, },
   {           NULL,          0,          0, },
};

struct GAME_MAIN game_snow_bros_alt_2 =
{
   snow_bros_alt_2_dirs,
   snow_bros_alt_2_roms,
   snow_bros_inputs,
   snow_bros_dsw,
   NULL,

   LoadSnowBrosAlt_2,
   ClearSnowBros,
   &snow_bros_video,
   ExecuteSnowBrosFrame,
   "snowbra2",
   "Snow Bros (alternate 2)",
   "Xm[uU[Y (alternate 2)",
   COMPANY_ID_TOAPLAN,
   "MIN16-02",
   1990,
   snow_bros_sound,
};

static struct DIR_INFO winter_bobble_dirs[] =
{
   { "winter_bobble", },
   { "wintbob", },
   { CLONEOF("snowbros"), },
   { NULL, },
};

static struct ROM_INFO winter_bobble_roms[] =
{
   {     "wb01.bin", 0x00010000, 0x05722F17, },
   {     "wb03.bin", 0x00010000, 0xDF56E168, },
   {     "wb05.bin", 0x00010000, 0x53FE59DF, },
   {     "wb06.bin", 0x00010000, 0x68204937, },
   {     "wb07.bin", 0x00010000, 0x53F40978, },
   {     "wb08.bin", 0x00010000, 0x9497B88C, },
   {     "wb09.bin", 0x00010000, 0x9BE718CA, },
   {     "wb10.bin", 0x00010000, 0x5FA22B1E, },
   {     "wb11.bin", 0x00010000, 0x41CB4563, },
   {     "wb12.bin", 0x00010000, 0xEF4E04C7, },
   {     "wb13.bin", 0x00010000, 0x426921DE, },
   {     "wb02.bin", 0x00010000, 0xFC8E292E, },
   {     "wb04.bin", 0x00010000, 0x53BE758D, },
   {           NULL,          0,          0, },
};

struct GAME_MAIN game_winter_bobble =
{
   winter_bobble_dirs,
   winter_bobble_roms,
   snow_bros_inputs,
   snow_bros_dsw,
   NULL,

   LoadWinterBobble,
   ClearSnowBros,
   &snow_bros_video,
   ExecuteSnowBrosFrame,
   "wintbob",
   "Winter Bobble",
   "Xm[uU[YiCj",
   COMPANY_ID_BOOTLEG,
   NULL,
   1990,
   snow_bros_sound,
};

static UBYTE *RAM_SPR;
static UBYTE *GFX_SPR;
static UBYTE *GFX_SPR_SOLID;

static int sport=0;

static void SoundWrite(ULONG offset, UBYTE data)
{
   sport = data;

   #ifdef HACKS
   print_debug("68000 Sends $%02x\n",sport);
   #endif
   cpu_int_nmi(CPU_Z80_0);
}

static UWORD SoundRead(ULONG offset)
{
   return 3;           // Z80_OK
}

void SnowBrosPort4w(UWORD address, UBYTE data)
{
   sport=data;
}

UWORD SnowBrosPort4r(UWORD address)
{
   #ifdef HACKS
   print_debug("Z80 Receives $%02x\n",sport);
   #endif
   return(sport);
}

static void LoadActual(int romset)
{
   UBYTE *TMP;
   int ta,tb;

   RAMSize=0x8000+0x10000;

   if(!(ROM=AllocateMem(0x40000))) return;
   if(!(RAM=AllocateMem(RAMSize))) return;      // $6420
   if(!(GFX=AllocateMem(0x100000))) return;
   if(!(TMP=AllocateMem(0x80000))) return;

   if(romset<4){       // SNOW BROS
   if(!load_rom_index(1, TMP+0x00000, 0x20000)) return;             // 68000 EVEN
   if(!load_rom_index(0, TMP+0x20000, 0x20000)) return;             // 68000 ODD
   }
   else{              // WINTER BOBBLE
   if(!load_rom_index(1, TMP+0x00000, 0x10000)) return;            // 68000 EVEN
   if(!load_rom("wb04.bin", TMP+0x10000, 0x10000)) return;         // 68000 EVEN
   if(!load_rom_index(0, TMP+0x20000, 0x10000)) return;            // 68000 ODD
   if(!load_rom("wb02.bin", TMP+0x30000, 0x10000)) return;         // 68000 ODD
   }

   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta+0]=TMP[ta+0x00000];
      ROM[ta+ta+1]=TMP[ta+0x20000];
   }

   Z80ROM=RAM+0x8000;
   if(!load_rom_index(2, Z80ROM, 0x8000)) return;          // Z80 SOUND ROM

   memset(Z80ROM+0x8000,0x00,0x8000);

   // Fix Checksum
   // ------------

   Z80ROM[0x0156]=0x00; // NOP
   Z80ROM[0x0157]=0x00; // NOP
   Z80ROM[0x0158]=0x00; // NOP

   // Apply Speed Patch
   // -----------------

   if(romset<4){
   Z80ROM[0x019D]=0xD3; // OUTA (AAh)
   Z80ROM[0x019E]=0xAA; //

   SetStopZ80Mode2(0x019C);
   }
   else{
   Z80ROM[0x0181]=0xD3; // OUTA (AAh)
   Z80ROM[0x0182]=0xAA; //

   SetStopZ80Mode2(0x0180);
   }

   // Setup Z80 memory map
   // --------------------

   AddZ80AROMBase(Z80ROM, 0x0038, 0x0066);

   AddZ80AReadByte(0x0000, 0x87FF, NULL,                        Z80ROM+0x0000); // Z80 ROM/RAM
   AddZ80AReadByte(0x0000, 0xFFFF, DefBadReadZ80,               NULL);
   AddZ80AReadByte(    -1,     -1, NULL,                        NULL);

   AddZ80AWriteByte(0x8000, 0x87FF, NULL,                       Z80ROM+0x8000); // Z80 RAM
   AddZ80AWriteByte(0x0000, 0xFFFF, DefBadWriteZ80,             NULL);
   AddZ80AWriteByte(    -1,     -1, NULL,                       NULL);

   AddZ80AReadPort(0x02, 0x03, YM3812ReadZ80,           NULL);
   AddZ80AReadPort(0x04, 0x04, SnowBrosPort4r,          NULL);
   AddZ80AReadPort(0x00, 0xFF, DefBadReadZ80,           NULL);
   AddZ80AReadPort(  -1,   -1, NULL,                    NULL);

   AddZ80AWritePort(0x02, 0x03, YM3812WriteZ80,         NULL);
   AddZ80AWritePort(0x04, 0x04, SnowBrosPort4w,         NULL);
   AddZ80AWritePort(0xAA, 0xAA, StopZ80Mode2,           NULL);
   AddZ80AWritePort(0x00, 0xFF, DefBadWriteZ80,         NULL);
   AddZ80AWritePort(  -1,   -1, NULL,                   NULL);

   AddZ80AInit();

   if(romset<4){       // SNOWBROS

   if(romset<3){

   if(!load_rom("Ch0", TMP+0x00000, 0x20000)) return;
   if(!load_rom("Ch1", TMP+0x20000, 0x20000)) return;
   if(!load_rom("Ch2", TMP+0x40000, 0x20000)) return;
   if(!load_rom("Ch3", TMP+0x60000, 0x20000)) return;

   }
   else{

   if(!load_rom("sbros1", TMP, 0x80000)) return;

   }

   tb=0;
   for(ta=0;ta<0x80000;ta+=4){
      GFX[tb+0]=TMP[ta+0]>>4;
      GFX[tb+1]=TMP[ta+0]&15;
      GFX[tb+2]=TMP[ta+1]>>4;
      GFX[tb+3]=TMP[ta+1]&15;
      GFX[tb+4]=TMP[ta+2]>>4;
      GFX[tb+5]=TMP[ta+2]&15;
      GFX[tb+6]=TMP[ta+3]>>4;
      GFX[tb+7]=TMP[ta+3]&15;
      tb+=16;
      if((tb&0x7F)==0){tb-=0x78;}
      else{if((tb&0x7F)==8){tb-=0x8;}}
   }

   }
   else{                // WINTER BOBBLE

   if(!load_rom("wb06.bin", TMP+0x00000, 0x10000)) return;
   if(!load_rom("wb07.bin", TMP+0x10000, 0x10000)) return;
   if(!load_rom("wb08.bin", TMP+0x20000, 0x10000)) return;
   if(!load_rom("wb09.bin", TMP+0x30000, 0x10000)) return;
   if(!load_rom("wb13.bin", TMP+0x40000, 0x10000)) return;
   if(!load_rom("wb12.bin", TMP+0x50000, 0x10000)) return;
   if(!load_rom("wb11.bin", TMP+0x60000, 0x10000)) return;
   if(!load_rom("wb10.bin", TMP+0x70000, 0x10000)) return;

   tb=0;
   for(ta=0;ta<0x40000;ta++){
      GFX[tb+1]=TMP[ta+0x00000]>>4;
      GFX[tb+0]=TMP[ta+0x00000]&15;
      GFX[tb+3]=TMP[ta+0x40000]>>4;
      GFX[tb+2]=TMP[ta+0x40000]&15;
      tb+=4;
   }

   }

   FreeMem(TMP);

   RAM_SPR = RAM+0x004000;
   GFX_SPR = GFX+0x000000;
   GFX_SPR_SOLID = make_solid_mask_16x16(GFX_SPR, 0x1000);

   memset(RAM+0x0000,0x00,0x6404);
   memset(RAM+0x6404,0xFF,0x001C);

   RAM[0x6400]=0x7F;
   RAM[0x6402]=0x7F;

   RAM[0x6412]=0x00;    // $300000: Sound Related
   RAM[0x6413]=0x03;

   // fix checksum

   WriteLong68k(&ROM[0x008EE],0x4E714E71);
   WriteLong68k(&ROM[0x008FE],0x4E714E71);

   // speed hack

   WriteLong68k(&ROM[0x003BA],0x027C7BFF);
   WriteLong68k(&ROM[0x003BE],0x13FC0000);
   WriteLong68k(&ROM[0x003C2],0x00AA0000);

   WriteLong68k(&ROM[0x00436],0x027C7AFF);
   WriteLong68k(&ROM[0x0043A],0x13FC0000);
   WriteLong68k(&ROM[0x0043E],0x00AA0000);

   WriteLong68k(&ROM[0x0045E],0x027C79FF);
   WriteLong68k(&ROM[0x00462],0x13FC0000);
   WriteLong68k(&ROM[0x00466],0x00AA0000);

   if(romset==4){               // WINTER BOBBLE
      WriteWord68k(&ROM[0x00DE],0x4EB9);
      WriteLong68k(&ROM[0x00E0],0x0000065A);

      WriteWord68k(&ROM[0x0142],0x4EB9);
      WriteLong68k(&ROM[0x0144],0x0000066C);

      WriteWord68k(&ROM[0x08B8],0x4EB9);
      WriteLong68k(&ROM[0x08BA],0x000009F4);
   }

   InitPaletteMap(RAM+0x06000, 0x10, 0x10, 0x8000);


/*
 *  StarScream Main 68000 Setup
 */

   ByteSwap(ROM,0x40000);
   ByteSwap(RAM,0x6420);

   AddMemFetch(0x000000, 0x03FFFF, ROM+0x000000-0x000000);
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadByte(0x100000, 0x103FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadByte(0x700000, 0x701FFF, NULL, RAM+0x004000);                 // SPRITE RAM
   AddReadByte(0x600000, 0x6003FF, NULL, RAM+0x006000);                 // COLOUR RAM
   AddReadByte(0x500000, 0x50000F, NULL, RAM+0x006400);                 // INPUT
   AddReadByte(0x200000, 0x200001, NULL, RAM+0x006410);                 // ???
   AddReadByte(0x300000, 0x300001, SoundRead, NULL);                    // SOUND COMM
   AddReadByte(0x400000, 0x400001, NULL, RAM+0x006414);                 // ???
   AddReadByte(0x800000, 0x800001, NULL, RAM+0x006416);                 // (Interrupt Repsonse)
   AddReadByte(0x900000, 0x900001, NULL, RAM+0x006418);                 // (Interrupt Repsonse)
   AddReadByte(0xA00000, 0xA00001, NULL, RAM+0x00641A);                 // (Interrupt Repsonse)
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);		// <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadWord(0x100000, 0x103FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadWord(0x700000, 0x701FFF, NULL, RAM+0x004000);
   AddReadWord(0x600000, 0x6003FF, NULL, RAM+0x006000);
   AddReadWord(0x500000, 0x50000F, NULL, RAM+0x006400);
   AddReadWord(0x200000, 0x200001, NULL, RAM+0x006410);
   AddReadWord(0x300000, 0x300001, SoundRead, NULL);
   AddReadWord(0x400000, 0x400001, NULL, RAM+0x006414);
   AddReadWord(0x800000, 0x800001, NULL, RAM+0x006416);
   AddReadWord(0x900000, 0x900001, NULL, RAM+0x006418);
   AddReadWord(0xA00000, 0xA00001, NULL, RAM+0x00641A);
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);		// <Bad Reads>
   AddReadWord(-1, -1, NULL, NULL);

   AddWriteByte(0x100000, 0x103FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteByte(0x700000, 0x701FFF, NULL, RAM+0x004000);
   AddWriteByte(0x600000, 0x6003FF, NULL, RAM+0x006000);
   AddWriteByte(0x500000, 0x50000F, NULL, RAM+0x006400);
   AddWriteByte(0x200000, 0x200001, NULL, RAM+0x006410);
   AddWriteByte(0x300000, 0x300001, SoundWrite, NULL);
   AddWriteByte(0x400000, 0x400001, NULL, RAM+0x006414);
   AddWriteByte(0x800000, 0x800001, NULL, RAM+0x006416);
   AddWriteByte(0x900000, 0x900001, NULL, RAM+0x006418);
   AddWriteByte(0xA00000, 0xA00001, NULL, RAM+0x00641A);
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);                   // Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);		// <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x100000, 0x103FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteWord(0x700000, 0x701FFF, NULL, RAM+0x004000);
   AddWriteWord(0x600000, 0x6003FF, NULL, RAM+0x006000);
   AddWriteWord(0x500000, 0x50000F, NULL, RAM+0x006400);
   AddWriteWord(0x200000, 0x200001, NULL, RAM+0x006410);
   AddWriteWord(0x300000, 0x300001, SoundWrite, NULL);
   AddWriteWord(0x400000, 0x400001, NULL, RAM+0x006414);
   AddWriteWord(0x800000, 0x800001, NULL, RAM+0x006416);
   AddWriteWord(0x900000, 0x900001, NULL, RAM+0x006418);
   AddWriteWord(0xA00000, 0xA00001, NULL, RAM+0x00641A);
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);		// <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();     // Set Starscream mem pointers...
}

void LoadSnowBros(void)
{
   LoadActual(0);
}

void LoadSnowBrosAlt(void)
{
   LoadActual(1);
}

void LoadSnowBrosJ(void)
{
   LoadActual(2);
}

void LoadSnowBrosAlt_2(void)
{
   LoadActual(3);
}

void LoadWinterBobble(void)
{
   LoadActual(4);
}

void ClearSnowBros(void)
{
   #ifdef HACKS
      //save_debug("ROM.bin",ROM,0x040000,1);
      //save_debug("RAM.bin",RAM,0x018000,1);
   #endif
}

void ExecuteSnowBrosFrame(void)
{
   static int ta;

   cpu_execute_cycles(CPU_68K_0, 80000);          		 // Main 68000
   cpu_interrupt(CPU_68K_0, 4);
   cpu_execute_cycles(CPU_68K_0, 80000);
   cpu_interrupt(CPU_68K_0, 3);
   cpu_execute_cycles(CPU_68K_0, 80000);
   cpu_interrupt(CPU_68K_0, 2);

   ta++;

   cpu_execute_cycles(CPU_Z80_0, CPU_FRAME_MHz(4,60));        // Sound Z80
   /*#ifdef HACKS
   print_debug("%04x\n",z80pc);
   #endif*/
   cpu_interrupt(CPU_Z80_0, 0x38);
   if(ta&1) cpu_interrupt(CPU_Z80_0, 0x38);
}

void DrawSnowBros(void)
{
   int x,y,ta,tb;
   int zz;
   UBYTE *map;

   ClearPaletteMap();

   clear_game_screen(0);

   x=0;
   y=0;

   for(zz=0;zz<0x1E00;zz+=16){

      tb = RAM_SPR[zz+6];

      if(tb&4){
         x = (x + ((RAM_SPR[zz+ 8]) | ((tb&1)<<8))) & 0x1FF;
         y = (y + ((RAM_SPR[zz+10]) | ((tb&2)<<7))) & 0x1FF;
      }
      else{
         x = (32 + ((RAM_SPR[zz+ 8]) | ((tb&1)<<8))) & 0x1FF;
         y = (16 + ((RAM_SPR[zz+10]) | ((tb&2)<<7))) & 0x1FF;
      }

      if((x>16)&&(y>16)&&(x<256+32)&&(y<240+32)){

         ta = ((RAM_SPR[zz+12]) | (RAM_SPR[zz+14]<<8))&0x0FFF;

         if(GFX_SPR_SOLID[ta]){				// No pixels; skip

         MAP_PALETTE_MAPPED(
            Map_15bit_xBGR,
            tb>>4,
            16,
            map
         );

         if(GFX_SPR_SOLID[ta]==1){			// Some pixels; trans
            switch(RAM_SPR[zz+14]&0xC0){
            case 0x00: Draw16x16_Trans_Mapped_Rot(&GFX_SPR[ta<<8], x, y, map);        break;
            case 0x80: Draw16x16_Trans_Mapped_FlipY_Rot(&GFX_SPR[ta<<8], x, y, map);  break;
            case 0x40: Draw16x16_Trans_Mapped_FlipX_Rot(&GFX_SPR[ta<<8], x, y, map);  break;
            case 0xC0: Draw16x16_Trans_Mapped_FlipXY_Rot(&GFX_SPR[ta<<8], x, y, map); break;
            }
         }
         else{						// all pixels; solid
            switch(RAM_SPR[zz+14]&0xC0){
            case 0x00: Draw16x16_Mapped_Rot(&GFX_SPR[ta<<8], x, y, map);        break;
            case 0x80: Draw16x16_Mapped_FlipY_Rot(&GFX_SPR[ta<<8], x, y, map);  break;
            case 0x40: Draw16x16_Mapped_FlipX_Rot(&GFX_SPR[ta<<8], x, y, map);  break;
            case 0xC0: Draw16x16_Mapped_FlipXY_Rot(&GFX_SPR[ta<<8], x, y, map); break;
            }
         }

         }

      }
   }

}


/*

 OBJECT RAM
 ----------

- 16 bytes per sprite definition (only odd)
- Sprites are 16x16; 16 colours

-----+--------+-----------------------------
Byte |Bit(s)  | Use
-----+76543210+-----------------------------
1/3/5|........| Unused
  7  |.......x| XPos - Sign Bit
  7  |......x.| YPos - Sign Bit
  7  |.....x..| Use Relative offsets
  7  |xxxx....| Palette Bank
  9  |xxxxxxxx| XPos
  B  |xxxxxxxx| YPos
  D  |xxxxxxxx| Sprite Number (low 8 bits)
  F  |....xxxx| Sprite Number (high 4 bits)
  F  |.x......| Flip Sprite X-Axis
  F  |x.......| Flip Sprite Y-Axis
-----+--------+-----------------------------

*/
