
#include "deftypes.h"
#include "tc140syt.h"

extern struct SOUND_INFO taito_ym2203_sound[];
extern struct SOUND_INFO taito_ym2203_m6295_sound[];

extern struct SOUND_INFO taito_ym2151_sound[];
extern struct SOUND_INFO taito_ym2151_msm5205_sound[];

extern struct SOUND_INFO taito_ym2610_sound[];
extern struct SOUND_INFO taito_ym2610b_sound[];

// Z80->YM2203A Communication Handlers
// -----------------------------------

UWORD YM2203AReadZ80(UWORD offset);
void YM2203AWriteZ80(UWORD offset, UBYTE data);
void YM2203ATaitoWriteZ80(UWORD offset, UBYTE data);

// Z80->YM2203B Communication Handlers
// -----------------------------------

UWORD YM2203BReadZ80(UWORD offset);
void YM2203BWriteZ80(UWORD offset, UBYTE data);

// Z80->YM2151 Communication Handlers
// ----------------------------------

UWORD YM2151ReadZ80(UWORD offset);
UWORD YM2151ReadZ80_Timer(UWORD offset);
void YM2151WriteZ80(UWORD offset, UBYTE data);
void YM2151TaitoWriteZ80(UWORD offset, UBYTE data);

// Z80->YM2610 Communication Handlers
// ----------------------------------

UWORD YM2610ReadZ80(UWORD offset);
void YM2610WriteZ80(UWORD offset, UBYTE data);

// Z80->YM2610 Volume Panning Handlers (TAITO)
// -------------------------------------------
void Pan2610WriteZ80(UWORD offset, UBYTE data);
void Pan2610_OffsetVol( int data );
void Pan2610WriteZ80_OffsetVol(UWORD offset, UBYTE data);

void YM2610SetBuffers(UBYTE *bufa, UBYTE *bufb, ULONG sizea, ULONG sizeb);

// TAITO Generic YM2610+Z80 Handler
// --------------------------------

void AddTaitoYM2610(ULONG p1, ULONG p2, ULONG romsize);
void AddTaitoYM2610B(ULONG p1, ULONG p2, ULONG romsize);
void AddTaitoYM2610B_OffsetVol(ULONG p1, ULONG p2, ULONG romsize, int vol);

void Taito2610_Frame(void);
void Taito2610_FramePan(void);
void Taito2610_FrameWarriorB(void);
void Taito2610_FrameNinjaW(void);
void Taito2610_FrameRastan2(void);
void Taito2610_FrameCameltry(void);
#if 0
void Taito2610_Frame_ChaseHQ(void);
#endif
void RemoveTaitoYM2610(void);

void YM2610_Init(void);
void YM2610B_Init(void);

// TAITO Generic YM2151+ADPCM+Z80 Handler
// --------------------------------------

void AddTaitoYM2151(ULONG p1, ULONG p2, ULONG romsize, UBYTE *adpcmA, UBYTE *adpcmB);
void AddTaitoYM2151_Timer(ULONG p1, ULONG p2, ULONG romsize, UBYTE *adpcmA, UBYTE *adpcmB);
void AddTaitoYM2151_xsystem(ULONG p1, ULONG p2, ULONG romsize);

void Taito2151_Frame(void);
void Taito2151_FrameFast(void);
void Taito2151_FrameRI(void);
void Taito2151_Frame_xsystem(void);

void RemoveTaitoYM2151(void);

// TAITO Generic YM2203+M6295+Z80 Handler
// --------------------------------------

void AddTaitoYM2203(ULONG p1, ULONG p2, ULONG romsize, UBYTE *adpcm, ULONG adpcmsize);
void AddTaitoYM2203_B(ULONG p1, ULONG p2, ULONG romsize, UBYTE *adpcm, ULONG adpcmsize);
void Taito2203_Frame(void);
void RemoveTaitoYM2203(void);

// Z80 rom banking system (TAITO)
// ------------------------------

void AddTaitoSoundBanking(UBYTE *romsrc, ULONG romsize);
void RemoveTaitoSoundBanking(void);

UWORD TaitoSoundReadBankData(UWORD offset);
void TaitoSoundSetBank(UWORD offset, UBYTE data);

// First OKI 6295

void M6295_A_Write_68k( ULONG address, UWORD data );
UWORD M6295_A_Read_68k( ULONG address );
void M6295_A_WriteBank_68k( ULONG address, UWORD data );

void M6295_A_Write_Z80(UWORD offset, UBYTE data);
UWORD M6295_A_Read_Z80(UWORD offset);
UWORD M6295_A_ReadFree_Z80(UWORD offset);
void M6295_A_WriteBank_Z80(UWORD offset, UBYTE data);

// Second OKI 6295

void M6295_B_Write_68k( ULONG address, UWORD data );
UWORD M6295_B_Read_68k( ULONG address );
void M6295_B_WriteBank_68k( ULONG address, UWORD data );

/**************** end of file ****************/
