/******************************************************************************/
/*                                                                            */
/*             tc0100scn: TAITO SCREEN LAYER CHIP (BG0/BG1/FG0)               */
/*                                                                            */
/******************************************************************************/

#include "deftypes.h"

#define MAX_CHIP		2		// F1-System has 2 chips

#define SCN_BG0			0		// Layer type values
#define SCN_BG1			0
#define SCN_FG0			1
#define SCN_FG0_SCROLL		2
#define SCN_FG0_SCROLL_2	3

int tc0100scn_layer_count;			// Init this to 0 each screen update.

typedef struct TC0100SCN_LAYER			// Information about 1 layer
{
   UBYTE *RAM;					// Layer RAM address
   UBYTE *GFX;					// Layer GFX address
   UBYTE *SCR;					// Layer SCROLL RAM address
   UBYTE *MASK;					// Layer GFX SOLID MASK address
   ULONG type;					// Type of layer (BG0/BG1/FG0)
   ULONG bmp_x,bmp_y;				// Game screen offset (32,32)
   ULONG bmp_w,bmp_h;				// Game screen size (320,224)
   mapper_direct *mapper;			// Colour mapper
   ULONG tile_mask;				// mask for tile number
   ULONG scr_x,scr_y;				// Scroll offset add (16,8)
} TC0100SCN_LAYER;

typedef struct TC0100SCN			// Information about 1 chip
{
   TC0100SCN_LAYER layer[3];			// BG0/BG1/FG0 layer data
   UWORD ctrl;					// Layer control data
   UBYTE *RAM;
   UBYTE *GFX_FG0;
} TC0100SCN;

struct TC0100SCN tc0100scn[MAX_CHIP];

void render_tc0100scn_layer_mapped(int chip, int layer);

void render_tc0100scn_layer_r180_mapped(int chip, int layer);

void render_tc0100scn_layer_mapped_r270(int chip, int layer);

void render_tc0100scn_layer_mapped_flipy(int chip, int layer);

/******************************************************************************/

UBYTE *GFX_FG0;
UBYTE *GFX_FG1;

void init_tc0100scn(int chip);

void tc0100scn_0_gfx_fg0_wb  (ULONG addr, UBYTE data);
void tc0100scn_0_gfx_fg0_ww  (ULONG addr, UWORD data);
void tc0100scn_0_copy_gfx_fg0(UBYTE *src, ULONG size);

void tc0100scn_1_gfx_fg0_wb  (ULONG addr, UBYTE data);
void tc0100scn_1_gfx_fg0_ww  (ULONG addr, UWORD data);
void tc0100scn_1_copy_gfx_fg0(UBYTE *src, ULONG size);

void tc0100scn_0_gfx_fg0_wb_r270  (ULONG addr, UBYTE data);
void tc0100scn_0_gfx_fg0_ww_r270  (ULONG addr, UWORD data);
void tc0100scn_0_copy_gfx_fg0_r270(UBYTE *src, ULONG size);

/******************************************************************************/
