/******************************************************************************/
/*                                                                            */
/*                          TOAPLAN 68000 SYSTEM#1                            */
/*                          ----------------------                            */
/*   CPU: 68000 Z80 <PROTECTED>                                               */
/* SOUND: YM3812                                                              */
/* VIDEO: 320x240 TOAPLAN CUSTOM <4xBG0 1xSPR>                                */
/*                                                                            */
/******************************************************************************/

#include "gameinc.h"
#include "toaplan1.h"
#include "3812intf.h"
#include "ym3812x.h"

static struct DIR_INFO vimana_dirs[] =
{
   { "vimana", },
   { NULL, },
};

static struct ROM_INFO vimana_roms[] =
{
   {     "vim1.bin", 0x00080000, 0xcdde26cd, },
   {     "vim2.bin", 0x00080000, 0x1dbfc118, },
   {     "vim3.bin", 0x00020000, 0x0b539131, },
   {     "vim4.bin", 0x00020000, 0xb0515768, },
   {     "vim5.bin", 0x00020000, 0x61a63d7a, },
   {     "vim6.bin", 0x00020000, 0x2886878d, },
   {    "vim07.bin", 0x00020000, 0x1efaea84, },
   {    "vim08.bin", 0x00020000, 0xe45b7def, },
   {           NULL,          0,          0, },
};

static struct INPUT_INFO vimana_inputs[] =
{
   { KB_DEF_COIN1,        MSG_COIN1,               0x023003, 0x08, BIT_ACTIVE_1 },
   { KB_DEF_COIN2,        MSG_COIN2,               0x023003, 0x10, BIT_ACTIVE_1 },
   { KB_DEF_TILT,         MSG_TILT,                0x023003, 0x02, BIT_ACTIVE_1 },
   { KB_DEF_SERVICE,      MSG_SERVICE,             0x023003, 0x01, BIT_ACTIVE_1 },

   { KB_DEF_P1_START,     MSG_P1_START,            0x023003, 0x20, BIT_ACTIVE_1 },
   { KB_DEF_P1_UP,        MSG_P1_UP,               0x023004, 0x01, BIT_ACTIVE_1 },
   { KB_DEF_P1_DOWN,      MSG_P1_DOWN,             0x023004, 0x02, BIT_ACTIVE_1 },
   { KB_DEF_P1_LEFT,      MSG_P1_LEFT,             0x023004, 0x04, BIT_ACTIVE_1 },
   { KB_DEF_P1_RIGHT,     MSG_P1_RIGHT,            0x023004, 0x08, BIT_ACTIVE_1 },
   { KB_DEF_P1_B1,        MSG_P1_B1,               0x023004, 0x10, BIT_ACTIVE_1 },
   { KB_DEF_P1_B2,        MSG_P1_B2,               0x023004, 0x20, BIT_ACTIVE_1 },

   { KB_DEF_P2_START,     MSG_P2_START,            0x023003, 0x40, BIT_ACTIVE_1 },
   { KB_DEF_P2_UP,        MSG_P2_UP,               0x023005, 0x01, BIT_ACTIVE_1 },
   { KB_DEF_P2_DOWN,      MSG_P2_DOWN,             0x023005, 0x02, BIT_ACTIVE_1 },
   { KB_DEF_P2_LEFT,      MSG_P2_LEFT,             0x023005, 0x04, BIT_ACTIVE_1 },
   { KB_DEF_P2_RIGHT,     MSG_P2_RIGHT,            0x023005, 0x08, BIT_ACTIVE_1 },
   { KB_DEF_P2_B1,        MSG_P2_B1,               0x023005, 0x10, BIT_ACTIVE_1 },
   { KB_DEF_P2_B2,        MSG_P2_B2,               0x023005, 0x20, BIT_ACTIVE_1 },

   { 0,                   NULL,                    0,        0,    0            },
};

static struct DSW_DATA dsw_data_vimana_2[] =
{
   { "Language",              0x0F, 0x10 },
   { "0",                     0x00, 0x00 },
   { "1",                     0x01, 0x00 },
   { "2",                     0x02, 0x00 },
   { "3",                     0x03, 0x00 },
   { "4",                     0x04, 0x00 },
   { "5",                     0x05, 0x00 },
   { "6",                     0x06, 0x00 },
   { "7",                     0x07, 0x00 },
   { "8",                     0x08, 0x00 },
   { "9",                     0x09, 0x00 },
   { "A",                     0x0A, 0x00 },
   { "B",                     0x0B, 0x00 },
   { "C",                     0x0C, 0x00 },
   { "D",                     0x0D, 0x00 },
   { "E",                     0x0E, 0x00 },
   { "F",                     0x0F, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO vimana_dsw[] =
{
   { 0x023000, 0x00, dsw_data_default_0 },
   { 0x023001, 0x00, dsw_data_default_1 },
   { 0x023002, 0x00, dsw_data_vimana_2 },
   { 0,        0,    NULL,      },
};

static struct VIDEO_INFO toaplan1_video =
{
   DrawToaplan1,
   320,
   240,
   32,
   VIDEO_ROTATE_NORMAL |
   VIDEO_ROTATABLE,
};

static struct VIDEO_INFO toaplan1_r270_video =
{
   DrawToaplan1,
   320,
   240,
   32,
   VIDEO_ROTATE_270 |
   VIDEO_ROTATABLE,
};

static struct YM3812interface ym3812_interface =
{
   1,              // 1 chip
   3600000,        // 3.6 MHz (3.5?)
   { 255 },        // volume
    NULL
};

static struct SOUND_INFO toaplan1_sound[] =
{
   { SOUND_YM3812,  &ym3812_interface,    },
   { 0,             NULL,                 },
};

struct GAME_MAIN game_vimana =
{
   vimana_dirs,
   vimana_roms,
   vimana_inputs,
   vimana_dsw,
   NULL,

   load_vimana,
   clear_vimana,
   &toaplan1_r270_video,
   ExecuteToaplan1Frame,
   "vimana",
   "Vimana",
   NULL,
   COMPANY_ID_TOAPLAN,
   "TP019",
   1991,
   NULL,
};

static struct DIR_INFO vimana_alt_dirs[] =
{
   { "vimana2", },
   { ROMOF("vimana"), },
   { CLONEOF("vimana"), },
   { NULL, },
};

static struct ROM_INFO vimana_alt_roms[] =
{
   {     "vim1.bin", 0x00080000, 0xcdde26cd, },
   {     "vim2.bin", 0x00080000, 0x1dbfc118, },
   {     "vim3.bin", 0x00020000, 0x0b539131, },
   {     "vim4.bin", 0x00020000, 0xb0515768, },
   {     "vim5.bin", 0x00020000, 0x61a63d7a, },
   {     "vim6.bin", 0x00020000, 0x2886878d, },
   { "vimana07.bin", 0x00020000, 0x5a4bf73e, },
   { "vimana08.bin", 0x00020000, 0x03ba27e8, },
   {           NULL,          0,          0, },
};

struct GAME_MAIN game_vimana_alt =
{
   vimana_alt_dirs,
   vimana_alt_roms,
   vimana_inputs,
   vimana_dsw,
   NULL,

   load_vimana_alt,
   clear_vimana_alt,
   &toaplan1_r270_video,
   ExecuteToaplan1Frame,
   "vimana2",
   "Vimana (alternate)",
   NULL,
   COMPANY_ID_TOAPLAN,
   "TP019",
   1991,
   NULL,
};

static struct DIR_INFO outzone_dirs[] =
{
   { "outzone", },
   { NULL, },
};

static struct ROM_INFO outzone_roms[] =
{
   {     "rom1.bin", 0x00020000, 0x0934782d, },
   {     "rom2.bin", 0x00020000, 0x6bb72d16, },
   {     "rom3.bin", 0x00020000, 0xec903c07, },
   {     "rom4.bin", 0x00020000, 0x50cbf1a8, },
   {     "rom5.bin", 0x00080000, 0xc64ec7b6, },
   {     "rom6.bin", 0x00080000, 0x64b6c5ac, },
   {     "rom7.bin", 0x00020000, 0x936e25d8, },
   {     "rom8.bin", 0x00020000, 0xd19b3ecf, },
   {     "rom9.bin", 0x00008000, 0x73d8e235, },
   {           NULL,          0,          0, },
};

static struct DSW_DATA dsw_data_outzone_2[] =
{
   { "Language",              0x0F, 0x10 },
   { "0",                     0x00, 0x00 },
   { "1",                     0x01, 0x00 },
   { "2",                     0x02, 0x00 },
   { "3",                     0x03, 0x00 },
   { "4",                     0x04, 0x00 },
   { "5",                     0x05, 0x00 },
   { "6",                     0x06, 0x00 },
   { "7",                     0x07, 0x00 },
   { "8",                     0x08, 0x00 },
   { "9",                     0x09, 0x00 },
   { "A",                     0x0A, 0x00 },
   { "B",                     0x0B, 0x00 },
   { "C",                     0x0C, 0x00 },
   { "D",                     0x0D, 0x00 },
   { "E",                     0x0E, 0x00 },
   { "F",                     0x0F, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO outzone_dsw[] =
{
   { 0x023000, 0x00, dsw_data_default_0 },
   { 0x023001, 0x00, dsw_data_default_1 },
   { 0x023002, 0x00, dsw_data_outzone_2 },
   { 0,        0,    NULL,      },
};

struct GAME_MAIN game_outzone =
{
   outzone_dirs,
   outzone_roms,
   vimana_inputs,
   outzone_dsw,
   NULL,

   load_outzone,
   clear_outzone,
   &toaplan1_r270_video,
   ExecuteToaplan1Frame_Sound,
   "outzone",
   "Outzone",
   NULL,
   COMPANY_ID_TOAPLAN,
   "TP018",
   1990,
   toaplan1_sound,
};

static struct DIR_INFO rally_bike_dirs[] =
{
   { "rally_bike", },
   { "rallybik", },
   { "rallyb", },
   { NULL, },
};

static struct ROM_INFO rally_bike_roms[] =
{
   {    "rb_01.bin", 0x00008000, 0x7602f6a7, },
   {    "rb_02.bin", 0x00008000, 0x383386d7, },
   {    "rb_03.bin", 0x00020000, 0x555344ce, },
   {    "rb_04.bin", 0x00020000, 0xe9b005b1, },
   {    "rb_05.bin", 0x00004000, 0x10814601, },
   {    "rb_06.bin", 0x00020000, 0x144b085c, },
   {    "rb_07.bin", 0x00020000, 0xcd3748b4, },
   {    "rb_08.bin", 0x00020000, 0xfab661ba, },
   {    "rb_09.bin", 0x00020000, 0x1dc7b010, },
   {    "rb_10.bin", 0x00010000, 0xdbb7c57e, },
   {    "rb_11.bin", 0x00010000, 0x0d56e8bb, },
   {    "rb_12.bin", 0x00010000, 0xcf5aae4e, },
   {    "rb_13.bin", 0x00010000, 0x1683b07c, },
   {           NULL,          0,          0, },
};

static struct DSW_DATA dsw_data_rally_bike_2[] =
{
   { "Language",              0x0F, 0x10 },
   { "0",                     0x00, 0x00 },
   { "1",                     0x01, 0x00 },
   { "2",                     0x02, 0x00 },
   { "3",                     0x03, 0x00 },
   { "4",                     0x04, 0x00 },
   { "5",                     0x05, 0x00 },
   { "6",                     0x06, 0x00 },
   { "7",                     0x07, 0x00 },
   { "8",                     0x08, 0x00 },
   { "9",                     0x09, 0x00 },
   { "A",                     0x0A, 0x00 },
   { "B",                     0x0B, 0x00 },
   { "C",                     0x0C, 0x00 },
   { "D",                     0x0D, 0x00 },
   { "E",                     0x0E, 0x00 },
   { "F",                     0x0F, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO rally_bike_dsw[] =
{
   { 0x023000, 0x00, dsw_data_default_0 },
   { 0x023001, 0x00, dsw_data_default_1 },
   { 0x023002, 0x00, dsw_data_rally_bike_2 },
   { 0,        0,    NULL,      },
};

struct GAME_MAIN game_rally_bike =
{
   rally_bike_dirs,
   rally_bike_roms,
   vimana_inputs,
   rally_bike_dsw,
   NULL,

   load_rally_bike,
   clear_rally_bike,
   &toaplan1_r270_video,
   ExecuteToaplan1Frame_Sound,
   "rallybik",
   "Rally Bike",
   NULL,
   COMPANY_ID_TOAPLAN,
   "TP012",
   1988,
   toaplan1_sound,
};

static struct DIR_INFO zero_wing_dirs[] =
{
   { "zero_wing", },
   { "zerowing", },
   { NULL, },
};

static struct ROM_INFO zero_wing_roms[] =
{
   {   "o15-01.rom", 0x00020000, 0x70570e43, },
   {   "o15-02.rom", 0x00020000, 0x724b487f, },
   {   "o15-03.rom", 0x00020000, 0x7f245fd3, },
   {   "o15-04.rom", 0x00020000, 0x0b1a1289, },
   {   "o15-05.rom", 0x00020000, 0x4e5dd246, },
   {   "o15-06.rom", 0x00020000, 0xc8c6d428, },
   {   "o15-07.rom", 0x00020000, 0xefc40e99, },
   {   "o15-08.rom", 0x00020000, 0x1b019eab, },
   {   "o15-09.rom", 0x00020000, 0x13764e95, },
   {   "o15-10.rom", 0x00020000, 0x351ba71a, },
   {   "o15-11.rom", 0x00008000, 0x6ff2b9a0, },
   {   "o15-12.rom", 0x00008000, 0x9773e60b, },
   {   "o15-13.rom", 0x00008000, 0xe7b72383, },
   {           NULL,          0,          0, },
};

static struct DSW_DATA dsw_data_zero_wing_2[] =
{
   { "Language",              0x0F, 0x10 },
   { "0",                     0x00, 0x00 },
   { "1",                     0x01, 0x00 },
   { "2",                     0x02, 0x00 },
   { "3",                     0x03, 0x00 },
   { "4",                     0x04, 0x00 },
   { "5",                     0x05, 0x00 },
   { "6",                     0x06, 0x00 },
   { "7",                     0x07, 0x00 },
   { "8",                     0x08, 0x00 },
   { "9",                     0x09, 0x00 },
   { "A",                     0x0A, 0x00 },
   { "B",                     0x0B, 0x00 },
   { "C",                     0x0C, 0x00 },
   { "D",                     0x0D, 0x00 },
   { "E",                     0x0E, 0x00 },
   { "F",                     0x0F, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO zero_wing_dsw[] =
{
   { 0x023000, 0x00, dsw_data_default_0 },
   { 0x023001, 0x00, dsw_data_default_1 },
   { 0x023002, 0x00, dsw_data_zero_wing_2 },
   { 0,        0,    NULL,      },
};

struct GAME_MAIN game_zero_wing =
{
   zero_wing_dirs,
   zero_wing_roms,
   vimana_inputs,
   zero_wing_dsw,
   NULL,

   load_zero_wing,
   clear_zero_wing,
   &toaplan1_video,
   ExecuteToaplan1Frame_Sound,
   "zerowing",
   "Zero Wing",
   NULL,
   COMPANY_ID_TOAPLAN,
   "TP015",
   1990,
   toaplan1_sound,
};

static struct DIR_INFO demons_world_dirs[] =
{
   { "demons_world", },
   { "demonwld", },
   { NULL, },
};

static struct ROM_INFO demons_world_roms[] =
{
   {        "rom01", 0x00020000, 0x1b3724e9, },
   {        "rom02", 0x00020000, 0x7b20a44d, },
   {        "rom03", 0x00020000, 0x2cacdcd0, },
   {        "rom04", 0x00020000, 0x76fd3201, },
   {        "rom05", 0x00020000, 0x6506c982, },
   {        "rom06", 0x00020000, 0x4fc5e5f3, },
   {        "rom07", 0x00020000, 0xa3a0d993, },
   {        "rom08", 0x00020000, 0xeb53ab09, },
   {        "rom09", 0x00020000, 0xbed746e3, },
   {        "rom10", 0x00020000, 0x036ee46c, },
   {        "rom11", 0x00008000, 0x397eca1b, },
   {           NULL,          0,          0, },
};

static struct DSW_DATA dsw_data_demons_world_2[] =
{
   { "Language",              0x0F, 0x10 },
   { "0",                     0x00, 0x00 },
   { "1",                     0x01, 0x00 },
   { "2",                     0x02, 0x00 },
   { "3",                     0x03, 0x00 },
   { "4",                     0x04, 0x00 },
   { "5",                     0x05, 0x00 },
   { "6",                     0x06, 0x00 },
   { "7",                     0x07, 0x00 },
   { "8",                     0x08, 0x00 },
   { "9",                     0x09, 0x00 },
   { "A",                     0x0A, 0x00 },
   { "B",                     0x0B, 0x00 },
   { "C",                     0x0C, 0x00 },
   { "D",                     0x0D, 0x00 },
   { "E",                     0x0E, 0x00 },
   { "F",                     0x0F, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO demons_world_dsw[] =
{
   { 0x023000, 0x00, dsw_data_default_0 },
   { 0x023001, 0x00, dsw_data_default_1 },
   { 0x023002, 0x00, dsw_data_demons_world_2 },
   { 0,        0,    NULL,      },
};

struct GAME_MAIN game_demons_world =
{
   demons_world_dirs,
   demons_world_roms,
   vimana_inputs,
   demons_world_dsw,
   NULL,

   load_demons_world,
   clear_demons_world,
   &toaplan1_video,
   ExecuteToaplan1Frame_Sound,
   "demonwld",
   "Demon's World",
   NULL,
   COMPANY_ID_TOAPLAN,
   "TP016",
   1990,
   toaplan1_sound,
};

static struct DIR_INFO fire_shark_dirs[] =
{
   { "fire_shark", },
   { "fireshrk", },
   { NULL, },
};

static struct ROM_INFO fire_shark_roms[] =
{
   {   "o17_01.bin", 0x00020000, 0xea12e491, },
   {   "o17_02.bin", 0x00020000, 0x32a13a9f, },
   {   "o17_03.bin", 0x00020000, 0x68723dc9, },
   {   "o17_04.bin", 0x00020000, 0xfe0ecb13, },
   {   "o17_05.bin", 0x00020000, 0x565315f8, },
   {   "o17_06.bin", 0x00020000, 0x95262d4c, },
   {   "o17_07.bin", 0x00020000, 0x4c4b735c, },
   {   "o17_08.bin", 0x00020000, 0x95c6586c, },
   {        "9.bin", 0x00008000, 0xf0c70e6f, },
   {       "10.bin", 0x00008000, 0x9d253d77, },
   {       "11.bin", 0x00020000, 0x6beac378, },
   {       "12.bin", 0x00020000, 0x6adb6eb5, },
   {           NULL,          0,          0, },
};

static struct DSW_DATA dsw_data_fire_shark_2[] =
{
   { "Language",              0x0F, 0x10 },
   { "0",                     0x00, 0x00 },
   { "1",                     0x01, 0x00 },
   { "2",                     0x02, 0x00 },
   { "3",                     0x03, 0x00 },
   { "4",                     0x04, 0x00 },
   { "5",                     0x05, 0x00 },
   { "6",                     0x06, 0x00 },
   { "7",                     0x07, 0x00 },
   { "8",                     0x08, 0x00 },
   { "9",                     0x09, 0x00 },
   { "A",                     0x0A, 0x00 },
   { "B",                     0x0B, 0x00 },
   { "C",                     0x0C, 0x00 },
   { "D",                     0x0D, 0x00 },
   { "E",                     0x0E, 0x00 },
   { "F",                     0x0F, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO fire_shark_dsw[] =
{
   { 0x023000, 0x00, dsw_data_default_0 },
   { 0x023001, 0x00, dsw_data_default_1 },
   { 0x023002, 0x00, dsw_data_fire_shark_2 },
   { 0,        0,    NULL,      },
};

struct GAME_MAIN game_fire_shark =
{
   fire_shark_dirs,
   fire_shark_roms,
   vimana_inputs,
   fire_shark_dsw,
   NULL,

   load_fire_shark,
   clear_fire_shark,
   &toaplan1_r270_video,
   ExecuteToaplan1Frame,
   "fireshrk",
   "Fire Shark",
   NULL,
   COMPANY_ID_TOAPLAN,
   "TP017",
   1990,
   NULL,
};

static struct DIR_INFO same_same_same_dirs[] =
{
   { "same_same_same", },
   { "same3", },
   { ROMOF("fireshrk"), },
   { CLONEOF("fireshrk"), },
   { NULL, },
};

static struct ROM_INFO same_same_same_roms[] =
{
   {   "o17_01.bin", 0x00020000, 0xea12e491, },
   {   "o17_02.bin", 0x00020000, 0x32a13a9f, },
   {   "o17_03.bin", 0x00020000, 0x68723dc9, },
   {   "o17_04.bin", 0x00020000, 0xfe0ecb13, },
   {   "o17_05.bin", 0x00020000, 0x565315f8, },
   {   "o17_06.bin", 0x00020000, 0x95262d4c, },
   {   "o17_07.bin", 0x00020000, 0x4c4b735c, },
   {   "o17_08.bin", 0x00020000, 0x95c6586c, },
   {   "o17_09.bin", 0x00008000, 0x3f69e437, },
   {   "o17_10.bin", 0x00008000, 0x4e723e0a, },
   {   "o17_11.bin", 0x00020000, 0xbe07d101, },
   {   "o17_12.bin", 0x00020000, 0xef698811, },
   {           NULL,          0,          0, },
};

static struct DSW_DATA dsw_data_same_same_same_2[] =
{
   { "Language",              0x0F, 0x10 },
   { "0",                     0x00, 0x00 },
   { "1",                     0x01, 0x00 },
   { "2",                     0x02, 0x00 },
   { "3",                     0x03, 0x00 },
   { "4",                     0x04, 0x00 },
   { "5",                     0x05, 0x00 },
   { "6",                     0x06, 0x00 },
   { "7",                     0x07, 0x00 },
   { "8",                     0x08, 0x00 },
   { "9",                     0x09, 0x00 },
   { "A",                     0x0A, 0x00 },
   { "B",                     0x0B, 0x00 },
   { "C",                     0x0C, 0x00 },
   { "D",                     0x0D, 0x00 },
   { "E",                     0x0E, 0x00 },
   { "F",                     0x0F, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO same_same_same_dsw[] =
{
   { 0x023000, 0x00, dsw_data_default_0 },
   { 0x023001, 0x00, dsw_data_default_1 },
   { 0x023002, 0x00, dsw_data_same_same_same_2 },
   { 0,        0,    NULL,      },
};

struct GAME_MAIN game_same_same_same =
{
   same_same_same_dirs,
   same_same_same_roms,
   vimana_inputs,
   same_same_same_dsw,
   NULL,

   load_same_same_same,
   clear_same_same_same,
   &toaplan1_r270_video,
   ExecuteToaplan1Frame,
   "same3",
   "Same Same Same",
   NULL,
   COMPANY_ID_TOAPLAN,
   "TP017",
   1990,
   NULL,
};

static struct DIR_INFO hell_fire_dirs[] =
{
   { "hell_fire", },
   { "hellfire", },
   { NULL, },
};

static struct ROM_INFO hell_fire_roms[] =
{
   {   "b90-03.bin", 0x00008000, 0x4058fa67, },
   {   "b90-04.bin", 0x00020000, 0xea6150fc, },
   {   "b90-05.bin", 0x00020000, 0xbb52c507, },
   {   "b90-06.bin", 0x00020000, 0xcf5b0252, },
   {   "b90-07.bin", 0x00020000, 0xb98af263, },
   {   "b90-08.bin", 0x00020000, 0x3404a5e3, },
   {   "b90-09.bin", 0x00020000, 0xcf01009e, },
   {   "b90-10.bin", 0x00020000, 0x35fd1092, },
   {   "b90-11.bin", 0x00020000, 0xc33e543c, },
   {   "b90-14.bin", 0x00020000, 0x101df9f5, },
   {   "b90-15.bin", 0x00020000, 0xe67fd452, },
   {           NULL,          0,          0, },
};

static struct DSW_DATA dsw_data_hell_fire_2[] =
{
   { "Language",              0x0F, 0x10 },
   { "0",                     0x00, 0x00 },
   { "1",                     0x01, 0x00 },
   { "2",                     0x02, 0x00 },
   { "3",                     0x03, 0x00 },
   { "4",                     0x04, 0x00 },
   { "5",                     0x05, 0x00 },
   { "6",                     0x06, 0x00 },
   { "7",                     0x07, 0x00 },
   { "8",                     0x08, 0x00 },
   { "9",                     0x09, 0x00 },
   { "A",                     0x0A, 0x00 },
   { "B",                     0x0B, 0x00 },
   { "C",                     0x0C, 0x00 },
   { "D",                     0x0D, 0x00 },
   { "E",                     0x0E, 0x00 },
   { "F",                     0x0F, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO hell_fire_dsw[] =
{
   { 0x023000, 0x00, dsw_data_default_0 },
   { 0x023001, 0x00, dsw_data_default_1 },
   { 0x023002, 0x00, dsw_data_hell_fire_2 },
   { 0,        0,    NULL,      },
};

struct GAME_MAIN game_hell_fire =
{
   hell_fire_dirs,
   hell_fire_roms,
   vimana_inputs,
   hell_fire_dsw,
   NULL,

   load_hell_fire,
   clear_hell_fire,
   &toaplan1_video,
   ExecuteToaplan1Frame_Sound,
   "hellfire",
   "Hell Fire",
   NULL,
   COMPANY_ID_TOAPLAN,
   "B90",
   1990,
   toaplan1_sound,
};

static struct DIR_INFO truxton_dirs[] =
{
   { "truxton", },
   { NULL, },
};

static struct ROM_INFO truxton_roms[] =
{
   {   "b65_01.bin", 0x00020000, 0xe974937f, },
   {   "b65_02.bin", 0x00020000, 0x1dd55161, },
   {   "b65_03.bin", 0x00020000, 0x58b1350b, },
   {   "b65_04.bin", 0x00020000, 0x8c6ff461, },
   {   "b65_05.bin", 0x00020000, 0x81cd95f1, },
   {   "b65_06.bin", 0x00020000, 0x36cedcbe, },
   {   "b65_07.bin", 0x00020000, 0xfb83252a, },
   {   "b65_08.bin", 0x00020000, 0xd2315b37, },
   {   "b65_09.bin", 0x00008000, 0xf1c0f410, },
   {   "b65_10.bin", 0x00020000, 0xaff5195d, },
   {   "b65_11.bin", 0x00020000, 0x1a62379a, },
   {           NULL,          0,          0, },
};

static struct DSW_DATA dsw_data_truxton_2[] =
{
   { "Language",              0x0F, 0x10 },
   { "0",                     0x00, 0x00 },
   { "1",                     0x01, 0x00 },
   { "2",                     0x02, 0x00 },
   { "3",                     0x03, 0x00 },
   { "4",                     0x04, 0x00 },
   { "5",                     0x05, 0x00 },
   { "6",                     0x06, 0x00 },
   { "7",                     0x07, 0x00 },
   { "8",                     0x08, 0x00 },
   { "9",                     0x09, 0x00 },
   { "A",                     0x0A, 0x00 },
   { "B",                     0x0B, 0x00 },
   { "C",                     0x0C, 0x00 },
   { "D",                     0x0D, 0x00 },
   { "E",                     0x0E, 0x00 },
   { "F",                     0x0F, 0x00 },
   { NULL,                    0,    0,   },
};

static struct DSW_INFO truxton_dsw[] =
{
   { 0x023000, 0x00, dsw_data_default_0 },
   { 0x023001, 0x00, dsw_data_default_1 },
   { 0x023002, 0x00, dsw_data_truxton_2 },
   { 0,        0,    NULL,      },
};

struct GAME_MAIN game_truxton =
{
   truxton_dirs,
   truxton_roms,
   vimana_inputs,
   truxton_dsw,
   NULL,

   load_truxton,
   clear_truxton,
   &toaplan1_r270_video,
   ExecuteToaplan1Frame_Sound,
   "truxton",
   "Truxton",
   NULL,
   COMPANY_ID_TOAPLAN,
   "B65",
   1988,
   toaplan1_sound,
};

// Store Tiles in priority lists (good idea from Shark!).

#define MAX_PRI         16              // 16 levels of priority
#define MAX_TILES       0x8000          // 0x4000*0x14=0x60000

static struct TILE_Q
{
   ULONG tile;                          // Tile number
   ULONG x,y;                           // X,Y position
   UBYTE *map;                          // Colour map data
   struct TILE_Q *next;                 // Next item with equal priority
} TILE_Q;

static struct TILE_Q *TileQueue;               // full list
static struct TILE_Q *last_tile;               // last tile in use
static struct TILE_Q *prilist[MAX_PRI];        // start node for each priority list

#define ROM_COUNT       10

/*

 Supported romsets:

  0 - Vimana              - 1991 - TOAPLAN
  1 - Vimana alt set      - 1991 - TOAPLAN
  2 - Outzone             - 1990 - TOAPLAN
  3 - Rally Bike          - 1988 - TOAPLAN
  4 - Zero Wing           - 1990 - TOAPLAN
  5 - Demon's World       - 1990 - TOAPLAN
  6 - Fire Shark          - 19.. - TOAPLAN
  7 - Same Same Same      - 19.. - TOAPLAN
  8 - Hell Fire           - 19.. - TOAPLAN
  9 - Truxton             - 1988 - TOAPLAN

 Todo:

 - Lots

Vimana:

 - Sound Z80 is not dumped

Rally Bike:

 - Object ram is directly accessed (no port) bootleg?

Fire Shark:

 - Sound Z80 is not dumped

Same Same Same:

 - Sound Z80 is not dumped

*/

static int romset;
static int y_ofs;

static UBYTE *GFX_BG0;
static UBYTE *GFX_BG0_SOLID;

static UBYTE *GFX_SPR;

/******************************************************************************/
/*  VCU CHIP                                                                  */
/******************************************************************************/

typedef struct TP1VCU				// information about 1 chip
{
   UBYTE *VRAM;					// layer ram (0x10000 bytes)
   UBYTE *RAM_OBJ;				// object ram (0x800 + 0x80 bytes)
   UBYTE *SCROLL;				// scroll ram (0x10 bytes)
   ULONG obj_pos;				// current offset in object ram
   ULONG bg_pos;				// current offset in layer ram
   ULONG scroll_pos;				// current offset in SCROLL
   ULONG status;				// some status read
   UBYTE *GFX_BG;				// GFX data
   UBYTE *MASK_BG;				// MASK data
   ULONG tile_mask_bg;				// tile count
   ULONG tile_mask_obj;				// tile count
   ULONG tile_ofs_obj;				// tile offset
} TP1VCU;

struct TP1VCU tp1vcu[1];			// max 1 chips

static void init_tp1vcu(ULONG num)
{
   tp1vcu[num].obj_pos    = 0x0000;
   tp1vcu[num].bg_pos     = 0x0000;
   tp1vcu[num].scroll_pos = 0x0000;
   tp1vcu[num].status     = 0x0000;
}

static void tp1vcu_obj_ww(ULONG offset, UWORD data)
{
   switch(offset&0x0E){
      case 0x02:                        // Port Address
         tp1vcu[0].obj_pos = (data<<1) & 0x7FE;
      break;
      case 0x04:                        // Port Data
         WriteWord(&tp1vcu[0].RAM_OBJ[tp1vcu[0].obj_pos], data);
         tp1vcu[0].obj_pos += 0x002;
         tp1vcu[0].obj_pos &= 0x7FE;
      break;
      case 0x06:                        // Port Data
         WriteWord(&tp1vcu[0].RAM_OBJ[tp1vcu[0].obj_pos+0x800], data);
         tp1vcu[0].obj_pos += 0x002;
         tp1vcu[0].obj_pos &= 0x7FE;
      break;
      default:
         #ifdef HACKS
         print_debug("tp1vcu[0] ww(%04x,%04x)\n", offset&0x0E, data);
         #endif
      break;
   }
}

static UWORD tp1vcu_obj_rw(ULONG offset)
{
   UWORD ret;

   switch(offset&0x0E){
      case 0x00:                        // Status
         ret = tp1vcu[0].status;
         tp1vcu[0].status ^= 1;
      break;
      case 0x04:                        // Port Data
         ret = ReadWord(&tp1vcu[0].RAM_OBJ[tp1vcu[0].obj_pos]);
         tp1vcu[0].obj_pos += 0x002;
         tp1vcu[0].obj_pos &= 0x7FE;
      break;
      case 0x06:                        // Port Data
         ret = ReadWord(&tp1vcu[0].RAM_OBJ[tp1vcu[0].obj_pos+0x800]);
         tp1vcu[0].obj_pos += 0x002;
         tp1vcu[0].obj_pos &= 0x7FE;
      break;
      default:
         ret = 0x0000;
         #ifdef HACKS
         print_debug("tp1vcu[0] rw(%04x)\n", offset&0x0E);
         #endif
      break;
   }

   return ret;
}

static UBYTE tp1vcu_obj_rb(ULONG offset)
{
   UWORD ret;

   ret = tp1vcu_obj_rw(offset);

   if((offset&1)==0)
      return (UBYTE) ((ret>>8)&0xFF);
   else
      return (UBYTE) ((ret>>0)&0xFF);
}

static void tp1vcu_bg_ww(ULONG offset, UWORD data)
{
   switch(offset&0x1E){
      case 0x02:                        // Port Address
         tp1vcu[0].bg_pos = (data<<2) & 0xFFFE;
      break;
      case 0x04:                        // Port Data
      case 0x06:                        // Port Data
         WriteWord(&tp1vcu[0].VRAM[tp1vcu[0].bg_pos], data);
         tp1vcu[0].bg_pos += 0x0002;
         tp1vcu[0].bg_pos &= 0xFFFE;
      break;
      case 0x10:                        // Scroll
      case 0x12:                        // Scroll
      case 0x14:                        // Scroll
      case 0x16:                        // Scroll
      case 0x18:                        // Scroll
      case 0x1A:                        // Scroll
      case 0x1C:                        // Scroll
      case 0x1E:                        // Scroll
         WriteWord(&tp1vcu[0].SCROLL[offset&0x0E], data);
      break;
      default:
         #ifdef HACKS
         print_debug("tp1vcu[0] ww(%04x,%04x)\n", offset&0x1E, data);
         #endif
      break;
   }
}

static UWORD tp1vcu_bg_rw(ULONG offset)
{
   UWORD ret;

   switch(offset&0x1E){
      case 0x00:                        // Status
         ret = tp1vcu[0].status;
         tp1vcu[0].status ^= 1;
      break;
      case 0x04:                        // Port Data
      case 0x06:                        // Port Data
         ret = ReadWord(&tp1vcu[0].VRAM[tp1vcu[0].bg_pos]);
         tp1vcu[0].bg_pos += 0x0002;
         tp1vcu[0].bg_pos &= 0xFFFE;
      break;
      case 0x10:                        // Scroll
      case 0x12:                        // Scroll
      case 0x14:                        // Scroll
      case 0x16:                        // Scroll
      case 0x18:                        // Scroll
      case 0x1A:                        // Scroll
      case 0x1C:                        // Scroll
      case 0x1E:                        // Scroll
         ret = ReadWord(&tp1vcu[0].SCROLL[offset&0x0E]);
      break;
      default:
         ret = 0x0000;
         #ifdef HACKS
         print_debug("tp1vcu[0] rw(%04x)\n", offset&0x0E);
         #endif
      break;
   }

   return ret;
}

static UBYTE tp1vcu_bg_rb(ULONG offset)
{
   UWORD ret;

   ret = tp1vcu_bg_rw(offset);

   if((offset&1)==0)
      return (UBYTE) ((ret>>8)&0xFF);
   else
      return (UBYTE) ((ret>>0)&0xFF);
}

/******************************************************************************/
/*  VBLANK/INTERRUPT THINGY                                                   */
/******************************************************************************/

static ULONG want_int;

static int x_ofs[4];

static UBYTE tp_vblank_rb(ULONG offset)
{
   static UBYTE status;

   switch(offset & 0x0E){
   case 0x00:
      status ^= 1;
      return status;
   break;
   default:
      return 0;
   break;
   }
}

static UWORD tp_vblank_rw(ULONG offset)
{
   return (UWORD) (tp_vblank_rb(offset));
}

static void tp_vblank_wb(ULONG offset, UBYTE data)
{
   switch(offset & 0x0E){
   case 0x02:
      want_int = data & 1;
      #ifdef HACKS
      print_debug("want_int: %d\n", want_int);
      #endif
   break;
   case 0x08:
      x_ofs[0] = (data&0xFF) - 0xd8;
   break;
   case 0x0A:
      x_ofs[1] = (data&0xFF) - 0x17;
   break;
   case 0x0C:
      x_ofs[2] = (data&0xFF) - 0x88;
   break;
   case 0x0E:
      x_ofs[3] = (data&0xFF) - 0x07;
   break;
   default:
   break;
   }
}

static void tp_vblank_ww(ULONG offset, UWORD data)
{
   tp_vblank_wb(offset, (UBYTE)(data&0xFF));
}

static int bg_x_ofs;
static int bg_y_ofs;

static void tp1vcu_ofsreg_ww(ULONG offset, UWORD data)
{
   if((offset & 2) == 0)
	bg_x_ofs = data;
   else
	bg_y_ofs = data;
}


/******************************************************************************/
/*  Z80 CUSTOM (not dumped)                                                   */
/******************************************************************************/

static UBYTE vimana_z80_rb(ULONG offset)
{
   UWORD ret;

   offset |= 1;

   switch(offset&0xFF){
      case 0x01:                        // sound comm?
         ret = 0xFF;
      break;
      case 0x03:                        // sound comm?
         ret = 0x00;
      break;
      case 0x05:                        // credits
         ret = 0x01;
      break;
      case 0x07:                        // dswa
         ret = get_dsw(0);
      break;
      case 0x09:                        // misc input
         ret = RAM[0x023003];
      break;
      case 0x0B:                        // p1 input
         ret = RAM[0x023004];
      break;
      case 0x0D:                        // p2 input
         ret = RAM[0x023005];
      break;
      case 0x0F:                        // dswb
         ret = get_dsw(1);
      break;
      case 0x11:                        // dswc
         ret = get_dsw(2);
      break;
      default:
         ret = 0x00;
         #ifdef HACKS
         print_debug("vimana_z80_rb(%04x)\n", offset&0xFF);
         #endif
      break;
   }

   return ret;
}

static UWORD vimana_z80_rw(ULONG offset)
{
   return (UWORD) (vimana_z80_rb(offset|1));
}

/******************************************************************************/
/*  Z80 NORMAL                                                                */
/******************************************************************************/

static UBYTE tp1_z80_rb(ULONG offset)
{
   offset  &= 0xFFF;
   offset >>= 1;

   return Z80ROM[0x8000 + offset];
}

static UWORD tp1_z80_rw(ULONG offset)
{
   offset  &= 0xFFF;
   offset >>= 1;

   return (UWORD) (Z80ROM[0x8000 + offset]);
}

static void tp1_z80_wb(ULONG offset, UBYTE data)
{
   offset  &= 0xFFF;
   offset >>= 1;

   Z80ROM[0x8000 + offset] = data;
}

static void tp1_z80_ww(ULONG offset, UWORD data)
{
   offset  &= 0xFFF;
   offset >>= 1;

   Z80ROM[0x8000 + offset] = (UBYTE) (data&0xFF);
}

static UBYTE zero_wing_port_rb(UWORD offset)
{
   switch(offset&0xFF){
      case 0x00:
         return RAM[0x023004];
      break;
      case 0x08:
         return RAM[0x023005];
      break;
      case 0x20:
         return get_dsw(0);
      break;
      case 0x28:
         return get_dsw(1);
      break;
      case 0x80:
         return RAM[0x023003];
      break;
      case 0x88:
         return get_dsw(2);
      break;
      case 0xA8:
         return YM3812ReadZ80(0);
      break;
      default:
         return 0x00;
      break;
   }
}

static void zero_wing_port_wb(UWORD offset, UBYTE data)
{
   switch(offset&0xFF){
      case 0xA8:
         YM3812WriteZ80(0,data);
      break;
      case 0xA9:
         YM3812WriteZ80(1,data);
      break;
      default:
      break;
   }
}

static UBYTE demons_world_port_rb(UWORD offset)
{
   switch(offset&0xFF){
      case 0x00:
         return YM3812ReadZ80(0);
      break;
      case 0x20:
         return get_dsw(2);
      break;
      case 0x60:
         return RAM[0x023003];
      break;
      case 0x80:
         return RAM[0x023004];
      break;
      case 0xA0:
         return get_dsw(1);
      break;
      case 0xC0:
         return RAM[0x023005];
      break;
      case 0xE0:
         return get_dsw(0);
      break;
      default:
         return 0xFF;
      break;
   }
}

static void demons_world_port_wb(UWORD offset, UBYTE data)
{
   switch(offset&0xFF){
      case 0x00:
         YM3812WriteZ80(0,data);
      break;
      case 0x01:
         YM3812WriteZ80(1,data);
      break;
      default:
      break;
   }
}

static UBYTE truxton_port_rb(UWORD offset)
{
   switch(offset&0xFF){
      case 0x00:
         return RAM[0x023004];
      break;
      case 0x10:
         return RAM[0x023005];
      break;
      case 0x20:
         return RAM[0x023003];
      break;
      case 0x40:
         return get_dsw(0);
      break;
      case 0x50:
         return get_dsw(1);
      break;
      case 0x60:
         return YM3812ReadZ80(0);
      break;
      case 0x70:
         return get_dsw(2);
      break;
      default:
         return 0xFF;
      break;
   }
}

static void truxton_port_wb(UWORD offset, UBYTE data)
{
   switch(offset&0xFF){
      case 0x60:
         YM3812WriteZ80(0,data);
      break;
      case 0x61:
         YM3812WriteZ80(1,data);
      break;
      default:
      break;
   }
}

static UBYTE outzone_port_rb(UWORD offset)
{
   switch(offset&0xFF){
      case 0x00:
         return YM3812ReadZ80(0);
      break;
      case 0x08:
         return get_dsw(0);
      break;
      case 0x0C:
         return get_dsw(1);
      break;
      case 0x10:
         return RAM[0x023003];
      break;
      case 0x14:
         return RAM[0x023004];
      break;
      case 0x18:
         return RAM[0x023005];
      break;
      case 0x1C:
         return get_dsw(2);
      break;
      default:
         return 0xFF;
      break;
   }
}

static void outzone_port_wb(UWORD offset, UBYTE data)
{
   switch(offset&0xFF){
      case 0x00:
         YM3812WriteZ80(0,data);
      break;
      case 0x01:
         YM3812WriteZ80(1,data);
      break;
      default:
      break;
   }
}


static UBYTE hell_fire_port_rb(UWORD offset)
{
   switch(offset&0xFF){
      case 0x00:
         return get_dsw(0);
      break;
      case 0x10:
         return get_dsw(1);
      break;
      case 0x20:
         return get_dsw(2);
      break;
      case 0x40:
         return RAM[0x023004];
      break;
      case 0x50:
         return RAM[0x023005];
      break;
      case 0x60:
         return RAM[0x023003];
      break;
      case 0x70:
         return YM3812ReadZ80(0);
      break;
      default:
         return 0xFF;
      break;
   }
}

static void hell_fire_port_wb(UWORD offset, UBYTE data)
{
   switch(offset&0xFF){
      case 0x70:
         YM3812WriteZ80(0,data);
      break;
      case 0x71:
         YM3812WriteZ80(1,data);
      break;
      default:
      break;
   }
}

/******************************************************************************/

static void load_vimana_actual(int set)
{
   int ta,tb,tc;

   romset = set;

   if(!(ROM=AllocateMem(0x40000)))return;
   if(!(RAM=AllocateMem(0x80000)))return;
   if(!(GFX=AllocateMem(0x100000+0x200000)))return;

   GFX_BG0 = GFX+0x000000;
   GFX_SPR = GFX+0x100000;

   if(!load_rom("vim1.bin", RAM, 0x80000))return;           // GFX SPR
   tb=0;
   for(ta=0;ta<0x80000;ta+=2){
      tc=RAM[ta];
      GFX_SPR[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_SPR[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_SPR[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_SPR[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_SPR[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_SPR[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_SPR[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_SPR[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+1];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("vim2.bin", RAM, 0x80000))return;           // GFX SPR
   tb=0;
   for(ta=0;ta<0x80000;ta+=2){
      tc=RAM[ta];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+1];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }


   if(!load_rom("vim6.bin", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("vim5.bin", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_BG0[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_BG0[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_BG0[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_BG0[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_BG0[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_BG0[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_BG0[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("vim4.bin", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("vim3.bin", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x04000+0x08000);

   y_ofs = 8;

   if(!(TileQueue = (struct TILE_Q *) AllocateMem(sizeof(TILE_Q)*MAX_TILES)))return;

   if(set==0){
   if(!load_rom("vim07.bin", RAM, 0x20000))return;             // MAIN 68000
   }
   else{
   if(!load_rom("vimana07.bin", RAM, 0x20000))return;          // MAIN 68000
   }

   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta]=RAM[ta];
   }

   if(set==0){
   if(!load_rom("vim08.bin", RAM, 0x20000))return;             // MAIN 68000
   }
   else{
   if(!load_rom("vimana08.bin", RAM, 0x20000))return;          // MAIN 68000
   }

   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x40000);
   RAMSize=0x40000+0x10000;

   tp1vcu[0].VRAM          = RAM+0x11000;
   tp1vcu[0].RAM_OBJ       = RAM+0x21000;
   tp1vcu[0].SCROLL        = RAM+0x22000;
   tp1vcu[0].GFX_BG        = GFX_BG0;
   tp1vcu[0].MASK_BG       = GFX_BG0_SOLID;
   tp1vcu[0].tile_mask_bg  = 0x03FFF;
   tp1vcu[0].tile_mask_obj = 0x07FFF;
   tp1vcu[0].tile_ofs_obj  = 0x04000;
   init_tp1vcu(0);

   InitPaletteMap(RAM+0x10000, 0x80, 0x10, 0x8000);

   if(set==0){

   // Kill the annoying reset instruction

   WriteWord68k(&ROM[0x1A830],0x4E71);          // nop

   // Something failed a test

   WriteWord68k(&ROM[0x1AA3E],0x4E71);          // nop

   // 68000 speed hack

   WriteLong68k(&ROM[0x17918],0x13FC0000);
   WriteLong68k(&ROM[0x1791C],0x00AA0000);

   }
   else{

   // Kill the annoying reset instruction

   WriteWord68k(&ROM[0x1ab1c],0x4E71);          // nop

   // Something failed a test

   WriteWord68k(&ROM[0x1ad2a],0x4E71);          // nop

   // 68000 speed hack

   WriteLong68k(&ROM[0x17c04],0x13FC0000);
   WriteLong68k(&ROM[0x17c08],0x00AA0000);

   }

/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x40000);
   ByteSwap(RAM,0x40000);

   AddMemFetch(0x000000, 0x03FFFF, ROM+0x000000-0x000000);      // 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadByte(0x480000, 0x483FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadByte(0x404000, 0x4047FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadByte(0x406000, 0x4067FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadByte(0x0C0000, 0x0C000F, tp1vcu_obj_rb, NULL);                // OBJECT
   AddReadByte(0x4C0000, 0x4C001F, tp1vcu_bg_rb, NULL);                 // LAYER
   AddReadByte(0x400000, 0x40000F, tp_vblank_rb, NULL);                 // VSYNC
   AddReadByte(0x440000, 0x440FFF, vimana_z80_rb, NULL);                // SOUND COMM
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);               // <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadWord(0x480000, 0x483FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadWord(0x404000, 0x4047FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadWord(0x406000, 0x4067FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadWord(0x0C0000, 0x0C000F, tp1vcu_obj_rw, NULL);                // OBJECT
   AddReadWord(0x4C0000, 0x4C001F, tp1vcu_bg_rw, NULL);                 // LAYER
   AddReadWord(0x400000, 0x40000F, tp_vblank_rw, NULL);                 // VSYNC
   AddReadWord(0x440000, 0x440FFF, vimana_z80_rw, NULL);                // SOUND COMM
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);               // <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x480000, 0x483FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteByte(0x404000, 0x4047FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteByte(0x406000, 0x4067FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteByte(0x400000, 0x40000F, tp_vblank_wb, NULL);                // VSYNC
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);                   // Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);             // <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x480000, 0x483FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteWord(0x404000, 0x4047FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteWord(0x406000, 0x4067FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteWord(0x0C0000, 0x0C000F, tp1vcu_obj_ww, NULL);               // OBJECT
   AddWriteWord(0x4C0000, 0x4C001F, tp1vcu_bg_ww, NULL);                // LAYER
   AddWriteWord(0x400000, 0x40000F, tp_vblank_ww, NULL);                // VSYNC
   AddWriteWord(0x080000, 0x080003, tp1vcu_ofsreg_ww, NULL);            // OFFSET
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);             // <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();     // Set Starscream mem pointers... 

}

void load_vimana(void)
{
   load_vimana_actual(0);
}

void load_vimana_alt(void)
{
   load_vimana_actual(1);
}

void load_outzone(void)
{
   int ta,tb,tc;

   romset = 2;

   if(!(ROM=AllocateMem(0x40000)))return;
   if(!(RAM=AllocateMem(0x80000)))return;
   if(!(GFX=AllocateMem(0x200000+0x100000)))return;

   GFX_BG0 = GFX+0x000000;
   GFX_SPR = GFX+0x200000;

   if(!load_rom("rom5.bin", RAM, 0x80000))return;           // GFX BG0
   tb=0;
   for(ta=0;ta<0x80000;ta+=2){
      tc=RAM[ta];
      GFX_BG0[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_BG0[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_BG0[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_BG0[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_BG0[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_BG0[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_BG0[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_BG0[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+1];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("rom6.bin", RAM, 0x80000))return;           // GFX BG0
   tb=0;
   for(ta=0;ta<0x80000;ta+=2){
      tc=RAM[ta];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+1];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }


   if(!load_rom("rom2.bin", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("rom1.bin", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_SPR[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_SPR[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_SPR[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_SPR[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_SPR[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_SPR[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_SPR[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("rom3.bin", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("rom4.bin", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x08000+0x04000);

   y_ofs = 8;

   if(!(TileQueue = (struct TILE_Q *) AllocateMem(sizeof(TILE_Q)*MAX_TILES)))return;

   if(!load_rom("rom7.bin", RAM, 0x20000))return;             // MAIN 68000
   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   if(!load_rom("rom8.bin", RAM, 0x20000))return;
   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x40000);
   RAMSize=0x40000+0x10000;

   /* Sound Setup */

   Z80ROM=RAM+0x40000;
   if(!load_rom("rom9.bin", Z80ROM, 0x08000)) return;
/*
   // Fix Checksum

   Z80ROM[0x0156]=0x00; // NOP
   Z80ROM[0x0157]=0x00; // NOP
   Z80ROM[0x0158]=0x00; // NOP

   // Apply Speed Patch

   Z80ROM[0x019D]=0xD3; // OUTA (AAh)
   Z80ROM[0x019E]=0xAA; //

   SetStopZ80Mode2(0x019C);
*/
   // Setup Z80 memory map
   // --------------------

   AddZ80AROMBase(Z80ROM, 0x0038, 0x0066);

   AddZ80AReadByte(0x0000, 0x87FF, NULL,                Z80ROM+0x0000); // Z80 ROM/RAM
   AddZ80AReadByte(0x0000, 0xFFFF, DefBadReadZ80,       NULL);
   AddZ80AReadByte(    -1,     -1, NULL,                NULL);

   AddZ80AWriteByte(0x8000, 0x87FF, NULL,               Z80ROM+0x8000); // Z80 RAM
   AddZ80AWriteByte(0x0000, 0xFFFF, DefBadWriteZ80,     NULL);
   AddZ80AWriteByte(    -1,     -1, NULL,               NULL);

   AddZ80AReadPort(0x00, 0xFF, outzone_port_rb,         NULL);
   AddZ80AReadPort(0x00, 0xFF, DefBadReadZ80,           NULL);
   AddZ80AReadPort(  -1,   -1, NULL,                    NULL);

   AddZ80AWritePort(0xAA, 0xAA, StopZ80Mode2,           NULL);
   AddZ80AWritePort(0x00, 0xFF, outzone_port_wb,        NULL);
   AddZ80AWritePort(0x00, 0xFF, DefBadWriteZ80,         NULL);
   AddZ80AWritePort(  -1,   -1, NULL,                   NULL);

   AddZ80AInit();

   /* ----------- */

   tp1vcu[0].VRAM          = RAM+0x11000;
   tp1vcu[0].RAM_OBJ       = RAM+0x21000;
   tp1vcu[0].SCROLL        = RAM+0x22000;
   tp1vcu[0].GFX_BG        = GFX_BG0;
   tp1vcu[0].MASK_BG       = GFX_BG0_SOLID;
   tp1vcu[0].tile_mask_bg  = 0x07FFF;
   tp1vcu[0].tile_mask_obj = 0x03FFF;
   tp1vcu[0].tile_ofs_obj  = 0x08000;
   init_tp1vcu(0);

   InitPaletteMap(RAM+0x10000, 0x80, 0x10, 0x8000);

   // Kill the annoying reset instruction

   WriteWord68k(&ROM[0x124bc],0x4E71);          // nop
/*
   // Something failed a test

   WriteWord68k(&ROM[0x1AA3E],0x4E71);          // nop

   // 68000 speed hack

   WriteLong68k(&ROM[0x17918],0x13FC0000);
   WriteLong68k(&ROM[0x1791C],0x00AA0000);
*/
/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x40000);
   ByteSwap(RAM,0x40000);

   AddMemFetch(0x000000, 0x03FFFF, ROM+0x000000-0x000000);      // 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadByte(0x240000, 0x243FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadByte(0x304000, 0x3047FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadByte(0x306000, 0x3067FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadByte(0x100000, 0x10000F, tp1vcu_obj_rb, NULL);                // OBJECT
   AddReadByte(0x200000, 0x20001F, tp1vcu_bg_rb, NULL);                 // LAYER
   AddReadByte(0x300000, 0x30000F, tp_vblank_rb, NULL);                 // VSYNC
   AddReadByte(0x140000, 0x140FFF, tp1_z80_rb, NULL);                   // SOUND COMM
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);               // <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadWord(0x240000, 0x243FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadWord(0x304000, 0x3047FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadWord(0x306000, 0x3067FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadWord(0x100000, 0x10000F, tp1vcu_obj_rw, NULL);                // OBJECT
   AddReadWord(0x200000, 0x20001F, tp1vcu_bg_rw, NULL);                 // LAYER
   AddReadWord(0x300000, 0x30000F, tp_vblank_rw, NULL);                 // VSYNC
   AddReadWord(0x140000, 0x140FFF, tp1_z80_rw, NULL);                   // SOUND COMM
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);               // <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x240000, 0x243FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteByte(0x304000, 0x3047FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteByte(0x306000, 0x3067FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteByte(0x300000, 0x30000F, tp_vblank_wb, NULL);                // VSYNC
   AddWriteByte(0x140000, 0x140FFF, tp1_z80_wb, NULL);                  // SOUND COMM
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);                   // Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);             // <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x240000, 0x243FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteWord(0x304000, 0x3047FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteWord(0x306000, 0x3067FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteWord(0x100000, 0x10000F, tp1vcu_obj_ww, NULL);               // OBJECT
   AddWriteWord(0x200000, 0x20001F, tp1vcu_bg_ww, NULL);                // LAYER
   AddWriteWord(0x300000, 0x30000F, tp_vblank_ww, NULL);                // VSYNC
   AddWriteWord(0x140000, 0x140FFF, tp1_z80_ww, NULL);                  // SOUND COMM
   AddWriteWord(0x340000, 0x340003, tp1vcu_ofsreg_ww, NULL);            // OFFSET
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);             // <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();     // Set Starscream mem pointers... 

}

void load_rally_bike(void)
{
   int ta,tb,tc;

   romset = 3;

   if(!(ROM=AllocateMem(0x80000)))return;
   if(!(RAM=AllocateMem(0x80000)))return;
   if(!(GFX=AllocateMem(0x100000+0x080000)))return;

   GFX_BG0 = GFX+0x000000;
   GFX_SPR = GFX+0x100000;

   if(!load_rom("rb_09.bin", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("rb_08.bin", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_BG0[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_BG0[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_BG0[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_BG0[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_BG0[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_BG0[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_BG0[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("rb_07.bin", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("rb_06.bin", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }


   if(!load_rom("rb_13.bin", RAM+0x00000, 0x10000))return;   // GFX SPR
   if(!load_rom("rb_12.bin", RAM+0x10000, 0x10000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x10000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_SPR[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_SPR[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_SPR[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_SPR[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_SPR[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_SPR[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_SPR[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x10000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("rb_10.bin", RAM+0x00000, 0x10000))return;   // GFX SPR
   if(!load_rom("rb_11.bin", RAM+0x10000, 0x10000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x10000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x10000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x04000+0x02000);

   y_ofs = 8;

   if(!(TileQueue = (struct TILE_Q *) AllocateMem(sizeof(TILE_Q)*MAX_TILES)))return;

   if(!load_rom("rb_02.bin", RAM+0x00000, 0x08000))return;     // MAIN 68000
   memset(RAM+0x08000,0x00,0x18000);                           // SPACE
   if(!load_rom("rb_04.bin", RAM+0x20000, 0x20000))return;     // MAIN 68000
   for(ta=0;ta<0x40000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   if(!load_rom("rb_01.bin", RAM+0x00000, 0x08000))return;     // MAIN 68000
   memset(RAM+0x08000,0x00,0x18000);                           // SPACE
   if(!load_rom("rb_03.bin", RAM+0x20000, 0x20000))return;     // MAIN 68000
   for(ta=0;ta<0x40000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x40000);
   RAMSize=0x40000+0x10000;

   /* Sound Setup */

   Z80ROM=RAM+0x40000;
   if(!load_rom("rb_05.bin", Z80ROM, 0x08000)) return;
/*
   // Fix Checksum

   Z80ROM[0x0156]=0x00; // NOP
   Z80ROM[0x0157]=0x00; // NOP
   Z80ROM[0x0158]=0x00; // NOP

   // Apply Speed Patch

   Z80ROM[0x019D]=0xD3; // OUTA (AAh)
   Z80ROM[0x019E]=0xAA; //

   SetStopZ80Mode2(0x019C);
*/
   // Setup Z80 memory map
   // --------------------

   AddZ80AROMBase(Z80ROM, 0x0038, 0x0066);

   AddZ80AReadByte(0x0000, 0x87FF, NULL,                Z80ROM+0x0000); // Z80 ROM/RAM
   AddZ80AReadByte(0x0000, 0xFFFF, DefBadReadZ80,       NULL);
   AddZ80AReadByte(    -1,     -1, NULL,                NULL);

   AddZ80AWriteByte(0x8000, 0x87FF, NULL,               Z80ROM+0x8000); // Z80 RAM
   AddZ80AWriteByte(0x0000, 0xFFFF, DefBadWriteZ80,     NULL);
   AddZ80AWriteByte(    -1,     -1, NULL,               NULL);

   AddZ80AReadPort(0x00, 0xFF, truxton_port_rb,         NULL);
   AddZ80AReadPort(0x00, 0xFF, DefBadReadZ80,           NULL);
   AddZ80AReadPort(  -1,   -1, NULL,                    NULL);

   AddZ80AWritePort(0xAA, 0xAA, StopZ80Mode2,           NULL);
   AddZ80AWritePort(0x00, 0xFF, truxton_port_wb,        NULL);
   AddZ80AWritePort(0x00, 0xFF, DefBadWriteZ80,         NULL);
   AddZ80AWritePort(  -1,   -1, NULL,                   NULL);

   AddZ80AInit();

   /* ----------- */

   tp1vcu[0].VRAM          = RAM+0x11000;
   tp1vcu[0].RAM_OBJ       = RAM+0x21000;
   tp1vcu[0].SCROLL        = RAM+0x22000;
   tp1vcu[0].GFX_BG        = GFX_BG0;
   tp1vcu[0].MASK_BG       = GFX_BG0_SOLID;
   tp1vcu[0].tile_mask_bg  = 0x03FFF;
   tp1vcu[0].tile_mask_obj = 0x01FFF;
   tp1vcu[0].tile_ofs_obj  = 0x04000;
   init_tp1vcu(0);

   InitPaletteMap(RAM+0x10000, 0x80, 0x10, 0x8000);

   // Something failed a test

   WriteWord68k(&ROM[0x0193E],0x6000);

   // Fix rom checksum

   WriteWord68k(&ROM[0x01A66],0x7400);

   // 68000 speed hack

   WriteLong68k(&ROM[0x01BA0],0x13FC0000);
   WriteLong68k(&ROM[0x01BA4],0x00AA0000);

/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x80000);
   ByteSwap(RAM,0x40000);

   AddMemFetch(0x000000, 0x07FFFF, ROM+0x000000-0x000000);      // 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x07FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadByte(0x080000, 0x083FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadByte(0x144000, 0x1447FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadByte(0x146000, 0x1467FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadByte(0x0C0000, 0x0C0FFF, NULL, RAM+0x021000);                 // OBJECT
   AddReadByte(0x100000, 0x10001F, tp1vcu_bg_rb, NULL);                 // LAYER
   AddReadByte(0x140000, 0x14000F, tp_vblank_rb, NULL);                 // VSYNC
   AddReadByte(0x180000, 0x180FFF, tp1_z80_rb, NULL);                   // SOUND COMM
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);               // <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x07FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadWord(0x080000, 0x083FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadWord(0x144000, 0x1447FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadWord(0x146000, 0x1467FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadWord(0x0C0000, 0x0C0FFF, NULL, RAM+0x021000);                 // OBJECT
   AddReadWord(0x100000, 0x10001F, tp1vcu_bg_rw, NULL);                 // LAYER
   AddReadWord(0x140000, 0x14000F, tp_vblank_rw, NULL);                 // VSYNC
   AddReadWord(0x180000, 0x180FFF, tp1_z80_rw, NULL);                   // SOUND COMM
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);               // <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x080000, 0x083FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteByte(0x144000, 0x1447FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteByte(0x146000, 0x1467FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteByte(0x0C0000, 0x0C0FFF, NULL, RAM+0x021000);                // OBJECT
   AddWriteByte(0x140000, 0x14000F, tp_vblank_wb, NULL);                // VSYNC
   AddWriteByte(0x180000, 0x180FFF, tp1_z80_wb, NULL);                  // SOUND COMM
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);                   // Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);             // <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x080000, 0x083FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteWord(0x144000, 0x1447FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteWord(0x146000, 0x1467FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteWord(0x0C0000, 0x0C0FFF, NULL, RAM+0x021000);                // OBJECT
   AddWriteWord(0x100000, 0x10001F, tp1vcu_bg_ww, NULL);                // LAYER
   AddWriteWord(0x140000, 0x14000F, tp_vblank_ww, NULL);                // VSYNC
   AddWriteWord(0x180000, 0x180FFF, tp1_z80_ww, NULL);                  // SOUND COMM
   AddWriteWord(0x1C0000, 0x1C0003, tp1vcu_ofsreg_ww, NULL);            // OFFSET
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);             // <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();     // Set Starscream mem pointers... 

}

void load_zero_wing(void)
{
   int ta,tb,tc;

   romset = 4;

   if(!(ROM=AllocateMem(0x80000)))return;
   if(!(RAM=AllocateMem(0x80000)))return;
   if(!(GFX=AllocateMem(0x100000+0x100000)))return;

   GFX_BG0 = GFX+0x000000;
   GFX_SPR = GFX+0x100000;

   if(!load_rom("o15-05.rom", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("o15-06.rom", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_BG0[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_BG0[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_BG0[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_BG0[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_BG0[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_BG0[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_BG0[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("o15-07.rom", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("o15-08.rom", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }


   if(!load_rom("o15-03.rom", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("o15-04.rom", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_SPR[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_SPR[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_SPR[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_SPR[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_SPR[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_SPR[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_SPR[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("o15-01.rom", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("o15-02.rom", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x04000+0x04000);

   y_ofs = 8;

   if(!(TileQueue = (struct TILE_Q *) AllocateMem(sizeof(TILE_Q)*MAX_TILES)))return;

   if(!load_rom("o15-11.rom", RAM+0x00000, 0x08000))return;     // MAIN 68000
   memset(RAM+0x08000,0x00,0x18000);                           // SPACE
   if(!load_rom("o15-09.rom", RAM+0x20000, 0x20000))return;     // MAIN 68000
   for(ta=0;ta<0x40000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   if(!load_rom("o15-12.rom", RAM+0x00000, 0x08000))return;     // MAIN 68000
   memset(RAM+0x08000,0x00,0x18000);                           // SPACE
   if(!load_rom("o15-10.rom", RAM+0x20000, 0x20000))return;     // MAIN 68000
   for(ta=0;ta<0x40000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   RAMSize=0x40000+0x10000;
   memset(RAM+0x00000,0x00,RAMSize);

   /* Sound Setup */

   Z80ROM=RAM+0x40000;
   if(!load_rom("o15-13.rom", Z80ROM, 0x08000)) return;
/*
   // Fix Checksum

   Z80ROM[0x0156]=0x00; // NOP
   Z80ROM[0x0157]=0x00; // NOP
   Z80ROM[0x0158]=0x00; // NOP

   // Apply Speed Patch

   Z80ROM[0x019D]=0xD3; // OUTA (AAh)
   Z80ROM[0x019E]=0xAA; //

   SetStopZ80Mode2(0x019C);
*/
   // Setup Z80 memory map
   // --------------------

   AddZ80AROMBase(Z80ROM, 0x0038, 0x0066);

   AddZ80AReadByte(0x0000, 0x87FF, NULL,                Z80ROM+0x0000); // Z80 ROM/RAM
   AddZ80AReadByte(0x0000, 0xFFFF, DefBadReadZ80,       NULL);
   AddZ80AReadByte(    -1,     -1, NULL,                NULL);

   AddZ80AWriteByte(0x8000, 0x87FF, NULL,               Z80ROM+0x8000); // Z80 RAM
   AddZ80AWriteByte(0x0000, 0xFFFF, DefBadWriteZ80,     NULL);
   AddZ80AWriteByte(    -1,     -1, NULL,               NULL);

   AddZ80AReadPort(0x00, 0xFF, zero_wing_port_rb,       NULL);
   AddZ80AReadPort(0x00, 0xFF, DefBadReadZ80,           NULL);
   AddZ80AReadPort(  -1,   -1, NULL,                    NULL);

   AddZ80AWritePort(0xAA, 0xAA, StopZ80Mode2,           NULL);
   AddZ80AWritePort(0x00, 0xFF, zero_wing_port_wb,      NULL);
   AddZ80AWritePort(0x00, 0xFF, DefBadWriteZ80,         NULL);
   AddZ80AWritePort(  -1,   -1, NULL,                   NULL);

   AddZ80AInit();

   /* ----------- */

   tp1vcu[0].VRAM          = RAM+0x11000;
   tp1vcu[0].RAM_OBJ       = RAM+0x21000;
   tp1vcu[0].SCROLL        = RAM+0x22000;
   tp1vcu[0].GFX_BG        = GFX_BG0;
   tp1vcu[0].MASK_BG       = GFX_BG0_SOLID;
   tp1vcu[0].tile_mask_bg  = 0x03FFF;
   tp1vcu[0].tile_mask_obj = 0x03FFF;
   tp1vcu[0].tile_ofs_obj  = 0x04000;
   init_tp1vcu(0);

   InitPaletteMap(RAM+0x10000, 0x80, 0x10, 0x8000);

   // Kill the annoying reset instruction

   WriteWord68k(&ROM[0x0023C],0x4E71);          // nop

   // Fix rom checksum

   WriteLong68k(&ROM[0x004DA],0x4E714E71);      // nop

   // Fix sound error

   WriteLong68k(&ROM[0x004E6],0x4E714E71);      // nop
   WriteLong68k(&ROM[0x0050C],0x4E714E71);      // nop

/*
   // 68000 speed hack

   WriteLong68k(&ROM[0x17918],0x13FC0000);
   WriteLong68k(&ROM[0x1791C],0x00AA0000);
*/
/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x80000);
   ByteSwap(RAM,0x40000);

   AddMemFetch(0x000000, 0x07FFFF, ROM+0x000000-0x000000);      // 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x07FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadByte(0x080000, 0x083FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadByte(0x404000, 0x4047FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadByte(0x406000, 0x4067FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadByte(0x4C0000, 0x4C000F, tp1vcu_obj_rb, NULL);                // OBJECT
   AddReadByte(0x480000, 0x48001F, tp1vcu_bg_rb, NULL);                 // LAYER
   AddReadByte(0x400000, 0x40000F, tp_vblank_rb, NULL);                 // VSYNC
   AddReadByte(0x440000, 0x440FFF, tp1_z80_rb, NULL);                   // SOUND COMM
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);               // <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x07FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadWord(0x080000, 0x083FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadWord(0x404000, 0x4047FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadWord(0x406000, 0x4067FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadWord(0x4C0000, 0x4C000F, tp1vcu_obj_rw, NULL);                // OBJECT
   AddReadWord(0x480000, 0x48001F, tp1vcu_bg_rw, NULL);                 // LAYER
   AddReadWord(0x400000, 0x40000F, tp_vblank_rw, NULL);                 // VSYNC
   AddReadWord(0x440000, 0x440FFF, tp1_z80_rw, NULL);                   // SOUND COMM
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);               // <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x080000, 0x083FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteByte(0x404000, 0x4047FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteByte(0x406000, 0x4067FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteByte(0x400000, 0x40000F, tp_vblank_wb, NULL);                // VSYNC
   AddWriteByte(0x440000, 0x440FFF, tp1_z80_wb, NULL);                  // SOUND COMM
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);                   // Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);             // <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x080000, 0x083FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteWord(0x404000, 0x4047FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteWord(0x406000, 0x4067FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteWord(0x4C0000, 0x4C000F, tp1vcu_obj_ww, NULL);               // OBJECT
   AddWriteWord(0x480000, 0x48001F, tp1vcu_bg_ww, NULL);                // LAYER
   AddWriteWord(0x400000, 0x40000F, tp_vblank_ww, NULL);                // VSYNC
   AddWriteWord(0x440000, 0x440FFF, tp1_z80_ww, NULL);                  // SOUND COMM
   AddWriteWord(0x080000, 0x080003, tp1vcu_ofsreg_ww, NULL);            // OFFSET
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);             // <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();     // Set Starscream mem pointers... 

}

void load_demons_world(void)
{
   int ta,tb,tc;

   romset = 5;

   if(!(ROM=AllocateMem(0x40000)))return;
   if(!(RAM=AllocateMem(0x80000)))return;
   if(!(GFX=AllocateMem(0x100000+0x100000)))return;

   GFX_BG0 = GFX+0x000000;
   GFX_SPR = GFX+0x100000;

   if(!load_rom("rom05", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("rom07", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_BG0[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_BG0[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_BG0[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_BG0[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_BG0[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_BG0[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_BG0[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("rom06", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("rom08", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }


   if(!load_rom("rom01", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("rom02", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_SPR[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_SPR[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_SPR[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_SPR[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_SPR[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_SPR[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_SPR[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("rom03", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("rom04", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x04000+0x04000);

   y_ofs = 8;

   if(!(TileQueue = (struct TILE_Q *) AllocateMem(sizeof(TILE_Q)*MAX_TILES)))return;

   if(!load_rom("rom10", RAM+0x00000, 0x20000))return;     // MAIN 68000
   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   if(!load_rom("rom09", RAM+0x00000, 0x20000))return;     // MAIN 68000
   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x40000);
   RAMSize=0x40000+0x10000;

   /* Sound Setup */

   Z80ROM=RAM+0x40000;
   if(!load_rom("rom11", Z80ROM, 0x08000)) return;
/*
   // Fix Checksum

   Z80ROM[0x0156]=0x00; // NOP
   Z80ROM[0x0157]=0x00; // NOP
   Z80ROM[0x0158]=0x00; // NOP

   // Apply Speed Patch

   Z80ROM[0x019D]=0xD3; // OUTA (AAh)
   Z80ROM[0x019E]=0xAA; //

   SetStopZ80Mode2(0x019C);
*/
   // Setup Z80 memory map
   // --------------------

   AddZ80AROMBase(Z80ROM, 0x0038, 0x0066);

   AddZ80AReadByte(0x0000, 0x87FF, NULL,                Z80ROM+0x0000); // Z80 ROM/RAM
   AddZ80AReadByte(0x0000, 0xFFFF, DefBadReadZ80,       NULL);
   AddZ80AReadByte(    -1,     -1, NULL,                NULL);

   AddZ80AWriteByte(0x8000, 0x87FF, NULL,               Z80ROM+0x8000); // Z80 RAM
   AddZ80AWriteByte(0x0000, 0xFFFF, DefBadWriteZ80,     NULL);
   AddZ80AWriteByte(    -1,     -1, NULL,               NULL);

   AddZ80AReadPort(0x00, 0xFF, demons_world_port_rb,    NULL);
   AddZ80AReadPort(0x00, 0xFF, DefBadReadZ80,           NULL);
   AddZ80AReadPort(  -1,   -1, NULL,                    NULL);

   AddZ80AWritePort(0xAA, 0xAA, StopZ80Mode2,           NULL);
   AddZ80AWritePort(0x00, 0xFF, demons_world_port_wb,   NULL);
   AddZ80AWritePort(0x00, 0xFF, DefBadWriteZ80,         NULL);
   AddZ80AWritePort(  -1,   -1, NULL,                   NULL);

   AddZ80AInit();

   /* ----------- */

   tp1vcu[0].VRAM          = RAM+0x11000;
   tp1vcu[0].RAM_OBJ       = RAM+0x21000;
   tp1vcu[0].SCROLL        = RAM+0x22000;
   tp1vcu[0].GFX_BG        = GFX_BG0;
   tp1vcu[0].MASK_BG       = GFX_BG0_SOLID;
   tp1vcu[0].tile_mask_bg  = 0x03FFF;
   tp1vcu[0].tile_mask_obj = 0x03FFF;
   tp1vcu[0].tile_ofs_obj  = 0x04000;
   init_tp1vcu(0);

   InitPaletteMap(RAM+0x10000, 0x80, 0x10, 0x8000);

   // Kill the annoying reset instruction
/*
   WriteWord68k(&ROM[0x0023C],0x4E71);          // nop

   // Fix rom checksum

   WriteLong68k(&ROM[0x004DA],0x4E714E71);      // nop
*/
   // Fix sound error

   WriteLong68k(&ROM[0x0188C],0x4E714E71);      // nop
   WriteLong68k(&ROM[0x018B2],0x4E714E71);      // nop

/*
   // 68000 speed hack

   WriteLong68k(&ROM[0x17918],0x13FC0000);
   WriteLong68k(&ROM[0x1791C],0x00AA0000);
*/
   // Fix protection/mcu?

   WriteWord68k(&ROM[0x0181C],0x4E71);
   WriteWord68k(&ROM[0x01824],0x600A);

/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x40000);
   ByteSwap(RAM,0x40000);

   AddMemFetch(0x000000, 0x03FFFF, ROM+0x000000-0x000000);      // 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadByte(0xC00000, 0xC03FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadByte(0x404000, 0x4047FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadByte(0x406000, 0x4067FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadByte(0xA00000, 0xA0000F, tp1vcu_obj_rb, NULL);                // OBJECT
   AddReadByte(0x800000, 0x80001F, tp1vcu_bg_rb, NULL);                 // LAYER
   AddReadByte(0x400000, 0x40000F, tp_vblank_rb, NULL);                 // VSYNC
   AddReadByte(0x600000, 0x600FFF, tp1_z80_rb, NULL);                   // SOUND COMM
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);               // <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadWord(0xC00000, 0xC03FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadWord(0x404000, 0x4047FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadWord(0x406000, 0x4067FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadWord(0xA00000, 0xA0000F, tp1vcu_obj_rw, NULL);                // OBJECT
   AddReadWord(0x800000, 0x80001F, tp1vcu_bg_rw, NULL);                 // LAYER
   AddReadWord(0x400000, 0x40000F, tp_vblank_rw, NULL);                 // VSYNC
   AddReadWord(0x600000, 0x600FFF, tp1_z80_rw, NULL);                   // SOUND COMM
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);               // <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0xC00000, 0xC03FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteByte(0x404000, 0x4047FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteByte(0x406000, 0x4067FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteByte(0x400000, 0x40000F, tp_vblank_wb, NULL);                // VSYNC
   AddWriteByte(0x600000, 0x600FFF, tp1_z80_wb, NULL);                  // SOUND COMM
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);                   // Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);             // <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0xC00000, 0xC03FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteWord(0x404000, 0x4047FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteWord(0x406000, 0x4067FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteWord(0xA00000, 0xA0000F, tp1vcu_obj_ww, NULL);               // OBJECT
   AddWriteWord(0x800000, 0x80001F, tp1vcu_bg_ww, NULL);                // LAYER
   AddWriteWord(0x400000, 0x40000F, tp_vblank_ww, NULL);                // VSYNC
   AddWriteWord(0x600000, 0x600FFF, tp1_z80_ww, NULL);                  // SOUND COMM
   AddWriteWord(0xE00000, 0xE00003, tp1vcu_ofsreg_ww, NULL);            // OFFSET
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);             // <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();     // Set Starscream mem pointers... 

}

static void load_fire_shark_actual(int set)
{
   int ta,tb,tc;

   romset = set + 6;

   if(!(ROM=AllocateMem(0x80000)))return;
   if(!(RAM=AllocateMem(0x80000)))return;
   if(!(GFX=AllocateMem(0x100000+0x100000)))return;

   GFX_BG0 = GFX+0x000000;
   GFX_SPR = GFX+0x100000;

   if(!load_rom("o17_05.bin", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("o17_06.bin", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_BG0[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_BG0[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_BG0[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_BG0[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_BG0[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_BG0[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_BG0[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("o17_07.bin", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("o17_08.bin", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }


   if(!load_rom("o17_01.bin", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("o17_02.bin", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_SPR[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_SPR[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_SPR[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_SPR[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_SPR[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_SPR[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_SPR[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("o17_03.bin", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("o17_04.bin", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x04000+0x04000);

   y_ofs = 8;

   if(!(TileQueue = (struct TILE_Q *) AllocateMem(sizeof(TILE_Q)*MAX_TILES)))return;

   if(set==0){
   if(!load_rom("9.bin", RAM+0x00000, 0x08000))return;          // MAIN 68000
   memset(RAM+0x08000,0x00,0x18000);                            // SPACE
   if(!load_rom("11.bin", RAM+0x20000, 0x20000))return;         // MAIN 68000
   }
   else{
   if(!load_rom("o17_09.bin", RAM+0x00000, 0x08000))return;     // MAIN 68000
   memset(RAM+0x08000,0x00,0x18000);                            // SPACE
   if(!load_rom("o17_11.bin", RAM+0x20000, 0x20000))return;     // MAIN 68000
   }

   for(ta=0;ta<0x40000;ta++){
      ROM[ta+ta]=RAM[ta];
   }

   if(set==0){
   if(!load_rom("10.bin", RAM+0x00000, 0x08000))return;         // MAIN 68000
   memset(RAM+0x08000,0x00,0x18000);                            // SPACE
   if(!load_rom("12.bin", RAM+0x20000, 0x20000))return;         // MAIN 68000
   }
   else{
   if(!load_rom("o17_10.bin", RAM+0x00000, 0x08000))return;     // MAIN 68000
   memset(RAM+0x08000,0x00,0x18000);                            // SPACE
   if(!load_rom("o17_12.bin", RAM+0x20000, 0x20000))return;     // MAIN 68000
   }

   for(ta=0;ta<0x40000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x40000);
   RAMSize=0x40000+0x10000;

   tp1vcu[0].VRAM          = RAM+0x11000;
   tp1vcu[0].RAM_OBJ       = RAM+0x21000;
   tp1vcu[0].SCROLL        = RAM+0x22000;
   tp1vcu[0].GFX_BG        = GFX_BG0;
   tp1vcu[0].MASK_BG       = GFX_BG0_SOLID;
   tp1vcu[0].tile_mask_bg  = 0x03FFF;
   tp1vcu[0].tile_mask_obj = 0x03FFF;
   tp1vcu[0].tile_ofs_obj  = 0x04000;
   init_tp1vcu(0);

   InitPaletteMap(RAM+0x10000, 0x80, 0x10, 0x8000);

   if(set==0){

   // Kill the annoying reset instruction

   WriteWord68k(&ROM[0x06f6c],0x4E71);          // nop
/*
   // Something failed a test

   WriteWord68k(&ROM[0x1AA3E],0x4E71);          // nop

   // 68000 speed hack

   WriteLong68k(&ROM[0x17918],0x13FC0000);
   WriteLong68k(&ROM[0x1791C],0x00AA0000);
*/
   }
   else{

   // Kill the annoying reset instruction

   WriteWord68k(&ROM[0x06c56],0x4E71);          // nop
/*
   // Something failed a test

   WriteWord68k(&ROM[0x1ad2a],0x4E71);          // nop

   // 68000 speed hack

   WriteLong68k(&ROM[0x17c04],0x13FC0000);
   WriteLong68k(&ROM[0x17c08],0x00AA0000);
*/
   }

/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x80000);
   ByteSwap(RAM,0x40000);

   AddMemFetch(0x000000, 0x07FFFF, ROM+0x000000-0x000000);      // 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x07FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadByte(0x0C0000, 0x0C3FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadByte(0x104000, 0x1047FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadByte(0x106000, 0x1067FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadByte(0x1C0000, 0x1C000F, tp1vcu_obj_rb, NULL);                // OBJECT
   AddReadByte(0x180000, 0x18001F, tp1vcu_bg_rb, NULL);                 // LAYER
   AddReadByte(0x100000, 0x10000F, tp_vblank_rb, NULL);                 // VSYNC
   AddReadByte(0x140000, 0x140FFF, vimana_z80_rb, NULL);                // SOUND COMM
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);               // <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x07FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadWord(0x0C0000, 0x0C3FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadWord(0x104000, 0x1047FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadWord(0x106000, 0x1067FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadWord(0x1C0000, 0x1C000F, tp1vcu_obj_rw, NULL);                // OBJECT
   AddReadWord(0x180000, 0x18001F, tp1vcu_bg_rw, NULL);                 // LAYER
   AddReadWord(0x100000, 0x10000F, tp_vblank_rw, NULL);                 // VSYNC
   AddReadWord(0x140000, 0x140FFF, vimana_z80_rw, NULL);                // SOUND COMM
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);               // <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x0C0000, 0x0C3FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteByte(0x104000, 0x1047FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteByte(0x106000, 0x1067FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteByte(0x100000, 0x10000F, tp_vblank_wb, NULL);                // VSYNC
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);                   // Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);             // <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x0C0000, 0x0C3FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteWord(0x104000, 0x1047FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteWord(0x106000, 0x1067FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteWord(0x1C0000, 0x1C000F, tp1vcu_obj_ww, NULL);               // OBJECT
   AddWriteWord(0x180000, 0x18001F, tp1vcu_bg_ww, NULL);                // LAYER
   AddWriteWord(0x100000, 0x10000F, tp_vblank_ww, NULL);                // VSYNC
   AddWriteWord(0x080000, 0x080003, tp1vcu_ofsreg_ww, NULL);            // OFFSET
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);             // <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();     // Set Starscream mem pointers... 

}

void load_fire_shark(void)
{
   load_fire_shark_actual(0);
}

void load_same_same_same(void)
{
   load_fire_shark_actual(1);
}

void load_hell_fire(void)
{
   int ta,tb,tc;

   romset = 8;

   if(!(ROM=AllocateMem(0x40000)))return;
   if(!(RAM=AllocateMem(0x80000)))return;
   if(!(GFX=AllocateMem(0x100000+0x100000)))return;

   GFX_BG0 = GFX+0x000000;
   GFX_SPR = GFX+0x100000;

   if(!load_rom("b90-04.bin", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("b90-05.bin", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_BG0[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_BG0[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_BG0[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_BG0[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_BG0[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_BG0[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_BG0[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("b90-06.bin", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("b90-07.bin", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }


   if(!load_rom("b90-11.bin", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("b90-10.bin", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_SPR[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_SPR[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_SPR[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_SPR[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_SPR[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_SPR[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_SPR[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("b90-09.bin", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("b90-08.bin", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x04000+0x04000);

   y_ofs = 8;

   if(!(TileQueue = (struct TILE_Q *) AllocateMem(sizeof(TILE_Q)*MAX_TILES)))return;

   if(!load_rom("b90-14.bin", RAM+0x00000, 0x20000))return;     // MAIN 68000
   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   if(!load_rom("b90-15.bin", RAM+0x00000, 0x20000))return;     // MAIN 68000
   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   memset(RAM+0x00000,0x00,0x40000);
   RAMSize=0x40000+0x10000;

   /* Sound Setup */

   Z80ROM=RAM+0x40000;
   if(!load_rom("b90-03.bin", Z80ROM, 0x08000)) return;
/*
   // Fix Checksum

   Z80ROM[0x0156]=0x00; // NOP
   Z80ROM[0x0157]=0x00; // NOP
   Z80ROM[0x0158]=0x00; // NOP

   // Apply Speed Patch

   Z80ROM[0x019D]=0xD3; // OUTA (AAh)
   Z80ROM[0x019E]=0xAA; //

   SetStopZ80Mode2(0x019C);
*/
   // Setup Z80 memory map
   // --------------------

   AddZ80AROMBase(Z80ROM, 0x0038, 0x0066);

   AddZ80AReadByte(0x0000, 0x87FF, NULL,                Z80ROM+0x0000); // Z80 ROM/RAM
   AddZ80AReadByte(0x0000, 0xFFFF, DefBadReadZ80,       NULL);
   AddZ80AReadByte(    -1,     -1, NULL,                NULL);

   AddZ80AWriteByte(0x8000, 0x87FF, NULL,               Z80ROM+0x8000); // Z80 RAM
   AddZ80AWriteByte(0x0000, 0xFFFF, DefBadWriteZ80,     NULL);
   AddZ80AWriteByte(    -1,     -1, NULL,               NULL);

   AddZ80AReadPort(0x00, 0xFF, hell_fire_port_rb,       NULL);
   AddZ80AReadPort(0x00, 0xFF, DefBadReadZ80,           NULL);
   AddZ80AReadPort(  -1,   -1, NULL,                    NULL);

   AddZ80AWritePort(0xAA, 0xAA, StopZ80Mode2,           NULL);
   AddZ80AWritePort(0x00, 0xFF, hell_fire_port_wb,      NULL);
   AddZ80AWritePort(0x00, 0xFF, DefBadWriteZ80,         NULL);
   AddZ80AWritePort(  -1,   -1, NULL,                   NULL);

   AddZ80AInit();

   /* ----------- */

   tp1vcu[0].VRAM          = RAM+0x11000;
   tp1vcu[0].RAM_OBJ       = RAM+0x21000;
   tp1vcu[0].SCROLL        = RAM+0x22000;
   tp1vcu[0].GFX_BG        = GFX_BG0;
   tp1vcu[0].MASK_BG       = GFX_BG0_SOLID;
   tp1vcu[0].tile_mask_bg  = 0x03FFF;
   tp1vcu[0].tile_mask_obj = 0x03FFF;
   tp1vcu[0].tile_ofs_obj  = 0x04000;
   init_tp1vcu(0);

   InitPaletteMap(RAM+0x10000, 0x80, 0x10, 0x8000);

   // Kill the annoying reset instruction
/*
   WriteWord68k(&ROM[0x0023C],0x4E71);          // nop

   // Fix rom checksum

   WriteLong68k(&ROM[0x004DA],0x4E714E71);      // nop
*/
   // Fix sound error
/*
   WriteLong68k(&ROM[0x0188C],0x4E714E71);      // nop
   WriteLong68k(&ROM[0x018B2],0x4E714E71);      // nop
*/
/*
   // 68000 speed hack

   WriteLong68k(&ROM[0x17918],0x13FC0000);
   WriteLong68k(&ROM[0x1791C],0x00AA0000);
*/
   // Fix protection/mcu?
/*
   WriteWord68k(&ROM[0x0181C],0x4E71);
   WriteWord68k(&ROM[0x01824],0x600A);
*/
/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x40000);
   ByteSwap(RAM,0x40000);

   AddMemFetch(0x000000, 0x03FFFF, ROM+0x000000-0x000000);      // 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadByte(0x040000, 0x043FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadByte(0x084000, 0x0847FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadByte(0x086000, 0x0867FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadByte(0x140000, 0x14000F, tp1vcu_obj_rb, NULL);                // OBJECT
   AddReadByte(0x100000, 0x10001F, tp1vcu_bg_rb, NULL);                 // LAYER
   AddReadByte(0x080000, 0x08000F, tp_vblank_rb, NULL);                 // VSYNC
   AddReadByte(0x0C0000, 0x0C0FFF, tp1_z80_rb, NULL);                   // SOUND COMM
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);               // <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadWord(0x040000, 0x043FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadWord(0x084000, 0x0847FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadWord(0x086000, 0x0867FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadWord(0x140000, 0x14000F, tp1vcu_obj_rw, NULL);                // OBJECT
   AddReadWord(0x100000, 0x10001F, tp1vcu_bg_rw, NULL);                 // LAYER
   AddReadWord(0x080000, 0x08000F, tp_vblank_rw, NULL);                 // VSYNC
   AddReadWord(0x0C0000, 0x0C0FFF, tp1_z80_rw, NULL);                   // SOUND COMM
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);               // <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x040000, 0x043FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteByte(0x084000, 0x0847FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteByte(0x086000, 0x0867FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteByte(0x080000, 0x08000F, tp_vblank_wb, NULL);                // VSYNC
   AddWriteByte(0x0C0000, 0x0C0FFF, tp1_z80_wb, NULL);                  // SOUND COMM
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);                   // Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);             // <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x040000, 0x043FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteWord(0x084000, 0x0847FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteWord(0x086000, 0x0867FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteWord(0x140000, 0x14000F, tp1vcu_obj_ww, NULL);               // OBJECT
   AddWriteWord(0x100000, 0x10001F, tp1vcu_bg_ww, NULL);                // LAYER
   AddWriteWord(0x080000, 0x08000F, tp_vblank_ww, NULL);                // VSYNC
   AddWriteWord(0x0C0000, 0x0C0FFF, tp1_z80_ww, NULL);                  // SOUND COMM
   AddWriteWord(0x180000, 0x180003, tp1vcu_ofsreg_ww, NULL);            // OFFSET
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);             // <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();     // Set Starscream mem pointers... 

}

void load_truxton(void)
{
   int ta,tb,tc;

   romset = 9;

   if(!(ROM=AllocateMem(0x40000)))return;
   if(!(RAM=AllocateMem(0x80000)))return;
   if(!(GFX=AllocateMem(0x100000+0x100000)))return;

   GFX_BG0 = GFX+0x000000;
   GFX_SPR = GFX+0x100000;

   if(!load_rom("b65_08.bin", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("b65_07.bin", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_BG0[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_BG0[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_BG0[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_BG0[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_BG0[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_BG0[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_BG0[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("b65_06.bin", RAM+0x00000, 0x20000))return;   // GFX BG0
   if(!load_rom("b65_05.bin", RAM+0x20000, 0x20000))return;   // GFX BG0
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_BG0[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_BG0[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_BG0[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_BG0[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_BG0[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_BG0[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_BG0[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_BG0[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }


   if(!load_rom("b65_04.bin", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("b65_03.bin", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] = (((tc&0x80)>>7)<<0);
      GFX_SPR[tb+1] = (((tc&0x40)>>6)<<0);
      GFX_SPR[tb+2] = (((tc&0x20)>>5)<<0);
      GFX_SPR[tb+3] = (((tc&0x10)>>4)<<0);
      GFX_SPR[tb+4] = (((tc&0x08)>>3)<<0);
      GFX_SPR[tb+5] = (((tc&0x04)>>2)<<0);
      GFX_SPR[tb+6] = (((tc&0x02)>>1)<<0);
      GFX_SPR[tb+7] = (((tc&0x01)>>0)<<0);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<1);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<1);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<1);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<1);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<1);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<1);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<1);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<1);
      tb+=8;
   }

   if(!load_rom("b65_02.bin", RAM+0x00000, 0x20000))return;   // GFX SPR
   if(!load_rom("b65_01.bin", RAM+0x20000, 0x20000))return;   // GFX SPR
   tb=0;
   for(ta=0;ta<0x20000;ta++){
      tc=RAM[ta];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<2);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<2);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<2);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<2);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<2);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<2);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<2);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<2);
      tc=RAM[ta+0x20000];
      GFX_SPR[tb+0] |= (((tc&0x80)>>7)<<3);
      GFX_SPR[tb+1] |= (((tc&0x40)>>6)<<3);
      GFX_SPR[tb+2] |= (((tc&0x20)>>5)<<3);
      GFX_SPR[tb+3] |= (((tc&0x10)>>4)<<3);
      GFX_SPR[tb+4] |= (((tc&0x08)>>3)<<3);
      GFX_SPR[tb+5] |= (((tc&0x04)>>2)<<3);
      GFX_SPR[tb+6] |= (((tc&0x02)>>1)<<3);
      GFX_SPR[tb+7] |= (((tc&0x01)>>0)<<3);
      tb+=8;
   }

   GFX_BG0_SOLID = make_solid_mask_8x8(GFX_BG0, 0x04000+0x04000);

   y_ofs = 8;

   if(!(TileQueue = (struct TILE_Q *) AllocateMem(sizeof(TILE_Q)*MAX_TILES)))return;

   if(!load_rom("b65_11.bin", RAM+0x00000, 0x20000))return;     // MAIN 68000
   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta]=RAM[ta];
   }
   if(!load_rom("b65_10.bin", RAM+0x00000, 0x20000))return;     // MAIN 68000
   for(ta=0;ta<0x20000;ta++){
      ROM[ta+ta+1]=RAM[ta];
   }

   RAMSize=0x40000+0x10000;
   memset(RAM+0x00000,0x00,RAMSize);

   /* Sound Setup */

   Z80ROM=RAM+0x40000;
   if(!load_rom("b65_09.bin", Z80ROM, 0x08000)) return;
/*
   // Fix Checksum

   Z80ROM[0x0156]=0x00; // NOP
   Z80ROM[0x0157]=0x00; // NOP
   Z80ROM[0x0158]=0x00; // NOP

   // Apply Speed Patch

   Z80ROM[0x019D]=0xD3; // OUTA (AAh)
   Z80ROM[0x019E]=0xAA; //

   SetStopZ80Mode2(0x019C);
*/
   // Setup Z80 memory map
   // --------------------

   AddZ80AROMBase(Z80ROM, 0x0038, 0x0066);

   AddZ80AReadByte(0x0000, 0x87FF, NULL,                        Z80ROM+0x0000); // Z80 ROM/RAM
   AddZ80AReadByte(0x0000, 0xFFFF, DefBadReadZ80,               NULL);
   AddZ80AReadByte(    -1,     -1, NULL,                        NULL);

   AddZ80AWriteByte(0x8000, 0x87FF, NULL,                       Z80ROM+0x8000); // Z80 RAM
   AddZ80AWriteByte(0x0000, 0xFFFF, DefBadWriteZ80,             NULL);
   AddZ80AWriteByte(    -1,     -1, NULL,                       NULL);

   AddZ80AReadPort(0x00, 0xFF, truxton_port_rb,         NULL);
   AddZ80AReadPort(0x00, 0xFF, DefBadReadZ80,           NULL);
   AddZ80AReadPort(  -1,   -1, NULL,                    NULL);

   AddZ80AWritePort(0xAA, 0xAA, StopZ80Mode2,           NULL);
   AddZ80AWritePort(0x00, 0xFF, truxton_port_wb,        NULL);
   AddZ80AWritePort(0x00, 0xFF, DefBadWriteZ80,         NULL);
   AddZ80AWritePort(  -1,   -1, NULL,                   NULL);

   AddZ80AInit();

   /* ----------- */

   tp1vcu[0].VRAM          = RAM+0x11000;
   tp1vcu[0].RAM_OBJ       = RAM+0x21000;
   tp1vcu[0].SCROLL        = RAM+0x22000;
   tp1vcu[0].GFX_BG        = GFX_BG0;
   tp1vcu[0].MASK_BG       = GFX_BG0_SOLID;
   tp1vcu[0].tile_mask_bg  = 0x03FFF;
   tp1vcu[0].tile_mask_obj = 0x03FFF;
   tp1vcu[0].tile_ofs_obj  = 0x04000;
   init_tp1vcu(0);

   InitPaletteMap(RAM+0x10000, 0x80, 0x10, 0x8000);

   // Kill the annoying reset instruction
/*
   WriteWord68k(&ROM[0x0023C],0x4E71);          // nop

   // Fix rom checksum

   WriteLong68k(&ROM[0x004DA],0x4E714E71);      // nop
*/
   // Fix sound error
/*
   WriteLong68k(&ROM[0x0188C],0x4E714E71);      // nop
   WriteLong68k(&ROM[0x018B2],0x4E714E71);      // nop
*/
/*
   // 68000 speed hack

   WriteLong68k(&ROM[0x17918],0x13FC0000);
   WriteLong68k(&ROM[0x1791C],0x00AA0000);
*/
   // Fix protection/mcu?
/*
   WriteWord68k(&ROM[0x0181C],0x4E71);
   WriteWord68k(&ROM[0x01824],0x600A);
*/
/*
 *  StarScream Stuff follows
 */

   ByteSwap(ROM,0x40000);
   ByteSwap(RAM,0x20000);

   AddMemFetch(0x000000, 0x03FFFF, ROM+0x000000-0x000000);      // 68000 ROM
   AddMemFetch(-1, -1, NULL);

   AddReadByte(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadByte(0x080000, 0x083FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadByte(0x144000, 0x1447FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadByte(0x146000, 0x1467FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadByte(0x0C0000, 0x0C000F, tp1vcu_obj_rb, NULL);                // OBJECT
   AddReadByte(0x100000, 0x10001F, tp1vcu_bg_rb, NULL);                 // LAYER
   AddReadByte(0x140000, 0x14000F, tp_vblank_rb, NULL);                 // VSYNC
   AddReadByte(0x180000, 0x180FFF, tp1_z80_rb, NULL);                   // SOUND COMM
   AddReadByte(0x000000, 0xFFFFFF, DefBadReadByte, NULL);               // <Bad Reads>
   AddReadByte(-1, -1, NULL, NULL);

   AddReadWord(0x000000, 0x03FFFF, NULL, ROM+0x000000);                 // 68000 ROM
   AddReadWord(0x080000, 0x083FFF, NULL, RAM+0x000000);                 // 68000 RAM
   AddReadWord(0x144000, 0x1447FF, NULL, RAM+0x010000);                 // COLOR RAM
   AddReadWord(0x146000, 0x1467FF, NULL, RAM+0x010800);                 // COLOR RAM
   AddReadWord(0x0C0000, 0x0C000F, tp1vcu_obj_rw, NULL);                // OBJECT
   AddReadWord(0x100000, 0x10001F, tp1vcu_bg_rw, NULL);                 // LAYER
   AddReadWord(0x140000, 0x14000F, tp_vblank_rw, NULL);                 // VSYNC
   AddReadWord(0x180000, 0x180FFF, tp1_z80_rw, NULL);                   // SOUND COMM
   AddReadWord(0x000000, 0xFFFFFF, DefBadReadWord, NULL);               // <Bad Reads>
   AddReadWord(-1, -1,NULL, NULL);

   AddWriteByte(0x080000, 0x083FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteByte(0x144000, 0x1447FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteByte(0x146000, 0x1467FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteByte(0x140000, 0x14000F, tp_vblank_wb, NULL);                // VSYNC
   AddWriteByte(0x180000, 0x180FFF, tp1_z80_wb, NULL);                  // SOUND COMM
   AddWriteByte(0xAA0000, 0xAA0001, Stop68000, NULL);                   // Trap Idle 68000
   AddWriteByte(0x000000, 0xFFFFFF, DefBadWriteByte, NULL);             // <Bad Writes>
   AddWriteByte(-1, -1, NULL, NULL);

   AddWriteWord(0x080000, 0x083FFF, NULL, RAM+0x000000);                // 68000 RAM
   AddWriteWord(0x144000, 0x1447FF, NULL, RAM+0x010000);                // COLOR RAM
   AddWriteWord(0x146000, 0x1467FF, NULL, RAM+0x010800);                // COLOR RAM
   AddWriteWord(0x0C0000, 0x0C000F, tp1vcu_obj_ww, NULL);               // OBJECT
   AddWriteWord(0x100000, 0x10001F, tp1vcu_bg_ww, NULL);                // LAYER
   AddWriteWord(0x140000, 0x14000F, tp_vblank_ww, NULL);                // VSYNC
   AddWriteWord(0x180000, 0x180FFF, tp1_z80_ww, NULL);                  // SOUND COMM
   AddWriteWord(0x1C0000, 0x1C0003, tp1vcu_ofsreg_ww, NULL);            // OFFSET
   AddWriteWord(0x000000, 0xFFFFFF, DefBadWriteWord, NULL);             // <Bad Writes>
   AddWriteWord(-1, -1, NULL, NULL);

   AddInitMemory();     // Set Starscream mem pointers... 

}

void ClearToaplan1(void)
{
   #ifdef HACKS
      switch(romset){
      case 3:
      case 4:
      case 6:
      case 7:
      save_debug("ROM.bin",ROM,0x80000,1);
      save_debug("RAM.bin",RAM,RAMSize,1);
      break;
      default:
      save_debug("ROM.bin",ROM,0x40000,1);
      save_debug("RAM.bin",RAM,RAMSize,1);
      break;
      }
   #endif
}

void ExecuteToaplan1Frame(void)
{
   cpu_execute_cycles(CPU_68K_0, CPU_FRAME_MHz(16,60));    // M68000 16MHz (60fps) (real game is only ??MHz)
   #ifdef HACKS
      print_debug("PC:%06x SR:%04x\n",s68000context.pc,s68000context.sr);
   #endif
   if(want_int)
      cpu_interrupt(CPU_68K_0, 4);
}

void ExecuteToaplan1Frame_Sound(void)
{
   cpu_execute_cycles(CPU_68K_0, CPU_FRAME_MHz(16,60));    // M68000 16MHz (60fps) (real game is only ??MHz)
   #ifdef HACKS
      print_debug("PC:%06x SR:%04x\n",s68000context.pc,s68000context.sr);
   #endif
   if(want_int)
      cpu_interrupt(CPU_68K_0, 4);

   cpu_execute_cycles(CPU_Z80_0, CPU_FRAME_MHz(4,60));        // Sound Z80
   #ifdef HACKS
      print_debug("PC_Z80:%04x\n",z80pc);
   #endif
   cpu_interrupt(CPU_Z80_0, 0x38);
}

static void ClearTileQueue(void)
{
   int ta;

   for(ta=0;ta<MAX_PRI;ta++){
     prilist[ta] = NULL;
   }

   last_tile = TileQueue;
}

static DEF_INLINE void QueueTile(int tile, int x, int y, UBYTE *map, int pri)
{
   last_tile->tile = tile;
   last_tile->x    = x;
   last_tile->y    = y;
   last_tile->map  = map;
   last_tile->next = prilist[pri];
   prilist[pri]    = last_tile;
   last_tile       = last_tile+1;
}

static void DrawTileQueue(void)
{
   struct TILE_Q *tile_ptr;
   ULONG ta,pri;
   //ULONG rend,rend2;

   //rend = 0;
   //rend2 = (last_tile - TileQueue);	// / sizeof(TILE_Q);

   for(pri=1; pri<MAX_PRI; pri++){
      tile_ptr = prilist[pri];

      while(tile_ptr){
         ta = tile_ptr->tile;
         if(GFX_BG0_SOLID[ta]!=0){                      // No pixels; skip
            if(GFX_BG0_SOLID[ta]==1)                    // Some pixels; trans
               Draw8x8_Trans_Mapped_Rot(&GFX_BG0[ta<<6],tile_ptr->x,tile_ptr->y,tile_ptr->map);
            else                                        // all pixels; solid
               Draw8x8_Mapped_Rot(&GFX_BG0[ta<<6],tile_ptr->x,tile_ptr->y,tile_ptr->map);
         }
         tile_ptr = tile_ptr->next;
         //rend++;
      }

   }

   //print_ingame(600,"%04x Tiles Listed", rend2 );
   //print_ingame(600,"%04x Tiles Drawn", rend );
}

static UWORD scr_ofs[ROM_COUNT][8] =
{
   { 0x1B4,0x1EB,0x1B6,0x1EB,0x1B8,0x1EB,0x1BA,0x1EB },
   { 0x1B4,0x1EB,0x1B6,0x1EB,0x1B8,0x1EB,0x1BA,0x1EB },
   { 0x1B4,0x1EB,0x1B6,0x1EB,0x1B8,0x1EB,0x1BA,0x1EB },
   { 0x1B4,0x1EB,0x1B6,0x1EB,0x1B8,0x1EB,0x1BA,0x1EB },
   { 0x1B4,0x1EB,0x1B6,0x1EB,0x1B8,0x1EB,0x1BA,0x1EB },
   { 0x1B4,0x1EB,0x1B6,0x1EB,0x1B8,0x1EB,0x1BA,0x1EB },
   { 0x1B4,0x1EB,0x1B6,0x1EB,0x1B8,0x1EB,0x1BA,0x1EB },
   { 0x1B4,0x1EB,0x1B6,0x1EB,0x1B8,0x1EB,0x1BA,0x1EB },
   { 0x1B4,0x1EB,0x1B6,0x1EB,0x1B8,0x1EB,0x1BA,0x1EB },
   { 0x1B4,0x1EB,0x1B6,0x1EB,0x1B8,0x1EB,0x1BA,0x1EB },
};

static UWORD vram_ofs[ROM_COUNT][4] =
{
   { 0x0000,0x4000,0x8000,0xC000 },
   { 0x0000,0x4000,0x8000,0xC000 },
   { 0x0000,0x4000,0x8000,0xC000 },
   { 0x0000,0x4000,0x8000,0xC000 },
   { 0x0000,0x4000,0x8000,0xC000 },
   { 0x0000,0x4000,0x8000,0xC000 },
   { 0x0000,0x4000,0x8000,0xC000 },
   { 0x0000,0x4000,0x8000,0xC000 },
   { 0x0000,0x4000,0x8000,0xC000 },
   { 0x0000,0x4000,0x8000,0xC000 }, // Truxton
};

#define OBJ_X_MASK	0x1FF
#define OBJ_Y_MASK	0x1FF

void DrawToaplan1(void)
{
   int x,y,ta,pri,layer;
   int zz,zzz,zzzz,x16,y16;
   int xx,xxx,xxxx,yyy,x1,y1;
   UBYTE *MAP,*RAM_BG,*RAM_BG2;
   ULONG tile_mask,tile_ofs;
   UWORD ctrl;

   ClearPaletteMap();

   clear_game_screen(0);

   ClearTileQueue();

   // BG0123

   for(layer=0; layer<4; layer++){

   tile_mask = tp1vcu[0].tile_mask_bg;

   RAM_BG = tp1vcu[0].VRAM + vram_ofs[romset][layer];

   x1 = ReadWord(&tp1vcu[0].SCROLL[0+(layer<<2)])>>7;
   x1 += (0x1F0 - bg_x_ofs);
   x1 += x_ofs[layer];
   x1 &= 0x1FF;

   y1 = ReadWord(&tp1vcu[0].SCROLL[2+(layer<<2)])>>7;
   y1 += (0x101 - bg_y_ofs);
   y1 &= 0x1FF;

//-(scr_ofs[romset][0+(layer<<1)]),
//-(scr_ofs[romset][1+(layer<<1)])

   MAKE_SCROLL_512x512_4_8(
      x1,
      y1
   );

   START_SCROLL_512x512_4_8(32,32,320,240);

      pri = ReadWord(&RAM_BG[zz]);

      if(pri){

      ta = ReadWord(&RAM_BG[2+zz]);

      ta &= tile_mask;

      if(GFX_BG0_SOLID[ta]!=0){

      MAP_PALETTE_MAPPED(
         Map_15bit_xBGR,
         pri&0x3F,
         16,
         MAP
      );

      pri >>= 12;

      //if(pri==1)
      //   Draw8x8_Mapped(&GFX_BG0[ta<<6], x, y, MAP);
      //else
         QueueTile(ta, x, y, MAP, pri);

      }
      }

   END_SCROLL_512x512_4_8();

   }

   // OBJECT

   tile_mask = tp1vcu[0].tile_mask_obj;
   tile_ofs  = tp1vcu[0].tile_ofs_obj;

   RAM_BG  = tp1vcu[0].RAM_OBJ;
   RAM_BG2 = tp1vcu[0].RAM_OBJ+0x800;

   zz = 0x7F8;

   do{
      ctrl = ReadWord(&RAM_BG[zz+2]);

      pri = ReadWord(&RAM_BG2[(ctrl>>5)&0x7E]);

      if(pri&0xFF){

      ta = (ReadWord(&RAM_BG[zz+0])&tile_mask)+tile_ofs;

      x = ((ReadWord(&RAM_BG[zz+4])>>7)+32) & OBJ_X_MASK;
      y = ((ReadWord(&RAM_BG[zz+6])>>7)+32) & OBJ_Y_MASK;

      MAP_PALETTE_MAPPED(
         Map_15bit_xBGR,
         (ctrl&0x3F)|0x40,
         16,
         MAP
      );

      pri = ReadWord(&RAM_BG2[(ctrl>>5)&0x7E]);

      xxx = (pri>>0) & 0x0F;
      yyy = (pri>>4) & 0x0F;

      pri = (ctrl&0xF000)>>12;

      xxxx=x;
      while((--yyy)>=0){
      x=xxxx;
      xx=xxx;
      while((--xx)>=0){

         if(GFX_BG0_SOLID[ta]!=0){

         if((x>24)&&(y>24)&&(x<320+32)&&(y<240+32)){
            QueueTile(ta, x, y, MAP, pri);
         }

         }

         ta++;
         x = (x + 8) & OBJ_X_MASK;
      };
      y = (y + 8) & OBJ_Y_MASK;
      };

      }

   }while((zz-=8)>=0);

   DrawTileQueue();
}
