/******************************************************************************/
/*                                                                            */
/*                  RAINE GUI: Based on allegro GUI routines                  */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "raine.h"
#include "emumain.h"		// main loops
#include "gui.h"
#include "gui2.h"
#include "default.h"
#include "dsw.h"
#include "ingame.h"
#include "arpro.h"
#include "control.h"
#include "sasound.h"
#include "config.h"
#include "games.h"		// Games List
#include "tilemod.h"

#ifdef RAINE_DOS
#include "arcmon.h"
#endif

//#ifndef RAINE_DOS

#include <sys/io.h>
#define outportb(a,b)	outb(b,a)
#define inportb		inb

#define outp(a,b)		outb(b,a)
#define inp		inb

#define outportw(a,b)	outw(b,a)
#define inportw		inw

#define strcasecmp	stricmp

//#endif

static int screen_valid;	// 1 if screen contains a valid image

extern int use_emulated_ym3812;

char S[128];

static int WantScreen;
static int WantQuit;
static int WantPlay;

int allocate_pens(int pens);
void free_pens(int pens);
void free_all_pens(void);

void MakeGUIBack(void);
void DestroyGUIBack(void);

void free_game_avail_list(void);

/*

Updated to allegro 3921 data

*/

static char *KeyName(char scancode)
{
   switch(scancode){
   case 0                     : return "<None>"; // #0
   case KEY_A                 : return "A"; // #1
   case KEY_B                 : return "B"; // #2
   case KEY_C                 : return "C"; // #3
   case KEY_D                 : return "D"; // #4
   case KEY_E                 : return "E"; // #5
   case KEY_F                 : return "F"; // #6
   case KEY_G                 : return "G"; // #7
   case KEY_H                 : return "H"; // #8
   case KEY_I                 : return "I"; // #9
   case KEY_J                 : return "J"; // #10
   case KEY_K                 : return "K"; // #11
   case KEY_L                 : return "L"; // #12
   case KEY_M                 : return "M"; // #13
   case KEY_N                 : return "N"; // #14
   case KEY_O                 : return "O"; // #15
   case KEY_P                 : return "P"; // #16
   case KEY_Q                 : return "Q"; // #17
   case KEY_R                 : return "R"; // #18
   case KEY_S                 : return "S"; // #19
   case KEY_T                 : return "T"; // #20
   case KEY_U                 : return "U"; // #21
   case KEY_V                 : return "V"; // #22
   case KEY_W                 : return "W"; // #23
   case KEY_X                 : return "X"; // #24
   case KEY_Y                 : return "Y"; // #25
   case KEY_Z                 : return "Z"; // #26
   case KEY_0                 : return "0"; // #27
   case KEY_1                 : return "1"; // #28
   case KEY_2                 : return "2"; // #29
   case KEY_3                 : return "3"; // #30
   case KEY_4                 : return "4"; // #31
   case KEY_5                 : return "5"; // #32
   case KEY_6                 : return "6"; // #33
   case KEY_7                 : return "7"; // #34
   case KEY_8                 : return "8"; // #35
   case KEY_9                 : return "9"; // #36
   case KEY_0_PAD             : return "0 (pad)"; // #37
   case KEY_1_PAD             : return "1 (pad)"; // #38
   case KEY_2_PAD             : return "2 (pad)"; // #39
   case KEY_3_PAD             : return "3 (pad)"; // #40
   case KEY_4_PAD             : return "4 (pad)"; // #41
   case KEY_5_PAD             : return "5 (pad)"; // #42
   case KEY_6_PAD             : return "6 (pad)"; // #43
   case KEY_7_PAD             : return "7 (pad)"; // #44
   case KEY_8_PAD             : return "8 (pad)"; // #45
   case KEY_9_PAD             : return "9 (pad)"; // #46
   case KEY_F1                : return "F1"; // #47
   case KEY_F2                : return "F2"; // #48
   case KEY_F3                : return "F3"; // #49
   case KEY_F4                : return "F4"; // #50
   case KEY_F5                : return "F5"; // #51
   case KEY_F6                : return "F6"; // #52
   case KEY_F7                : return "F7"; // #53
   case KEY_F8                : return "F8"; // #54
   case KEY_F9                : return "F9"; // #55
   case KEY_F10               : return "F10"; // #56
   case KEY_F11               : return "F11"; // #57
   case KEY_F12               : return "F12"; // #58
   case KEY_ESC               : return "Esc"; // #59
   case KEY_TILDE             : return ""; // #60
   case KEY_MINUS             : return "-"; // #61
   case KEY_EQUALS            : return "="; // #62
   case KEY_BACKSPACE         : return "Backspace"; // #63
   case KEY_TAB               : return "Tab"; // #64
   case KEY_OPENBRACE         : return "{"; // #65
   case KEY_CLOSEBRACE        : return "}"; // #66
   case KEY_ENTER             : return "Enter"; // #67
   case KEY_COLON             : return ":"; // #68
   case KEY_QUOTE             : return "'"; // #69
   case KEY_BACKSLASH         : return "#"; // #70
   case KEY_BACKSLASH2        : return "\\"; // #71
   case KEY_COMMA             : return ","; // #72
   case KEY_STOP              : return "."; // #73
   case KEY_SLASH             : return "/"; // #74
   case KEY_SPACE             : return "Space"; // #75
   case KEY_INSERT            : return "Insert"; // #76
   case KEY_DEL               : return "Delete"; // #77
   case KEY_HOME              : return "Home"; // #78
   case KEY_END               : return "End"; // #79
   case KEY_PGUP              : return "Page Up"; // #80
   case KEY_PGDN              : return "Page Down"; // #81
   case KEY_LEFT              : return "Left"; // #82
   case KEY_RIGHT             : return "Right"; // #83
   case KEY_UP                : return "Up"; // #84
   case KEY_DOWN              : return "Down"; // #85
   case KEY_SLASH_PAD         : return "/ (pad)"; // #86
   case KEY_ASTERISK          : return "* (pad)"; // #87
   case KEY_MINUS_PAD         : return "- (pad)"; // #88
   case KEY_PLUS_PAD          : return "+ (pad)"; // #89
   case KEY_DEL_PAD           : return "Del (pad)"; // #90
   case KEY_ENTER_PAD         : return "Enter (pad)"; // #91
   case KEY_PRTSCR            : return "Print Scr"; // #92
   case KEY_PAUSE             : return "Pause"; // #93
   case KEY_LSHIFT            : return "Left Shift"; // #94
   case KEY_RSHIFT            : return "Right Shift"; // #95
   case KEY_LCONTROL          : return "Left Ctrl"; // #96
   case KEY_RCONTROL          : return "Right Ctrl"; // #97
   case KEY_ALT               : return "Left Alt"; // #98
   case KEY_ALTGR             : return "Right Alt"; // #99
   case KEY_LWIN              : return "Left Win"; // #100
   case KEY_RWIN              : return "Right Win"; // #101
   case KEY_MENU              : return "Menu Win"; // #102
   case KEY_SCRLOCK           : return "Scr Lock"; // #103
   case KEY_NUMLOCK           : return "Num Lock"; // #104
   case KEY_CAPSLOCK          : return "Caps Lock"; // #105
   default                    : return "<Undefined>"; // #???
   }
}

static char *JoyName(int scancode)
{
   static char *joy_switches[] =
   {
   "<None>",

   "Joy1 Up",
   "Joy1 Down",
   "Joy1 Left",
   "Joy1 Right",

   "Joy1 Button 1",
   "Joy1 Button 2",
   "Joy1 Button 3",
   "Joy1 Button 4",
   "Joy1 Button 5",
   "Joy1 Button 6",
   "Joy1 Button 7",
   "Joy1 Button 8",

   "Joy2 Up",
   "Joy2 Down",
   "Joy2 Left",
   "Joy2 Right",

   "Joy2 Button 1",
   "Joy2 Button 2",
   "Joy2 Button 3",
   "Joy2 Button 4",
   "Joy2 Button 5",
   "Joy2 Button 6",
   "Joy2 Button 7",
   "Joy2 Button 8",

   "Joy3 Up",
   "Joy3 Down",
   "Joy3 Left",
   "Joy3 Right",

   "Joy3 Button 1",
   "Joy3 Button 2",
   "Joy3 Button 3",
   "Joy3 Button 4",
   "Joy3 Button 5",
   "Joy3 Button 6",
   "Joy3 Button 7",
   "Joy3 Button 8",

   "Joy4 Up",
   "Joy4 Down",
   "Joy4 Left",
   "Joy4 Right",

   "Joy4 Button 1",
   "Joy4 Button 2",
   "Joy4 Button 3",
   "Joy4 Button 4",
   "Joy4 Button 5",
   "Joy4 Button 6",
   "Joy4 Button 7",
   "Joy4 Button 8",

   "Joy1 Button 9",
   "Joy1 Button 10",
   "Joy1 Button 11",
   "Joy1 Button 12",

   "Joy2 Button 9",
   "Joy2 Button 10",
   "Joy2 Button 11",
   "Joy2 Button 12",

   "Joy3 Button 9",
   "Joy3 Button 10",
   "Joy3 Button 11",
   "Joy3 Button 12",

   "Joy4 Button 9",
   "Joy4 Button 10",
   "Joy4 Button 11",
   "Joy4 Button 12",

   };

   if(scancode>=RJOY_COUNT) scancode=0;

   return joy_switches[scancode];
}

// centre_all_dialogs():
// Moves all dialogs so they are centered, used when we change
// screen size.

static int scale = 8;

void raine_centre_dialog(DIALOG *dialog)
{
   int min_x, min_y;
   int max_x, max_y;
   int dx, dy;
   int i, new_scale;

   /*

   place window at (0,0)

   */

   dx = dialog[0].x;
   dy = dialog[0].y;

   for(i=0; dialog[i].proc; i++){
  
      dialog[i].x -= dx;
      dialog[i].y -= dy;

   }

   /*

   scale window

   */

   new_scale = text_height(gui_main_font);

   for(i=0; dialog[i].proc; i++){
  
      dialog[i].x = (dialog[i].x * new_scale) / scale;
      dialog[i].y = (dialog[i].y * new_scale) / scale;
      dialog[i].w = (dialog[i].w * new_scale) / scale;
      dialog[i].h = (dialog[i].h * new_scale) / scale;

   }

   /*

   centre window

   */

   min_x = dialog[0].x;
   min_y = dialog[0].y;
   max_x = dialog[0].x + dialog[0].w;
   max_y = dialog[0].y + dialog[0].h;

   dx = (SCREEN_W - (max_x - min_x)) / 2 - min_x;
   dy = (SCREEN_H - (max_y - min_y)) / 2 - min_y;

   for(i=0; dialog[i].proc; i++){

      dialog[i].x += dx;
      dialog[i].y += dy;

   }
}

static void centre_all_dialogs(void)
{
   ULONG i;

   DestroyGUIBack();

   i = 0;

   while(raine_dialog_list[i])

      raine_centre_dialog( raine_dialog_list[i++] );

   scale = text_height(gui_main_font);

   MakeGUIBack();
}

// Unselect_Button():
// Force a Button object to redraw itself unselected+unfocused
// Otherwise, they don't get redrawn till after the button's action
// (which looks ugly...).

void Unselect_Button(DIALOG *d)
{
   d->flags &= ~(D_SELECTED|D_GOTFOCUS);
   scare_mouse();
   x_raine_button_proc(MSG_DRAW,d,0);
   unscare_mouse();
   //dialog_oxygen();
}


ULONG darken(ULONG src)
{
   ULONG r,g,b;

   r = (src >> 16) & 0xFF;
   g = (src >>  8) & 0xFF;
   b = (src >>  0) & 0xFF;

   r = (r * 0x100) / 0x148;
   g = (g * 0x100) / 0x148;
   b = (b * 0x100) / 0x148;

   return (r<<16) | (g<<8) | (b<<0);
}

void set_gui_palette(void)
{
   gui_pal[GUI_COL_TEXT_1].r=(rgui_cfg.gui_col_text_1>>18)&0x3F;
   gui_pal[GUI_COL_TEXT_1].g=(rgui_cfg.gui_col_text_1>>10)&0x3F;
   gui_pal[GUI_COL_TEXT_1].b=(rgui_cfg.gui_col_text_1>>2)&0x3F;

   gui_pal[GUI_COL_TEXT_2].r=(rgui_cfg.gui_col_text_2>>18)&0x3F;
   gui_pal[GUI_COL_TEXT_2].g=(rgui_cfg.gui_col_text_2>>10)&0x3F;
   gui_pal[GUI_COL_TEXT_2].b=(rgui_cfg.gui_col_text_2>>2)&0x3F;

   gui_pal[GUI_COL_BLACK].r=(rgui_cfg.gui_col_black>>18)&0x3F;
   gui_pal[GUI_COL_BLACK].g=(rgui_cfg.gui_col_black>>10)&0x3F;
   gui_pal[GUI_COL_BLACK].b=(rgui_cfg.gui_col_black>>2)&0x3F;

   gui_pal[GUI_COL_SELECT].r=(rgui_cfg.gui_col_select>>18)&0x3F;
   gui_pal[GUI_COL_SELECT].g=(rgui_cfg.gui_col_select>>10)&0x3F;
   gui_pal[GUI_COL_SELECT].b=(rgui_cfg.gui_col_select>>2)&0x3F;

   gui_pal[GUI_BOX_COL_HIGH_2].r=(rgui_cfg.gui_box_col_high_2>>18)&0x3F;
   gui_pal[GUI_BOX_COL_HIGH_2].g=(rgui_cfg.gui_box_col_high_2>>10)&0x3F;
   gui_pal[GUI_BOX_COL_HIGH_2].b=(rgui_cfg.gui_box_col_high_2>>2)&0x3F;

   gui_pal[GUI_BOX_COL_HIGH_1].r=(rgui_cfg.gui_box_col_high_1>>18)&0x3F;
   gui_pal[GUI_BOX_COL_HIGH_1].g=(rgui_cfg.gui_box_col_high_1>>10)&0x3F;
   gui_pal[GUI_BOX_COL_HIGH_1].b=(rgui_cfg.gui_box_col_high_1>>2)&0x3F;

   gui_pal[GUI_BOX_COL_MIDDLE].r=(rgui_cfg.gui_box_col_middle>>18)&0x3F;
   gui_pal[GUI_BOX_COL_MIDDLE].g=(rgui_cfg.gui_box_col_middle>>10)&0x3F;
   gui_pal[GUI_BOX_COL_MIDDLE].b=(rgui_cfg.gui_box_col_middle>>2)&0x3F;

   gui_pal[GUI_BOX_COL_LOW_1].r=(rgui_cfg.gui_box_col_low_1>>18)&0x3F;
   gui_pal[GUI_BOX_COL_LOW_1].g=(rgui_cfg.gui_box_col_low_1>>10)&0x3F;
   gui_pal[GUI_BOX_COL_LOW_1].b=(rgui_cfg.gui_box_col_low_1>>2)&0x3F;

   gui_pal[GUI_BOX_COL_LOW_2].r=(rgui_cfg.gui_box_col_low_2>>18)&0x3F;
   gui_pal[GUI_BOX_COL_LOW_2].g=(rgui_cfg.gui_box_col_low_2>>10)&0x3F;
   gui_pal[GUI_BOX_COL_LOW_2].b=(rgui_cfg.gui_box_col_low_2>>2)&0x3F;

   gui_pal[DGUI_COL_TEXT_1].r=(darken(rgui_cfg.gui_col_text_1)>>18)&0x3F;
   gui_pal[DGUI_COL_TEXT_1].g=(darken(rgui_cfg.gui_col_text_1)>>10)&0x3F;
   gui_pal[DGUI_COL_TEXT_1].b=(darken(rgui_cfg.gui_col_text_1)>>2)&0x3F;

   gui_pal[DGUI_COL_TEXT_2].r=(darken(rgui_cfg.gui_col_text_2)>>18)&0x3F;
   gui_pal[DGUI_COL_TEXT_2].g=(darken(rgui_cfg.gui_col_text_2)>>10)&0x3F;
   gui_pal[DGUI_COL_TEXT_2].b=(darken(rgui_cfg.gui_col_text_2)>>2)&0x3F;

   gui_pal[DGUI_BOX_COL_HIGH_2].r=(darken(rgui_cfg.gui_box_col_high_2)>>18)&0x3F;
   gui_pal[DGUI_BOX_COL_HIGH_2].g=(darken(rgui_cfg.gui_box_col_high_2)>>10)&0x3F;
   gui_pal[DGUI_BOX_COL_HIGH_2].b=(darken(rgui_cfg.gui_box_col_high_2)>>2)&0x3F;

   gui_pal[DGUI_BOX_COL_HIGH_1].r=(darken(rgui_cfg.gui_box_col_high_1)>>18)&0x3F;
   gui_pal[DGUI_BOX_COL_HIGH_1].g=(darken(rgui_cfg.gui_box_col_high_1)>>10)&0x3F;
   gui_pal[DGUI_BOX_COL_HIGH_1].b=(darken(rgui_cfg.gui_box_col_high_1)>>2)&0x3F;

   gui_pal[DGUI_BOX_COL_MIDDLE].r=(darken(rgui_cfg.gui_box_col_middle)>>18)&0x3F;
   gui_pal[DGUI_BOX_COL_MIDDLE].g=(darken(rgui_cfg.gui_box_col_middle)>>10)&0x3F;
   gui_pal[DGUI_BOX_COL_MIDDLE].b=(darken(rgui_cfg.gui_box_col_middle)>>2)&0x3F;

   gui_pal[DGUI_BOX_COL_LOW_1].r=(darken(rgui_cfg.gui_box_col_low_1)>>18)&0x3F;
   gui_pal[DGUI_BOX_COL_LOW_1].g=(darken(rgui_cfg.gui_box_col_low_1)>>10)&0x3F;
   gui_pal[DGUI_BOX_COL_LOW_1].b=(darken(rgui_cfg.gui_box_col_low_1)>>2)&0x3F;

   gui_pal[DGUI_BOX_COL_LOW_2].r=(darken(rgui_cfg.gui_box_col_low_2)>>18)&0x3F;
   gui_pal[DGUI_BOX_COL_LOW_2].g=(darken(rgui_cfg.gui_box_col_low_2)>>10)&0x3F;
   gui_pal[DGUI_BOX_COL_LOW_2].b=(darken(rgui_cfg.gui_box_col_low_2)>>2)&0x3F;
}

// FadeGUI():
// Remaps screen image of gui to darker colours. Used to 'fade out'
// Inactive gui areas

void FadeGUI(void)
{
   BITMAP *fade_backdrop;
   UBYTE *BIT;
   ULONG ta,tb;
   UBYTE convert[256];

   for(ta=0;ta<256;ta++){
      convert[ta]=ta;
   }
   convert[GUI_COL_TEXT_1]      = DGUI_COL_TEXT_1;
   convert[GUI_COL_TEXT_2]      = DGUI_COL_TEXT_2;
   convert[GUI_COL_BLACK]       = GUI_COL_BLACK;
   convert[GUI_COL_SELECT]      = DGUI_BOX_COL_MIDDLE;
   convert[GUI_BOX_COL_HIGH_2]  = DGUI_BOX_COL_HIGH_2;
   convert[GUI_BOX_COL_HIGH_1]  = DGUI_BOX_COL_HIGH_1;
   convert[GUI_BOX_COL_MIDDLE]  = DGUI_BOX_COL_MIDDLE;
   convert[GUI_BOX_COL_LOW_1]   = DGUI_BOX_COL_LOW_1;
   convert[GUI_BOX_COL_LOW_2]   = DGUI_BOX_COL_LOW_2;

   fade_backdrop = create_bitmap_ex(8, display_cfg.screen_x,display_cfg.screen_y);

   scare_mouse();
   blit(screen,fade_backdrop,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);

   for(tb=0;tb<display_cfg.screen_y;tb++){
      BIT=fade_backdrop->line[tb];
      for(ta=0;ta<display_cfg.screen_x;ta++){
         BIT[ta]=convert[BIT[ta]];
      }
   }

   blit(fade_backdrop,screen,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);
   unscare_mouse();

   destroy_bitmap(fade_backdrop);

   //dialog_oxygen();
}

#ifdef RAINE_DOS

static DEF_INLINE void WaitVBlank(void)
{
  while(inportb(0x3DA)&0x08);
  while(!(inportb(0x3DA)&0x08));
}

static int Stretch256x240(int SType)
{
   static UBYTE fwnesdata[] = {
   0x4F,0x3F,0x40,0x92,0x46,0x10,0x0D,0x3E,
   0x00,0x41,0x00,0x00,0x00,0x00,0x00,0x00,
   0xEA,0xAC,0xDF,0x20,0x00,0xE7,0x06,0xE3
   };

   int ta,ret;

   if(display_cfg.triple_buffer)
      ret = set_gfx_mode(SType,256,240,256,240*3);
   else
      ret = set_gfx_mode(SType,256,240,256,240);

   if(!ret){

   //disable();
   WaitVBlank();

   //outportw(0x3C4,0x100);
   outportb(0x3C2,0xE3);
   //outportw(0x3C4,0x300);

   outportw(0x3D4,((fwnesdata[0x11]&0x7F)<<8)+0x11);
   for(ta=0;ta<24;ta++){
      outportw(0x3D4,(fwnesdata[ta]<<8)+ta);
   }
   //inportb(0x3DA);
   //outportb(0x3c0,0x31);
   //outportb(0x3c0,0x3f);
   inportb(0x3DA);
   //outportb(0x3C0,0x11  | 0x20);
   //outportb(0x3C0,0x40);
   //enable();

   }
   return ret;
}

// SetDarkForceMode():
// Set a normal vesa mode (via allegro), then divide
// the number of screen lines...

static int SetDarkForceMode(int SType, int SX, int SY, int NewSY)
{
   int ret,lines;
   UBYTE scanline;

   if(display_cfg.triple_buffer)
      ret = set_gfx_mode(SType,SX,SY,SX,NewSY*3);
   else
      ret = set_gfx_mode(SType,SX,SY,SX,SY);

   if(!ret){
      lines=SY/NewSY;
      if(lines<1)lines=1;
      if(lines>4)lines=4;
      outp(0x3D4,0x09);		// CRTC Register index 9
      scanline=inp(0x3D5);	// get scanline height
      scanline&=0x60;		// clear any existing bits
      scanline|=(lines-1);	// set multiply bits
      outp(0x3D5,scanline); 	// set new line height      
   }

   return ret;
}

// Set320x256():
// Set a normal mode-x mode (via allegro), then twiddle
// the vga registers and the vtable

static int Set320x256(int SType)
{
   int n,ret;
   UWORD port;
   UBYTE *q,reg,data;
   static unsigned char gfx_320x256_data[] = {
      194,3,0,227,212,3,0,95,212,3,1,79,212,3,2,80,212,3,3,130,212,3,4,84,212,3,5,154,212,3,6,35,212,
      3,7,178,212,3,8,0,212,3,9,97,212,3,16,10,212,3,17,172,212,3,18,255,212,3,19,40,212,3,20,0,212,3,
      21,7,212,3,22,26,212,3,23,227,196,3,1,1,196,3,4,6,206,3,5,64,206,3,6,5,192,3,16,65,192,3,19,
      0,};

   if(display_cfg.triple_buffer)
      ret = set_gfx_mode(SType,320,480,320,256*3);
   else
      ret = set_gfx_mode(SType,320,480,320,256);

   if(ret)
      return ret;

   //disable();
   WaitVBlank();

   /* ready the vga registers */
   outp(0x3d4,0x11);
   data = inp(0x3d5) & 0x7f;
   outp(0x3d4,0x11);
   outp(0x3d5,data);

   n = 25;
   q = &gfx_320x256_data[0];
   do {
      port = *q++;
      port |= ((*q++) << 8);
      reg = *q++;
      data = *q++;
      switch(port)
      {
      case 0x3c0:
         inp(0x3da);         /* reset read/write flip-flop */
         outp(0x3c0, reg | 0x20);
                                 /* ensure VGA output is enabled */
         outp(0x3c0, data);
         break;

      case 0x3c2:
      case 0x3c3:
         outp(port, data);   /* directly to the port */
			break;

      case 0x3c4:
      case 0x3ce:
      case 0x3d4:
      default:                   /* This is the default method: */
         outp(port, reg);    /* index to port               */
         outp(port+1, data); /* value to port+1             */
			break;
      }
   } while(--n);

   //enable();
   screen->h = 256;
   set_clip(screen, 0,0,319,255);
   return ret;
}

static int set_arcade_mode(int SType, int SX, int SY)
{
   int ret;

   // Currently only accept 320x240

   if((SX!=320)||(SY!=240))
      return -1;

   // Set Mode X as the base mode

   if(display_cfg.triple_buffer)
      ret = set_gfx_mode(GFX_MODEX,SX,SY,SX,SY*3);
   else
      ret = set_gfx_mode(GFX_MODEX,SX,SY,SX,SY);

   // Apply hacks for arcade monitor rates

   if(!ret){
      arcade_monitor_mode(
         display_cfg.arcade_h_timing,
         display_cfg.arcade_v_timing,
         display_cfg.arcade_center_x,
         display_cfg.arcade_center_y
      );
   }

   return ret;
}
#endif

static int raine_set_gfx_mode(int SType,int SX,int SY)
{
   set_color_depth(display_cfg.bpp);

#ifdef RAINE_DOS
   if((SX==256)&&(SY==240)) return Stretch256x240(SType);
   if((SX==320)&&(SY==256)) return Set320x256(SType);

   if((SX==400)&&(SY==150)) return SetDarkForceMode(SType,SX,300,SY);
   if((SX==640)&&(SY==100)) return SetDarkForceMode(SType,SX,400,SY);
   if((SX==640)&&(SY==150)) return SetDarkForceMode(SType,SX,400,SY);
   if((SX==640)&&(SY==200)) return SetDarkForceMode(SType,SX,400,SY);
   if((SX==640)&&(SY==120)) return SetDarkForceMode(SType,SX,480,SY);
   if((SX==640)&&(SY==160)) return SetDarkForceMode(SType,SX,480,SY);
   if((SX==640)&&(SY==240)) return SetDarkForceMode(SType,SX,480,SY);
   if((SX==800)&&(SY==150)) return SetDarkForceMode(SType,SX,600,SY);
   if((SX==800)&&(SY==200)) return SetDarkForceMode(SType,SX,600,SY);
   if((SX==800)&&(SY==300)) return SetDarkForceMode(SType,SX,600,SY);
   if((SX==1024)&&(SY==192)) return SetDarkForceMode(SType,SX,768,SY);
   if((SX==1024)&&(SY==256)) return SetDarkForceMode(SType,SX,768,SY);
   if((SX==1024)&&(SY==384)) return SetDarkForceMode(SType,SX,768,SY);

   //if(SType==GFX_ARCMON) return set_arcade_mode(SType,SX,SY);
#endif

   if(display_cfg.triple_buffer)
      return set_gfx_mode(SType,SX,SY,SX,SY*3);
   else
      return set_gfx_mode(SType,SX,SY,SX,SY);
}

static int new_set_gfx_mode(int SType,int SX,int SY,int SLines)
{
   int ta,tb;

   screen_valid=0;

   if(SLines==0){				// Set a Normal Mode
      ta=raine_set_gfx_mode(SType,SX,SY);
      if(!ta){
         screen->w=SX;
         screen->h=SY;
         screen->cr=screen->w;
         screen->cb=screen->h;
         gfx_driver->w=screen->w;
         gfx_driver->h=screen->h;
         centre_all_dialogs();
      }
      return ta;
   }
   else{					// Set a Mode with Scanlines Hack
      ta=raine_set_gfx_mode(SType,SX,SY<<1);
      if(!ta){
         screen->w=SX;
         screen->h=SY;
         screen->cr=screen->w;
         screen->cb=screen->h;
         gfx_driver->w=screen->w;
         gfx_driver->h=screen->h;
         if(display_cfg.triple_buffer){
         for(tb=0;tb<SY*3;tb++){
            screen->line[tb]=screen->line[tb+tb];
         }
         }
         else{
         for(tb=0;tb<SY;tb++){
            screen->line[tb]=screen->line[tb+tb];
         }
         }
         centre_all_dialogs();
      }
      return ta;
   }
}

static char *OSName(void)
{
   switch(os_type){
      case OSTYPE_WIN3:		return "Windows 3.1";
      case OSTYPE_WIN95:	return "Windows 95";
      case OSTYPE_WIN98:	return "Windows 98";
      case OSTYPE_WINNT:	return "Windows NT";
      case OSTYPE_OS2:		return "OS/2";
      case OSTYPE_WARP:		return "OS/2 Warp 3";
      case OSTYPE_DOSEMU:	return "Linux DOSEMU";
      case OSTYPE_OPENDOS:	return "Caldera OpenDOS";
      default:			return "MSDOS";
   }
}

static char *CPUFamily(void)
{
    switch(cpu_family){
       case 3:	return "386";
       case 4:	return "486";
       case 5:	return "Pentium";
       case 6:	return "PPro";
       default:	return "x86";
   }
}

static char *CPUFPU(void)
{
   if(cpu_fpu)
      return "FPU";
   else
      return "NoFPU";
}

static char *CPUMMX(void)
{
   if(cpu_mmx)
      return "MMX";
   else
      return "NoMMX";
}

BITMAP *make_mapped_bitmap(BITMAP *src, int *start, PALETTE *src_pal, ULONG cols)
{
   BITMAP *dest;
   UBYTE *line,*line_2;
   int ta,x,y;

   ta = allocate_pens(cols);

   dest = create_bitmap_ex(8, src->w, src->h);

   if(ta>0){

      for(y = 0; y < src->h; y++){
         line   = src->line[y];
         line_2 = dest->line[y];
         for(x = 0; x < src->w; x++){
               if(line[x])
                  line_2[x] = line[x] + ta;
               else
                  line_2[x] = GUI_BOX_COL_MIDDLE;
         }
      }

      memcpy(&gui_pal[ta], src_pal, cols*4);
      set_palette_range(gui_pal, ta, ta+cols-1, 0);

   }
   else{
      ta = 0;
   }

   *start = ta;

   return dest;
}

BITMAP *make_mapped_bitmap_2(BITMAP *src, int *start, PALETTE *src_pal, ULONG cols)
{
   BITMAP *dest;
   UBYTE *line,*line_2;
   int ta,x,y;

      ta = allocate_pens(cols);

      dest = create_bitmap_ex(8, src->w, src->h);

      if(ta>0){

         for(y=0;y<src->h;y++){
            line   = src->line[y];
            line_2 = dest->line[y];
            for(x=0;x<src->w;x++){
               if(line[x]==0)
                  line_2[x] = 0;
               else
                  line_2[x] = line[x]+ta;
            }
         }
         memcpy(&gui_pal[ta], src_pal, cols*4);
         set_palette_range(gui_pal, ta, ta+cols-1, 0);

      }
      else{
         ta = 0;
      }

   *start = ta;

   return dest;
}

void destroy_mapped_bitmap(BITMAP *dest, ULONG cols)
{
   free_pens(cols);
   destroy_bitmap(dest);
}

// about_proc():
// Setup and display the about dialog

static int cycle_logo_pal;

int about_proc(int msg, DIALOG *d, int c)
{
   BITMAP *logo;

   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();
      check_cpu();
      sprintf(about0, EMUNAME " " VERSION " (c)1998-2000 " HANDLE);
      sprintf(about1, "Compiled on " __DATE__ " (" __TIME__ ")");
#ifdef RAINE_DOS
      sprintf(about2, "DJGPP (GCC 2.95.2) by DJ Delorie");
      sprintf(about3, "Allegro 3.9.32 wip by Shawn Hargreaves");
#else
      sprintf(about2, "MSVC++ 5.0 by Microsoft Corporation");
      sprintf(about3, "Allegro 3932 WIP by Shawn Hargreaves");
#endif
      sprintf(about4, "Starscream 0.26r by Neill Corlett");
      sprintf(about5, " ");
      sprintf(about6, "Host System: %s", OSName());
      sprintf(about7, "CPU: %s %s %s %s", CPUFamily(), CPUFPU(), CPUMMX(), cpu_vendor);
      sprintf(about8, " ");
      sprintf(about9, "http://www.rainemu.com/ ?");

      logo = make_mapped_bitmap(RaineData[RaineLogo].dat, &cycle_logo_pal, RaineData[GUIPalette].dat, 7);

      about_raine[1].proc = &d_bitmap_proc;
      about_raine[1].dp   = logo;
      about_raine[1].x    = about_raine[0].x + ((120 * text_height(gui_main_font)) / 8) - (logo->w / 2);
      about_raine[1].y    = about_raine[0].y + (( 32 * text_height(gui_main_font)) / 8) - (logo->h / 2);
      about_raine[1].w    = logo->w;
      about_raine[1].h    = logo->h;

      ret = raine_do_dialog(about_raine, -1);

      destroy_mapped_bitmap(logo, 7);

      return D_REDRAW;
   }

   return ret;
}

// cycle_proc():
// Colour Cycling for the 'About Raine' logo

int cycle_proc(int msg, DIALOG *d, int c)
{
   static int tt;
   static ULONG ts;
   static UBYTE *CYC;
   static PALETTE cyc_pal;
   static int dirty_palette;

   ULONG ta;

   switch(msg){
      case MSG_START:	// Startup
         CYC = malloc(4*192*2);

         for(ta=0;ta<64;ta++){
            CYC[(  0<<2)+(ta<<2)+0] = 63 - ta;
            CYC[(  0<<2)+(ta<<2)+1] = 63;
            CYC[(  0<<2)+(ta<<2)+2] = ta;

            CYC[( 64<<2)+(ta<<2)+0] = ta;
            CYC[( 64<<2)+(ta<<2)+1] = 63 - ta;
            CYC[( 64<<2)+(ta<<2)+2] = 63;

            CYC[(128<<2)+(ta<<2)+0] = 63;
            CYC[(128<<2)+(ta<<2)+1] = ta;
            CYC[(128<<2)+(ta<<2)+2] = 63 - ta;
         }

         for(ta=0;ta<192;ta++){
            CYC[((ta+192)<<2)+0] = CYC[(ta<<2)+0];
            CYC[((ta+192)<<2)+1] = CYC[(ta<<2)+1];
            CYC[((ta+192)<<2)+2] = CYC[(ta<<2)+2];
         }

         memcpy(&cyc_pal, &gui_pal, 256*4);

         tt=0;
         ts=0;
         start_ingame_timer();
         dirty_palette=1;
      break;
      case MSG_END:	// Shutdown
         gui_pal[cycle_logo_pal+1].r=cyc_pal[cycle_logo_pal+1].r;
         gui_pal[cycle_logo_pal+1].g=cyc_pal[cycle_logo_pal+1].g;
         gui_pal[cycle_logo_pal+1].b=cyc_pal[cycle_logo_pal+1].b;
         gui_pal[cycle_logo_pal+2].r=cyc_pal[cycle_logo_pal+2].r;
         gui_pal[cycle_logo_pal+2].g=cyc_pal[cycle_logo_pal+2].g;
         gui_pal[cycle_logo_pal+2].b=cyc_pal[cycle_logo_pal+2].b;
         gui_pal[cycle_logo_pal+3].r=cyc_pal[cycle_logo_pal+3].r;
         gui_pal[cycle_logo_pal+3].g=cyc_pal[cycle_logo_pal+3].g;
         gui_pal[cycle_logo_pal+3].b=cyc_pal[cycle_logo_pal+3].b;
         gui_pal[cycle_logo_pal+4].r=cyc_pal[cycle_logo_pal+4].r;
         gui_pal[cycle_logo_pal+4].g=cyc_pal[cycle_logo_pal+4].g;
         gui_pal[cycle_logo_pal+4].b=cyc_pal[cycle_logo_pal+4].b;
         gui_pal[cycle_logo_pal+5].r=cyc_pal[cycle_logo_pal+5].r;
         gui_pal[cycle_logo_pal+5].g=cyc_pal[cycle_logo_pal+5].g;
         gui_pal[cycle_logo_pal+5].b=cyc_pal[cycle_logo_pal+5].b;
         set_palette_range(gui_pal,cycle_logo_pal+1,cycle_logo_pal+5,0);
         stop_ingame_timer();
         free(CYC);
      break;
      case MSG_IDLE:	// Update
         if( read_ingame_timer() != ts){
         ts = read_ingame_timer();
         gui_pal[cycle_logo_pal+1].r=CYC[((tt+0)<<2)];
         gui_pal[cycle_logo_pal+1].g=CYC[((tt+0)<<2)+1];
         gui_pal[cycle_logo_pal+1].b=CYC[((tt+0)<<2)+2];
         gui_pal[cycle_logo_pal+2].r=CYC[((tt+32)<<2)];
         gui_pal[cycle_logo_pal+2].g=CYC[((tt+32)<<2)+1];
         gui_pal[cycle_logo_pal+2].b=CYC[((tt+32)<<2)+2];
         gui_pal[cycle_logo_pal+3].r=CYC[((tt+64)<<2)];
         gui_pal[cycle_logo_pal+3].g=CYC[((tt+64)<<2)+1];
         gui_pal[cycle_logo_pal+3].b=CYC[((tt+64)<<2)+2];
         gui_pal[cycle_logo_pal+4].r=CYC[((tt+96)<<2)];
         gui_pal[cycle_logo_pal+4].g=CYC[((tt+96)<<2)+1];
         gui_pal[cycle_logo_pal+4].b=CYC[((tt+96)<<2)+2];
         gui_pal[cycle_logo_pal+5].r=CYC[((tt+128)<<2)];
         gui_pal[cycle_logo_pal+5].g=CYC[((tt+128)<<2)+1];
         gui_pal[cycle_logo_pal+5].b=CYC[((tt+128)<<2)+2];
         tt++;
         if(tt>=192) tt=0;
         dirty_palette=1;
         SEND_MESSAGE(d,MSG_DRAW,0);
         }
      break;
      case MSG_DRAW:	// Draw
         if(dirty_palette){
            dirty_palette=0;
            set_palette_range(gui_pal,cycle_logo_pal+1,cycle_logo_pal+5,0);
         }
         //textout(screen,font,gamename,64,tt,255);
      break;
      case MSG_CLICK:	// Draw
         rgui_cfg.gui_col_text_1     = ((rgui_cfg.gui_col_text_1>>8)&0xFFFF)     | ((rgui_cfg.gui_col_text_1<<16)&0xFF0000);
         rgui_cfg.gui_col_text_2     = ((rgui_cfg.gui_col_text_2>>8)&0xFFFF)     | ((rgui_cfg.gui_col_text_2<<16)&0xFF0000);
         rgui_cfg.gui_col_black      = ((rgui_cfg.gui_col_black>>8)&0xFFFF)      | ((rgui_cfg.gui_col_black<<16)&0xFF0000);
         rgui_cfg.gui_col_select     = ((rgui_cfg.gui_col_select>>8)&0xFFFF)     | ((rgui_cfg.gui_col_select<<16)&0xFF0000);
         rgui_cfg.gui_box_col_high_2 = ((rgui_cfg.gui_box_col_high_2>>8)&0xFFFF) | ((rgui_cfg.gui_box_col_high_2<<16)&0xFF0000);
         rgui_cfg.gui_box_col_high_1 = ((rgui_cfg.gui_box_col_high_1>>8)&0xFFFF) | ((rgui_cfg.gui_box_col_high_1<<16)&0xFF0000);
         rgui_cfg.gui_box_col_middle = ((rgui_cfg.gui_box_col_middle>>8)&0xFFFF) | ((rgui_cfg.gui_box_col_middle<<16)&0xFF0000);
         rgui_cfg.gui_box_col_low_1  = ((rgui_cfg.gui_box_col_low_1>>8)&0xFFFF)  | ((rgui_cfg.gui_box_col_low_1<<16)&0xFF0000);
         rgui_cfg.gui_box_col_low_2  = ((rgui_cfg.gui_box_col_low_2>>8)&0xFFFF)  | ((rgui_cfg.gui_box_col_low_2<<16)&0xFF0000);

         set_gui_palette();

         set_palette_range(gui_pal,0,255,0);
      break;
      default:
      break;
   }

   return D_O_K;
}

// about_game_proc():
// Setup and display the about game dialog

int about_game_proc(int msg, DIALOG *d, int c)
{
   char *text_data;
   BITMAP *logo,*logo_2;
   PALETTE *logo_pal;
   DATAFILE *company_logo;
   DIR_INFO *dir_list;
   ROM_INFO *rom_list;
   VIDEO_INFO *vid_info;
   int ret,romof,cloneof;
   ULONG i;
   GAME_MAIN *game_info;

   ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){

   if(current_game){

      Unselect_Button(d);
      FadeGUI();

      text_data = malloc(0x10000);

      text_data[0] = 0;

      game_info = current_game;

      sprintf(text_data+strlen(text_data),"Game:\n   %s\n\n", game_info->long_name);

      sprintf(text_data+strlen(text_data),"Company:\n   %s\n\n", game_company_name(game_info->company_id));

      sprintf(text_data+strlen(text_data),"Year:\n   %4d\n\n", game_info->year);

      /*

      board number [optional]

      */

      if(game_info->board)

         sprintf(text_data+strlen(text_data),"Board:\n   %s\n\n", game_info->board);

      /*

      video %d x %d

      */

      vid_info = game_info->video_info;

      sprintf(text_data+strlen(text_data),"Video:\n   %d x %d (8bpp)", vid_info->screen_x, vid_info->screen_y);

      if(vid_info->flags & VIDEO_ROTATABLE){

      switch(VIDEO_ROTATE( vid_info->flags )){
         case VIDEO_ROTATE_NORMAL:
            sprintf(text_data+strlen(text_data)," horizontal");
         break;
         case VIDEO_ROTATE_90:
            sprintf(text_data+strlen(text_data)," vertical");
         break;
         case VIDEO_ROTATE_180:
            sprintf(text_data+strlen(text_data)," horizontal");
         break;
         case VIDEO_ROTATE_270:
            sprintf(text_data+strlen(text_data)," vertical");
         break;
      }

      }

      sprintf(text_data+strlen(text_data),"\n\n");

      /*

      tiles %s

      */

      if(tile_list_count){

         sprintf(text_data+strlen(text_data),"Tiles:\n");

         for( i = 0; i < tile_list_count; i ++){

            sprintf(text_data+strlen(text_data),"   %-14s", tile_type[tile_list[i].type]);
            sprintf(text_data+strlen(text_data)," %6d tiles\n", tile_list[i].count);

         }

         sprintf(text_data+strlen(text_data),"\n");

      }

      /*

      archive %s

      */

      sprintf(text_data+strlen(text_data),"Archive:\n");

      dir_list = game_info->dir_list;

      while(dir_list->maindir){

         if((dir_list->maindir[0] != '#') && (dir_list->maindir[0] != '$'))

            sprintf(text_data+strlen(text_data),"   %s\n", dir_list->maindir);

         dir_list++;

      }

      sprintf(text_data+strlen(text_data),"\n");

      /*

      romof %s [optional]

      */

      romof = 0;

      dir_list = game_info->dir_list;

      while(dir_list->maindir){

         if(dir_list->maindir[0] == '#'){

               if(!romof)

                  sprintf(text_data+strlen(text_data),"Romof:\n");

               sprintf(text_data+strlen(text_data),"   %s\n", (dir_list->maindir) + 1 );

               romof ++;

         }

         dir_list ++;

      }

      if(romof)

         sprintf(text_data+strlen(text_data),"\n");

      /*

      cloneof %s [optional]

      */

      cloneof = 0;

      dir_list = game_info->dir_list;

      while(dir_list->maindir){

         if(dir_list->maindir[0] == '$'){

               if(!cloneof)

                  sprintf(text_data+strlen(text_data),"Cloneof:\n");

               sprintf(text_data+strlen(text_data),"   %s\n", (dir_list->maindir) + 1 );

               cloneof ++;

         }

         dir_list ++;

      }

      if(cloneof)

         sprintf(text_data+strlen(text_data),"\n");

      sprintf(text_data+strlen(text_data),"Roms:\n");

      rom_list = game_info->rom_list;

      while(rom_list->name){
         sprintf(text_data+strlen(text_data),"   %-12s %7d bytes\n", rom_list->name, rom_list->size);
         rom_list++;
      }

      sprintf(text_data+strlen(text_data),"\n");

      about_game[3].d1 = 0;
      about_game[3].d2 = 0;
      about_game[3].dp = text_data;

      company_logo = (DATAFILE *) RaineData[company].dat;

      switch(game_info->company_id){
         case COMPANY_ID_EX_SYSTEM:
            logo     = company_logo[company_ex_system].dat;
            logo_pal = company_logo[company_ex_system_pal].dat;
         break;
         case COMPANY_ID_JALECO:
            logo     = company_logo[company_jaleco].dat;
            logo_pal = company_logo[company_jaleco_pal].dat;
         break;
         case COMPANY_ID_NICHIBUTSU:
            logo     = company_logo[company_nichibutsu].dat;
            logo_pal = company_logo[company_nichibutsu_pal].dat;
         break;
         case COMPANY_ID_NMK:
            logo     = company_logo[company_nmk].dat;
            logo_pal = company_logo[company_nmk_pal].dat;
         break;
         case COMPANY_ID_RAIZING:
            logo     = company_logo[company_raizing].dat;
            logo_pal = company_logo[company_raizing_pal].dat;
         break;
         case COMPANY_ID_SETA:
            logo     = company_logo[company_seta].dat;
            logo_pal = company_logo[company_seta_pal].dat;
         break;
         case COMPANY_ID_TAD:
            logo     = company_logo[company_tad].dat;
            logo_pal = company_logo[company_tad_pal].dat;
         break;
         case COMPANY_ID_TAITO:
            logo     = company_logo[company_taito].dat;
            logo_pal = company_logo[company_taito_pal].dat;
         break;
         case COMPANY_ID_TECHNOS:
            logo     = company_logo[company_technos].dat;
            logo_pal = company_logo[company_technos_pal].dat;
         break;
         case COMPANY_ID_TECMO:
            logo     = company_logo[company_tecmo].dat;
            logo_pal = company_logo[company_tecmo_pal].dat;
         break;
         case COMPANY_ID_TOAPLAN:
            logo     = company_logo[company_toaplan].dat;
            logo_pal = company_logo[company_toaplan_pal].dat;
         break;
         case COMPANY_ID_UPL:
            logo     = company_logo[company_upl].dat;
            logo_pal = company_logo[company_upl_pal].dat;
         break;
         case COMPANY_ID_VISCO:
            logo     = company_logo[company_visco].dat;
            logo_pal = company_logo[company_visco_pal].dat;
         break;
         default:
            logo     = NULL;
            logo_pal = NULL;
         break;
      }

      if(logo){
         logo_2 = make_mapped_bitmap(logo, &i, logo_pal, 16);

         about_game[1].proc = &d_bitmap_proc;
         about_game[1].dp   = logo_2;
         about_game[1].x    = about_game[0].x + ((120 * text_height(gui_main_font)) / 8) - (logo_2->w / 2);
         about_game[1].y    = about_game[0].y + (( 32 * text_height(gui_main_font)) / 8) - (logo_2->h / 2);
         about_game[1].w    = logo_2->w;
         about_game[1].h    = logo_2->h;
      }
      else{
         about_game[1].proc = &d_null_proc;
         about_game[1].dp   = NULL;
         about_game[1].x    = about_game[0].x;
         about_game[1].y    = about_game[0].y;
         about_game[1].w    = 0;
         about_game[1].h    = 0;
      }

      ret = raine_do_dialog(about_game,-1);

      if(logo)
         destroy_mapped_bitmap(logo_2, 16);

      free(text_data);
   }
   ret = D_REDRAW;
   }
   return ret;
}

void ScreenChange(void)
{
   show_mouse(NULL);

   if(display_cfg.scanlines==2)
      display_cfg.screen_y >>= 1;

   if(new_set_gfx_mode(display_cfg.screen_type,display_cfg.screen_x,display_cfg.screen_y,display_cfg.scanlines)){	// Didn't work:
      memcpy(&display_cfg, &prev_display_cfg, sizeof(DISPLAY_CFG));
      new_set_gfx_mode(display_cfg.screen_type,display_cfg.screen_x,display_cfg.screen_y,display_cfg.scanlines);	// Revert to old mode
   }
   else{
      ReClipScreen();
   }

   set_palette_range(gui_pal,0,255,0);
   sprintf(screenname,  raine_translate_text("%dx%d %dbpp; %s"), display_cfg.screen_x, display_cfg.screen_y, display_cfg.bpp, gfx_card_name(display_cfg.screen_type));
   show_mouse(screen);
}

// screen_select_proc():
// Display then process Screen Select dialog

int screen_select_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      memcpy(&prev_display_cfg, &display_cfg, sizeof(DISPLAY_CFG));

      ret = raine_gfx_mode_select();

      if(ret){
         WantScreen=1;
         return D_EXIT;
      }
      else{
         memcpy(&display_cfg, &prev_display_cfg, sizeof(DISPLAY_CFG));
         return D_REDRAW;
      }
   }
   return ret;
}

// Data for game list and available/missing games
 
static int *GameAvail = NULL;
static ULONG GameAvailCount = 0;
static int *GameMissing = NULL;
static ULONG GameMissingCount = 0;

// build_game_avail_list():
// Updates available and missing game lists, called
// before doing game select dialog.


void set_game_list_mode(int i)
{
   int ta, tb;

   switch(rgui_cfg.game_list_mode){
      case 0x00:
         tb = game_select[1].d1;
      break;
      case 0x01:
         tb = GameAvail[game_select[1].d1];
      break;
      case 0x02:
         tb = GameMissing[game_select[1].d1];
      break;
   }

   rgui_cfg.game_list_mode = i;

   switch(rgui_cfg.game_list_mode){
      case 0x00:
         sprintf(game_select[10].dp,"%d/%d Games Listed", game_count, game_count);
         game_select[1].d1 = tb;
      break;
      case 0x01:
         sprintf(game_select[10].dp,"%d/%d Games Listed", GameAvailCount, game_count);
         game_select[1].d1 = 0;
         for(ta = 0; ta < GameAvailCount; ta++)
            if(GameAvail[ta] == tb)
               game_select[1].d1 = ta;
      break;
      case 0x02:
         sprintf(game_select[10].dp,"%d/%d Games Listed", GameMissingCount, game_count);
         game_select[1].d1 = 0;
         for(ta = 0; ta < GameMissingCount; ta++)
            if(GameMissing[ta] == tb)
               game_select[1].d1 = ta;
      break;
   }
}

int game_exists(int num)
{
   DIR_INFO *dir_list;
   UBYTE str[256];
   ULONG ta;

   dir_list = game_list[num]->dir_list;

   while( (dir_list->maindir) ){

      if((dir_list->maindir[0] != '#')&&(dir_list->maindir[0] != '$')){

         for(ta = 0; ta < 4; ta ++){

            if(dir_cfg.rom_dir[ta][0]){

               sprintf(str,"%s%s.zip", dir_cfg.rom_dir[ta], dir_list->maindir);
               if((file_exists(str,255,NULL))) return 1;

               sprintf(str,"%s%s", dir_cfg.rom_dir[ta], dir_list->maindir);
               if((file_exists(str,255,NULL))) return 1;

            }

         }

      }

      dir_list++;
   }

   return 0;

}

void build_game_avail_list(void)
{
   int ta;

   if(GameAvail) free_game_avail_list();

   GameAvail   = (int *) malloc(sizeof(int) * game_count);
   GameMissing = (int *) malloc(sizeof(int) * game_count);

   GameAvailCount = 0;
   GameMissingCount = 0;

   for(ta=0; ta<game_count; ta++){

      if( game_exists(ta) )

         GameAvail[GameAvailCount++] = ta;

      else

         GameMissing[GameMissingCount++] = ta;

   }

   ta = rgui_cfg.game_list_mode;
   rgui_cfg.game_list_mode = 0;
   game_select[1].d1 = raine_cfg.req_game_index;

   set_game_list_mode(ta);
}

void free_game_avail_list(void)
{
   if(GameAvail) free(GameAvail);
   GameAvail = NULL;

   if(GameMissing) free(GameMissing);
   GameMissing = NULL;
}

// listbox_getter():
// Get game names for game select list, three
// different listing modes.

char *listbox_getter(int index, int *list_size)
{
   int ta;
   char tb[2];
   char tc[2];

   switch(rgui_cfg.game_list_mode){
   case 0x00:			// Mode 0: List all games
      switch(index){
      case -1:			// Return List Size
         *list_size = game_count;
         return NULL;
      break;
      case -2:			// Act upon List Object Selection
      return NULL;
      break;
      case -3:			// Act Keyboard Input
         tb[0] = *list_size & 0xFF;
         tb[1] = 0;
         tc[1] = 0;
         for(ta=0;ta<game_count;ta++){
            tc[0] = game_list[ta]->long_name[0];
            if(!(strcasecmp(tb,tc))){
               game_select[1].d1 = ta;
               if(ta < (game_select[1].d2     )) game_select[1].d2 = ta;
               if(ta > (game_select[1].d2 + 13)) game_select[1].d2 = ta - 13;
               *list_size = D_USED_CHAR;
               return NULL;
            }
         }
      return NULL;
      break;
      default:
         if((index >= 0)&&(index<game_count))
            return game_list[index]->long_name;
         else
            return NULL;
      break;
      }
   break;
   case 0x01:			// Mode 1: List available games
      switch(index){
      case -1:			// Return List Size
         *list_size = GameAvailCount;
         return NULL;
      break;
      case -2:			// Act upon List Object Selection
      return NULL;
      break;
      case -3:			// Act Keyboard Input
	 tb[0] = *list_size & 0xFF;
         tb[1] = 0;
         tc[1] = 0;
         for(ta=0; ta<GameAvailCount; ta++){
            tc[0] = game_list[GameAvail[ta]]->long_name[0];
            if(!(strcasecmp(tb,tc))){
               game_select[1].d1 = ta;
               if(ta < (game_select[1].d2     )) game_select[1].d2 = ta;
               if(ta > (game_select[1].d2 + 13)) game_select[1].d2 = ta - 13;
               *list_size = D_USED_CHAR;
               return NULL;
            }
         }
      return NULL;
      break;
      default:
         if((index >= 0)&&(index<GameAvailCount))
            return game_list[GameAvail[index]]->long_name;
         else
            return NULL;
      break;
      }
   break;
   case 0x02:			// Mode 2: List missing games
      switch(index){
      case -1:			// Return List Size
         *list_size=GameMissingCount;
         return NULL;
      break;
      case -2:			// Act upon List Object Selection
      return NULL;
      break;
      case -3:			// Act Keyboard Input
	 tb[0] = *list_size & 0xFF;
         tb[1] = 0;
         tc[1] = 0;
         for(ta=0;ta<GameMissingCount;ta++){
            tc[0] = game_list[GameMissing[ta]]->long_name[0];
            if(!(strcasecmp(tb,tc))){
               game_select[1].d1 = ta;
               if(ta < (game_select[1].d2     )) game_select[1].d2 = ta;
               if(ta > (game_select[1].d2 + 13)) game_select[1].d2 = ta - 13;
               *list_size = D_USED_CHAR;
               return NULL;
            }
         }
      return NULL;
      break;
      default:
         if((index >= 0)&&(index<GameMissingCount))
            return game_list[GameMissing[index]]->long_name;
         else
            return NULL;
      break;
      }
   break;
   }
   return NULL;
}

// game_radio_proc():
// Radio buttons for game list mode, need to be custom
// to incorporate automatic update on button selection.

int game_radio_proc(int msg, DIALOG *d, int c)
{
   int i;

   int ret=x_raine_radio_proc(msg, d, c);

   if((msg==MSG_KEY)||(msg==MSG_CLICK)){

      // Get selected List Mode
      // ----------------------

      if((game_select[6+0].flags)&D_SELECTED) i = 0;
      if((game_select[6+1].flags)&D_SELECTED) i = 1;
      if((game_select[6+2].flags)&D_SELECTED) i = 2;

      set_game_list_mode(i);

      // Refresh the Game List
      // ---------------------

      SEND_MESSAGE(&game_select[1], MSG_END, 0);
      SEND_MESSAGE(&game_select[1], MSG_START, 0);

      scare_mouse();
      broadcast_dialog_message(MSG_DRAW, 0);
      unscare_mouse();

      // Wait for release
      // ----------------

      while(gui_mouse_b()){
         dialog_oxygen();
      };
   }
   return ret;
}

int load_game_proc(int msg, DIALOG *d, int c)
{
   switch(msg){
      case MSG_IDLE:

         mouse_on_real_screen();	// this is a hack to keep the mouse movable

         LoadDefault();

         load_error = 0;
         load_debug = malloc(0x10000);

         sprintf(load_debug,"Load Game\n---------\n\n");

         current_game = game_list[raine_cfg.req_game_index];

         current_game->load_game();

         init_inputs();
         init_dsw();
         init_romsw();
         init_sound_list();

         mouse_on_buffer_screen();

         return D_EXIT;
      break;
      default:
         return D_O_K;
      break;
   }
}


void do_load_game(void)
{
   BITMAP *load_mouse;
   int ta;

   /*

   satisfy the request

   */

   raine_cfg.req_load_game = 0;

   /*

   don't reload if the game is already in memory

   */

   if(current_game == game_list[raine_cfg.req_game_index])

      return;

   /*

   close down the current game (if there is one)

   */

   if(current_game){

      save_game_config();

      current_game->clear_game();

      ClearDefault();

      current_game = NULL;

   }

      // Do the load (via a loading window)

      load_mouse = make_mapped_bitmap_2(RaineData[mouse_busy].dat, &ta, RaineData[mouse_busy_pal].dat, 16);

      set_mouse_sprite(load_mouse);

      FadeGUI();
      raine_do_dialog(loading_dialog,-1);

      destroy_mapped_bitmap(load_mouse, 16);

      set_mouse_sprite(RaineData[Mouse].dat);

      // Check if the load was successful, if ok, load game config, if bad clear the mess

      switch(load_error&3){
      case 0x00:				// SUCCESS

         memcpy(&prev_display_cfg, &display_cfg, sizeof(DISPLAY_CFG));

         load_game_config();
         reset_game_hardware();
         sprintf(gamename, "%s", current_game->long_name);

         if(raine_cfg.save_game_screen_settings)

            WantScreen=1;

      break;
      case LOAD_FATAL_ERROR:			// FATAL ERROR - REMOVE GAME
      case LOAD_FATAL_ERROR|LOAD_WARNING:

         FadeGUI();
         load_problem[2].dp = load_debug;
         raine_do_dialog(load_problem,-1);

         current_game->clear_game();

         ClearDefault();

         current_game = NULL;

         sprintf(gamename,raine_translate_text("<No Game Loaded>"));
      break;
      case LOAD_WARNING:			// WARNING - IT MIGHT RUN OK

         FadeGUI();
         load_problem[2].dp = load_debug;
         raine_do_dialog(load_problem,-1);

         memcpy(&prev_display_cfg, &display_cfg, sizeof(DISPLAY_CFG));

         load_game_config();
         reset_game_hardware();
         sprintf(gamename, "%s", current_game->long_name);

         if(raine_cfg.save_game_screen_settings)

            WantScreen=1;


      break;
      }

      free(load_debug);
}

// game_select_proc():
// Display then process Game Select dialog

int game_select_proc(int msg, DIALOG *d, int c)
{
   int ta;
   int ret=x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){

      Unselect_Button(d);
      FadeGUI();

      /*

      set radio buttons

      */

      if(rgui_cfg.game_list_mode>2) rgui_cfg.game_list_mode = 2;

      game_select[6+0].flags=0;
      game_select[6+1].flags=0;
      game_select[6+2].flags=0;

      game_select[6+rgui_cfg.game_list_mode].flags = D_SELECTED;

      build_game_avail_list();

      // Do the Dialog
      // -------------

      ret=raine_do_dialog(game_select,-1);

      // Load game (if user didn't click cancel)
      // ---------------------------------------

      if((ret!=3)&&(ret!=-1)){

         switch(rgui_cfg.game_list_mode){
         case 0x00:
            ta = game_select[1].d1;
            if((ta >= 0)&&(ta < game_count)){
               raine_cfg.req_game_index = ta;
               do_load_game();
            }
         break;
         case 0x01:
            ta = game_select[1].d1;
            if((ta >= 0)&&(ta < GameAvailCount)){
               raine_cfg.req_game_index = GameAvail[ta];
               do_load_game();
            }
         break;
         case 0x02:
            ta = game_select[1].d1;
            if((ta >= 0)&&(ta < GameMissingCount)){
               raine_cfg.req_game_index = GameMissing[ta];
               do_load_game();
            }
         break;
         }

      }

      free_game_avail_list();

      if(WantScreen)

         return D_EXIT;

      else

         return D_REDRAW;

   }
   return ret;
}

int romdir_edit_proc(int msg, DIALOG *d, int c)
{
   int i;
   int ret=x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      sprintf(romedit[0],"%s",dir_cfg.rom_dir[0]);
      sprintf(romedit[1],"%s",dir_cfg.rom_dir[1]);
      sprintf(romedit[2],"%s",dir_cfg.rom_dir[2]);
      sprintf(romedit[3],"%s",dir_cfg.rom_dir[3]);

      ret=raine_do_dialog(romdir_edit_dialog,-1);
      if(ret==1){

      sprintf(dir_cfg.rom_dir[0],"%s",romedit[0]);
      sprintf(dir_cfg.rom_dir[1],"%s",romedit[1]);
      sprintf(dir_cfg.rom_dir[2],"%s",romedit[2]);
      sprintf(dir_cfg.rom_dir[3],"%s",romedit[3]);

      for(i = 0; i < 4; i ++){
         if(dir_cfg.rom_dir[i][0]){
            put_backslash(dir_cfg.rom_dir[i]);
            strlwr(dir_cfg.rom_dir[i]);
         }
      }

      build_game_avail_list();

      scare_mouse();
      SEND_MESSAGE(&game_select[1],  MSG_START, 0);
      SEND_MESSAGE(&game_select[1],  MSG_DRAW,  0);
      SEND_MESSAGE(&game_select[10], MSG_DRAW,  0);
      unscare_mouse();
      dialog_oxygen();

      }
      return D_REDRAW;
   }
   return ret;
}

int dir_edit_proc(int msg, DIALOG *d, int c)
{
   char dir_path[256];
   char *ta;
   int ret;

   ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      sprintf(dir_path,"%s",romedit[d->d1]);

      if(raine_file_select("Select directory", dir_path, NULL)){

      // Cut any filename

      ta = get_filename(dir_path);
      if(ta) ta[0] = 0;

      sprintf(romedit[d->d1],"%s",dir_path);

      }

      return D_REDRAW;
   }
   return ret;
}

static int key_view_mode = 0;

void set_use_custom(int ta)
{
   use_custom_keys = ta;

   input_select[KEY_TYPE_LIST+0].flags=0;
   input_select[KEY_TYPE_LIST+1].flags=0;

   input_select[KEY_TYPE_LIST+use_custom_keys].flags=D_SELECTED;
}

void set_key_view_mode(int ta)
{
   key_view_mode = ta;

   input_select[KEY_VIEW_LIST+0].flags=0;
   input_select[KEY_VIEW_LIST+1].flags=0;
   input_select[KEY_VIEW_LIST+2].flags=0;

   input_select[KEY_VIEW_LIST+key_view_mode].flags=D_SELECTED;
}

void do_key_edit(void)
{
   switch(key_view_mode){
      case 0:
         def_input_list[input_select[1].d1].scancode = read_new_keycode();
      break;
      case 1:
         def_input_list_emu[input_select[1].d1].scancode = read_new_keycode();
      break;
      case 2:
         if((InputCount)&&(use_custom_keys)){
            InputList[input_select[1].d1].Key = read_new_keycode();
            set_use_custom(1);
         }
      break;
   }
}

char *inputlist_getter(int index, int *list_size)
{
   switch(key_view_mode){

   case 0:

   switch(index){
   case -1:			// Return List Size
      *list_size = KB_DEF_COUNT;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      do_key_edit();
      *list_size = D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < KB_DEF_COUNT)){
         sprintf(S,"%-21s %s",def_input_list[index].name, KeyName(def_input_list[index].scancode));
         return S;
      }
      else
         return NULL;
   break;
   }

   break;
   case 1:

   switch(index){
   case -1:			// Return List Size
      *list_size = KB_EMU_DEF_COUNT;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      do_key_edit();
      *list_size = D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < KB_EMU_DEF_COUNT)){
         sprintf(S,"%-21s %s",def_input_list_emu[index].name, KeyName(def_input_list_emu[index].scancode));
         return S;
      }
      else
         return NULL;
   break;
   }

   break;
   case 2:

   if(!use_custom_keys){

      if(index==-1){
         *list_size=1;
         return NULL;
      }
      if(index==0){
         return raine_translate_text("<Using Default Settings>");
      }
      return NULL;

   }

   if(!InputCount){

      if(index==-1){
         *list_size=1;
         return NULL;
      }
      if(index==0){
         return raine_translate_text("<No Inputs Mapped>");
      }
      return NULL;

   }

   switch(index){
   case -1:			// Return List Size
      *list_size = InputCount;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      do_key_edit();
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < InputCount)){
         sprintf(S,"%-21s %s",InputList[index].InputName, KeyName(InputList[index].Key));
         return S;
      }
      else
         return NULL;
   break;
   }

   break;
   }

}


/*

key edit custom mode

*/

int keys_radio_proc(int msg, DIALOG *d, int c)
{
   int i;
   int ret=x_raine_radio_proc(msg, d, c);

   if((msg==MSG_KEY)||(msg==MSG_CLICK)){

      i = use_custom_keys;

      if((input_select[KEY_TYPE_LIST+0].flags)&D_SELECTED) use_custom_keys=0;
      if((input_select[KEY_TYPE_LIST+1].flags)&D_SELECTED) use_custom_keys=1;

      SEND_MESSAGE(&input_select[1], MSG_END, 0);

      if((use_custom_keys) && (!i)) reset_game_keys();

      SEND_MESSAGE(&input_select[1], MSG_START, 0);

      scare_mouse();
      broadcast_dialog_message(MSG_DRAW, 0);
      unscare_mouse();

      while(gui_mouse_b()){
         dialog_oxygen();
      };
   }

   return ret;
}


/*

key edit view mode

*/

int keys_radio_proc_2(int msg, DIALOG *d, int c)
{
   int ret=x_raine_radio_proc(msg, d, c);

   if((msg==MSG_KEY)||(msg==MSG_CLICK)){

      if((input_select[KEY_VIEW_LIST+0].flags)&D_SELECTED) key_view_mode=0;
      if((input_select[KEY_VIEW_LIST+1].flags)&D_SELECTED) key_view_mode=1;
      if((input_select[KEY_VIEW_LIST+2].flags)&D_SELECTED) key_view_mode=2;

      SEND_MESSAGE(&input_select[1], MSG_END, 0);

      SEND_MESSAGE(&input_select[1], MSG_START, 0);

      scare_mouse();
      broadcast_dialog_message(MSG_DRAW, 0);
      unscare_mouse();

      while(gui_mouse_b()){
         dialog_oxygen();
      };
   }

   return ret;
}

/*

keyboard edit dialog

*/

int input_select_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){

      Unselect_Button(d);
      FadeGUI();

      set_use_custom(use_custom_keys);
      set_key_view_mode(key_view_mode);

      ret=0;
      while((ret!=2)&&(ret!=-1)){
         ret=raine_do_dialog(input_select,-1);
         if((ret!=2)&&(ret!=-1)){

            do_key_edit();

         }
      }

      ret = D_REDRAW;

   }

   return ret;
}

static UBYTE selected_key;

int key_select_proc(int msg, DIALOG *d, int c)
{
   int ta,tb;

   switch(msg){
      case MSG_START:
      case MSG_END:

         ta = 1;
         while(ta!=0){
            ta = 0;
            for(tb=1;tb<KEY_MAX;tb++){
               if(key[tb]) ta = 1;
            }
         }
         clear_keybuf();

         return D_O_K;

      break;
      case MSG_IDLE:
         for(tb=1;tb<KEY_MAX;tb++){
            if(key[tb]){
               selected_key = tb;
               return D_EXIT;
            }
         }
         return D_O_K;
      break;
      default:
         return D_O_K;
      break;
   }
}

UBYTE read_new_keycode(void)
{
   FadeGUI();
   raine_do_dialog(key_editing,-1);
   return selected_key;
}

int key_edit_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      do_key_edit();
      ret = D_REDRAW;
   }
   return ret;
}

/******************************************************************************/

static int joy_key_view_mode = 0;

void joy_set_use_custom(int ta)
{
   joy_use_custom_keys = ta;

   joy_input_select[JOY_KEY_TYPE_LIST+0].flags=0;
   joy_input_select[JOY_KEY_TYPE_LIST+1].flags=0;

   joy_input_select[JOY_KEY_TYPE_LIST+joy_use_custom_keys].flags=D_SELECTED;
}

void set_joy_key_view_mode(int ta)
{
   joy_key_view_mode = ta;

   joy_input_select[JOY_KEY_VIEW_LIST+0].flags=0;
   joy_input_select[JOY_KEY_VIEW_LIST+1].flags=0;
   joy_input_select[JOY_KEY_VIEW_LIST+2].flags=0;

   joy_input_select[JOY_KEY_VIEW_LIST+joy_key_view_mode].flags=D_SELECTED;
}

void do_joy_edit(void)
{
   switch(joy_key_view_mode){
      case 0:
         def_input_list[joy_input_select[1].d1].joycode = joy_read_new_keycode();
      break;
      case 1:
         def_input_list_emu[joy_input_select[1].d1].joycode = joy_read_new_keycode();
      break;
      case 2:
         if((InputCount)&&(joy_use_custom_keys)){
            InputList[joy_input_select[1].d1].Joy = joy_read_new_keycode();
            joy_set_use_custom(1);
         }
      break;
   }
}

char *joy_inputlist_getter(int index, int *list_size)
{
   switch(joy_key_view_mode){

   case 0:

   switch(index){
   case -1:			// Return List Size
      *list_size = KB_DEF_COUNT;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      do_joy_edit();
      *list_size = D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < KB_DEF_COUNT)){
         sprintf(S,"%-21s %s",def_input_list[index].name, JoyName(def_input_list[index].joycode));
         return S;
      }
      else
         return NULL;
   break;
   }

   break;
   case 1:

   switch(index){
   case -1:			// Return List Size
      *list_size = KB_EMU_DEF_COUNT;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      do_joy_edit();
      *list_size = D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < KB_EMU_DEF_COUNT)){
         sprintf(S,"%-21s %s",def_input_list_emu[index].name, JoyName(def_input_list_emu[index].joycode));
         return S;
      }
      else
         return NULL;
   break;
   }

   break;
   case 2:

   if(!joy_use_custom_keys){

      if(index==-1){
         *list_size=1;
         return NULL;
      }
      if(index==0){
         return raine_translate_text("<Using Default Settings>");
      }
      return NULL;

   }

   if(!InputCount){

      if(index==-1){
         *list_size=1;
         return NULL;
      }
      if(index==0){
         return raine_translate_text("<No Inputs Mapped>");
      }
      return NULL;

   }

   switch(index){
   case -1:			// Return List Size
      *list_size = InputCount;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      do_joy_edit();
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < InputCount)){
         sprintf(S,"%-21s %s",InputList[index].InputName, JoyName(InputList[index].Joy));
         return S;
      }
      else
         return NULL;
   break;
   }

   break;
   }

}

int joy_keys_radio_proc(int msg, DIALOG *d, int c)
{
   int i;
   int ret=x_raine_radio_proc(msg, d, c);

   if((msg==MSG_KEY)||(msg==MSG_CLICK)){

      i = joy_use_custom_keys;

      if((joy_input_select[JOY_KEY_TYPE_LIST+0].flags)&D_SELECTED) joy_use_custom_keys=0;
      if((joy_input_select[JOY_KEY_TYPE_LIST+1].flags)&D_SELECTED) joy_use_custom_keys=1;

      SEND_MESSAGE(&joy_input_select[1], MSG_END, 0);

      if((joy_use_custom_keys) && (!i)) joy_reset_game_keys();

      SEND_MESSAGE(&joy_input_select[1], MSG_START, 0);

      scare_mouse();
      broadcast_dialog_message(MSG_DRAW, 0);
      unscare_mouse();

      while(gui_mouse_b()){
         dialog_oxygen();
      };
   }

   return ret;
}

int joy_keys_radio_proc_2(int msg, DIALOG *d, int c)
{
   int ret=x_raine_radio_proc(msg, d, c);

   if((msg==MSG_KEY)||(msg==MSG_CLICK)){

      if((joy_input_select[JOY_KEY_VIEW_LIST+0].flags)&D_SELECTED) joy_key_view_mode=0;
      if((joy_input_select[JOY_KEY_VIEW_LIST+1].flags)&D_SELECTED) joy_key_view_mode=1;
      if((joy_input_select[JOY_KEY_VIEW_LIST+2].flags)&D_SELECTED) joy_key_view_mode=2;

      SEND_MESSAGE(&joy_input_select[1], MSG_END, 0);

      SEND_MESSAGE(&joy_input_select[1], MSG_START, 0);

      scare_mouse();
      broadcast_dialog_message(MSG_DRAW, 0);
      unscare_mouse();

      while(gui_mouse_b()){
         dialog_oxygen();
      };
   }

   return ret;
}

int joy_select_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      joy_set_use_custom(joy_use_custom_keys);
      set_joy_key_view_mode(joy_key_view_mode);

      ret=0;
      while((ret!=2)&&(ret!=-1)){
         ret=raine_do_dialog(joy_input_select,-1);
         if((ret!=2)&&(ret!=-1)){

            do_joy_edit();

         }
      }
      return D_REDRAW;
   }
   return ret;
}

static UBYTE joy_selected_key;

int joy_key_select_proc(int msg, DIALOG *d, int c)
{
   int ta,tb;

   switch(msg){
      case MSG_START:
      case MSG_END:

         ta = 1;
         while(ta){
            ta = 0;
            update_rjoy_list();
            for(tb=0;tb<RJOY_COUNT;tb++){
               if(rjoy[tb]) ta = 1;
            }
         }

         ta = 1;
         while(ta){
            ta = 0;
            for(tb=1;tb<KEY_MAX;tb++){
               if(key[tb]) ta = 1;
            }
         }
         clear_keybuf();

         return D_O_K;

      break;
      case MSG_IDLE:

         update_rjoy_list();
         for(tb=0;tb<RJOY_COUNT;tb++){
            if(rjoy[tb]){
               joy_selected_key = tb;
               return D_EXIT;
            }
         }

         if(key[KEY_ESC]){
            joy_selected_key = 0;
            return D_EXIT;
         }

         return D_O_K;
      break;
      default:
         return D_O_K;
      break;
   }
}

UBYTE joy_read_new_keycode(void)
{
   FadeGUI();
   raine_do_dialog(joy_key_editing,-1);
   return joy_selected_key;
}

int joy_key_edit_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      do_joy_edit();
      ret = D_REDRAW;
   }
   return ret;
}

/******************************************************************************/

int auto_edit_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      if(InputCount){
         InputList[auto_select[1].d1].auto_rate ++;
         if(InputList[auto_select[1].d1].auto_rate>5) InputList[auto_select[1].d1].auto_rate = 0;
      }
      ret = D_REDRAW;
   }
   return ret;
}

static char *AutoName(char rate)
{
   switch(rate){
   case 0:  return "Off";
   case 1:  return "1 (30 fps)";
   case 2:  return "2 (15 fps)";
   case 3:  return "3 (10 fps)";
   case 4:  return "4 (07 fps)";
   case 5:  return "5 (06 fps)";
   default: return "Off";
   }
}

char *autolist_getter(int index, int *list_size)
{
   if(InputCount){

   switch(index){
   case -1:			// Return List Size
      *list_size=InputCount;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      if(InputCount){
         InputList[auto_select[1].d1].auto_rate ++;
         if(InputList[auto_select[1].d1].auto_rate>5) InputList[auto_select[1].d1].auto_rate = 0;
      }
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < InputCount)){
         sprintf(S,"%-21s %s",InputList[index].InputName, AutoName(InputList[index].auto_rate));
         return S;
      }
      else
         return NULL;
   break;
   }

   }
   else{

   if(index==-1){
      *list_size=1;
      return NULL;
   }
   else{
      if(index==0){
         return(raine_translate_text("<No Inputs Mapped>"));
      }
      else{
         return NULL;
      }
   }

   }
}

int auto_select_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      ret=0;
      while((ret!=2)&&(ret!=-1)){
         ret=raine_do_dialog(auto_select,-1);
         if((ret!=2)&&(ret!=-1)){
            if(InputCount){
               InputList[auto_select[1].d1].auto_rate ++;
               if(InputList[auto_select[1].d1].auto_rate>5) InputList[auto_select[1].d1].auto_rate = 0;
            }
         }
      }

      ret = D_REDRAW;
   }
   return ret;
}

/******************************************************************************/

typedef struct GUI_JOY
{
   char *name;		// Mode name/string
   ULONG id;		// ID_ Value for this mode
} GUI_JOY;

static GUI_JOY joystick_data[] =
{
   { "Autodetect",		JOY_TYPE_AUTODETECT,	},
   { "None",			JOY_TYPE_NONE,		},
#ifdef JOYSTICK_DRIVER_STANDARD
   { "Joystick (2 buttons)",	JOY_TYPE_STANDARD,	}, 
   { "Joystick (4 buttons)",	JOY_TYPE_4BUTTON, 	},
   { "Joystick (6 buttons)",	JOY_TYPE_6BUTTON, 	},
   { "Joystick (8 buttons)",	JOY_TYPE_8BUTTON, 	},
   { "Joystick (2 pads)",	JOY_TYPE_2PADS, 	},
   { "Flightstick Pro",		JOY_TYPE_FSPRO, 	},
   { "Wingman Extreme",		JOY_TYPE_WINGEX,	},
#endif
#ifdef JOYSTICK_DRIVER_WINGWARRIOR
   { "Wingman Warrior",		JOY_TYPE_WINGWARRIOR,	}, // [RTB:26.02.99]: Allegro 3.11
#endif
#ifdef JOYSTICK_DRIVER_GAMEPAD_PRO
   { "Gamepad Pro",		JOY_TYPE_GAMEPAD_PRO,	},
#endif
#ifdef JOYSTICK_DRIVER_GRIP
   { "Grip",			JOY_TYPE_GRIP,		}, // [RTB:19.11.99]: Allegro 3927
   { "Grip 4 way",		JOY_TYPE_GRIP4,		},
#endif
#ifdef JOYSTICK_DRIVER_SIDEWINDER
   { "Sidewinder",		JOY_TYPE_SIDEWINDER,	},
#endif
#ifdef JOYSTICK_DRIVER_IFSEGA_ISA
   { "SEGA pads (isa)",		JOY_TYPE_IFSEGA_ISA,	}, // [RTB:22.05.99]: Allegro driver by E.Watanabe
#endif
#ifdef JOYSTICK_DRIVER_IFSEGA_PCI
   { "SEGA pads (pci)",		JOY_TYPE_IFSEGA_PCI,	}, // [RTB:22.05.99]: Allegro driver by E.Watanabe
#endif
#ifdef JOYSTICK_DRIVER_SNESPAD
   { "SNES pads (port 1)",	JOY_TYPE_SNESPAD_LPT1,	},
   { "SNES pads (port 2)",	JOY_TYPE_SNESPAD_LPT2,	},
   { "SNES pads (port 3)",	JOY_TYPE_SNESPAD_LPT3,	},
#endif
#ifdef JOYSTICK_DRIVER_PSXPAD
   { "PSX pads (port 1)",	JOY_TYPE_PSXPAD_LPT1,	}, // [RTB:17.06.99]: Allegro driver by Richard Davies
   { "PSX pads (port 2)",	JOY_TYPE_PSXPAD_LPT2,	},
   { "PSX pads (port 3)",	JOY_TYPE_PSXPAD_LPT3,	},
#endif
#ifdef JOYSTICK_DRIVER_N64PAD
   { "N64 pads (port 1)",	JOY_TYPE_N64PAD_LPT1,	}, // [RTB:19.11.99]: Allegro 3927
   { "N64 pads (port 2)",	JOY_TYPE_N64PAD_LPT2,	},
   { "N64 pads (port 3)",	JOY_TYPE_N64PAD_LPT3,	},
#endif
#ifdef JOYSTICK_DRIVER_WIN32
   { "Win32 Joystick Driver",	JOY_TYPE_WIN32		}, // [RTB:30.06.99]: Allegro driver for Win32/DirectX port
#endif
};

/* gfx_card_getter:
 *  Listbox data getter routine for the graphics card list.
 */

char *joylist_getter(int index, int *list_size)
{
   switch(index){
   case -1:			// Return List Size
      if(list_size){
	 *list_size = (sizeof(joystick_data) / sizeof(GUI_JOY));
      }
      return NULL;
   break;
   case -2:
      FadeGUI();
      DoJoyEdit();
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < (sizeof(joystick_data) / sizeof(GUI_JOY))))
         return joystick_data[index].name;
      else
         return NULL;
   break;
   }
}

// joy_name():
// Get card name for card ID

char *joy_name(int index)
{
   int ta,tb;

   tb = (sizeof(joystick_data) / sizeof(GUI_JOY));

   for(ta=0;ta<tb;ta++){
      if(joystick_data[ta].id == index)
         return joystick_data[ta].name;
   }

   return joystick_data[1].name;	// None
}

// gfx_card_number():
// Get card number for card ID

int joystick_number(int index)
{
   int ta,tb;

   tb = (sizeof(joystick_data) / sizeof(GUI_JOY));

   for(ta=0;ta<tb;ta++){
      if(joystick_data[ta].id == index)
         return ta;
   }

   return 1;				// None
}

// joystick_id():
// Get joy ID for joy number

int joystick_id(int index)
{
   int tb;

   tb = (sizeof(joystick_data) / sizeof(GUI_JOY));

   if((index>=0)&&(index<tb)){
      return joystick_data[index].id;
   }

   return joystick_data[1].id;		// None
}

void DoJoyEdit(void)
{
   int ta,tb;
   char *msg;

   ta = joystick_id(joy_select[1].d1);

   remove_joystick();

   JoystickType = JOY_TYPE_NONE;

   switch(ta){
   case JOY_TYPE_NONE:
      JoystickType=JOY_TYPE_NONE;
      return;
   break;
   default:		// Allegro

   show_mouse(NULL);
   if(install_joystick(ta)){
      show_mouse(screen);
      raine_alert(raine_translate_text("Control Error"),raine_translate_text("Unable to initialise Joystick"),allegro_error,NULL,raine_translate_text("&Ok"),NULL,'O',0);
      JoystickType=JOY_TYPE_NONE;
      return;
   }

   if(!(num_joysticks)){
      show_mouse(screen);
      raine_alert(raine_translate_text("Control Error"),raine_translate_text("No Joysticks of type"),joy_name(ta),raine_translate_text("could be detected"),raine_translate_text("&Ok"),NULL,'O',0);
      JoystickType=JOY_TYPE_NONE;
      return;
   }

   for(tb=0; tb<num_joysticks; tb++){

   while(joy[tb].flags & JOYFLAG_CALIBRATE){
      msg=calibrate_joystick_name(tb);
      show_mouse(screen);
      raine_alert(raine_translate_text("Control Setup"),raine_translate_text("Joystick Calibration:"),msg,raine_translate_text("and press space..."),raine_translate_text("Space"),NULL,' ',0);
      show_mouse(NULL);
      if(calibrate_joystick(tb)){
         show_mouse(screen);
         raine_alert(raine_translate_text("Control Error"),NULL,raine_translate_text("Joystick Calibration error"),NULL,raine_translate_text("&Ok"),NULL,'O',0);
         JoystickType=JOY_TYPE_NONE;
         return;
      }
   }

   }

   show_mouse(screen);
   JoystickType=ta;
   return;
   break;
   }
}

int joy_edit_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();
      DoJoyEdit();
      ret = D_REDRAW;
   }
   return ret;
}

int joy_calib_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();
      ret=0;
      while((ret!=2)&&(ret!=-1)){
         joy_select[1].d1 = joystick_number(JoystickType);
         ret=raine_do_dialog(joy_select,-1);
         if((ret!=2)&&(ret!=-1)){
            FadeGUI();
            DoJoyEdit();
         }
      }
      ret = D_REDRAW;
   }
   return ret;
}

void make_dsw_strings(void)
{
   static UBYTE str_dswa[32];
   static UBYTE str_dswb[32];

   int ta, tb;

   // Generate DSW bytes

   make_dipswitch_bytes();

   // Build Text

   sprintf(str_dswa, "DIPSW A: HHHHHHHH L:ON");
   sprintf(str_dswb, "DIPSW B: HHHHHHHH H:OFF");

   tb=1;
   for(ta=7; ta>=0; ta--){

      if(get_dsw(0) & tb)
		 str_dswa[ta+9] = (UBYTE) 'H';
	  else
		 str_dswa[ta+9] = (UBYTE) 'L';

      if(get_dsw(1) & tb)
		 str_dswb[ta+9] = (UBYTE) 'H';
	  else
		 str_dswb[ta+9] = (UBYTE) 'L';

      tb <<= 1;
   }

   // Insert

   dsw_select[6].dp = str_dswa;
   dsw_select[7].dp = str_dswb;

}

char *dswlist_getter(int index, int *list_size)
{
   int tcount;
   int num,numb,n;

   tcount=0;

   for(num=0;num<MAX_DIPSWITCHES;num++){
      tcount += dipswitch[num].statcount;
   }

   if(tcount>0){

   switch(index){
   case -1:			// Return List Size
      *list_size=tcount;
      return NULL;
   break;
   case -2:
      DoDSWEdit();
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index<tcount)){

      n = 0;
      numb = 0;
      for(num=0;num<MAX_DIPSWITCHES;num++){
         if(index >= n){
            numb=num;
         }
         n += dipswitch[num].statcount;
      }
      for(num=0;num<numb;num++){
         index -= dipswitch[num].statcount;
      }
         sprintf(
            S,
            "%-19s %s",
            dipswitch[numb].diplist[dipswitch[numb].statlist[index].pos].DSWName,
            dipswitch[numb].diplist[dipswitch[numb].statlist[index].pos+dipswitch[numb].statlist[index].offset+1].DSWName
         );
         return S;

      }
      else
         return NULL;
   break;
   }

   }
   else{

   if(index==-1){
      *list_size = 1;
      return NULL;
   }
   else{
      if(index==0)
         return raine_translate_text("<No Dipswitches Mapped>");
      else
         return NULL;
   }

   }
}

int dsw_select_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();
      ret=0;
      make_dsw_strings();
      while((ret!=2)&&(ret!=-1)){
         ret=raine_do_dialog(dsw_select,-1);
         if((ret!=2)&&(ret!=-1)){
            DoDSWEdit();
         }
      }
      make_dipswitch_bytes();
      ret = D_REDRAW;
   }
   return ret;
}

void DoDSWEdit(void)
{
   int ta,tb,tc;
   int num,numb,n;

   ta=dsw_select[1].d1;

      n = 0;
      numb = 0;
      for(num=0;num<MAX_DIPSWITCHES;num++){
         if(ta >= n){
            numb=num;
         }
         n += dipswitch[num].statcount;
      }
      for(num=0;num<numb;num++){
         ta -= dipswitch[num].statcount;
      }

      tb=dipswitch[numb].statlist[ta].pos;
      tc=dipswitch[numb].statlist[ta].offset+1;
      if(tc>=(dipswitch[numb].diplist[tb].values)){tc=0;}
      dipswitch[numb].statlist[ta].offset=tc;

   make_dsw_strings();

   clear_keybuf();
}

int dsw_edit_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      DoDSWEdit();
      ret = D_REDRAW;
   }
   return ret;
}

int dsw_reset_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      RestoreDSWDefault();
      make_dipswitch_statlist();
      make_dsw_strings();
      ret = D_REDRAW;
   }
   return ret;
}

int game_reset_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();
      reset_game_hardware();
      ret = D_REDRAW;
   }
   return ret;
}

static ULONG hex_start;
static ULONG hex_end;
static ULONG hex_lines;

char *hex_view_getter(int index, int *list_size)
{
   int ta;
   UBYTE data[8];
   UBYTE state[8+1];

   switch(index){
   case -1:			// Return List Size
      *list_size = hex_lines;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index<hex_lines)){
         sprintf(S,"%06X:",hex_start+(index<<3));
         state[0] = 0;
         for(ta=0;ta<8;ta++){
            data[ta] = gen_cpu_read_byte_2(hex_start+(index<<3)+ta);
            state[ta+1] = get_address_status(hex_start+(index<<3)+ta);
         }
         for(ta=0;ta<8;ta++){
            if((state[ta])||(state[ta+1]))
               sprintf(S+strlen(S),"*");
            else
               sprintf(S+strlen(S)," ");
            sprintf(S+strlen(S),"%02X",data[ta]);
         }
         if(state[8])
            sprintf(S+strlen(S),"*");
         else
            sprintf(S+strlen(S)," ");
         for(ta=0;ta<8;ta++){
            if((data[ta]>=0x20)&&(data[ta]<=0x7F))
               sprintf(S+strlen(S),"%c",data[ta]);
            else
               sprintf(S+strlen(S),".");
         }
         return S;
      }
      else
         return NULL;
   break;
   }
}

void do_hex_view(ULONG offset)
{
   int ret;

   hex_start = get_search_range_start(offset);
   hex_end   = get_search_range_end  (offset);
   hex_lines = (hex_end + 1) - (hex_start);
   hex_lines = hex_lines >> 3;

   hex_view_dialog[2].d1 = (offset - hex_start) >> 3;

   if(hex_view_dialog[2].d1 >= 9)
      hex_view_dialog[2].d2 = hex_view_dialog[2].d1 - 9;

   ret = raine_do_dialog(hex_view_dialog,-1);
}

static ULONG *s_list;

char *search_list_getter(int index, int *list_size)
{
   int ta;

   if(match_all_count>0){

   switch(index){
   case -1:			// Return List Size
      if(match_all_count>512){
         *list_size=512;
      }
      else{
         *list_size=match_all_count;
      }
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      ta = search_main_dialog[4].d1;
      if(ta<match_all_count){
         do_hex_view(s_list[ta]);
      }
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index<511)){
         sprintf(S,"%06x: %02x",s_list[index],gen_cpu_read_byte_2(s_list[index]));
         return(S);
      }
      if(index==511){
         return("......  ..");
      }
      return NULL;
   break;
   }
   }
   else{

   switch(index){
   case -1:			// Return List Size
      *list_size=1;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if(index == 0)
         return raine_translate_text("<No Search Results>");
      else
         return NULL;
   break;
   }

   }
}

static int search_data;
static int ar_window = 0;

char *search_type_2b_getter(int index, int *list_size)
{
   static char *str_list[2] =
   {
      "The bit is unchanged",
      "The bit has flipped",
   };

   switch(index){
   case -1:			// Return List Size
      *list_size = 2;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      *list_size = D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < 2))
         return raine_translate_text(str_list[index]);
      else
         return NULL;
   break;
   }
}

char *search_type_3b_getter(int index, int *list_size)
{
   static char *str_list[6] =
   {
      "New Byte == Last Byte",
      "New Byte != Last Byte",
      "New Byte <  Last Byte",
      "New Byte <= Last Byte",
      "New Byte >  Last Byte",
      "New Byte >= Last Byte",
   };

   switch(index){
   case -1:			// Return List Size
      *list_size = 6;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      *list_size = D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < 6))
         return raine_translate_text(str_list[index]);
      else
         return NULL;
   break;
   }
}

char *search_mode_getter(int index, int *list_size)
{
   switch(index){
   case -1:			// Return List Size
      *list_size = SEARCH_MODE_COUNT;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      *list_size = D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index < SEARCH_MODE_COUNT))
         return raine_translate_text(get_search_mode_name(index));
      else
         return NULL;
   break;
   }
}

int byte_hex_edit_proc(int msg, DIALOG *d, int c)
{
   int ta,ret;

   if(msg == MSG_START)
      byte_hex_dialog = d;

   ret = d_edit_proc(msg, d, c);

   if(ret!=D_O_K){

      // Grab byte value

      if(byte_hex_data[0]){
         sscanf(byte_hex_data,"%x",&ta);
         ta &= 0xFF;
      }
      else{
         ta = 0;
      }

      // Fill in alternate byte strings

      sprintf(byte_dec_data,"%d",ta);
      sprintf(byte_sgn_data,"%d",(BYTE) ta);
      sprintf(byte_asc_data,"%c",ta);

      // Draw them

      scare_mouse();
      SEND_MESSAGE(byte_dec_dialog, MSG_DRAW, 0);
      SEND_MESSAGE(byte_sgn_dialog, MSG_DRAW, 0);
      SEND_MESSAGE(byte_asc_dialog, MSG_DRAW, 0);
      unscare_mouse();
      dialog_oxygen();

   }
   if(ret==D_CLOSE) ret = D_O_K;

   return ret;
}

int byte_dec_edit_proc(int msg, DIALOG *d, int c)
{
   int ta,ret;

   if(msg == MSG_START)
      byte_dec_dialog = d;

   ret = d_edit_proc(msg, d, c);

   if(ret!=D_O_K){

      // Grab byte value

      if(byte_dec_data[0]){
         sscanf(byte_dec_data,"%d",&ta);
         ta &= 0xFF;
      }
      else{
         ta = 0;
      }

      // Fill in alternate byte strings

      sprintf(byte_hex_data,"%02x", (UBYTE) ta);
      sprintf(byte_sgn_data,"%d",   (BYTE)  ta);
      sprintf(byte_asc_data,"%c",   (UBYTE) ta);

      // Draw them

      scare_mouse();
      SEND_MESSAGE(byte_hex_dialog, MSG_DRAW, 0);
      SEND_MESSAGE(byte_sgn_dialog, MSG_DRAW, 0);
      SEND_MESSAGE(byte_asc_dialog, MSG_DRAW, 0);
      unscare_mouse();
      dialog_oxygen();

   }
   if(ret==D_CLOSE) ret = D_O_K;

   return ret;
}

int byte_sgn_edit_proc(int msg, DIALOG *d, int c)
{
   int ta,ret;

   if(msg == MSG_START)
      byte_sgn_dialog = d;

   ret = d_edit_proc(msg, d, c);

   if(ret!=D_O_K){

      // Grab byte value

      if(byte_sgn_data[0]){
         sscanf(byte_sgn_data,"%d",&ta);
         ta &= 0xFF;
      }
      else{
         ta = 0;
      }

      // Fill in alternate byte strings

      sprintf(byte_hex_data,"%02x", (UBYTE) ta);
      sprintf(byte_dec_data,"%d",   (UBYTE) ta);
      sprintf(byte_asc_data,"%c",   (UBYTE) ta);

      // Draw them

      scare_mouse();
      SEND_MESSAGE(byte_hex_dialog, MSG_DRAW, 0);
      SEND_MESSAGE(byte_dec_dialog, MSG_DRAW, 0);
      SEND_MESSAGE(byte_asc_dialog, MSG_DRAW, 0);
      unscare_mouse();
      dialog_oxygen();

   }
   if(ret==D_CLOSE) ret = D_O_K;

   return ret;
}

int byte_asc_edit_proc(int msg, DIALOG *d, int c)
{
   int ta,ret;

   if(msg == MSG_START)
      byte_asc_dialog = d;

   ret = d_edit_proc(msg, d, c);

   if(ret!=D_O_K){

      // Grab byte value

      ta = byte_asc_data[0];

      // Fill in alternate byte strings

      sprintf(byte_hex_data,"%02x",(UBYTE) ta);
      sprintf(byte_dec_data,"%d",  (UBYTE) ta);
      sprintf(byte_sgn_data,"%d",  (BYTE)  ta);

      // Draw them

      scare_mouse();
      SEND_MESSAGE(byte_hex_dialog, MSG_DRAW, 0);
      SEND_MESSAGE(byte_dec_dialog, MSG_DRAW, 0);
      SEND_MESSAGE(byte_sgn_dialog, MSG_DRAW, 0);
      unscare_mouse();
      dialog_oxygen();

   }
   if(ret==D_CLOSE) ret = D_O_K;

   return ret;
}

int cheat_search_new_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      ret = raine_do_dialog(search_mode_dialog,-1);

      search_mode = search_mode_dialog[2].d1;

      switch(search_mode){
      case SEARCH_MODE_ABSOLUTE:
      ret = raine_do_dialog(search_type_0a_dialog,-1);

      sscanf(byte_hex_data,"%x",&search_data);
      search_data &= 0xFF;

      start_arpro_search( (UBYTE) search_data);
      break;
      case SEARCH_MODE_RELATIVE:
      ret = raine_do_dialog(search_type_1a_dialog,-1);

      start_arpro_search(0x00);
      break;
      case SEARCH_MODE_SINGLE_BIT:
      ret = raine_do_dialog(search_type_2a_dialog,-1);

      start_arpro_search(0x00);
      break;
      case SEARCH_MODE_SLOW:
      ret = raine_do_dialog(search_type_3a_dialog,-1);

      start_arpro_search(0x00);
      break;
      default:
      break;
      }

      if(match_all_count>0){
         if(match_all_count>512){
            s_list = get_arpro_results(512);
         }
         else{
            s_list = get_arpro_results(match_all_count);
         }
      }
      SEND_MESSAGE(&search_main_dialog[4], MSG_START, 0);

      return D_REDRAW;
   }
   return ret;
}

int cheat_search_cont_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      switch(search_mode){
      case SEARCH_MODE_ABSOLUTE:
      ret = raine_do_dialog(search_type_0b_dialog,-1);

      sscanf(byte_hex_data,"%x",&search_data);
      search_data &= 0xFF;

      continue_arpro_search( (UBYTE) search_data);
      break;
      case SEARCH_MODE_RELATIVE:
      ret = raine_do_dialog(search_type_1b_dialog,-1);

      sscanf(byte_hex_data,"%x",&search_data);
      search_data &= 0xFF;

      continue_arpro_search( (UBYTE) search_data);
      break;
      case SEARCH_MODE_SINGLE_BIT:
      ret = raine_do_dialog(search_type_2b_dialog,-1);

      search_data = search_type_2b_dialog[2].d1;
      if(search_data>1) search_data=0;
      if(search_data<0) search_data=0;

      continue_arpro_search( (UBYTE) search_data);
      break;
      case SEARCH_MODE_SLOW:
      ret = raine_do_dialog(search_type_3b_dialog,-1);

      search_data = search_type_3b_dialog[2].d1;
      if(search_data>5) search_data=0;
      if(search_data<0) search_data=0;

      continue_arpro_search( (UBYTE) search_data);
      break;
      default:
      break;
      }

      if(match_all_count>0){
         if(match_all_count>512){
            s_list = get_arpro_results(512);
         }
         else{
            s_list = get_arpro_results(match_all_count);
         }
      }
      SEND_MESSAGE(&search_main_dialog[4], MSG_START, 0);

      return D_REDRAW;
   }
   return ret;
}

int cheat_main_proc(void)
{
   int ret;

   ret=0;

   while(ret>=0){

      ret = raine_do_dialog(actionreplay_dialog,-1);

      switch(ret){
      case 0x02:	// [Done]
         return -1;
      break;
      case 0x01:	// List Click
         DoCheatEdit();
      break;
      case 0x06:	// [Search>>]
         ar_window = 1;
         ret = -1;
      break;
      }

   }
   return 0;
}

int cheat_search_proc(void)
{
   int ret;

   ret=0;

   while(ret>=0){

      sprintf(cheateditdata,"%02x",search_data);

      if(match_all_count>0){
         if(match_all_count>512){
            s_list = get_arpro_results(512);
         }
         else{
            s_list = get_arpro_results(match_all_count);
         }
      }

      ret = raine_do_dialog(search_main_dialog,-1);

      sscanf(cheateditdata,"%x",&search_data);
      search_data&=0xFF;

      switch(ret){
      case 0x01:	// [Done]
         return -1;
      break;
      case 0x06:	// [Edit>>]
         ar_window = 0;
         ret = -1;
      break;
      }

   }
   return 0;
}

int do_cheat_data_edit(int ta)
{
   int ret;

      if(ta<CheatCount){
         sprintf(cheateditname,"%s",CheatList[ta].name);
         sprintf(cheateditaddress,"%08x",CheatList[ta].address);
         sprintf(cheateditdata,"%02x",CheatList[ta].data);
         sprintf(cheateditinfo,"%s",CheatList[ta].info);

         ret = raine_do_dialog(cheatdata_edit_dialog,-1);

         if(ret==1){

            sprintf(CheatList[ta].name,"%s",cheateditname);
            sscanf(cheateditaddress,"%x",&CheatList[ta].address);
            sscanf(cheateditdata,"%x",&CheatList[ta].data);
            sprintf(CheatList[ta].info,"%s",cheateditinfo);

            return 1;
         }
         else{
            return 0;
         }
      }
      else{
         return 0;
      }
}

int cheatdata_edit_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      ret = do_cheat_data_edit(actionreplay_dialog[1].d1);

      return D_REDRAW;
   }
   return ret;
}

char *cheatlist_getter(int index, int *list_size)
{
   if(CheatCount){

   switch(index){
   case -1:			// Return List Size
      *list_size=CheatCount;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      DoCheatEdit();
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index<CheatCount)){

         sprintf(S,"%-29s", CheatList[index].name);
         S[29] = 0;

         if(CheatList[index].active)
            sprintf(S+strlen(S)," %s", raine_translate_text("On"));
         else
            sprintf(S+strlen(S)," %s", raine_translate_text("Off"));

         return S;
      }
      else
         return NULL;
   break;
   }
   }
   else{

   switch(index){
   case -1:			// Return List Size
      *list_size=1;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if(index == 0)
         return raine_translate_text("<No Cheats Mapped>");
      else
         return NULL;
   break;
   }

   }
}

void DoCheatEdit(void)
{
   int ta;

   ta=actionreplay_dialog[1].d1;

   if(ta<CheatCount){
      CheatList[ta].active^=1;
   }
}

int cheat_edit_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      DoCheatEdit();
      return D_REDRAW;
   }
   return ret;
}

void do_cheat_add(ULONG addr, UBYTE data)
{
   int ta, ret;

      if(CheatCount<CHEAT_MAX){
         ta = CheatCount;
         CheatCount++;

         sprintf(CheatList[ta].name,"New Cheat %d",ta);
         CheatList[ta].address = addr;
         CheatList[ta].data    = data;
         CheatList[ta].mask    = 0xFF;
         CheatList[ta].type    = 0x00;
         CheatList[ta].info[0] = 0;
         CheatList[ta].active  = 0;

         ret = do_cheat_data_edit(ta);

         if( ret != 1) CheatCount--;
      }
}

int cheat_add_proc(int msg, DIALOG *d, int c)
{
   int ta;
   int ret=x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      ta = search_main_dialog[4].d1;

      if(ta<match_all_count){

      do_cheat_add(s_list[ta], gen_cpu_read_byte_2(s_list[ta]));

      }

      return D_REDRAW;
   }
   return ret;
}

int cheat_new_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      do_cheat_add(0,0);

      return D_REDRAW;
   }
   return ret;
}

void DoCheatDelete(void)
{
   int ta,tb;

   ta=actionreplay_dialog[1].d1;

   if(ta<CheatCount){
      if((raine_alert(raine_translate_text("Warning"),raine_translate_text("Really delete this cheat"),raine_translate_text("permanently?"),NULL,raine_translate_text("&Yes"),raine_translate_text("&No"),'Y','N'))==1){
         for(tb=ta; tb<(CheatCount-1); tb++){
            memcpy(&CheatList[tb], &CheatList[tb+1], sizeof(CHEAT));
         }
         CheatCount--;
      }
   }
}

int cheat_del_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      DoCheatDelete();
      return D_REDRAW;
   }
   return ret;
}

int actionreplay_proc(int msg, DIALOG *d, int c)
{
   int ta;

   int ret=x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      ta=0;
      while(ta==0){

      switch(ar_window){
         case 0x00:
            ta = cheat_main_proc();
         break;
         case 0x01:
            ta = cheat_search_proc();
         break;
      }

      }
      return D_REDRAW;
   }
   return ret;
}


void DoLangEdit(void)
{
   int ta;
   ta=language_dialog[1].d1;
   if(ta<LanguageSw.Count){
      SetLanguageSwitch(ta);
      sprintf(langname,"%s",LanguageSw.Mode[ta]);
   }
}

// language_getter():
// Get mode names for langauge/version select list

char *language_getter(int index, int *list_size)
{
   switch(index){
   case -1:
      *list_size=LanguageSw.Count;
      return NULL;
   break;
   case -2:			// Act upon List Object Selection
      DoLangEdit();
      *list_size=D_REDRAW;
      return NULL;
   break;
   case -3:			// Act Keyboard Input
      return NULL;
   break;
   default:
      if((index >= 0)&&(index<LanguageSw.Count))
         return LanguageSw.Mode[index];
      else
         return NULL;
   break;
   }
}

int language_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();
      ret=GetLanguageSwitch();
      if(LanguageSw.Count){
         sprintf(langname,"%s",LanguageSw.Mode[ret]);
         language_dialog[1].d1=ret;
      }
      else{
         sprintf(langname,raine_translate_text("<No language switches>"));
      }
      ret=0;
      while((ret!=2)&&(ret!=-1)){
         ret = raine_do_dialog(language_dialog,-1);
         if((ret!=2)&&(ret!=-1)){
            DoLangEdit();
         }
      }
      return D_REDRAW;
   }
   return ret;
}

int language_sel_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){
      DoLangEdit();
      return D_REDRAW;
   }
   return ret;
}

// game_setup_proc():
// Does game setup options dialog

int game_setup_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();
      ret=0;
      while((ret!=2)&&(ret!=-1)){
         ret = raine_do_dialog(game_setup_dialog,-1);
      }
      return D_REDRAW;
   }
   return ret;
}

// control_setup_proc():
// Does game setup options dialog

int control_setup_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();
      ret=0;
      while((ret!=2)&&(ret!=-1)){
         ret=raine_do_dialog(control_setup_dialog,-1);
      }
      return D_REDRAW;
   }
   return ret;
}

// raine_quit_proc():
// Process a click on 'quit raine' button

int raine_quit_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();
      if((raine_alert(raine_translate_text("Warning"),NULL,raine_translate_text("Quit Raine?"),NULL,raine_translate_text("&Yes"),raine_translate_text("&No"),'Y','N'))!=1){
         return D_REDRAW;
      }
      else{
         WantQuit=1;
         return D_EXIT;
      }
   }
   return ret;
}

// game_play_proc():
// Process a click on 'play game' button

int game_play_proc(int msg, DIALOG *d, int c)
{
   int ret = x_raine_button_proc(msg, d, c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();
      if(!current_game){
         raine_alert(raine_translate_text("Error"),raine_translate_text("No Game Loaded"),raine_translate_text("Click on Change Game"),raine_translate_text("to Select a Game."),raine_translate_text("&Ok"),NULL,'O',0);
         ret = D_REDRAW;
      }
      else{
         WantPlay = 1;
         ret = D_EXIT;
      }
   }
   return ret;
}


char *soundcardlist_getter(int index, int *list_size)
{
   if(index==-1){
      *list_size = RaineSoundCardTotal;
      return NULL;
   }
   else{
      if((index >= 0)&&(index<RaineSoundCardTotal))
         return sound_card_name(index);
      else
         return NULL;
   }
}

void get_sound_variables(void)
{
      // Initialize Sample Rate switch
      // -----------------------------

      sound_setup_dialog[SND_RATE_LIST].flags=0;
      sound_setup_dialog[SND_RATE_LIST+1].flags=0;
      sound_setup_dialog[SND_RATE_LIST+2].flags=0;

      switch(SampleRate){
          case 11025:
             sound_setup_dialog[SND_RATE_LIST+0].flags=D_SELECTED;
          break;
          case 44100:
             sound_setup_dialog[SND_RATE_LIST+2].flags=D_SELECTED;
          break;
          case 22050:
          default:
             SampleRate=22050;
             sound_setup_dialog[SND_RATE_LIST+1].flags=D_SELECTED;
          break;
      }

      // Initialize YM3812 switch
      // ------------------------

      if(use_emulated_ym3812!=0) use_emulated_ym3812=1;

      sound_setup_dialog[SND_3812_LIST].flags=0;
      sound_setup_dialog[SND_3812_LIST+1].flags=0;

      sound_setup_dialog[SND_3812_LIST+use_emulated_ym3812].flags=D_SELECTED;

      // Initialize Soundcard switch
      // ---------------------------

      sound_setup_dialog[1].d1=RaineSoundCard;
}

void set_sound_variables(UBYTE update)
{
   if(update){
   RaineSoundCard=sound_setup_dialog[1].d1;

   if((sound_setup_dialog[SND_RATE_LIST].flags)&D_SELECTED)   SampleRate=11025;
   if((sound_setup_dialog[SND_RATE_LIST+1].flags)&D_SELECTED) SampleRate=22050;
   if((sound_setup_dialog[SND_RATE_LIST+2].flags)&D_SELECTED) SampleRate=44100;

   if((sound_setup_dialog[SND_3812_LIST].flags)&D_SELECTED)   use_emulated_ym3812=0;
   if((sound_setup_dialog[SND_3812_LIST+1].flags)&D_SELECTED) use_emulated_ym3812=1;

   change_sample_rate = 1;
   }

   sprintf(soundname,raine_translate_text(sound_card_name(RaineSoundCard)));
   sprintf(soundname,"%s; %02dkHz",soundname,SampleRate/1000);
}

// sound_setup_proc():
// Does sound setup options dialog

int sound_setup_proc(int msg, DIALOG *d, int c)
{
   int ret=x_raine_button_proc(msg,d,c);

   if(ret==D_CLOSE){
      Unselect_Button(d);
      FadeGUI();

      get_sound_variables();

      ret=raine_do_dialog(sound_setup_dialog,-1);

      if((ret!=3)&&(ret!=-1)){
         set_sound_variables(1);
      }

      return D_REDRAW;
   }
   return ret;
}

// real primitive pen allocation

static int used_pens;

int allocate_pens(int pens)
{
   int ret;

   if((used_pens+pens)<=256){
      ret = used_pens;
      used_pens += pens;
   }
   else{
      ret = -1;
   }
   return ret;
}

void free_pens(int pens)
{
   if((used_pens-pens)>=0){
      used_pens -= pens;
   }
}

void free_all_pens(void)
{
   used_pens = 0;
}

// MakeGUIBack():
// Make a faded copy of the game screen to go behind the gui.

void MakeGUIBack(void)
{
   static BITMAP *backdrop = NULL;
   static PALETTE backdrop_pal;
   PALETTE new_pal;
   UBYTE *BIT;
   UBYTE mapcol[256];
   int ccnt,ta,tb,tc,td,te,tcr,tcg,tcb;

   if(screen_valid){

   gui_backdrop = create_bitmap_ex(display_cfg.bpp, display_cfg.screen_x, display_cfg.screen_y);

   free_all_pens();

   memset(new_pal,0x00, 256*4);
   memset(mapcol, 0xFF, 256);

   ccnt=1;	// Black=Colour0

      blit(screen,gui_backdrop,0,0,0,0,display_cfg.screen_x, display_cfg.screen_y);

      if(display_cfg.bpp == 8){

      for(tb=0;tb<display_cfg.screen_y;tb++){
      BIT=gui_backdrop->line[tb];
      for(ta=0;ta<display_cfg.screen_x;ta++){
      tc=BIT[ta];
      if((te=mapcol[tc])==255){
         tcr=pal[tc].r;
         tcg=pal[tc].g;
         tcb=pal[tc].b;
         tcr = (tcr * 64) / 96;
         if(tcr<0) tcr=0;
         if(tcr>63) tcr=63;
         tcg = (tcg * 64) / 80;
         if(tcg<0) tcg=0;
         if(tcg>63) tcg=63;
         tcb = (tcb * 64) / 64;
         if(tcb<0) tcb=0;
         if(tcb>63) tcb=63;
         te=255;
         for(td=0;td<ccnt;td++){
            if((tcr==new_pal[td].r)&&(tcg==new_pal[td].g)&&(tcb==new_pal[td].b)){
               te=td;
               td=ccnt;
               mapcol[tc]=te;
            }
         }
         if(te==255){
            if(ccnt<216){
               new_pal[ccnt].r=tcr;
               new_pal[ccnt].g=tcg;
               new_pal[ccnt].b=tcb;
               te=ccnt;
               ccnt++;
            }
            else{			// No room! Find a close match... (allegro: slow)
               te=bestfit_color(new_pal,tcr,tcg,tcb);
            }
         }
      }
      BIT[ta]=te;
      }

      }

      }
      memcpy(&gui_pal, &new_pal, ccnt*4);
      allocate_pens(ccnt);

   }
   else{

   gui_backdrop = create_bitmap_ex(8, display_cfg.screen_x, display_cfg.screen_y);

   free_all_pens();

   memset(new_pal,0x00, 256*4);
   memset(mapcol, 0xFF, 256);

   ccnt=1;	// Black=Colour0

      if(!backdrop){

         if(rgui_cfg.bg_image[0])

            backdrop = load_pcx(rgui_cfg.bg_image, backdrop_pal);

         if(!backdrop){

            backdrop = RaineData[Backdrop].dat;
            memcpy(&backdrop_pal, RaineData[Backdrop_pal].dat, 256*4);
         }

      }

      clear_to_color(gui_backdrop,getpixel(backdrop,0,0));
      blit(backdrop,gui_backdrop,0,0,(display_cfg.screen_x - backdrop->w)/2,(display_cfg.screen_y - backdrop->h)/2,backdrop->w,backdrop->h);

      for(tb=0;tb<display_cfg.screen_y;tb++){
      BIT=gui_backdrop->line[tb];
      for(ta=0;ta<display_cfg.screen_x;ta++){
      tc=BIT[ta];
      if((te=mapcol[tc])==255){
         tcr=backdrop_pal[tc].r;
         tcg=backdrop_pal[tc].g;
         tcb=backdrop_pal[tc].b;
         te=255;
         for(td=0;td<ccnt;td++){
            if((tcr==new_pal[td].r)&&(tcg==new_pal[td].g)&&(tcb==new_pal[td].b)){
               te=td;
               td=ccnt;
               mapcol[tc]=te;
            }
         }
         if(te==255){
            if(ccnt<216){
               new_pal[ccnt].r=tcr;
               new_pal[ccnt].g=tcg;
               new_pal[ccnt].b=tcb;
               te=ccnt;
               ccnt++;
            }
            else{			// No room! Find a close match... (allegro: slow)
               te=bestfit_color(new_pal,tcr,tcg,tcb);
            }
         }
      }
      BIT[ta]=te;
      }

      }
      memcpy(&gui_pal, &new_pal, ccnt*4);
      allocate_pens(ccnt);
   }

   set_palette_range(gui_pal,0,255,1);
   blit(gui_backdrop,screen,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);

   the_dialog[0].proc= &d_bitmap_proc;
   the_dialog[0].dp  = gui_backdrop;
   the_dialog[0].x   = 0;
   the_dialog[0].y   = 0;
   the_dialog[0].w   = display_cfg.screen_x;
   the_dialog[0].h   = display_cfg.screen_y;

   if(raine_cfg.hide){
   
      clear(gui_backdrop);
      clear(screen);

   }
}

// DestroyGUIBack():
// Free resources used by the gui backdrop image.

void DestroyGUIBack(void)
{
   destroy_bitmap(gui_backdrop);

   the_dialog[0].proc = &d_clear_proc;
   the_dialog[0].dp   = NULL;
   the_dialog[0].x    = the_dialog[4].x;
   the_dialog[0].y    = the_dialog[4].y;
   the_dialog[0].w    = (the_dialog[1].x - the_dialog[4].x) + the_dialog[1].w;
   the_dialog[0].h    = (the_dialog[1].y - the_dialog[4].y) + the_dialog[1].h;

}

#include "shiftjis.c"

// StartGUI():
// Actually does the gui/game loop, called from raine.c
// Caters for command line load and no gui options

static DATAFILE *font_data;

int StartGUI(void)
{
   char str[256];

   #ifdef HACKS
   print_debug("StartGUI(): START\n");
   #endif

   gui_main_font     = RaineData[font_gui_main].dat;
   gui_fixed_font    = RaineData[font_gui_fixed].dat;
   ingame_font       = RaineData[font_8x8_ingame].dat;
   font              = gui_main_font;

   if(rgui_cfg.font_datafile[0]){
	   font_data = load_datafile(rgui_cfg.font_datafile);
	   if(font_data){
   		gui_main_font     = font_data[2].dat;
   		gui_fixed_font    = font_data[1].dat;
   		ingame_font       = font_data[0].dat;
   		font              = gui_main_font;
	   }
   }

   SpriteGun1        = RaineData[Gun1].dat;
   SpriteGun2        = RaineData[Gun2].dat;
   the_dialog[14].dp = RaineData[Antiriad].dat;

   sprintf(str, "%sconfig/language/%s", dir_cfg.exe_path, dir_cfg.language_file);
   raine_push_config_state();
   raine_set_config_file(str);

   if(new_set_gfx_mode(display_cfg.screen_type,display_cfg.screen_x,display_cfg.screen_y,display_cfg.scanlines)!=0){
      if(new_set_gfx_mode(GFX_AUTODETECT,320,240,0)){
         printf("%s\n",raine_translate_text("Unable to detect any screenmode."));
         printf("%s\n",raine_translate_text("Get Display Doctor from www.scitechsoft.com to fix this."));
         printf("%s\n\n",raine_translate_text("Mail about this problem is not appreciated!"));
         exit(1);
      }
      else{
         display_cfg.screen_type= GFX_AUTODETECT;
         display_cfg.screen_x   = 320;
         display_cfg.screen_y   = 240;
         display_cfg.scanlines  = 0;
      }
   }

   sprintf(gamename,    raine_translate_text("<No Game Loaded>"));
   sprintf(screenname,  raine_translate_text("%dx%d %dbpp; %s"), display_cfg.screen_x, display_cfg.screen_y, display_cfg.bpp, gfx_card_name(display_cfg.screen_type));

   set_sound_variables(0);

   memset(&gui_pal, 0x00, 256*4);

   set_gui_palette();

   WantScreen=0;
   WantQuit=0;
   WantPlay=0;

   register_uformat(U_SJIS, sjis_getc, sjis_getx, sjis_setc, sjis_width, sjis_cwidth, sjis_isok);
   set_uformat(U_SJIS);

   while(!WantQuit){		// ----- Main Loop ------

   install_timer();
   install_keyboard();
   install_mouse();

   MakeGUIBack();

   set_mouse_sprite(RaineData[Mouse].dat);
   show_mouse(screen);

   /*

   check for load game request (from the command line)

   */

   if(raine_cfg.req_load_game)

      do_load_game();

   if(!raine_cfg.no_gui){	// GUI MODE

   if(!WantScreen){

   while(raine_do_dialog(the_dialog,-1)==-1){
   }

   }

   }
   else{			// NO GUI MODE

   WantQuit=1;

   if(current_game)

      WantPlay = 1;

   else

      WantPlay = 0;

   }

   if(WantScreen){		// Are we here for a screenchange?
      WantScreen=0;
      ScreenChange();
   }

   remove_mouse();
   clear_keybuf();
   remove_keyboard();
   remove_timer();

   DestroyGUIBack();

   if(display_cfg.scanlines){
      if(display_cfg.triple_buffer){
         BITMAP *buf;
         buf = create_bitmap(display_cfg.screen_x,display_cfg.screen_y*3);
         clear(buf);
         blit(buf,screen,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y*3);
         destroy_bitmap(buf);
      }
      else{
         BITMAP *buf;
         buf = create_bitmap(display_cfg.screen_x,display_cfg.screen_y);
         clear(buf);
         blit(buf,screen,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);
         destroy_bitmap(buf);
      }
   }
   else{
      clear(screen);
   }

   text_mode(-1);

   if(WantPlay){		// Are we able to and wanting to play?
      WantPlay = 0;
      if(run_game_emulation()){	// In no gui mode, tab will reactivate the gui (req. by bubbles)
         raine_cfg.no_gui = 0;
         WantQuit = 0;
      }
      screen_valid=1;
   }

   }				// ----- Main Loop -----

   /*

   close down the current game (if there is one)

   */

   if(current_game){

      save_game_config();

      current_game->clear_game();

      ClearDefault();

      current_game = NULL;
   }

   raine_pop_config_state();

   #ifdef HACKS
   print_debug("StartGUI(): OK\n");
   #endif

   return 0;
}
