/******************************************************************************/
/*                                                                            */
/*        RAINE (680x0 arcade emulation) v0.28 (C) 1998/99 Raine Team         */
/*                                                                            */
/******************************************************************************/

#ifndef __rainedefs__
#define __rainedefs__

#include <stdio.h>

#ifndef RAINE_DOS
#include <string.h>
#endif

#define alleg_flic_unused
#define alleg_sound_unused
#define alleg_vidmem_unused

#include <allegro.h>

#ifdef _WINDOWS
#include <winalleg.h>
#endif

#ifdef GFX_DRIVER_MODEX
#define GFX_ARCMON               AL_ID('A','R','C','M')
#endif

#ifdef MEMORY_DEBUG
#include "newmem.h"		// This must be placed before other includes
#endif

#include "deftypes.h"		// Raine data types
#include "68020.h"		// UAE 68020 engine (ripped)
#include "dejap.h"		// Autogenerated defines for contents of datafile
#include "confile.h"

// Compilation Options

//#define HACKS        // Debugging Mode (write to debug.txt, saving ROM/RAM/GFX/SMP)
//#define HACKS_SOUND     // Extra Sound Debugging (Hiro-Shi sound code)

#ifdef HACKS_SOUND		// If sound debug on, HACKS is defined.
  #define HACKS
#endif

#define FASTF3MEM		// Use Faster (no bad address checking) memory access in F3-System
#define F3SINGLE		// Directly use F3-System memory access in 68020 engine

#define BGAMES			// Beta/Unreleased games included

#define RDTSC_PROFILE		// Rdtsc profiling included

#ifdef RAINE_DOS
#define TRIPLE_BUFFER		// Triple buffer support included
#endif

//#define HAVE_6502

// Useful String Constants

#define VERSION "20000819" // Version string
#define HANDLE  "Raine-Team"	// Credit string
#ifdef RAINE_DOS
#define EMUNAME "Raine"		// Dos Name
#else
#define EMUNAME "RaineX"	// Win32/DirectX Name
#endif

UBYTE *R24[0x100];		// Memory Addresses for UAE020 Read
UBYTE *W24[0x100];		// Memory Addresses for UAE020 Read
UBYTE *ROM;			// Pointer for ROM memory
UBYTE *RAM;			// Pointer for RAM memory
UBYTE *GFX;			// Pointer for Graphics memory
UBYTE *SMP;			// Pointer for Sample memory
UBYTE *EEPROM;			// Pointer for EEPROM memory
UBYTE *Z80ROM;			// Pointer for Z80ROM memory

#ifdef HAVE_6502
UBYTE *M6502ROM;		// Pointer for Z80ROM memory
#endif

// Intel Read macros
// -----------------

#define ReadByte(a)	(*(UBYTE *)(a))
#define ReadWord(a)	(*(UWORD *)(a))
#define ReadLong(a)	(*(ULONG *)(a))

// Intel Write macros
// ------------------

#define WriteByte(a,d)	(*(UBYTE *)(a) = (d))
#define WriteWord(a,d)	(*(UWORD *)(a) = (d))
#define WriteLong(a,d)	(*(ULONG *)(a) = (d))

// 680x0 Read macros
// -----------------

static DEF_INLINE ULONG GetWord(UWORD *a)
{
   ULONG retval;

#ifndef RAINE_DOS
   retval =	(((UBYTE *) a)[0] << 8) | (((UBYTE *) a)[1] << 0);
#else
   asm("xorl %k0,%k0\n\tmovw %w1,%w0\n\trolw $8,%w0" : "=&r" (retval) : "m" (*a) : "cc");
#endif
	return retval;
}
static DEF_INLINE ULONG GetLong(ULONG *a)
{
   ULONG retval;

#ifndef RAINE_DOS
   retval =	(((UBYTE *) a)[0] << 24) | (((UBYTE *) a)[1] << 16) | (((UBYTE *) a)[2] << 8) | (((UBYTE *) a)[3] << 0);
#else
   asm("bswapl %0" : "=r" (retval) : "0" (*a) : "cc");
#endif
   return retval;
}

#define ReadByte68k(a) ReadByte(a)
#define ReadWord68k(a) GetWord((UWORD *)a)
#define ReadLong68k(a) GetLong((ULONG *)a)

// 680x0 Write macros
// ------------------

static DEF_INLINE void PutWord(UWORD *a, ULONG v)
{
#ifndef RAINE_DOS
   ((UBYTE *) a)[0] = (v >> 8) & 0xFF;
   ((UBYTE *) a)[1] = (v >> 0) & 0xFF;
#else
   asm("rolw $8,%0" : "=r" (v) : "0" (v) : "cc");
   *a = v;
#endif
}
static DEF_INLINE void PutLong(ULONG *a, ULONG v)
{
#ifndef RAINE_DOS
   ((UBYTE *) a)[0] = (v >>24) & 0xFF;
   ((UBYTE *) a)[1] = (v >>16) & 0xFF;
   ((UBYTE *) a)[2] = (v >> 8) & 0xFF;
   ((UBYTE *) a)[3] = (v >> 0) & 0xFF;
#else
   asm("bswapl %0" : "=r" (v) : "0" (v) : "cc");
   *a = v;
#endif
}

#define WriteByte68k(a,d) WriteByte(a,d)
#define WriteWord68k(a,d) PutWord((UWORD *)a,d)
#define WriteLong68k(a,d) PutLong((ULONG *)a,d)

// UAE 68020 Read Pointers
// -----------------------

UBYTE (*ByteRead)(ULONG address);
UWORD (*WordRead)(ULONG address);
ULONG (*LongRead)(ULONG address);

// UAE 68020 Write Pointers
// ------------------------

void (*ByteWrite)(ULONG address, UBYTE data);
void (*WordWrite)(ULONG address, UWORD data);
void (*LongWrite)(ULONG address, ULONG data);

void (*F3SystemEEPROMAccess)(UBYTE data);

#ifdef F3SINGLE

static DEF_INLINE UBYTE ReadF3SystemByte(ULONG i)
{
   return(R24[(i>>16)&0xFF][i&0xFFFF]);
}

static DEF_INLINE UWORD ReadF3SystemWord(ULONG i)
{
   return(ReadWord68k(&R24[(i>>16)&0xFF][i&0xFFFF]));
}

static DEF_INLINE ULONG ReadF3SystemLong(ULONG i)
{
   return(ReadLong68k(&R24[(i>>16)&0xFF][i&0xFFFF]));
}

static DEF_INLINE void WriteF3SystemByte(ULONG i, UBYTE data)
{
   W24[(i>>16)&0xFF][i&0xFFFF]=data;
}

static DEF_INLINE void WriteF3SystemWord(ULONG i, UWORD data)
{
   WriteWord68k(&W24[(i>>16)&0xFF][i&0xFFFF],data);
}

static DEF_INLINE void WriteF3SystemLong(ULONG i, ULONG data)
{
   WriteLong68k(&W24[(i>>16)&0xFF][i&0xFFFF],data);
}

// Define UAE 680x0 Read routines
// ------------------------------

#define cpu_readmem24(address)			ReadF3SystemByte(address)
#define cpu_readmem24_word(address)		ReadF3SystemWord(address)
#define cpu_readmem24_dword(address)		ReadF3SystemLong(address)

// Define UAE 680x0 Write routines
// -------------------------------

#define cpu_writemem24(address,data)		WriteF3SystemByte(address,data)
#define cpu_writemem24_word(address,data)	WriteF3SystemWord(address,data)
#define cpu_writemem24_dword(address,data)	WriteF3SystemLong(address,data)

#else

// Define UAE 680x0 Read routines
// ------------------------------

#define cpu_readmem24(address)			ByteRead(address)
#define cpu_readmem24_word(address)		WordRead(address)
#define cpu_readmem24_dword(address)		LongRead(address)

// Define UAE 680x0 Write routines
// -------------------------------

#define cpu_writemem24(address,data)		ByteWrite(address,data)
#define cpu_writemem24_word(address,data)	WordWrite(address,data)
#define cpu_writemem24_dword(address,data)	LongWrite(address,data)

#endif

#define	CPU_FRAME_MHz(speed,slice) ((speed*1000000)/slice)

void (*ExecuteSoundFrame)();	// Pointer to ExecuteSoundFrame rountine (sound cpu work for 1 frame), used for pausegame + playsound

// Load Error Report

#define LOAD_FATAL_ERROR (0x0001)
#define LOAD_WARNING     (0x0002)
int load_error;
char *load_debug;

// Sound

int GameSound;			// 1 if game supports sound
int SoundEnabled;		// 1 if user wants sound emulation

// Misc Stuff

int GameMouse;			// 1 if game wants to use mouse driver

// Game Screen

PALETTE pal;			// 256 Colour palette for the game
BITMAP *GameBitmap;		// *Full* Bitmap for generating the game screen (eg. 0,0 to 320+32,240+32)
BITMAP *GameViewBitmap;		// *Viewable* Bitmap for generating the game screen (eg. 16,16 to 320+16,240+16)

int StarScreamEngine;		// Number of M68000's in use
int MZ80Engine;			// Number of Z80's in use

#ifdef HAVE_6502
int M6502Engine;		// Number of M6502's in use
#endif

int RAMSize;			// Size of RAM[] block

UBYTE *SpriteGun1;		// Raw data for player 1 mouse target
UBYTE *SpriteGun2;		// Raw data for player 2 mouse target
UBYTE *ingame_font;		// Raw data for ingame font

DATAFILE *RaineData;		// Resource datafile

int RefreshBuffers;

/*

general configuration

*/

typedef struct RAINE_CFG
{
   ULONG show_fps_mode;			// type of fps output
   ULONG no_gui;			// gui is disabled
   ULONG req_load_game;			// request to load game (from the command line)
   ULONG req_game_index;		// game number for req_load_game
   ULONG version_no;			// what version of raine is running
   ULONG run_count;			// how many times raine has run
   ULONG wibble;			// wibble the screen
   ULONG extra_games;			// extra games enable (currently unused)
   ULONG save_game_screen_settings;	// save display config for each game
   ULONG hide;
   ULONG req_pause_game;
   ULONG req_save_screen;
   ULONG req_pause_scroll;
} RAINE_CFG;

struct RAINE_CFG raine_cfg;

/*

display configuration

*/

typedef struct DISPLAY_CFG
{
   ULONG screen_type;			// pc screen display type {..MODX,VB2L,VBAF..}
    LONG screen_x;			// pc screen display width
    LONG screen_y;			// pc screen display height
   ULONG scanlines;			// scanlines mode {0,1,2}
   ULONG vsync;				// vsync mode {0,1}
   ULONG triple_buffer;			// triple buffer mode {0,1}
   ULONG eagle_fx;			// use eagle effect {0,1}
   ULONG fast_set_pal;			// fast or slow(compatable) set pal {0,1}
   ULONG pixel_double;			// double pixels effect {0,1,2}
   ULONG limit_speed;			// don't exceed 60 fps {0,1}
   ULONG frame_skip;			// skip frames number {0..9} (0=auto frame skip)
#ifdef RAINE_DOS
   ULONG arcade_h_timing;		// arcade monitor mode
   ULONG arcade_v_timing;
   ULONG arcade_center_x;
   ULONG arcade_center_y;
#endif
   ULONG rotate;			// real rotation setting {0,1,2,3}
   ULONG flip;				// real flipping setting {0,1,2,3}
   ULONG user_rotate;			// user preferred rotation {0,1,2,3}
   ULONG user_flip;			// user preferred flipping {0,1,2,3}
   ULONG no_rotate;			// ignore default rotation in game drivers
   ULONG no_flip;			// ignore default rotation in game drivers
   ULONG bpp;                           // colour depth {8,16}
} DISPLAY_CFG;

struct DISPLAY_CFG display_cfg, game_display_cfg, prev_display_cfg;

int use_custom_screen;

/*

file & directory configuration

*/

typedef struct DIR_CFG
{
   UBYTE exe_path[256];			// current path of raine.exe
   UBYTE language_file[256];		// language translation file (no path)
   UBYTE screen_dir[256];		// screenshot path
   UBYTE rom_dir[4][256];		// rom paths
   UBYTE cheat_file[256];		// cheat file (no path)
   UBYTE config_file[256];		// config file (no path)
   ULONG long_file_names;		// enable long file name support
   ULONG last_screenshot_num;		// counter for last screenshot
} DIR_CFG;

struct DIR_CFG dir_cfg;

#endif
