/******************************************************************************/
/*                                                                            */
/*                              R-GUI: RAINE GUI                              */
/*                                                                            */
/******************************************************************************/

#include "raine.h"
#include "rgui.h"
#include "rguiproc.h"

extern PALETTE gui_pal;

#ifdef TEST_WINGUI

/* hook functions for reading the mouse state */
int rgui_mouse_x(void) { return mouse_x - win_list[0].back_x; }
int rgui_mouse_y(void) { return mouse_y - win_list[0].back_y; }
int rgui_mouse_b(void) { return mouse_b; }

DIALOG_PLAYER *rgui_init_dialog(DIALOG *dialog, int focus_obj);
int rgui_update_dialog(DIALOG_PLAYER *player);
int rgui_shutdown_dialog(DIALOG_PLAYER *player);


/* Checking for double clicks is complicated. The user could release the
 * mouse button at almost any point, and I might miss it if I am doing some
 * other processing at the same time (eg. sending the single-click message).
 * To get around this I install a timer routine to do the checking for me,
 * so it will notice double clicks whenever they happen.
 */

static volatile int dclick_status, dclick_time;
static int dclick_install_count = 0;
//static int dclick_install_time = 0;

#define DCLICK_START      0
#define DCLICK_RELEASE    1
#define DCLICK_AGAIN      2
#define DCLICK_NOT        3

/* dclick_check:
 *  Double click checking user timer routine.
 */
static void dclick_check()
{
   if (dclick_status==DCLICK_START) {              /* first click... */
      if (!gui_mouse_b()) {
	 dclick_status = DCLICK_RELEASE;           /* aah! released first */
	 dclick_time = 0;
	 return;
      }
   }
   else if (dclick_status==DCLICK_RELEASE) {       /* wait for second click */
      if (gui_mouse_b()) {
	 dclick_status = DCLICK_AGAIN;             /* yes! the second click */
	 dclick_time = 0;
	 return;
      }
   }
   else
      return;

   /* timeout? */
   if (dclick_time++ > 10)
      dclick_status = DCLICK_NOT;
}

static END_OF_FUNCTION(dclick_check);

#endif

#define GFX_CANCEL         3
#define GFX_DRIVER_LIST    4
#define GFX_MODE_LIST      5
#define GFX_SCAN_LIST      7
#define GFX_FSKIP_LIST     11
#define GFX_VSYNC_LIST     13
#define GFX_SPEED_LIST     16
#define GFX_EAGLE_LIST     19
#define GFX_TRIPLE_LIST    21

extern DIALOG gfx_mode_dialog[];

/*

- Built on Allegro gui routines, like the bgui library, but it doesn't suck
  as much in my opinion :). But it still sucks a lot...

Todo:

- Anything/everything/something... some day maybe... boring...

*/

typedef struct GFX_MODE_DATA
{
   int w;
   int h;
   char *s;
} GFX_MODE_DATA;

static GFX_MODE_DATA gfx_mode_data_autodetect[] =
{
#ifdef RAINE_DOS
   { 256,   200,  " 256 x 200"  },
   { 256,   224,  " 256 x 224"  },
   { 256,   240,  " 256 x 240"  },
   { 256,   256,  " 256 x 256"  },
   { 320,   200,  " 320 x 200"  },
   { 320,   240,  " 320 x 240"  },
   { 320,   256,  " 320 x 256"  },
   { 320,   350,  " 320 x 350"  },
   { 320,   400,  " 320 x 400"  },
   { 320,   480,  " 320 x 480"  },
   { 320,   600,  " 320 x 600"  },
   { 360,   200,  " 360 x 200"  },
   { 360,   240,  " 360 x 240"  },
   { 360,   270,  " 360 x 270"  },
   { 360,   360,  " 360 x 360"  },
   { 360,   400,  " 360 x 400"  },
   { 360,   480,  " 360 x 480"  },
   { 360,   600,  " 360 x 600"  },
   { 376,   282,  " 376 x 282"  },
   { 376,   308,  " 376 x 308"  },
   { 376,   564,  " 376 x 564"  },
   { 400,   300,  " 400 x 300"  },
   { 400,   600,  " 400 x 600"  },
   { 512,   384,  " 512 x 384"  },
   { 640,   200,  " 640 x 200"  },
   { 640,   240,  " 640 x 240"  },
   { 640,   400,  " 640 x 400"  },
   { 640,   480,  " 640 x 480"  },
   { 800,   200,  " 800 x 200"  },
   { 800,   300,  " 800 x 300"  },
   { 800,   600,  " 800 x 600"  },
   { 1024,  256,  "1024 x 256"  },
   { 1024,  384,  "1024 x 384"  },
   { 1024,  768,  "1024 x 768"  },
   { 1280,  1024, "1280 x 1024" },
   { 1600,  1200, "1600 x 1200" },
#else
   { 320,   200,  " 320 x 200"  },
   { 320,   240,  " 320 x 240"  },
   { 400,   300,  " 400 x 300"  },
   { 512,   384,  " 512 x 384"  },
   { 640,   400,  " 640 x 400"  },
   { 640,   480,  " 640 x 480"  },
   { 800,   600,  " 800 x 600"  },
   { 1024,  768,  "1024 x 768"  },
   { 1280,  1024, "1280 x 1024" },
   { 1600,  1200, "1600 x 1200" },
#endif
   { 0,     0,    NULL          }
};

static GFX_MODE_DATA gfx_mode_data_modex[] =
{
   { 256,   200,  " 256 x 200"  },
   { 256,   224,  " 256 x 224"  },
   { 256,   240,  " 256 x 240"  },
   { 256,   256,  " 256 x 256"  },
   { 320,   200,  " 320 x 200"  },
   { 320,   240,  " 320 x 240"  },
   { 320,   256,  " 320 x 256"  },
   { 320,   350,  " 320 x 350"  },
   { 320,   400,  " 320 x 400"  },
   { 320,   480,  " 320 x 480"  },
   { 320,   600,  " 320 x 600"  },
   { 360,   200,  " 360 x 200"  },
   { 360,   240,  " 360 x 240"  },
   { 360,   270,  " 360 x 270"  },
   { 360,   360,  " 360 x 360"  },
   { 360,   400,  " 360 x 400"  },
   { 360,   480,  " 360 x 480"  },
   { 360,   600,  " 360 x 600"  },
   { 376,   282,  " 376 x 282"  },
   { 376,   308,  " 376 x 308"  },
   { 376,   564,  " 376 x 564"  },
   { 400,   300,  " 400 x 300"  },
   { 400,   600,  " 400 x 600"  },

   { 0,     0,    NULL          }
};

static GFX_MODE_DATA gfx_mode_data_arcade_monitor[] =
{
   { 320,   240,  " 320 x 240"  },
   { 0,     0,    NULL          }
};

static GFX_MODE_DATA gfx_mode_data_vga[] =
{
   { 320,   200,  " 320 x 200"  },
   { 0,     0,    NULL          }
};

static GFX_MODE_DATA gfx_mode_data_extended[] =
{
   { 640,   400,  " 640 x 400"  },
   { 0,     0,    NULL          }
};

static GFX_MODE_DATA gfx_mode_data_vesa[] =
{
   { 320,   200,  " 320 x 200"  },
   { 320,   240,  " 320 x 240"  },
   { 400,   300,  " 400 x 300"  },
   { 512,   384,  " 512 x 384"  },
   { 640,   400,  " 640 x 400"  },
   { 640,   480,  " 640 x 480"  },
   { 800,   600,  " 800 x 600"  },
   { 1024,  768,  "1024 x 768"  },
   { 1280,  1024, "1280 x 1024" },
   { 1600,  1200, "1600 x 1200" },

   { 320,   400,  " 320 x 400"  },
   { 320,   480,  " 320 x 480"  },
   { 400,   600,  " 400 x 600"  },
   { 640,   200,  " 640 x 200"  },
   { 640,   240,  " 640 x 240"  },
   { 800,   200,  " 800 x 200"  },
   { 800,   300,  " 800 x 300"  },
   { 1024,  256,  "1024 x 256"  },
   { 1024,  384,  "1024 x 384"  },

   { 0,     0,    NULL          }
};
#ifndef RAINE_DOS
static GFX_MODE_DATA gfx_mode_data_directx[] =
{
   { 320,   200,  " 320 x 200"  },
   { 320,   240,  " 320 x 240"  },
   { 400,   300,  " 400 x 300"  },
   { 512,   384,  " 512 x 384"  },
   { 640,   400,  " 640 x 400"  },
   { 640,   480,  " 640 x 480"  },
   { 800,   600,  " 800 x 600"  },
   { 1024,  768,  "1024 x 768"  },
   { 1280,  1024, "1280 x 1024" },
   { 1600,  1200, "1600 x 1200" },

   { 0,     0,    NULL          }
};
#endif
// gfx_mode_getter():
// Listbox data getter routine for the graphics mode list.

GFX_MODE_DATA *res_data;
ULONG res_data_length;

static char *gfx_mode_getter(int index, int *list_size)
{
   ULONG ta;

   if(index == -1){
      if(list_size){
         if(!res_data_length){
            ta = 0;
            while(res_data[ta].s){
               ta++;
            }
            res_data_length = ta;
         }
	 *list_size = res_data_length;
      }
      return NULL;
   }

   if((index >= 0)&&(index < res_data_length))
      return raine_translate_text(res_data[index].s);
   else
      return NULL;
}

typedef struct GFX_CARD_DATA
{
   ULONG id;
   UBYTE *name;
   GFX_MODE_DATA *res_list;
} GFX_CARD_DATA;

struct GFX_CARD_DATA gfx_card_data[] =
{
	{ GFX_AUTODETECT,     "Autodetect",           gfx_mode_data_autodetect },
#ifdef GFX_DRIVER_VGA
	{ GFX_VGA,            "VGA",                  gfx_mode_data_vga },
#endif
#ifdef GFX_DRIVER_MODEX
	{ GFX_MODEX,          "Mode-X",               gfx_mode_data_modex },
	{ GFX_ARCMON,         "Arcade Monitor",       gfx_mode_data_arcade_monitor },
#endif
#ifdef GFX_DRIVER_VESA1
	{ GFX_VESA1,          "VESA 1.2",             gfx_mode_data_vesa },
#endif
#ifdef GFX_DRIVER_VESA2B
	{ GFX_VESA2B,         "VESA 2.0 (banked)",    gfx_mode_data_vesa },
#endif
#ifdef GFX_DRIVER_VESA2L
	{ GFX_VESA2L,         "VESA 2.0 (linear)",    gfx_mode_data_vesa },
#endif
#ifdef GFX_DRIVER_VESA3
	{ GFX_VESA3,          "VESA 3.0",             gfx_mode_data_vesa },
#endif
#ifdef GFX_DRIVER_VBEAF
	{ GFX_VBEAF,          "VBE/AF",               gfx_mode_data_vesa },
#endif
#ifdef GFX_DRIVER_XTENDED
	{ GFX_XTENDED,        "Extended Mode",        gfx_mode_data_extended },
#endif
#ifdef GFX_DRIVER_DIRECTX
	{ GFX_DIRECTX_ACCEL,  "Direct X Accelerated", gfx_mode_data_directx },
	{ GFX_DIRECTX_SAFE,   "Direct X Safe",        gfx_mode_data_directx },
	{ GFX_DIRECTX_SOFT,   "Direct X Soft",        gfx_mode_data_directx },
	{ GFX_DIRECTX_OVL,    "Direct X Overlay",     gfx_mode_data_directx },
	{ GFX_DIRECTX_DBLBUF, "Direct X Double Buf",  gfx_mode_data_directx },
	{ GFX_GDI,            "Direct X GDI",         gfx_mode_data_directx },
#endif
};

void switch_gfx_res_data(int screen_mode, int x, int y)
{
   int i;

   res_data = gfx_card_data[screen_mode].res_list;
   res_data_length = 0;

   gfx_mode_dialog[GFX_MODE_LIST].d1 = 0;

   for(i=0; res_data[i].s; i++){
      if((res_data[i].w == x)&&(res_data[i].h == y)){
	 gfx_mode_dialog[GFX_MODE_LIST].d1 = i;
	 return; 
      }
   }

}

/* gfx_card_getter:
 *  Listbox data getter routine for the graphics card list.
 */
static char *gfx_card_getter(int index, int *list_size)
{
   if(index == -1){
      if(list_size){
	 *list_size = (sizeof(gfx_card_data) / sizeof(GFX_CARD_DATA));
      }
      return NULL;
   }

   if(index == -4){

      SEND_MESSAGE(&gfx_mode_dialog[GFX_MODE_LIST], MSG_END, 0);

      switch_gfx_res_data(
         *list_size,
         res_data[gfx_mode_dialog[GFX_MODE_LIST].d1].w,
         res_data[gfx_mode_dialog[GFX_MODE_LIST].d1].h
      );

      SEND_MESSAGE(&gfx_mode_dialog[GFX_MODE_LIST], MSG_START, 0);
      SEND_MESSAGE(&gfx_mode_dialog[GFX_MODE_LIST], MSG_DRAW, 0);

      return NULL;
   }

   if(index >= 0)
      return raine_translate_text(gfx_card_data[index].name);
   else
      return NULL;
}


// gfx_card_name():
// Get card name for card ID

char *gfx_card_name(ULONG index)
{
   int ta,tb;

   tb = (sizeof(gfx_card_data) / sizeof(GFX_CARD_DATA));
   for(ta=0;ta<tb;ta++){
      if(gfx_card_data[ta].id == index){
         return gfx_card_data[ta].name;
      }
   }

   return gfx_card_data[0].name;			// Autodetect
}

// gfx_card_number():
// Get card number for card ID

int gfx_card_number(ULONG index)
{
   int ta,tb;

   tb = (sizeof(gfx_card_data) / sizeof(GFX_CARD_DATA));
   for(ta=0;ta<tb;ta++){
      if(gfx_card_data[ta].id == index){
         return ta;
      }
   }

   return 0;				// Autodetect
}

// gfx_card_id():
// Get card ID for card number

int gfx_card_id(int index)
{
   if((index>=0)&&(index<(sizeof(gfx_card_data) / sizeof(GFX_CARD_DATA)))){
      return gfx_card_data[index].id;
   }

   return gfx_card_data[0].id;		// Autodetect
}

static char *frame_skip_getter(int index, int *list_size)
{
   static char *list_data[] =
   {
      "Automatic",
      "1",
      "2",
      "3",
      "4",
      "5",
      "6",
      "7",
      "8",
      "9",
   };

   if(index == -1){
      if(list_size)
	 *list_size = 10;
      return NULL;
   }

   if(index >= 0)
      return raine_translate_text(list_data[index]);
   else
      return NULL;
}

static char *screen_effect_getter(int index, int *list_size)
{
   static char *list_data[] =
   {
      "None",
      "Eagle",
      "Pixel Double",
      "Double Width",
   };

   if(index == -1){
      if(list_size)
	 *list_size = sizeof(list_data) / sizeof(char *);
      return NULL;
   }

   if(index >= 0)
      return raine_translate_text(list_data[index]);
   else
      return NULL;
}

int screen_radio_proc(int msg, DIALOG *d, int c);
int screen_def_edit_proc(int msg, DIALOG *d, int c);

DIALOG gfx_mode_dialog[] =
{
   /* (dialog proc)          (x)   (y)   (w)   (h)   (fg)  (bg)  (key) (flags)  (d1)  (d2)  (dp) */
   { d_raine_window_proc,     0,    0,    304,  224-20,  255,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    "Screen Setup"},
   { x_text_proc,             8,    16,   1,    1,    254,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    "Screen Mode:"},
   { x_raine_button_proc,     8,    200-20,  48,   16,   255,  GUI_BOX_COL_MIDDLE,    'O',  D_EXIT,  0,    0,    "&Ok"},
   { x_raine_button_proc,     64,   200-20,  48,   16,   255,  GUI_BOX_COL_MIDDLE,    'C',  D_EXIT,  0,    0,    "&Cancel"},
   { d_raine_list_proc,       8,    28,   140,  68,   255,  GUI_BOX_COL_MIDDLE,    0,    D_EXIT,  0,    0,    gfx_card_getter},
   { d_raine_list_proc,       156,  28,   140,  68,   255,  GUI_BOX_COL_MIDDLE,    0,    D_EXIT,  3,    0,    gfx_mode_getter},
   { x_text_proc,             196,  132+4,1,    1,    254,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    "Scanlines:"},
   { x_raine_radio_proc,      196,  142+4,64,   9,    255,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    1,    "Off"},
   { x_raine_radio_proc,      196,  152+4,64,   9,    255,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    1,    "Fullheight"},
   { x_raine_radio_proc,      196,  162+4,64,   9,    255,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    1,    "Halfheight"},
   { x_text_proc,             8,    102,  1,    1,    254,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    "Frame Skip:"},
   { d_raine_list_proc,       8,    112,  80,   19,   255,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    frame_skip_getter},
   { x_text_proc,             196,  102,  1,    1,    254,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    "VSync Palette:"},
   { x_raine_radio_proc,      196,  112,  64,   9,    255,  GUI_BOX_COL_MIDDLE,    0,    0,       2,    1,    "Off"},
   { x_raine_radio_proc,      196,  122,  64,   9,    255,  GUI_BOX_COL_MIDDLE,    0,    0,       2,    1,    "On"},
   { x_text_proc,             106,  102,  1,    1,    254,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    "Limit Speed:"},
   { x_raine_radio_proc,      106,  112,  64,   9,    255,  GUI_BOX_COL_MIDDLE,    0,    0,       3,    1,    "Off"},
   { x_raine_radio_proc,      106,  122,  64,   9,    255,  GUI_BOX_COL_MIDDLE,    0,    0,       3,    1,    "On"},
   { x_text_proc,             106,  132+4,1,    1,    254,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    "Screen Effect:"},
   { d_raine_list_proc,       106,  142+4,64,   19,   255,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    screen_effect_getter},
   { x_text_proc,             8,    132+4,1,    1,    254,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    "Triple Buffer:"},
   { x_raine_radio_proc,      8,    142+4,64,   9,    255,  GUI_BOX_COL_MIDDLE,    0,    0,       5,    1,    "Off"},
   { x_raine_radio_proc,      8,    152+4,64,   9,    255,  GUI_BOX_COL_MIDDLE,    0,    0,       5,    1,    "On"},
   { x_text_proc,             156,  16,   1,    1,    254,  GUI_BOX_COL_MIDDLE,    0,    0,       0,    0,    "Screen Size:"},
   { NULL }
};

// raine_gfx_mode_select():
// Displays the Allegro graphics mode selection dialog, which allows the
// user to select a screen mode and graphics card. Stores the selection
// in the three variables, and returns zero if it was closed with the 
// Cancel button, or non-zero if it was OK'd.

int raine_gfx_mode_select(void)
{
   int ret,i;

   // Wait for user

   clear_keybuf();
   while(gui_mouse_b()){
      dialog_oxygen();
   };

   // Init stuff

   if(display_cfg.scanlines>2)display_cfg.scanlines=2;

   gfx_mode_dialog[GFX_SCAN_LIST].flags=0;
   gfx_mode_dialog[GFX_SCAN_LIST+1].flags=0;
   gfx_mode_dialog[GFX_SCAN_LIST+2].flags=0;

   gfx_mode_dialog[GFX_SCAN_LIST+display_cfg.scanlines].flags=D_SELECTED;

   if(display_cfg.frame_skip>9) display_cfg.frame_skip = 0;

   gfx_mode_dialog[GFX_FSKIP_LIST].d1 = display_cfg.frame_skip;

   if(display_cfg.vsync>1)display_cfg.vsync=0;

   gfx_mode_dialog[GFX_VSYNC_LIST].flags=0;
   gfx_mode_dialog[GFX_VSYNC_LIST+1].flags=0;

   gfx_mode_dialog[GFX_VSYNC_LIST+display_cfg.vsync].flags=D_SELECTED;

   if(display_cfg.limit_speed>1) display_cfg.limit_speed=1;

   gfx_mode_dialog[GFX_SPEED_LIST].flags=0;
   gfx_mode_dialog[GFX_SPEED_LIST+1].flags=0;

   gfx_mode_dialog[GFX_SPEED_LIST+display_cfg.limit_speed].flags=D_SELECTED;

   gfx_mode_dialog[GFX_EAGLE_LIST].d1 = 0;

   if(display_cfg.eagle_fx)

      gfx_mode_dialog[GFX_EAGLE_LIST].d1 = 1;

   if(display_cfg.pixel_double==1)

      gfx_mode_dialog[GFX_EAGLE_LIST].d1 = 2;

   if(display_cfg.pixel_double==2)

      gfx_mode_dialog[GFX_EAGLE_LIST].d1 = 3;

   if(display_cfg.triple_buffer>1)display_cfg.triple_buffer=1;

   gfx_mode_dialog[GFX_TRIPLE_LIST].flags=0;
   gfx_mode_dialog[GFX_TRIPLE_LIST+1].flags=0;

   gfx_mode_dialog[GFX_TRIPLE_LIST+display_cfg.triple_buffer].flags=D_SELECTED;

   gfx_mode_dialog[GFX_DRIVER_LIST].d1 = gfx_card_number(display_cfg.screen_type);

   if(display_cfg.scanlines == 2) display_cfg.screen_y <<= 1;

   switch_gfx_res_data(
      gfx_card_number(display_cfg.screen_type),
      display_cfg.screen_x,
      display_cfg.screen_y
   );

   if(display_cfg.scanlines == 2) display_cfg.screen_y >>= 1;

   // Do the selection

   ret = raine_do_dialog(gfx_mode_dialog, GFX_DRIVER_LIST);

   // Set the results

   display_cfg.screen_type = gfx_card_id(gfx_mode_dialog[GFX_DRIVER_LIST].d1);

   display_cfg.screen_x = res_data[gfx_mode_dialog[GFX_MODE_LIST].d1].w;
   display_cfg.screen_y = res_data[gfx_mode_dialog[GFX_MODE_LIST].d1].h;

   if((gfx_mode_dialog[GFX_SCAN_LIST].flags)&D_SELECTED){display_cfg.scanlines=0;};
   if((gfx_mode_dialog[GFX_SCAN_LIST+1].flags)&D_SELECTED){display_cfg.scanlines=1;};
   if((gfx_mode_dialog[GFX_SCAN_LIST+2].flags)&D_SELECTED){display_cfg.scanlines=2;};

   display_cfg.frame_skip = gfx_mode_dialog[GFX_FSKIP_LIST].d1;

   if((gfx_mode_dialog[GFX_VSYNC_LIST].flags)&D_SELECTED){display_cfg.vsync=0;};
   if((gfx_mode_dialog[GFX_VSYNC_LIST+1].flags)&D_SELECTED){display_cfg.vsync=1;};

   if((gfx_mode_dialog[GFX_SPEED_LIST].flags)&D_SELECTED){display_cfg.limit_speed=0;};
   if((gfx_mode_dialog[GFX_SPEED_LIST+1].flags)&D_SELECTED){display_cfg.limit_speed=1;};

   display_cfg.eagle_fx = 0;
   display_cfg.pixel_double = 0;

   switch(gfx_mode_dialog[GFX_EAGLE_LIST].d1){
      case 1:
         display_cfg.eagle_fx = 1;
      break;
      case 2:
         display_cfg.pixel_double = 1;
      break;
      case 3:
         display_cfg.pixel_double = 2;
      break;
   }

   if((gfx_mode_dialog[GFX_TRIPLE_LIST].flags)&D_SELECTED){display_cfg.triple_buffer=0;};
   if((gfx_mode_dialog[GFX_TRIPLE_LIST+1].flags)&D_SELECTED){display_cfg.triple_buffer=1;};

   if(ret==GFX_CANCEL){return(FALSE);}
   else{return(TRUE);}
}



static DIALOG alert_dialog[] =
{
   /* (dialog proc)     (x)   (y)   (w)   (h)   (fg)  (bg)  (key) (flags)  (d1)  (d2)  (dp) */
   { d_raine_window_proc,0,   0,    0,    0,    255,  GUI_BOX_COL_MIDDLE,  0,    0,       0,    0,    NULL },
   { d_ctext_proc,       0,   0,    0,    0,    254,  GUI_BOX_COL_MIDDLE,  0,    0,       0,    0,    NULL },
   { d_ctext_proc,       0,   0,    0,    0,    254,  GUI_BOX_COL_MIDDLE,  0,    0,       0,    0,    NULL },
   { d_ctext_proc,       0,   0,    0,    0,    254,  GUI_BOX_COL_MIDDLE,  0,    0,       0,    0,    NULL },
   { d_raine_button_proc,0,   0,    0,    0,    255,  GUI_BOX_COL_MIDDLE,  0,    D_EXIT,  0,    0,    NULL },
   { d_raine_button_proc,0,   0,    0,    0,    255,  GUI_BOX_COL_MIDDLE,  0,    D_EXIT,  0,    0,    NULL },
   { d_raine_button_proc,0,   0,    0,    0,    255,  GUI_BOX_COL_MIDDLE,  0,    D_EXIT,  0,    0,    NULL },
   { NULL }
};

#define A_T1  0
#define A_S1  1
#define A_S2  2
#define A_S3  3
#define A_B1  4
#define A_B2  5
#define A_B3  6

// raine_alert3():
// Displays a simple alert box, containing three lines of text (s1-s3),
// and with either one, two, or three buttons. The text for these buttons 
// is passed in b1, b2, and b3 (NULL for buttons which are not used), and
// the keyboard shortcuts in c1 and c2. Returns 1, 2, or 3 depending on 
// which button was selected.

int raine_alert3(char *title, char *s1, char *s2, char *s3, char *b1, char *b2, char *b3, int c1, int c2, int c3)
{
   int maxlen = 0;
   int len1, len2, len3;
   int avg_w = text_length(font, " ");
   int avg_h = text_height(font);
   int buttons = 0;
   int b[3];
   int c;

   #define SORT_OUT_BUTTON(x) {                                            \
      if (b##x) {                                                          \
	 alert_dialog[A_B##x].flags &= ~D_HIDDEN;                          \
	 alert_dialog[A_B##x].key = c##x;                                  \
	 alert_dialog[A_B##x].dp = b##x;                                   \
	 len##x = gui_strlen(b##x);                                        \
	 b[buttons++] = A_B##x;                                            \
      }                                                                    \
      else {                                                               \
	 alert_dialog[A_B##x].flags |= D_HIDDEN;                           \
	 len##x = 0;                                                       \
      }                                                                    \
   }

   alert_dialog[A_S1].dp = alert_dialog[A_S2].dp = alert_dialog[A_S3].dp = 
			   alert_dialog[A_B1].dp = alert_dialog[A_B2].dp = 
                           alert_dialog[A_T1].dp = "";

   if (s1) {
      alert_dialog[A_S1].dp = s1;
      maxlen = text_length(font, s1);
   }

   if (s2) {
      alert_dialog[A_S2].dp = s2;
      len1 = text_length(font, s2);
      if (len1 > maxlen)
	 maxlen = len1;
   }

   if (s3) {
      alert_dialog[A_S3].dp = s3;
      len1 = text_length(font, s3);
      if (len1 > maxlen)
	 maxlen = len1;
   }

   if (title) {
      alert_dialog[A_T1].dp = title;
      len1 = text_length(font, title);
      if (len1 > maxlen)
	 maxlen = len1;
   }

   SORT_OUT_BUTTON(1);
   SORT_OUT_BUTTON(2);
   SORT_OUT_BUTTON(3);

   len1 = MAX(len1, MAX(len2, len3)) + avg_w*3;
   if (len1*buttons > maxlen)
      maxlen = len1*buttons;

   maxlen += avg_w*4;
   alert_dialog[0].w = maxlen;
   alert_dialog[A_S1].x = alert_dialog[A_S2].x = alert_dialog[A_S3].x = 
						alert_dialog[0].x + maxlen/2;

   alert_dialog[A_B1].w = alert_dialog[A_B2].w = alert_dialog[A_B3].w = len1;

   alert_dialog[A_B1].x = alert_dialog[A_B2].x = alert_dialog[A_B3].x = 
				       alert_dialog[0].x + maxlen/2 - len1/2;

   if (buttons == 3) {
      alert_dialog[b[0]].x = alert_dialog[0].x + maxlen/2 - len1*3/2 - avg_w;
      alert_dialog[b[2]].x = alert_dialog[0].x + maxlen/2 + len1/2 + avg_w;
   }
   else if (buttons == 2) {
      alert_dialog[b[0]].x = alert_dialog[0].x + maxlen/2 - len1 - avg_w;
      alert_dialog[b[1]].x = alert_dialog[0].x + maxlen/2 + avg_w;
   }

   alert_dialog[0].y += 10;
   alert_dialog[0].h = 10 + avg_h*8;
   alert_dialog[A_S1].y = alert_dialog[0].y + avg_h;
   alert_dialog[A_S2].y = alert_dialog[0].y + avg_h*2;
   alert_dialog[A_S3].y = alert_dialog[0].y + avg_h*3;
   alert_dialog[A_S1].h = alert_dialog[A_S2].h = alert_dialog[A_S2].h = avg_h;

   alert_dialog[A_B1].y = alert_dialog[A_B2].y = alert_dialog[A_B3].y = 
						alert_dialog[0].y + avg_h*5;

   alert_dialog[A_B1].h = alert_dialog[A_B2].h = alert_dialog[A_B3].h = avg_h*2;

   alert_dialog[0].y -= 10;

   centre_dialog(alert_dialog);

   clear_keybuf();
   while(gui_mouse_b()){
      dialog_oxygen();
   };

   //c = raine_do_dialog(alert_dialog, A_B1);
   c = raine_window_dialog(alert_dialog, A_B1);

   if (c == A_B1)
      return 1;
   else if (c == A_B2)
      return 2;
   else
      return 3;
}



// raine_alert():
// Displays a simple alert box, containing three lines of text (s1-s3),
// and with either one or two buttons. The text for these buttons is passed
// in b1 and b2 (b2 may be null), and the keyboard shortcuts in c1 and c2.
// Returns 1 or 2 depending on which button was selected.

int raine_alert(char *title, char *s1, char *s2, char *s3, char *b1, char *b2, int c1, int c2)
{
   int ret;

   ret = raine_alert3(title, s1, s2, s3, b1, b2, NULL, c1, c2, 0);

   if (ret > 2)
      ret = 2;

   return ret;
}

/* raine_do_dialog:
 *  The basic dialog manager. The list of dialog objects should be
 *  terminated by one with a null dialog procedure. Returns the index of 
 *  the object which caused it to exit.
 */

#define BUFFER_GUI

#ifdef BUFFER_GUI
static BITMAP *screen_old;
static BITMAP *screen_new;
#ifdef TEST_WINGUI
static BITMAP *back_buffer;
#endif
static int oxygen_level;
#endif

int raine_do_dialog(DIALOG *dialog, int focus_obj)
{
   void *player;
   int kill,ta;
#ifdef BUFFER_GUI
   if(!(screen_new)){
      show_mouse(NULL);
      screen_new = create_bitmap_ex(8, display_cfg.screen_x,display_cfg.screen_y);
      screen_old = screen;
      screen     = screen_new;
      kill       = 1;
      blit(screen_old,screen_new,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);
#ifdef TEST_WINGUI
      back_buffer = create_bitmap_ex(8, display_cfg.screen_x,display_cfg.screen_y);
      blit(screen_old,back_buffer,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);
#endif
   }
   else{
      kill       = 0;
   }
#endif
   show_mouse(screen);

   player = init_dialog(dialog, focus_obj);

   oxygen_level=0;

   while(update_dialog(player)){
      if(oxygen_level==0){		// Gasp! Ack!
         dialog_oxygen();
      }
      oxygen_level=0;
   }

   //show_mouse(mouse_screen);

   ta = shutdown_dialog(player);
#ifdef BUFFER_GUI
   if(kill){
      show_mouse(NULL);
      dialog_oxygen();
      destroy_bitmap(screen_new);
      screen     = screen_old;
      screen_new = NULL;
#ifdef TEST_WINGUI
      destroy_bitmap(back_buffer);
      back_buffer = NULL;
#endif
   }
#endif
   return ta;
}

// Stop the mouse freezing, when the game is loading and there
// is no oxygen coming in.

void mouse_on_real_screen(void)
{
   show_mouse(NULL);
   dialog_oxygen();
   show_mouse(screen_old);
}

// Restore the mouse to the buffer screen, when oxygen supply
// returns.

void mouse_on_buffer_screen(void)
{
   //show_mouse(NULL);
   //dialog_oxygen();
   show_mouse(screen);
   dialog_oxygen();
}

#ifndef TEST_WINGUI

static void save_gui_screen(void)
{
   UBYTE full_name[256];
   UBYTE file_name[32];

   raine_cfg.req_save_screen = 0;

   /*

   first try gui.pcx

   */

   sprintf(file_name, "%s.pcx", "gui");

   sprintf(full_name, "%s%s", dir_cfg.screen_dir, file_name);

   /*

   otherwise, find the next gui_NNN.pcx (or guiNNN.pcx)

   */

   while( exists(full_name) ){

      if(dir_cfg.last_screenshot_num > 999)

         return;

      if(dir_cfg.long_file_names)

         sprintf(file_name, "%s_%03d.pcx", "gui", dir_cfg.last_screenshot_num++);

      else

         sprintf(file_name, "%.5s%03d.pcx", "gui", dir_cfg.last_screenshot_num++);

      sprintf(full_name, "%s%s", dir_cfg.screen_dir, file_name);

   };

   save_pcx(full_name, screen, gui_pal);

   //print_ingame(120, "Screen Saved to %s", file_name);
}

void dialog_oxygen(void)
{

#ifdef BUFFER_GUI
   static int z,z1,z2,z3;
   int ta;
   int x;

   if(raine_cfg.hide){

       oxygen_level=1;

   }
   else{

    if(screen_new){		// Ahhhhhhh.... That's better :)
       oxygen_level=1;

       if(!raine_cfg.wibble){

       blit(screen,screen_old,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);

       if(key[KEY_0]) save_gui_screen();

       }
       else{
/*
       for(ta=0;ta<display_cfg.screen_y;ta++){
       blit(screen,screen_old,0,ta,0,display_cfg.screen_y-(ta+1),display_cfg.screen_x,1);
       }
*/
       z++;

       z1=z/7;
       z1&=255;

       z2=z/5;
       z2&=255;

       z3=z/1;
       z3&=255;
       z3^=255;

       for(ta=0;ta<display_cfg.screen_y;ta++){
       x = fixtoi ( (fsin(itofix(ta+z1)) * 16) + (fsin(itofix(z2-(ta*16))) * 2) + (fsin(itofix(z3-(ta*3))) * 4));
       blit(screen,screen_old,0,ta,x,ta,display_cfg.screen_x,1);
       }

       }
    }
    
    }
#endif

}

/******************************************************************************/

#else

/* find_mouse_object:
 *  Finds which object the mouse is on top of.
 */
static int find_mouse_object(DIALOG *d)
{
   /* finds which object the mouse is on top of */

   int mouse_object = -1;
   int c;

   for (c=0; d[c].proc; c++)
      if ((rgui_mouse_x() >= d[c].x) && (rgui_mouse_y() >= d[c].y) &&
	  (rgui_mouse_x() < d[c].x + d[c].w) && (rgui_mouse_y() < d[c].y + d[c].h) &&
	  (!(d[c].flags & (D_HIDDEN | D_DISABLED))))
	 mouse_object = c;

   return mouse_object;
}



/* offer_focus:
 *  Offers the input focus to a particular object.
 */
static int offer_focus(DIALOG *d, int obj, int *focus_obj, int force)
{
   int res = D_O_K;

   if ((obj == *focus_obj) || 
       ((obj >= 0) && (d[obj].flags & (D_HIDDEN | D_DISABLED))))
      return D_O_K;

   /* check if object wants the focus */
   if (obj >= 0) {
      res = SEND_MESSAGE(d+obj, MSG_WANTFOCUS, 0);
      if (res & D_WANTFOCUS)
	 res ^= D_WANTFOCUS;
      else
	 obj = -1;
   }

   if ((obj >= 0) || (force)) {
      /* take focus away from old object */
      if (*focus_obj >= 0) {
	 res |= SEND_MESSAGE(d+*focus_obj, MSG_LOSTFOCUS, 0);
	 if (res & D_WANTFOCUS) {
	    if (obj < 0)
	       return D_O_K;
	    else
	       res &= ~D_WANTFOCUS;
	 }
	 d[*focus_obj].flags &= ~D_GOTFOCUS;
	 scare_mouse();
	 res |= SEND_MESSAGE(d+*focus_obj, MSG_DRAW, 0);
	 unscare_mouse();
      }

      *focus_obj = obj;

      /* give focus to new object */
      if (obj >= 0) {
	 scare_mouse();
	 d[obj].flags |= D_GOTFOCUS;
	 res |= SEND_MESSAGE(d+obj, MSG_GOTFOCUS, 0);
	 res |= SEND_MESSAGE(d+obj, MSG_DRAW, 0);
	 unscare_mouse();
      }
   }

   return res;
}



#define MAX_OBJECTS     512

typedef struct OBJ_LIST
{
   int index;
   int diff;
} OBJ_LIST;



/* obj_list_cmp:
 *  Callback function for qsort().
 */
static int obj_list_cmp(const void *e1, const void *e2)
{
   return (((OBJ_LIST *)e1)->diff - ((OBJ_LIST *)e2)->diff);
}



/* cmp_tab:
 *  Comparison function for tab key movement.
 */
static int cmp_tab(DIALOG *d1, DIALOG *d2)
{
   int ret = (int)d2 - (int)d1;

   if (ret < 0)
      ret += 0x10000;

   return ret;
}



/* cmp_shift_tab:
 *  Comparison function for shift+tab key movement.
 */
static int cmp_shift_tab(DIALOG *d1, DIALOG *d2)
{
   int ret = (int)d1 - (int)d2;

   if (ret < 0)
      ret += 0x10000;

   return ret;
}



/* cmp_right:
 *  Comparison function for right arrow key movement.
 */
static int cmp_right(DIALOG *d1, DIALOG *d2)
{
   int ret = (d2->x - d1->x) + ABS(d1->y - d2->y) * 8;

   if (d1->x >= d2->x)
      ret += 0x10000;

   return ret;
}



/* cmp_left:
 *  Comparison function for left arrow key movement.
 */
static int cmp_left(DIALOG *d1, DIALOG *d2)
{
   int ret = (d1->x - d2->x) + ABS(d1->y - d2->y) * 8;

   if (d1->x <= d2->x)
      ret += 0x10000;

   return ret;
}



/* cmp_down:
 *  Comparison function for down arrow key movement.
 */
static int cmp_down(DIALOG *d1, DIALOG *d2)
{
   int ret = (d2->y - d1->y) + ABS(d1->x - d2->x) * 8;

   if (d1->y >= d2->y)
      ret += 0x10000;

   return ret;
}



/* cmp_up:
 *  Comparison function for up arrow key movement.
 */
static int cmp_up(DIALOG *d1, DIALOG *d2)
{
   int ret = (d1->y - d2->y) + ABS(d1->x - d2->x) * 8;

   if (d1->y <= d2->y)
      ret += 0x10000;

   return ret;
}



/* move_focus:
 *  Handles arrow key and tab movement through a dialog, deciding which
 *  object should be given the input focus.
 */
static int move_focus(DIALOG *d, long ch, int *focus_obj)
{
   int (*cmp)(DIALOG *d1, DIALOG *d2);
   OBJ_LIST obj[MAX_OBJECTS];
   int obj_count = 0;
   int fobj, c;
   int res = D_O_K;

   /* choose a comparison function */ 
   switch (ch >> 8) {
      case KEY_TAB:     cmp = ((ch&0xFF) == '\t') ? cmp_tab : cmp_shift_tab;  break;
      case KEY_RIGHT:   cmp = cmp_right;  break;
      case KEY_LEFT:    cmp = cmp_left;   break;
      case KEY_DOWN:    cmp = cmp_down;   break;
      case KEY_UP:      cmp = cmp_up;     break;
      default:          return D_O_K;
   }

   /* fill temporary table */
   for (c=0; d[c].proc; c++) {
      if ((*focus_obj < 0) || (c != *focus_obj)) {
	 obj[obj_count].index = c;
	 if (*focus_obj >= 0)
	    obj[obj_count].diff = cmp(d+*focus_obj, d+c);
	 else
	    obj[obj_count].diff = c;
	 obj_count++;
	 if (obj_count >= MAX_OBJECTS)
	    break;
      }
   }

   /* sort table */
   qsort(obj, obj_count, sizeof(OBJ_LIST), obj_list_cmp);

   /* find an object to give the focus to */
   fobj = *focus_obj;
   for (c=0; c<obj_count; c++) {
      res |= offer_focus(d, obj[c].index, focus_obj, FALSE);
      if (fobj != *focus_obj)
	 break;
   } 

   return res;
}



#define MESSAGE(i, msg, c) {                       \
   r = SEND_MESSAGE(player->dialog+i, msg, c);     \
   if (r != D_O_K) {                               \
      player->res |= r;                            \
      player->obj = i;                             \
   }                                               \
}

int raine_window_dialog(DIALOG *dialog, int focus_obj)
{
   int count;

   void *player;
   int kill,ta;
#ifdef BUFFER_GUI
   if(screen_new==NULL){
      show_mouse(NULL);
      screen_new = create_bitmap_ex(8, display_cfg.screen_x,display_cfg.screen_y);
      screen_old = screen;
      screen     = screen_new;
      kill       = 1;
      blit(screen_old,screen_new,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);
      back_buffer = create_bitmap_ex(8, display_cfg.screen_x,display_cfg.screen_y);
      blit(screen_old,back_buffer,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);
   }
   else{
      kill       = 0;
   }
#endif

   win_list[0].back_x      = dialog->x;
   win_list[0].back_y      = dialog->y;
   win_list[0].back_w      = dialog->w+1;
   win_list[0].back_h      = dialog->h+1;
   win_list[0].window      = create_bitmap_ex(8, win_list[0].back_w, win_list[0].back_h);

   screen = win_list[0].window;

   for(count=0; dialog[count].proc; count++){ 
      dialog[count].x -= win_list[0].back_x;
      dialog[count].y -= win_list[0].back_y;
   }

   win_count++;

   show_mouse(screen_new);

   player = rgui_init_dialog(dialog, focus_obj);

   oxygen_level=0;

   while(rgui_update_dialog(player)){
      if(oxygen_level==0){		// Gasp! Ack!
         dialog_oxygen();
      }
      oxygen_level=0;
   }

   win_count--;

   //show_mouse(mouse_screen);

   ta = rgui_shutdown_dialog(player);
#ifdef BUFFER_GUI
   if(kill!=0){
      show_mouse(NULL);
      dialog_oxygen();
      destroy_bitmap(screen_new);
      destroy_bitmap(back_buffer);
      screen     = screen_old;
      screen_new = NULL;
      back_buffer = NULL;
   }
#endif
   return ta;
}


// Make sure the gui doesn't suffocate
// this is the downside to a buffered gui

void dialog_oxygen(void)
{

#ifdef BUFFER_GUI
    if(win_count==0){

    if(screen_new){		// Ahhhhhhh.... That's better :)
       oxygen_level=1;
       blit(screen,screen_old,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);
    }

    }
    else{

    if(screen_new){		// Ahhhhhhh.... That's better :)

    show_mouse(NULL);

    if(back_buffer){		// Ahhhhhhh.... That's better :)
    blit(back_buffer,screen_new,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);
    }

    if(win_list[0].window){
       oxygen_level=1;
       blit(win_list[0].window,screen_new,0,0,win_list[0].back_x,win_list[0].back_y,win_list[0].back_w,win_list[0].back_h);
    }

    show_mouse(screen_new);

       oxygen_level=1;
       blit(screen_new,screen_old,0,0,0,0,display_cfg.screen_x,display_cfg.screen_y);
    }

    }
#endif

}



/* init_dialog:
 *  Sets up a dialog, returning a player object that can be used with
 *  the update_dialog() and shutdown_dialog() functions.
 */
DIALOG_PLAYER *rgui_init_dialog(DIALOG *dialog, int focus_obj)
{
   DIALOG_PLAYER *player = malloc(sizeof(DIALOG_PLAYER));
   int c;

   player->res = D_REDRAW;
   player->joy_on = TRUE;
   player->click_wait = FALSE;
   player->dialog = dialog;

   /* set up the global  dialog pointer */
   player->previous_dialog = active_dialog;
   active_dialog = dialog;

   /* set up dclick checking code */
   if (dclick_install_count <= 0) {
      LOCK_VARIABLE(dclick_status);
      LOCK_VARIABLE(dclick_time);
      LOCK_FUNCTION(dclick_check);
      install_int(dclick_check, 20);
      dclick_install_count = 1;
      //dclick_install_time = _allegro_count;
   }
   else
      dclick_install_count++;

   /* initialise the dialog */
   set_clip(screen, 0, 0, SCREEN_W-1, SCREEN_H-1);
   player->res |= dialog_message(dialog, MSG_START, 0, &player->obj);

   player->mouse_obj = find_mouse_object(dialog);
   if (player->mouse_obj >= 0)
      dialog[player->mouse_obj].flags |= D_GOTMOUSE;

   for (c=0; dialog[c].proc; c++)
      dialog[c].flags &= ~D_GOTFOCUS;

   if (focus_obj >= 0)
      c = focus_obj;
   else
      c = player->mouse_obj;

   if ((c >= 0) && ((SEND_MESSAGE(dialog+c, MSG_WANTFOCUS, 0)) & D_WANTFOCUS)) {
      dialog[c].flags |= D_GOTFOCUS;
      player->focus_obj = c;
   }
   else
      player->focus_obj = -1;

   return player;
}



/* update_dialog:
 *  Updates the status of a dialog object returned by init_dialog(),
 *  returning TRUE if it is still active or FALSE if it has finished.
 */
int rgui_update_dialog(DIALOG_PLAYER *player)
{
   int c, ch, r, ret, nowhere;

   if (player->res & D_CLOSE)
      return FALSE;

   /* need to reinstall the dclick handler? */
   /*if (dclick_install_time != _allegro_count) {
      install_int(dclick_check, 20);
      dclick_install_time = _allegro_count;
   }*/

   /* are we dealing with a mouse click? */
   if (player->click_wait) {
      if ((ABS(player->mouse_ox-rgui_mouse_x()) > 8) || 
	  (ABS(player->mouse_oy-rgui_mouse_y()) > 8))
	 dclick_status = DCLICK_NOT;

      /* waiting... */
      if ((dclick_status != DCLICK_AGAIN) && (dclick_status != DCLICK_NOT)) {
	 dialog_message(player->dialog, MSG_IDLE, 0, &nowhere);
	 return TRUE;
      }

      player->click_wait = FALSE;

      /* double click! */
      if ((dclick_status==DCLICK_AGAIN) &&
	  (rgui_mouse_x() >= player->dialog[player->mouse_obj].x) && 
	  (rgui_mouse_y() >= player->dialog[player->mouse_obj].y) &&
	  (rgui_mouse_x() <= player->dialog[player->mouse_obj].x + player->dialog[player->mouse_obj].w) &&
	  (rgui_mouse_y() <= player->dialog[player->mouse_obj].y + player->dialog[player->mouse_obj].h)) {
	 MESSAGE(player->mouse_obj, MSG_DCLICK, 0);
      }

      goto getout;
   }

   player->res &= ~D_USED_CHAR;

   /* need to draw it? */
   if (player->res & D_REDRAW) {
      player->res ^= D_REDRAW;
      player->res |= dialog_message(player->dialog, MSG_DRAW, 0, &player->obj);
   }

   /* need to give the input focus to someone? */
   if (player->res & D_WANTFOCUS) {
      player->res ^= D_WANTFOCUS;
      player->res |= offer_focus(player->dialog, player->obj, &player->focus_obj, FALSE);
   }

   /* has mouse object changed? */
   c = find_mouse_object(player->dialog);
   if (c != player->mouse_obj) {
      if (player->mouse_obj >= 0) {
	 player->dialog[player->mouse_obj].flags &= ~D_GOTMOUSE;
	 MESSAGE(player->mouse_obj, MSG_LOSTMOUSE, 0);
      }
      if (c >= 0) {
	 player->dialog[c].flags |= D_GOTMOUSE;
	 MESSAGE(c, MSG_GOTMOUSE, 0);
      }
      player->mouse_obj = c; 

      /* move the input focus as well? */
      if ((gui_mouse_focus) && (player->mouse_obj != player->focus_obj))
	 player->res |= offer_focus(player->dialog, player->mouse_obj, &player->focus_obj, TRUE);
   }

   /* deal with mouse button clicks */
   if (gui_mouse_b()) {
      player->res |= offer_focus(player->dialog, player->mouse_obj, &player->focus_obj, FALSE);

      if (player->mouse_obj >= 0) {
	 dclick_time = 0;
	 dclick_status = DCLICK_START;
	 player->mouse_ox = rgui_mouse_x();
	 player->mouse_oy = rgui_mouse_y();

	 /* send click message */
	 MESSAGE(player->mouse_obj, MSG_CLICK, gui_mouse_b());

	 if (player->res==D_O_K)
	    player->click_wait = TRUE;
      }
      else
	 dialog_message(player->dialog, MSG_IDLE, 0, &nowhere);

      goto getout;
   }

   /* fake joystick input by converting it to key presses */
   if (player->joy_on)
      rest(20);

   poll_joystick();

   if (player->joy_on) {
      if ((!joy[0].stick[0].axis[0].d1) && (!joy[0].stick[0].axis[0].d2) && 
	  (!joy[0].stick[0].axis[1].d1) && (!joy[0].stick[0].axis[1].d2) &&
	  (!joy[0].button[0].b) && (!joy[0].button[1].b)) {
	 player->joy_on = FALSE;
	 rest(20);
      }
      ch = 0;
   }
   else {
      if (joy[0].stick[0].axis[0].d1) {
	 ch = KEY_LEFT << 8;
	 player->joy_on = TRUE;
      }
      else if (joy[0].stick[0].axis[0].d2) {
	 ch = KEY_RIGHT << 8;
	 player->joy_on = TRUE;
      }
      else if (joy[0].stick[0].axis[1].d1) {
	 ch = KEY_UP << 8;
	 player->joy_on = TRUE;
      }
      else if (joy[0].stick[0].axis[1].d2) {
	 ch = KEY_DOWN << 8;
	 player->joy_on = TRUE;
      }
      else if ((joy[0].button[0].b) || (joy[0].button[1].b)) {
	 ch = (KEY_SPACE << 8) + ' ';
	 player->joy_on = TRUE;
      }
      else
	 ch = 0;
   }

   /* deal with keyboard input */
   if ((ch) || (keypressed())) {
      if (!ch)
	 ch = readkey();

      /* let object deal with the key? */
      if (player->focus_obj >= 0) {
	 MESSAGE(player->focus_obj, MSG_CHAR, ch);
	 if (player->res & D_USED_CHAR)
	    goto getout;
      }

      /* keyboard shortcut? */
      if (ch & 0xff) {
	 for (c=0; player->dialog[c].proc; c++) {
	    if (((tolower(player->dialog[c].key) == tolower((ch & 0xff))) ||
		 (((ch & 0xff) == 0) && (player->dialog[c].key == ch))) && 
		(!(player->dialog[c].flags & (D_HIDDEN | D_DISABLED)))) {
	       MESSAGE(c, MSG_KEY, ch);
	       ch = 0;
	       break;
	    }
	 }
	 if (!ch)
	    goto getout;
      }

      /* broadcast in case any other objects want it */
      for (c=0; player->dialog[c].proc; c++) {
	 if (!(player->dialog[c].flags & (D_HIDDEN | D_DISABLED))) {
	    MESSAGE(c, MSG_XCHAR, ch);
	    if (player->res & D_USED_CHAR)
	       goto getout;
	 }
      }

      /* pass <CR> or <SPACE> to selected object? */
      if ((((ch & 0xff) == 10) || ((ch & 0xff) == 13) || 
	   ((ch & 0xff) == 32)) && (player->focus_obj >= 0)) {
	 MESSAGE(player->focus_obj, MSG_KEY, ch);
	 goto getout;
      }

      /* ESC closes dialog? */
      if ((ch & 0xff) == 27) {
	 player->res |= D_CLOSE;
	 player->obj = -1;
	 goto getout;
      }

      /* move focus around the dialog? */
      player->res |= move_focus(player->dialog, ch, &player->focus_obj);
   }

   /* send idle messages */
   player->res |= dialog_message(player->dialog, MSG_IDLE, 0, &player->obj);

   getout:

   ret = (!(player->res & D_CLOSE));
   player->res &= ~D_CLOSE;
   return ret;
}



/* shutdown_dialog:
 *  Destroys a dialog object returned by init_dialog(), returning the index
 *  of the object that caused it to exit.
 */
int rgui_shutdown_dialog(DIALOG_PLAYER *player)
{
   int obj;

   /* send the finish messages */
   dialog_message(player->dialog, MSG_END, 0, &player->obj);

   /* remove the double click handler */
   dclick_install_count--;
   if (dclick_install_count <= 0)
      remove_int(dclick_check);

   if (player->mouse_obj >= 0)
      player->dialog[player->mouse_obj].flags &= ~D_GOTMOUSE;

   active_dialog = player->previous_dialog;

   obj = player->obj;

   free(player);

   return obj;
}

#endif
