/***************************************************************************

  2610intf.c

  The YM2610 emulator supports up to 2 chips.
  Each chip has the following connections:
  - Status Read / Control Write A
  - Port Read / Data Write A
  - Control Write B
  - Data Write B

***************************************************************************/

#include "sasound.h"

#ifdef BUILD_YM2610

//#define TEST_OPN

#ifdef TEST_OPN
#include "opn.h"
#endif

/* use FM.C with stream system */

extern unsigned char No_FM;

static int stream[MAX_2610];

static FMSAMPLE *Buf[YM2610_NUMBUF];

/* Global Interface holder */
static const struct YM2610interface *intf;

static void *Timer[MAX_2610][2];

/*------------------------- TM2610 -------------------------------*/
/* IRQ Handler */
static void IRQHandler(int n,int irq)
{
	if(intf->handler[n]) intf->handler[n](irq);
}

/* Timer overflow callback from timer.c */
static void timer_callback_2610(int param)
{
	int n=param&0x7f;
	int c=param>>7;

//	if(errorlog) fprintf(errorlog,"2610 TimerOver %d\n",c);
	Timer[n][c] = 0;
	YM2610TimerOver(n,c);
}

#ifndef __RAINE__
/* TimerHandler from fm.c */
static void TimerHandler(int n,int c,int count,double stepTime)
{
	if( count == 0 )
	{	/* Reset FM Timer */
		if( Timer[n][c] )
		{
//			if(errorlog) fprintf(errorlog,"2610 TimerReset %d\n",c);
	 		timer_remove (Timer[n][c]);
			Timer[n][c] = 0;
		}
	}
	else
	{	/* Start FM Timer */
		double timeSec = (double)count * stepTime;

		if( Timer[n][c] == 0 )
		{
			Timer[n][c] = timer_set (timeSec , (c<<7)|n, timer_callback_2610 );
		}
	}
}
#endif
static void FMTimerInit( void )
{
	int i;

	for( i = 0 ; i < MAX_2610 ; i++ )
		Timer[i][0] = Timer[i][1] = 0;
}

/* update request from fm.c */
void YM2610UpdateReq(int chip)
{
	stream_update(stream[chip],100);
}

/**** hardware start (YM2610) ****/
int YM2610_sh_start(const struct YM2610interface *interface ){

#ifdef TEST_OPN

   intf = interface;

   OPN_SetDebugHandler( NULL );

   if(OPN_Init(OPN_YM2610, 0x388, intf->num) == OPN_SUCCESS){
      OPN_SetClock(0, intf->baseclock);
      OPN_Reset(0);
      return 0;
   }
   else{
      return 1;
   }

#else

	int i,j;
	//int rate = Machine->sample_rate;
	int rate = audio_sample_rate;
	char buf[YM2610_NUMBUF][40];
	const char *name[YM2610_NUMBUF];
	int mixed_vol,vol[YM2610_NUMBUF],pan[YM2610_NUMBUF];

	intf = interface;
	if( intf->num > MAX_2610 ) return 1;

	if( AY8910_sh_start_ex((struct AY8910interface *)interface,"YM2610(SSG)") ) return 1;
        #ifdef HACKS
	print_debug("PSG OK\n");
        #endif

	/* FM init */
#if 0
	if( No_FM ) FMMode = CHIP_YM2610_DAC;
	else        FMMode = CHIP_YM2610_OPL;
#endif

	/* Timer Handler set */
	FMTimerInit();

	/* stream system initialize */
	for (i = 0;i < intf->num;i++)
	{
		/* stream setup */
		mixed_vol = intf->volumeFM[i];
		/* stream setup */
		for (j = 0 ; j < YM2610_NUMBUF ; j++)
		{
			char *chname;
			name[j]=buf[j];
			vol[j] = mixed_vol & 0xff;
			pan[j] = (mixed_vol>>8) & 0xff;
			mixed_vol>>=16;
			switch( pan[j] ){
			case OSD_PAN_CENTER:chname="Ct";break;
			case OSD_PAN_LEFT:  chname="Lt";break;
			case OSD_PAN_RIGHT: chname="Rt";break;
			default:            chname="??";break;
			}
			sprintf(buf[j],"YM2610 #%d Ch%d(%s)",i,j+1,chname);
		}
		stream[i] = stream_init_multi(YM2610_NUMBUF,name,rate,FM_OUTPUT_BIT,i,YM2610UpdateOne);
		/* volume setup */
		for (j = 0 ; j < YM2610_NUMBUF ; j++)
		{
			stream_set_volume(stream[i]+j,vol[j]);
			stream_set_pan(stream[i]+j,pan[j]);
		}
	}

	/**** initialize YM2610 ****/
	if (YM2610Init(intf->num,intf->baseclock,rate, intf->pcmroma, intf->pcmromb,0,0) == 0)
		return 0;

	/* error */
	return 1;

#endif
}

/**** hardware start (YM2610B) ****/
int YM2610B_sh_start(const struct YM2610interface *interface ){

#ifdef TEST_OPN

   intf = interface;

   OPN_SetDebugHandler( NULL );

   if(OPN_Init(OPN_YM2610, 0x388, intf->num) == OPN_SUCCESS){
      OPN_SetClock(0, intf->baseclock);
      OPN_Reset(0);
      return 0;
   }
   else{
      return 1;
   }

#else
	int i,j;
	//int rate = Machine->sample_rate;
	int rate = audio_sample_rate;
	char buf[YM2610_NUMBUF][40];
	const char *name[YM2610_NUMBUF];
	int mixed_vol,vol[YM2610_NUMBUF],pan[YM2610_NUMBUF];

	intf = interface;
	if( intf->num > MAX_2610 ) return 1;

	if( AY8910_sh_start_ex((struct AY8910interface *)interface,"YM2610(SSG)") ) return 1;
        #ifdef HACKS
	print_debug("PSG OK\n");
        #endif

	/* FM init */
#if 0
	if( No_FM ) FMMode = CHIP_YM2610_DAC;
	else        FMMode = CHIP_YM2610_OPL;
#endif

	/* Timer Handler set */
	FMTimerInit();

	/* stream system initialize */
	for (i = 0;i < intf->num;i++)
	{
		/* stream setup */
		mixed_vol = intf->volumeFM[i];
		/* stream setup */
		for (j = 0 ; j < YM2610_NUMBUF ; j++)
		{
			char *chname;
			name[j]=buf[j];
			vol[j] = mixed_vol & 0xff;
			pan[j] = (mixed_vol>>8) & 0xff;
			mixed_vol>>=16;
			switch( pan[j] ){
			case OSD_PAN_CENTER:chname="Ct";break;
			case OSD_PAN_LEFT:  chname="Lt";break;
			case OSD_PAN_RIGHT: chname="Rt";break;
			default:            chname="??";break;
			}
			sprintf(buf[j],"YM2610 #%d Ch%d(%s)",i,j+1,chname);
		}
		stream[i] = stream_init_multi(YM2610_NUMBUF,name,rate,FM_OUTPUT_BIT,i,YM2610BUpdateOne);
		/* volume setup */
		for (j = 0 ; j < YM2610_NUMBUF ; j++)
		{
			stream_set_volume(stream[i]+j,vol[j]);
			stream_set_pan(stream[i]+j,pan[j]);
		}
	}

	/**** initialize YM2610 ****/
	if (YM2610Init(intf->num,intf->baseclock,rate, intf->pcmroma, intf->pcmromb,0,0) == 0)
		return 0;

	/* error */
	return 1;
#endif
}

/************************************************/
/* Sound Hardware Stop                          */
/************************************************/
void YM2610_sh_stop(void)
{
#ifdef TEST_OPN

   OPN_Shutdown();

#else

  //AY8910_sh_stop();
  YM2610Shutdown();

#endif
}

/************************************************/
/* Status Read for YM2610 - Chip 0              */
/************************************************/
int YM2610_status_port_0_A_r( int offset )
{
#ifdef TEST_OPN

   static int ta;
   ta^=255;
   return ta;

#else

   static int ta;

   if(RaineSoundCard>0){
      return YM2610Read(0,0);
   }
   else{
      ta^=255;
      return ta;
   }

#endif
}

int YM2610_status_port_0_B_r( int offset )
{
#ifdef TEST_OPN

   static int ta;
   ta^=255;
   return ta;

#else

   static int ta;
   if(RaineSoundCard>0){
      return YM2610Read(0,2);
   }
   else{
      ta^=255;
      return ta;
   }

#endif
}

/************************************************/
/* Port Read for YM2610 - Chip 0                */
/************************************************/
int YM2610_read_port_0_r( int offset )
{
#ifdef TEST_OPN

   static int ta;
   ta^=255;
   return ta;

#else

   static int ta;
   if(RaineSoundCard>0){
      return YM2610Read(0,1);
   }
   else{
      ta^=255;
      return ta;
   }

#endif
}

/************************************************/
/* Control Write for YM2610 - Chip 0            */
/* Consists of 2 addresses                      */
/************************************************/
void YM2610_control_port_0_A_w(int offset,int data)
{
#ifdef TEST_OPN

   YM2610_WriteReg(0,data);

#else

   if(RaineSoundCard>0) YM2610Write(0,0,data);

#endif
}

void YM2610_control_port_0_B_w(int offset,int data)
{
#ifdef TEST_OPN

   YM2610_WriteReg(2,data);

#else

   if(RaineSoundCard>0) YM2610Write(0,2,data);

#endif
}

/************************************************/
/* Data Write for YM2610 - Chip 0               */
/* Consists of 2 addresses                      */
/************************************************/
void YM2610_data_port_0_A_w(int offset,int data)
{
#ifdef TEST_OPN

   YM2610_WriteReg(1,data);

#else

   if(RaineSoundCard>0) YM2610Write(0,1,data);

#endif
}

void YM2610_data_port_0_B_w(int offset,int data)
{
#ifdef TEST_OPN

   YM2610_WriteReg(3,data);

#else

   if(RaineSoundCard>0)	YM2610Write(0,3,data);

#endif
}

/************************************************/
/* Sound Hardware Update                        */
/************************************************/
void YM2610_sh_update(void)
{
}

/**************************************************/
/*   YM2610 left/right position change (TAITO)    */
/**************************************************/
int YM2610_get_stream_num( int num )
{
  return stream[num];
}

/**************** end of file ****************/

#endif
