/* esp.h
 */

#ifndef __ESP_INCLUDED__

#ifdef __cplusplus
extern "C" {
#endif

/* 24-bit Word (bit 31-24 not used) */

typedef unsigned long	WORD;

/* 48-bit Instruction (bit 63-48 not used) */

typedef struct {

	unsigned long	most_significant_bits;
	unsigned long	less_significant_bits;

} INSTRUCTION;

/* Ensoniq Signal Processor State */

typedef struct {

	WORD	        registers[256];
	INSTRUCTION     instructions[160];

        int     host_control;
        int     ram_access_control;

        int     program_counter;
        int     internal_refresh_counter;
        int     host_serial_control;

        int     frame_counter;

        int     register_23_16;
        int     register_15_8;
        int     register_7_0;

        int     instruction_47_40;
        int     instruction_39_32;
        int     instruction_31_24;
        int     instruction_23_16;
        int     instruction_15_8;
        int     instruction_7_3;

        int     dil_23_16;
        int     dil_15_8;
        int     dil_7_0;

        int     dol_23_16;
        int     dol_15_8;
        int     dol_7_0;

        int     dadr_23_16;
        int     dadr_15_8;
        int     dadr_7_4;

        int     read_select_register_instruction;
        int     write_select_register;
        int     write_select_instruction;
        int     write_select_register_instruction;

} ESP;

extern void     InitializeEnsoniqSignalProcessor (ESP *esp);
extern void     ShutdownEnsoniqSignalProcessor (ESP *esp);

extern int      ReadEnsoniqSignalProcessor (ESP *esp, int address);
extern void     WriteEnsoniqSignalProcessor (ESP *esp, int address, int data);
extern int 	ExecuteInstruction (ESP *esp);
extern void     RunEnsoniqSignalProcessor (ESP *esp);

#ifdef __cplusplus
}
#endif

#define __ESP_INCLUDED__
#endif