/**************************************************************/
/*   M6295 control                                            */
/**************************************************************/

#include "raine.h"
#include "sasound.h"
#include "m6295.h"

/*

30-06-99:

Replaced ADSAMPLE with something smaller - M6295_SAMPLE.

20-05-99:

Changed external interface, bankcount is removed and rom info is moved into
M6295_ROM structs.
Changed internal interface, samp info is moved to M6295_ROM_DECODE structs.
Now, there is no real limit on number of rom banks! Bomb Jack Twin is a good
example of why this is needed (2 M6295, with 16 banks each).

18-04-99:

Changed rom decode volume boost, originally it was *8, which is too quiet and
was only giving a range of -0x4000 to 0x3FFF.
I think it might be nice to put a volume control in M6295buffer_interface[].

*/

#define DECODE_VOL_SCALE	(24)

static INLINE int aLimit( int val, int max, int min )
{
  if ( val > max )
    val = max;
  else if ( val < min )
    val = min;
  return val;
}

M6295Rec m6295_entry[M6295BF_MAX], *m6295;

static int chip_max;
static int sample_len;
static int emu_rate;
static int stream[M6295BF_MAX];

#define UPDATE_SHIFT (8)

/**************************************************/
/*    decode control                              */
/**************************************************/
static int ad_sample;
static int ad_index;
static int ad_index_table[8] = {
  -1, -1, -1, -1, 2, 4, 6, 8
};
static int ad_offset_table[49]= {
  0x0010, 0x0011, 0x0013, 0x0015, 0x0017, 0x0019, 0x001c, 0x001f,
  0x0022, 0x0025, 0x0029, 0x002d, 0x0032, 0x0037, 0x003c, 0x0042,
  0x0049, 0x0050, 0x0058, 0x0061, 0x006b, 0x0076, 0x0082, 0x008f,
  0x009d, 0x00ad, 0x00be, 0x00d1, 0x00e6, 0x00fd, 0x0117, 0x0133,
  0x0151, 0x0173, 0x0198, 0x01c1, 0x01ee, 0x0220, 0x0256, 0x0292,
  0x02d4, 0x031c, 0x036c, 0x03c3, 0x0424, 0x048e, 0x0502, 0x0583,
  0x0610
};

static void ad_decode(int ad_code){
  int ad_base;
  int ad_offset;

  ad_base   = ad_offset_table[ad_index];
  ad_offset = ad_base/8;
  
  if( ad_code&0x01 )  ad_offset += ad_base>>2;
  if( ad_code&0x02 )  ad_offset += ad_base>>1;
  if( ad_code&0x04 )  ad_offset += ad_base;
  if( ad_code&0x08 )  ad_offset  = 0 - ad_offset;

  ad_sample = aLimit( ad_sample + ad_offset , 0x7FF, -0x800 );
  ad_index = aLimit( ad_index + ad_index_table[ad_code&7], 48, 0 );
}

/**************************************************/
/*    rom decode                                  */
/**************************************************/
static unsigned int err_table;
static int   decode_table_count;
static unsigned int *decode_table;
static int   decode_area_size;
static WORD *decode_area;

void M6295_decode( int i, struct M6295_SAMPLE *samp, UBYTE *rom, ULONG romsize )
{
  ULONG start, stop;
  UBYTE *base;
  signed short *ad_data;

  /**** table initial ****/
  samp->len        = 0;
  samp->data       = NULL;

  /*** decode start ****/
  //if( i*8 >= err_table )  return;/* not sample table */
  base  = rom + i*8;
  start = (base[0] << 16) + (base[1] << 8) + base[2];
  stop  = (base[3] << 16) + (base[4] << 8) + base[5];
  //if( start >= romsize )  start = romsize - 1;
  if( stop >= romsize )  stop = romsize;

#ifdef HACKS
  print_debug( "Now decode No.%d %08x-%08x\n", i, start, stop );
#endif
  if((start<romsize) && (stop<romsize)){	// Add Antiriad (Kick Off needs it)
    
    if( (start < stop) &&
	!(start == 0x00808080 || start == 0x00080808 ||
	  stop == 0x00808080 || stop == 0x00080808) ){
      /**** low limit address check ****/
      if( start < err_table )  err_table = start;
      samp->len = (stop - start)*2; /* get length (1byte = 2 data) */
      /**** new decode routine ****/
      {
	int  count, j;
	count = decode_table_count;
	for( j = 0; j < count; j++ ){
#if 0
	  if( start == *(decode_table + j*3) && stop <= *(decode_table + j*3 + 1) ){
#ifdef HACKS
	    print_debug( "Get decode table (%08x - %08x) [%08x-%08x]\n", start, stop, *(decode_table + j*3),  *(decode_table + j*3 + 1) );
#endif
	    samp->data = (void *)(*(decode_table + j*3 + 2)|0x80000000);
	    return;
	  }
#else
	  if( start >= *(decode_table + j*3) && stop <= *(decode_table + j*3 + 1) ){
#ifdef HACKS
	    print_debug( "Get decode table (%08x - %08x) [%08x-%08x]\n", start, stop, *(decode_table + j*3),  *(decode_table + j*3 + 1) );
#endif
	    samp->data = (void *)((*(decode_table + j*3 + 2) + (start - *(decode_table + j*3))*2)|0x80000000 );
	    return;
	  }
#endif
	}
	/**** decode start ****/
	*(decode_table + count*3)     = start;
	*(decode_table + count*3 + 1) = stop;
	*(decode_table + count*3 + 2) = decode_area_size;
	decode_table_count++;
#ifdef HACKS
	print_debug( "New decode Address (%02x:%08x - %08x)\n", count, *(decode_table + count*3), *(decode_table + count*3 + 1) );
#endif
	decode_area = realloc( decode_area, (decode_area_size + samp->len) * sizeof(WORD) );
	ad_data = (WORD *)(decode_area + decode_area_size);
	samp->data = (void *)(decode_area_size|0x80000000); /* rom offset */
	decode_area_size += samp->len;

	rom += start;
	ad_sample = 0;		/* work init */
	ad_index = 0;		/* work init */
	for( start = 0; start < samp->len/2; start++ ){
	  ad_decode( (*rom)>>4 );
	  *ad_data++ = (WORD) aLimit(ad_sample * DECODE_VOL_SCALE, 0x7FFF, -0x8000);
	  ad_decode( (*rom)&0x0f );
	  *ad_data++ = (WORD) aLimit(ad_sample * DECODE_VOL_SCALE, 0x7FFF, -0x8000);
	  rom++;
	}
      }
    }
  }
}

/**************************************************/
/*    sh start                                    */
/**************************************************/
int M6295buffer_sh_start( struct M6295buffer_interface *intf )
{
  int  i, j, k;
  char buf[40];
  int ta;
  M6295_ROM *rom;
  M6295_ROM_DECODE *rom_decode;

  int rate = audio_sample_rate;

  chip_max = intf->num;
  sample_len  = rate / 60;
  emu_rate = sample_len * 60;

  for( i = 0; i < intf->num; i++ ){
    m6295_entry[i].rate    = intf->sample_rate[i];
    m6295_entry[i].command = -1;
    m6295_entry[i].add     = (unsigned int)(((float)intf->sample_rate[i] / (float)emu_rate)*(1<<UPDATE_SHIFT));
    for( j = 0; j < M6295BF_CHANNEL; j++ ){
      m6295_entry[i].offset[j]  = (unsigned int)-1;
      m6295_entry[i].req_len[j] = 0;
      m6295_entry[i].req_buf[j] = NULL;
    }
    m6295_entry[i].bank = 0;

    // rom bank base pointers

    rom = intf->rom[i];

    // Calculate rom bank count

    ta=0;
    while(rom[ta].data){
       ta++;
    }
    m6295_entry[i].bankmax = ta;

    // Make decode rom data

    m6295_entry[i].rom_decode = (M6295_ROM_DECODE *) malloc(m6295_entry[i].bankmax * sizeof(M6295_ROM_DECODE));

    rom_decode = m6295_entry[i].rom_decode;

    // malloc error

    if(!(rom_decode)){
       m6295_entry[i].bankmax = 0;
    }

    // Decode rom banks

    for( k = 0; k < m6295_entry[i].bankmax; k++ ){

      err_table = rom[k].size; /* initial error limit address */

      rom_decode[k].decode_table = malloc( (sizeof(unsigned int) * 0x80 * 3) );

      if( rom_decode[k].decode_table ){
	decode_table = (unsigned int *) rom_decode[k].decode_table;
	decode_table_count = 0;
	decode_area  = NULL;
	decode_area_size = 0;
	for( j = 0; j < 0x80; j++ ){
	  *(decode_table + j*3)     = 0xffffffff;
	  *(decode_table + j*3 + 1) = 0xffffffff;
	  *(decode_table + j*3 + 2) = 0x00000000;
	}
	for( j = 0; j < 0x80; j++ )  M6295_decode( j, &(rom_decode[k].samp[j]), rom[k].data, rom[k].size );

	rom_decode[k].decode_rom = decode_area;

	if( decode_area ){
	  for( j = 0; j < 0x80; j++ ){
	    if( ((unsigned int) rom_decode[k].samp[j].data)&0x80000000 )
	      rom_decode[k].samp[j].data = decode_area + (((unsigned int) rom_decode[k].samp[j].data )&0x7fffffff);
	    /**** 0x8xxxxxxx is use flag ****/
	  }
#ifdef HACKS
	  print_debug( "Now Get decode area size = %08x[#%d:#%d]\n", decode_area_size * sizeof(WORD), i, k );
#endif
	}
      }
    }
    /* stream setup */
    sprintf( buf, "M6295 #%d", i );
    stream[i] = stream_init( buf, emu_rate, 16, i, M6295buffer_UpdateOne );
  }
  return 0;
}

/**************************************************/
/*    sh end                                      */
/**************************************************/
void M6295buffer_sh_stop( void )
{
  M6295_ROM_DECODE *rom_decode;
  int i, k;

  for( i = 0; i < chip_max; i++ ){

    rom_decode = m6295_entry[i].rom_decode;

    if(rom_decode){

    for( k = 0; k < m6295_entry[i].bankmax; k++ ){

      if( rom_decode[k].decode_table ) free( rom_decode[k].decode_table );
      if(rom_decode[k].decode_rom)     free( rom_decode[k].decode_rom );

    }

    free(rom_decode);

    }

  }

  chip_max = 0;			// ADDED FOR SAFETY!!!
}

/**************************************************/
/*    request                                     */
/**************************************************/
void M6295buffer_request( int num, int code )
{
  M6295_ROM_DECODE *rom_decode;
  int bank;
  if( num >= chip_max ) return;
  m6295 = &m6295_entry[num];
  bank = m6295->bank;

  if( m6295->command != -1 ){
    int i, temp = code >> 4;
    for( i = 0; i < M6295BF_CHANNEL; i++ ){
      if( temp&1 ){

        rom_decode = m6295->rom_decode;

        if(rom_decode){
	   m6295->req_buf[i] = (signed short *)(rom_decode[bank].samp[m6295->command]).data;
	   m6295->req_len[i] = (rom_decode[bank].samp[m6295->command]).len << UPDATE_SHIFT;
	   m6295->offset[i]  = 0;	/* request */
        }
        else{
           m6295->req_buf[i] = NULL;
        }
	if( m6295->req_buf[i] == NULL ){
	  m6295->offset[i]  = -1;
#ifdef HACKS
	  if( m6295->command )  print_ingame(120,"Now M6295(%d) not samp[%02x]", num, m6295->command );
#endif
	}
      }
      temp >>= 1;
    }
    m6295->command = -1;
  } else if( code&0x80 ){
    m6295->command = code&0x7f;
  } else{
    /**** stop command ****/
    int i, temp = code >> 3;
    for( i = 0; i < M6295BF_CHANNEL; i++ ){
      if( temp&1 )  m6295->offset[i]  = (unsigned int)-1;	/* stop */
      temp >>= 1;
    }
  }
}

/**************************************************/
/*    read status                                 */
/**************************************************/
int  M6295buffer_status( int num ){
  int i, ret;
  /* range check the numbers */
  if( num >= chip_max )    return 0x0f;

  m6295 = &m6295_entry[num];
  ret = 0;
  for( i = 0; i < 4; i++ ){
    if( m6295->offset[i] != (unsigned int)-1 )  ret |= 1<<i;
  }
  return ret;
}
/**************************************************/
/*    bank change                                 */
/**************************************************/
void M6295buffer_bankswitch( int num, int bank ){
  /* range check the numbers */
  if( num >= chip_max )    return;
  m6295 = &m6295_entry[num];
  m6295->bank = bank < m6295->bankmax ? bank : m6295->bankmax;
}

/**************************************************/
/*    update                                      */
/**************************************************/
void M6295buffer_UpdateOne( int num, void *buffer, int length ){
  int  chan;
  int  i;
  int  data;
  signed short *sp;

  sp = (signed short *)buffer;
  m6295 = &m6295_entry[num];

  for( i = 0; i < length; i++ ){
    data = 0;
    for( chan = 0; chan < M6295BF_CHANNEL; chan++ ){
      if( m6295->offset[chan] != (unsigned int)-1 ){
	data += *( m6295->req_buf[chan] + (m6295->offset[chan] >> UPDATE_SHIFT) );
	if( (m6295->offset[chan] += m6295->add) >= m6295->req_len[chan] )
	  m6295->offset[chan] = (unsigned int)-1;
      }
    }
    *sp++ = aLimit( data, 32767, -32768 );
  }
}
/**************** end of file ****************/
