/* Yamaha FM/OP-N emulator v1.01
 *
 * Copyright (c) 1999 Richard Mitton
 * All unauthorised distribution prohibited.
 */

/******* CUSTOMISE THIS BIT *******/

#ifdef __RAINE__

   #include "deftypes.h"

   #define byte UBYTE
   #define SCALE_FREQUENCY

#else

   /* 'byte' should be an 8-bit unsigned integer type */
   typedef unsigned char byte;

   /* TRUE should be define to boolean true (non-zero) */
   #ifndef TRUE
   #define TRUE 1
   #endif

   /* FALSE should be define to boolean false (zero) */
   #ifndef FALSE
   #define FALSE 0
   #endif

   /* If SCALE_FREQUENCY is #defined, then the emulator will attempt to
    * compensate for the frequency differences between OPL and OPN.
    * Un-#define to remove support for this.
    */
   #define SCALE_FREQUENCY

#endif

/******* END OF CUSTOMISATION *******/


/* The OPL3 clock rate in Hz */
#define OPL3_CLOCK 3579545

/* The chips we can emulate */
#define OPN_YM2203   0
#define OPN_YM2610   1
#define OPN_YM2612   2

/* Return value from OPN_Init */
#define OPN_SUCCESS  0
#define OPN_TOOMANY  1
#define OPN_NO_OPL3  2

/* A virtual OPN */
typedef struct {
   byte reg[0x100], reg_latch;
   int hertz, scale_factor, ch_offset, op_offset;
   void *timer_a_event, *timer_b_event;
   void (*IRQHandler)(int n);
} OPN;

/* Generic functions */
int OPN_Init(int type, int base, int num);
void OPN_SetClock(int num, int hz);
byte OPN_ReadStatus(int num);
void OPN_Reset(int num);
void OPN_Shutdown(void);
void OPN_SetDebugHandler(void (*func)(char *));

/* Chip specific functions */
void YM2203_WriteReg(int num, int addr, byte value);
void YM2610_WriteReg(int addr, byte value);
void YM2612_WriteReg(int addr, byte value);
