/******************************************************************************/
/*                                                                            */
/*                     SOME EXTRA STUFF FOR HARDWARE YM3812                   */
/*                                                                            */
/******************************************************************************/

#include "deftypes.h"
#include "ym3812x.h"
#include "3812intf.h"

//#ifndef RAINE_DOS
#include <sys/io.h>
#define outportb(a,b)	outb(b,a)
#define inportb		inb

//#endif

#define ADLIB_BASE	0x388

static UBYTE raine_ym3812_regs[0x100];
static UBYTE raine_ym3812_port;
static UBYTE raine_ym3812_active;
static UBYTE raine_ym3812_paused;

static void fm_write(UBYTE reg, UBYTE data)
{
   int i;

   i=35;
   do{
      inportb(ADLIB_BASE);
   }while(--i);

   outportb(ADLIB_BASE, reg);

   i=6;
   do{
      inportb(ADLIB_BASE);
   }while(--i);

   outportb(ADLIB_BASE+1, data);
}

static void mute_raine_ym3812(void)
{
   UBYTE i;

   for(i=0x40;i<0x55;i++)
      fm_write(i, 0x3F);	// clear volume (0=max 0x3F=min) [thx cgfm]

   for(i=0;i<0x40;i++)
      fm_write(i, 0x00);	// clear registers

   for(i=0x55;i<0xFF;i++)
      fm_write(i, 0x00);	// clear registers
}

void start_raine_ym3812(void)
{
   int i;

   mute_raine_ym3812();

   raine_ym3812_active = 1;
   raine_ym3812_port = 0;

   for(i=0;i<0xFF;i++)
      raine_ym3812_regs[i]= 0;	// clear register cache

   raine_ym3812_paused = 0;
}

void stop_raine_ym3812(void)
{
   mute_raine_ym3812();

   raine_ym3812_active = 0;
}

void pause_raine_ym3812(void)
{
   if(raine_ym3812_active!=0){

      mute_raine_ym3812();

      raine_ym3812_paused = 1;
   }
}

void unpause_raine_ym3812(void)
{
   UBYTE i;

   if(raine_ym3812_active){

      for(i=0;i<0x40;i++)
         fm_write(i, raine_ym3812_regs[i]);	// restore registers

      for(i=0x55;i<0xFF;i++)
         fm_write(i, raine_ym3812_regs[i]);	// restore registers

      for(i=0x40;i<0x55;i++)
         fm_write(i, raine_ym3812_regs[i]);	// restore volume (0=max 0x3F=min) [thx cgfm]

      raine_ym3812_paused = 0;
   }
}

void osd_ym3812_control(UBYTE data)
{
   raine_ym3812_port = data;

   if(raine_ym3812_paused!=0)
      unpause_raine_ym3812();

   outportb(ADLIB_BASE, data);
}

void osd_ym3812_write(UBYTE data)
{
   raine_ym3812_regs[raine_ym3812_port] = data;

   outportb(ADLIB_BASE+1, data);
}

// YM3812 Handler
// --------------

UBYTE YM3812ReadZ80(UWORD offset)
{
   if((offset&1)==0)
      return YM3812_status_port_0_r(offset)|0x60;	// Snow Bros fix/no timer emulation :(.
   else
      return YM3812_read_port_0_r(offset);
}

void YM3812WriteZ80(UWORD offset, UBYTE data)
{
   if((offset&1)==0)
      YM3812_control_port_0_w(offset,data);
   else
      YM3812_write_port_0_w(offset,data);
}

