/******************************************************************************/
/*                                                                            */
/*            YAMAHA YMZ280B - PCMD8 - 8 CHANNEL PCM/ADPCM DECODER            */
/*                                                                            */
/******************************************************************************/

#include "raine.h"
#include "sasound.h"
#include "ymz280b.h"
#include "smp16bit.h"

static DEF_INLINE int aLimit( int val, int max, int min )
{
  if ( val > max )
    val = max;
  else if ( val < min )
    val = min;
  return val;
}

ymz280bRec ymz280b_entry[YMZ280B_MAX], *ymz280b;

static int chip_max;
static int sample_len;
static int emu_rate;
static int stream[YMZ280B_MAX];

#define UPDATE_SHIFT (8)

/**************************************************/
/*    sh start                                    */
/**************************************************/

int ymz280b_sh_start( struct ymz280b_interface *intf )
{
  char buf[40];
  int ta, i;
  int rate;
  ymz280b_ROM *rom;
  ymz280b_SAMPLE *sample_decode;

  rate       = audio_sample_rate;
  chip_max   = intf->num;
  sample_len = rate / 60;
  emu_rate   = sample_len * 60;

  for( i = 0; i < chip_max; i++ ){

    ymz280b_entry[i].rate    = intf->sample_rate[i];
    ymz280b_entry[i].add     = (ULONG) (((float)intf->sample_rate[i] / (float)emu_rate)*(1<<UPDATE_SHIFT));
    ymz280b_entry[i].offset  = (ULONG) -1;
    ymz280b_entry[i].req_len = 0;
    ymz280b_entry[i].req_buf = NULL;
    ymz280b_entry[i].current_bank = 0;

    // rom bank base pointers

    rom = intf->rom[i];

    // Calculate sample bank count

    ta=0;
    while(rom[ta].data){
       ta++;
    }
    ymz280b_entry[i].banks = ta;

    // Make decode rom data

    sample_decode = (ymz280b_SAMPLE *) malloc(ymz280b_entry[i].banks * sizeof(ymz280b_SAMPLE));

    ymz280b_entry[i].sample_decode = sample_decode;

    // malloc error

    if(!sample_decode)
       ymz280b_entry[i].banks = 0;

    // Decode them

    for(ta = 0; ta < ymz280b_entry[i].banks; ta++){

       sample_decode[ta].len  = (ULONG)   (rom[ta].size);
       sample_decode[ta].data = (UBYTE *) (rom[ta].data);

    }

    // Stream setup

    sprintf( buf, "ymz280b #%d", i );
    stream[i] = stream_init( buf, emu_rate, 16, i, ymz280b_UpdateOne );
  }

  return 0;
}

/**************************************************/
/*    sh end                                      */
/**************************************************/

void ymz280b_sh_stop( void )
{
  int ta,i;
  ymz280b_SAMPLE *sample_decode;

  for(i = 0; i < chip_max; i++){

    sample_decode = ymz280b_entry[i].sample_decode;

    for(ta = 0; ta < ymz280b_entry[i].banks; ta++){

       if(sample_decode[ta].data) free(sample_decode[ta].data);

    }

  }

  chip_max = 0;
}

static int ad_sample;
static int ad_index;

/**************************************************/
/*    rom bank                                    */
/**************************************************/

void ymz280b_bank_write(int num, int bank)
{
   if( num >= chip_max ) return;

   ymz280b = &ymz280b_entry[num];

   if(bank >= ymz280b->banks){
      #ifdef HACKS
      print_ingame(1200,"ymz280b(%d) Bank out of range:%02x", num, bank );
      #endif
      return;
   }

   ymz280b->current_bank = bank;
}

/**************************************************/
/*    request                                     */
/**************************************************/

void ymz280b_request(int num, int offset, int data)
{
   int ta,tb;

   ymz280b_SAMPLE *sample_decode;

   if( num >= chip_max ) return;

   ymz280b = &ymz280b_entry[num];

   offset &= 7;

   ymz280b->command[offset] = data;

   if(offset==6){	// Trigger

   sample_decode = ymz280b->sample_decode;

   ymz280b->req_buf = NULL;

   ta=0;
   tb=0;

   if(sample_decode){
      ad_sample = 0;
      ad_index = 0;
      ta = ((ymz280b->command[2]<<16) | (ymz280b->command[1]<<8) | (ymz280b->command[0]<<0)) & 0x0FFFFF;
      tb = ((ymz280b->command[5]<<16) | (ymz280b->command[4]<<8) | (ymz280b->command[3]<<0)) & 0x0FFFFF;

      if( (tb > ta) &&
      (ta < sample_decode[ymz280b->current_bank].len) &&
      (tb < sample_decode[ymz280b->current_bank].len) ){

      ymz280b->req_buf = ((UBYTE *) (sample_decode[ymz280b->current_bank].data)) + ta;

      ymz280b->offset  = 0;

      ymz280b->req_len = tb - ta;

      ymz280b->req_len <<= 1;

      ymz280b->req_len <<= UPDATE_SHIFT;

      }

   }

   if(!(ymz280b->req_buf)){
      ymz280b->offset = (ULONG) -1;
      #ifdef HACKS
      print_ingame(1200,"ymz280b(%d) out of range: %06x %06x", num, ta, tb );
      #endif
   }
   else{
      /*#ifdef HACKS
      print_ingame(1200,"ymz280b(%d) %06x %06x", num, ta, tb );
      #endif*/
   }

   }
}

/**************************************************/
/*    read status                                 */
/**************************************************/

// 0 = not playing
// 1 = playing

int ymz280b_status( int num )
{
  int ret;

  if( num >= chip_max ) return 0;

  ret = 0;

  ymz280b = &ymz280b_entry[num];

  if( ymz280b->offset != (ULONG) -1 ) ret |= 1;

  return ret;
}

/**************************************************/
/*    update                                      */
/**************************************************/

static int ad_index_table[8] =
{
  -1, -1, -1, -1, 2, 4, 6, 8
};

static int ad_offset_table[49] =
{
  0x0010, 0x0011, 0x0013, 0x0015, 0x0017, 0x0019, 0x001c, 0x001f,
  0x0022, 0x0025, 0x0029, 0x002d, 0x0032, 0x0037, 0x003c, 0x0042,
  0x0049, 0x0050, 0x0058, 0x0061, 0x006b, 0x0076, 0x0082, 0x008f,
  0x009d, 0x00ad, 0x00be, 0x00d1, 0x00e6, 0x00fd, 0x0117, 0x0133,
  0x0151, 0x0173, 0x0198, 0x01c1, 0x01ee, 0x0220, 0x0256, 0x0292,
  0x02d4, 0x031c, 0x036c, 0x03c3, 0x0424, 0x048e, 0x0502, 0x0583,
  0x0610
};

static void ad_decode(int ad_code)
{
  int ad_base;
  int ad_offset;

  ad_base   = ad_offset_table[ad_index];
  ad_offset = ad_base/8;
  
  if( ad_code&0x01 ) ad_offset += ad_base>>2;
  if( ad_code&0x02 ) ad_offset += ad_base>>1;
  if( ad_code&0x04 ) ad_offset += ad_base;
  if( ad_code&0x08 ) ad_offset  = 0 - ad_offset;

  ad_sample = aLimit( ad_sample + ad_offset , 0x7FF, -0x800 );
  ad_index = aLimit( ad_index + ad_index_table[ad_code&7], 48, 0 );
}

void ymz280b_UpdateOne( int num, void *buffer, int length )
{
   int data,ta;
   WORD *sp;

   if(!length) return;

   sp = (WORD *) buffer;

   ymz280b = &ymz280b_entry[num];

   do{
      data = 0;
      if(ymz280b->offset != (ULONG) -1){

         if( (ymz280b->offset + ymz280b->add) >= ymz280b->req_len ){
            ymz280b->offset = (ULONG) -1;
         }
         else{

         for(ta = (ymz280b->offset >> UPDATE_SHIFT); ta<((ymz280b->offset + ymz280b->add) >> UPDATE_SHIFT); ta++){

         if((ta&1)==0)
            ad_decode((ymz280b->req_buf[ta>>1]>>4)&0x0F);
         else
            ad_decode((ymz280b->req_buf[ta>>1]>>0)&0x0F);

         }

         ymz280b->offset += ymz280b->add;

         data += (((ad_sample * 16) * 80) / 16);

         }

      }
      *sp++ = aLimit( data, 0x7FFF, -0x8000 );
   }while(--length);
}

/**************** end of file ****************/
