/******************************************************************************/
/*                                                                            */
/*                           RAINE COLOUR MAPPING                             */
/*                                                                            */
/******************************************************************************/

#ifndef _palette_h_
#define _palette_h_

#include "deftypes.h"

typedef void CMAP_FUNC(ULONG bank, ULONG cols);

typedef struct COLOUR_MAPPER
{
   UBYTE     *name;
   CMAP_FUNC *mapper_8bpp;
   CMAP_FUNC *mapper_15bpp;
   CMAP_FUNC *mapper_16bpp;
   CMAP_FUNC *mapper_24bpp;
   CMAP_FUNC *mapper_32bpp;
} COLOUR_MAPPER;

CMAP_FUNC *current_mapper;

/*

16 bit

*/

extern struct COLOUR_MAPPER col_map_xxxx_rrrr_gggg_bbbb;
extern struct COLOUR_MAPPER col_map_xggg_ggrr_rrrb_bbbb;
extern struct COLOUR_MAPPER col_map_xrrr_rrgg_gggb_bbbb;
extern struct COLOUR_MAPPER col_map_xbbb_bbgg_gggr_rrrr;
extern struct COLOUR_MAPPER col_map_rrrr_gggg_bbbb_rgbx_rev;
extern struct COLOUR_MAPPER col_map_rrrr_rggg_ggbb_bbbx_rev;
extern struct COLOUR_MAPPER col_map_rrrr_gggg_bbbb_xxxx_68k;

/*

32 bit

*/

extern struct COLOUR_MAPPER col_map_bbbb_bggg_ggrr_rrrx_xxxx_xxxx_xxxx_xxxx;
extern struct COLOUR_MAPPER col_map_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb;

void set_colour_mapper(COLOUR_MAPPER *colour_mapper);

#define MAX_COLBANKS	512

/*-----[Init Section]-----*/

// Call this to setup colour mapping (eg. during load game)

void InitPaletteMap(UBYTE *src, int banks, int bankcols, int mapsize);

/*

In the rare case that colour 255 is used and cannot be white
(eg. direct mapped, american horseshoes - the man on a bike)

*/

void set_white_pen(int pen);

/*

Used by the ingame gui, to get the white pen (nearly always 255)

*/

UBYTE get_white_pen(void);

// Call this to remove resources (eg. in clear default)

void destroy_palette_map(void);

// Call this at the start of each screen redraw

void ClearPaletteMap(void);

// Call this to clear caches (eg. when coming back into game from gui)

void reset_palette_map(void);

// Call this to update colours on the pc screen

void update_screen_palette(PALETTE palette);

// Call this to permanently allocate some pens

int Reserve_Pens(int num);

void silly_wgp_colour_ram(UBYTE *src);

// Call these to change permanently allocated pen colours

void Set_Pens_12bit_RGBx(int talc, int start, int cols);
void Set_Pens_15bit_xRGB(int talc, int start, int cols);

/*-----[Macro Section]-----*/

UWORD bank_status[MAX_COLBANKS];
UBYTE *coltab[MAX_COLBANKS];

// MAP_PALETTE_MAPPED(mapper, bank, cols, dst)

#define MAP_PALETTE_MAPPED(mapper,bank,cols,dst) \
   dst=coltab[bank];                             \
   if(bank_status[bank] == 0)                    \
      mapper(bank, cols);                        \

// MAP_PALETTE_MAPPED_NEW(mapper, bank, cols, dst)

#define MAP_PALETTE_MAPPED_NEW(bank,cols,dst) \
   dst = coltab[bank];                        \
   if(bank_status[bank] == 0)                 \
      current_mapper(bank, cols);             \

// MAP_PALETTE_MULTI_MAPPED(mapper, bank, cols, dst)

#define MAP_PALETTE_MULTI_MAPPED(mapper,bank,cols,dst) \
   dst=coltab[bank];                                   \
   if(bank_status[bank] < cols)                        \
      mapper(bank, cols);                              \

// MAP_PALETTE_MULTI_MAPPED_NEW(mapper, bank, cols, dst)

#define MAP_PALETTE_MULTI_MAPPED_NEW(bank,cols,dst) \
   dst = coltab[bank];                              \
   if(bank_status[bank] < cols)                     \
      current_mapper(bank, cols);                   \

// MAP_PALETTE(mapper, bank, cols)

#define MAP_PALETTE(mapper,bank,cols) \
   coltab[bank];                      \
   if(bank_status[bank] == 0)         \
      mapper(bank, cols);             \

/*-----[Macro Mappers]-----*/

void Map_12bit_RGBx(int bank, int cols);
void Map_12bit_RGBx_68k(int bank, int cols);

void Map_12bit_xRGB(int bank, int cols);
void Map_12bit_xxxxRRRRGGGGBBBB_Rev(int bank, int cols);

void Map_12bit_xBGR(int bank, int cols);
void Map_12bit_xBGR_wf(int bank, int cols);
void Map_12bit_xBGR_Rev(int bank, int cols);

void Map_15bit_RRRRGGGGBBBBRGBx(int bank, int cols);
void Map_15bit_RRRRGGGGBBBBRGBx_Rev(int bank, int cols);
void Map_15bit_RRRRRGGGGGBBBBBx_Rev(int bank, int cols);

void Map_15bit_xRGB(int bank, int cols);

void Map_15bit_xBGR(int bank, int cols);

void Map_15bit_xBGR_4bit(int bank, int cols);

// F3-System

void Map_12bit_RGBxxxxx(int bank, int cols);

void Map_21bit_xRGB(int bank, int cols);

void Map_24bit_xRGB(int bank, int cols);

void Map_24bit_xRGB_bswap(int bank, int cols);

void raine_blit_x2(BITMAP *src, BITMAP *dest, int s_x, int s_y, int d_x, int d_y, int w, int h);

void raine_blit_x2_x(BITMAP *src, BITMAP *dest, int s_x, int s_y, int d_x, int d_y, int w, int h);

#if 0

void fake_12_bit_blit(BITMAP *src, BITMAP *dest, int s_x, int s_y, int d_x, int d_y, int w, int h);
PALETTE pal_12bit;

#endif

#endif
