/******************************************************************************/
/*                                                                            */
/*                               TILE MODIFIERS                               */
/*                                                                            */
/******************************************************************************/

#include "gameinc.h"
#include "tilemod.h"

UBYTE tile_8x8_map[8 * 8];

static struct TILE_INFO tile_8x8_template =
{
   1,
   TILE_TYPE_8x8,
   8,
   8,
   tile_8x8_map,
   NULL,
   0,
   0,
};

UBYTE tile_16x16_map[16 * 16];

static struct TILE_INFO tile_16x16_template =
{
   1,
   TILE_TYPE_16x16,
   16,
   16,
   tile_16x16_map,
   NULL,
   0,
   0,
};

static void do_rotate(TILE_INFO *tile, ULONG rotate)
{
   ULONG count,size,x,y,ta,tb,tc;
   UBYTE *src,*buffer;

   if(!rotate) return;

   x     = tile->width;
   y     = tile->height;
   count = tile->count;
   size  = x * y;
   src   = tile->data;

   if(!(buffer=AllocateMem(size))) return;

   for(tc=0;tc<count;tc++){

   memcpy(buffer, src, size);

   switch(rotate){
   case 1:
      for(tb=0; tb<y; tb++){

         for(ta=0; ta<x; ta++){

            src[ ta + (tb*x) ] = buffer[ (((x-1)-ta)*y) + tb ];

         }

      }
   break;
   case 2:
      for(tb=0; tb<y; tb++){

         for(ta=0; ta<x; ta++){

            src[ ta + (tb*x) ] = buffer[ ((x-1)-ta) + (((y-1)-tb)*x) ];

         }

      }
   break;
   case 3:
      for(tb=0; tb<y; tb++){

         for(ta=0; ta<x; ta++){

            src[ ta + (tb*x) ] = buffer[ (ta*y) + ((y-1)-tb) ];

         }

      }
   break;
   }

   src += size;

   }

   FreeMem(buffer);
}

static void do_flip(TILE_INFO *tile, ULONG flip)
{
   ULONG count,size,x,y,ta,tb,tc;
   UBYTE *src,*buffer;

   if(!flip) return;

   x     = tile->width;
   y     = tile->height;
   count = tile->count;
   size  = x * y;
   src   = tile->data;

   if(!(buffer=AllocateMem(size))) return;

   for(tc=0;tc<count;tc++){

   memcpy(buffer, src, size);

   switch(flip){
   case 1:
      for(tb=0; tb<y; tb++){

         for(ta=0; ta<x; ta++){

            src[ ta + (tb*x) ] = buffer[ ((x-1)-ta) + (tb*x) ];

         }

      }
   break;
   case 2:
      for(tb=0; tb<y; tb++){

         for(ta=0; ta<x; ta++){

            src[ ta + (tb*x) ] = buffer[ ta + (((y-1)-tb)*x) ];

         }

      }
   break;
   case 3:
      for(tb=0; tb<y; tb++){

         for(ta=0; ta<x; ta++){

            src[ ta + (tb*x) ] = buffer[ ((x-1)-ta) + (((y-1)-tb)*x) ];

         }

      }
   break;
   }

   src += size;

   }

   FreeMem(buffer);
}

void Rotate8x8(UBYTE *source, ULONG count)
{
   UBYTE *TMP,*SRC;
   ULONG ta,tb,tc;

   if(!(TMP=AllocateMem(0x40))) return;

   SRC=source;

   for(tc=0;tc<count;tc++){

   for(tb=0;tb<0x40;tb++){
      TMP[tb]=SRC[tb];
   }

   for(tb=0;tb<8;tb++){
   for(ta=0;ta<8;ta++){
      SRC[ta+(tb<<3)]=TMP[(ta<<3)+tb];
   }
   }   

   SRC+=0x40;

   }

   FreeMem(TMP);
}

void Rotate8x8_4bpp(UBYTE *source, ULONG count)
{
   UBYTE TMP[0x40],TMP2[0x40],*SRC;
   ULONG ta,tb,tc;

   SRC=source;

   for(tc=0;tc<count;tc++){

   for(tb=0;tb<0x40;tb+=2){
      TMP[tb+0]=(SRC[tb>>1]>>4)&15;
      TMP[tb+1]=(SRC[tb>>1]>>0)&15;
   }

   for(tb=0;tb<8;tb++){
   for(ta=0;ta<8;ta++){
      TMP2[ta+(tb<<3)]=TMP[(ta<<3)+tb];
   }
   }   

   for(tb=0;tb<0x40;tb+=2){
      SRC[tb>>1]=(TMP2[tb+0]&15)<<4;
      SRC[tb>>1]|=(TMP2[tb+1]&15)<<0;
   }

   SRC+=0x20;

   }

}

void Rotate8x8_2pl(UBYTE *source, ULONG count)
{
   UBYTE TMP[0x40],TMP2[0x40],*SRC;
   ULONG ta,tb,tc,td;

   SRC=source;

   for(tc=0;tc<count;tc++){

   ta=0;
   for(tb=0;tb<16;tb+=2,ta+=8){
      td=SRC[tb+0];
      TMP[ta+0] = (((td&0x80)>>7)<<0);
      TMP[ta+1] = (((td&0x40)>>6)<<0);
      TMP[ta+2] = (((td&0x20)>>5)<<0);
      TMP[ta+3] = (((td&0x10)>>4)<<0);
      TMP[ta+4] = (((td&0x08)>>3)<<0);
      TMP[ta+5] = (((td&0x04)>>2)<<0);
      TMP[ta+6] = (((td&0x02)>>1)<<0);
      TMP[ta+7] = (((td&0x01)>>0)<<0);
      td=SRC[tb+1];
      TMP[ta+0] |= (((td&0x80)>>7)<<1);
      TMP[ta+1] |= (((td&0x40)>>6)<<1);
      TMP[ta+2] |= (((td&0x20)>>5)<<1);
      TMP[ta+3] |= (((td&0x10)>>4)<<1);
      TMP[ta+4] |= (((td&0x08)>>3)<<1);
      TMP[ta+5] |= (((td&0x04)>>2)<<1);
      TMP[ta+6] |= (((td&0x02)>>1)<<1);
      TMP[ta+7] |= (((td&0x01)>>0)<<1);
   }

   for(tb=0;tb<8;tb++){
   for(ta=0;ta<8;ta++){
      TMP2[ta+(tb<<3)]=TMP[(ta<<3)+tb];
   }
   }   

   ta=0;
   for(tb=0;tb<16;tb+=2,ta+=8){
      td  = (((TMP2[ta+0]&1)>>0)<<7);
      td |= (((TMP2[ta+1]&1)>>0)<<6);
      td |= (((TMP2[ta+2]&1)>>0)<<5);
      td |= (((TMP2[ta+3]&1)>>0)<<4);
      td |= (((TMP2[ta+4]&1)>>0)<<3);
      td |= (((TMP2[ta+5]&1)>>0)<<2);
      td |= (((TMP2[ta+6]&1)>>0)<<1);
      td |= (((TMP2[ta+7]&1)>>0)<<0);
      SRC[tb+0]=td;
      td  = (((TMP2[ta+0]&2)>>1)<<7);
      td |= (((TMP2[ta+1]&2)>>1)<<6);
      td |= (((TMP2[ta+2]&2)>>1)<<5);
      td |= (((TMP2[ta+3]&2)>>1)<<4);
      td |= (((TMP2[ta+4]&2)>>1)<<3);
      td |= (((TMP2[ta+5]&2)>>1)<<2);
      td |= (((TMP2[ta+6]&2)>>1)<<1);
      td |= (((TMP2[ta+7]&2)>>1)<<0);
      SRC[tb+1]=td;
   }

   SRC+=0x10;

   }

}

void Rotate16x16(UBYTE *source, ULONG count)
{
   UBYTE *TMP,*SRC;
   ULONG ta,tb,tc;

   if(!(TMP=AllocateMem(0x100))){return;}

   SRC=source;

   for(tc=0;tc<count;tc++){

   for(tb=0;tb<0x100;tb++){
      TMP[tb]=SRC[tb];
   }

   for(tb=0;tb<16;tb++){
   for(ta=0;ta<16;ta++){
      SRC[ta+(tb<<4)]=TMP[(ta<<4)+tb];
   }
   }   

   SRC+=0x100;

   }

   FreeMem(TMP);
}

void Flip8x8_XY(UBYTE *source, ULONG count)
{
   UBYTE *TMP,*SRC;
   ULONG ta,tb,tc;

   if(!(TMP=AllocateMem(0x40))){return;}

   SRC=source;

   for(tc=0;tc<count;tc++){

   for(tb=0;tb<0x40;tb++){
      TMP[tb]=SRC[tb];
   }

   for(tb=0;tb<8;tb++){
   for(ta=0;ta<8;ta++){
      SRC[ta+(tb<<3)]=TMP[(7-ta)+((7-tb)<<3)];
   }
   }   

   SRC+=0x40;

   }

   FreeMem(TMP);
}

void Flip16x16_XY(UBYTE *source, ULONG count)
{
   UBYTE *TMP,*SRC;
   ULONG ta,tb,tc;

   if(!(TMP=AllocateMem(0x100))){return;}

   SRC=source;

   for(tc=0;tc<count;tc++){

   for(tb=0;tb<0x100;tb++){
      TMP[tb]=SRC[tb];
   }

   for(tb=0;tb<16;tb++){
   for(ta=0;ta<16;ta++){
      SRC[ta+(tb<<4)]=TMP[(15-ta)+((15-tb)<<4)];
   }
   }   

   SRC+=0x100;

   }

   FreeMem(TMP);
}

void Flip8x8_X(UBYTE *source, ULONG count)
{
   UBYTE *TMP,*SRC;
   ULONG ta,tb,tc;

   if(!(TMP=AllocateMem(0x40))){return;}

   SRC=source;

   for(tc=0;tc<count;tc++){

   for(tb=0;tb<0x40;tb++){
      TMP[tb]=SRC[tb];
   }

   for(tb=0;tb<8;tb++){
   for(ta=0;ta<8;ta++){
      SRC[ta+(tb<<3)]=TMP[(ta)+((7-tb)<<3)];
   }
   }   

   SRC+=0x40;

   }

   FreeMem(TMP);
}

void Flip8x8_4bpp_X(UBYTE *source, ULONG count)
{
   UBYTE TMP[0x40],TMP2[0x40],*SRC;
   ULONG ta,tb,tc;

   SRC=source;

   for(tc=0;tc<count;tc++){

   for(tb=0;tb<0x40;tb+=2){
      TMP[tb+0]=(SRC[tb>>1]>>4)&15;
      TMP[tb+1]=(SRC[tb>>1]>>0)&15;
   }

   for(tb=0;tb<8;tb++){
   for(ta=0;ta<8;ta++){
      TMP2[ta+(tb<<3)]=TMP[(ta)+((7-tb)<<3)];
   }
   }   

   for(tb=0;tb<0x40;tb+=2){
      SRC[tb>>1]=(TMP2[tb+0]&15)<<4;
      SRC[tb>>1]|=(TMP2[tb+1]&15)<<0;
   }

   SRC+=0x20;

   }

}

void Flip8x8_2pl_X(UBYTE *source, ULONG count)
{
   UBYTE TMP[0x40],TMP2[0x40],*SRC;
   ULONG ta,tb,tc,td;

   SRC=source;

   for(tc=0;tc<count;tc++){

   ta=0;
   for(tb=0;tb<16;tb+=2,ta+=8){
      td=SRC[tb+0];
      TMP[ta+0] = (((td&0x80)>>7)<<0);
      TMP[ta+1] = (((td&0x40)>>6)<<0);
      TMP[ta+2] = (((td&0x20)>>5)<<0);
      TMP[ta+3] = (((td&0x10)>>4)<<0);
      TMP[ta+4] = (((td&0x08)>>3)<<0);
      TMP[ta+5] = (((td&0x04)>>2)<<0);
      TMP[ta+6] = (((td&0x02)>>1)<<0);
      TMP[ta+7] = (((td&0x01)>>0)<<0);
      td=SRC[tb+1];
      TMP[ta+0] |= (((td&0x80)>>7)<<1);
      TMP[ta+1] |= (((td&0x40)>>6)<<1);
      TMP[ta+2] |= (((td&0x20)>>5)<<1);
      TMP[ta+3] |= (((td&0x10)>>4)<<1);
      TMP[ta+4] |= (((td&0x08)>>3)<<1);
      TMP[ta+5] |= (((td&0x04)>>2)<<1);
      TMP[ta+6] |= (((td&0x02)>>1)<<1);
      TMP[ta+7] |= (((td&0x01)>>0)<<1);
   }

   for(tb=0;tb<8;tb++){
   for(ta=0;ta<8;ta++){
      TMP2[ta+(tb<<3)]=TMP[(ta)+((7-tb)<<3)];
   }
   }   

   ta=0;
   for(tb=0;tb<16;tb+=2,ta+=8){
      td  = (((TMP2[ta+0]&1)>>0)<<7);
      td |= (((TMP2[ta+1]&1)>>0)<<6);
      td |= (((TMP2[ta+2]&1)>>0)<<5);
      td |= (((TMP2[ta+3]&1)>>0)<<4);
      td |= (((TMP2[ta+4]&1)>>0)<<3);
      td |= (((TMP2[ta+5]&1)>>0)<<2);
      td |= (((TMP2[ta+6]&1)>>0)<<1);
      td |= (((TMP2[ta+7]&1)>>0)<<0);
      SRC[tb+0]=td;
      td  = (((TMP2[ta+0]&2)>>1)<<7);
      td |= (((TMP2[ta+1]&2)>>1)<<6);
      td |= (((TMP2[ta+2]&2)>>1)<<5);
      td |= (((TMP2[ta+3]&2)>>1)<<4);
      td |= (((TMP2[ta+4]&2)>>1)<<3);
      td |= (((TMP2[ta+5]&2)>>1)<<2);
      td |= (((TMP2[ta+6]&2)>>1)<<1);
      td |= (((TMP2[ta+7]&2)>>1)<<0);
      SRC[tb+1]=td;
   }

   SRC+=0x10;

   }

}

void Flip16x16_X(UBYTE *source, ULONG count)
{
   UBYTE *TMP,*SRC;
   ULONG ta,tb,tc;

   if(!(TMP=AllocateMem(0x100))){return;}

   SRC=source;

   for(tc=0;tc<count;tc++){

   for(tb=0;tb<0x100;tb++){
      TMP[tb]=SRC[tb];
   }

   for(tb=0;tb<16;tb++){
   for(ta=0;ta<16;ta++){
      SRC[ta+(tb<<4)]=TMP[(ta)+((15-tb)<<4)];
   }
   }   

   SRC+=0x100;

   }

   FreeMem(TMP);
}

void Flip8x8_Y(UBYTE *source, ULONG count)
{
   UBYTE *TMP,*SRC;
   ULONG ta,tb,tc;

   if(!(TMP=AllocateMem(0x40))){return;}

   SRC=source;

   for(tc=0;tc<count;tc++){

   for(tb=0;tb<0x40;tb++){
      TMP[tb]=SRC[tb];
   }

   for(tb=0;tb<8;tb++){
   for(ta=0;ta<8;ta++){
      SRC[ta+(tb<<3)]=TMP[(7-ta)+((tb)<<3)];
   }
   }   

   SRC+=0x40;

   }

   FreeMem(TMP);
}

void Flip16x16_Y(UBYTE *source, ULONG count)
{
   UBYTE *TMP,*SRC;
   ULONG ta,tb,tc;

   if(!(TMP=AllocateMem(0x100))){return;}

   SRC=source;

   for(tc=0;tc<count;tc++){

   for(tb=0;tb<0x100;tb++){
      TMP[tb]=SRC[tb];
   }

   for(tb=0;tb<16;tb++){
   for(ta=0;ta<16;ta++){
      SRC[ta+(tb<<4)]=TMP[(15-ta)+((tb)<<4)];
   }
   }   

   SRC+=0x100;

   }

   FreeMem(TMP);
}

void CountColours(UBYTE *source, ULONG count)
{
   ULONG ta,tb;

   tb=0;
   for(ta=0;ta<count;ta++){
      if((source[ta]&0x20)!=0)tb++;
   }
   #ifdef HACKS
   print_debug("64 Colour Bytes: $%08x\n",tb);
   #endif
}


UBYTE check_tile_solid(UBYTE *src, ULONG size)
{
   ULONG i, a, b;

   for(i = 0, a = 0, b = 0; i < size; i++){
      if(src[i])
         a = 1;
      else
         b = 1;
   }

   if((!a) &&  (b)) return 0; // Draw Nothing
   if( (a) &&  (b)) return 1; // Draw Transparent
   if( (a) && (!b)) return 2; // Draw Solid

   return 0;
}


UBYTE *MakeSolidTileMap(UBYTE *source, ULONG count, ULONG size)
{
   ULONG ta,tb,tc;
   UBYTE *map;

   if(!(map=AllocateMem(count)))return(NULL);

   tb=0;
   for(ta=0;ta<count;ta++){
      map[ta]=1;			// This tile is solid!
      for(tc=0;tc<size;tc++,tb++){
         if(source[tb]==0)map[ta]=0;	// This tile is not solid!
      }
   }

   return(map);
}

char *tile_type[TILE_TYPE_MAX] =
{
   " X x  Y (Nbpp)",
   " 8 x  8 (8bpp)",
   "16 x 16 (8bpp)",
};

void check_tile_rotation(void)
{
   ULONG i,j,k;

   /*

   rotate and flip all tile_list tiles

   */

   for( i = 0; i < tile_list_count; i ++){

      j  = display_cfg.rotate;
      j -= tile_list[i].rotate;
      j &= 3;

      k  = display_cfg.flip;
      k -= tile_list[i].flip;
      k &= 3;

      if(j || k){

         do_flip(&tile_list[i], tile_list[i].flip);

         do_rotate(&tile_list[i], j);

         do_flip(&tile_list[i], display_cfg.flip);

         tile_list[i].rotate = display_cfg.rotate;
         tile_list[i].flip   = display_cfg.flip;

      }

   }

   /*

   rotate and flip the template tiles (used for building correctly
   rotated GFX RAM tiles)

   */

   for(i = 0; i < (8 * 8); i++)

       tile_8x8_map[i] = i;

   tile_8x8_template.rotate = 0;
   tile_8x8_template.flip = 0;

   do_rotate(&tile_8x8_template, (0 - display_cfg.rotate) & 3);
   do_flip(&tile_8x8_template, display_cfg.flip);

   tile_8x8_template.rotate = display_cfg.rotate;
   tile_8x8_template.flip = display_cfg.flip;


   for(i = 0; i < (16 * 16); i++)

       tile_16x16_map[i] = i;

   tile_16x16_template.rotate = 0;
   tile_16x16_template.flip = 0;

   do_rotate(&tile_16x16_template, (0 - display_cfg.rotate) & 3);
   do_flip(&tile_16x16_template, display_cfg.flip);

   tile_16x16_template.rotate = display_cfg.rotate;
   tile_16x16_template.flip = display_cfg.flip;

}

UBYTE *MakeSolidTileMap16x16(UBYTE *source, ULONG count)
{
   return MakeSolidTileMap(source,count,16*16);
}

UBYTE *MakeSolidTileMap64x64(UBYTE *source, ULONG count)
{
   return MakeSolidTileMap(source,count,64*64);
}

UBYTE *make_solid_mask(UBYTE *source, ULONG count, ULONG size, ULONG pad_size)
{
   ULONG ta,tb,tc,td,te;
   UBYTE *map;

   if(pad_size<count){

   pad_size = 0x1000;
   if(count>0x01000) pad_size = 0x02000;
   if(count>0x02000) pad_size = 0x04000;
   if(count>0x04000) pad_size = 0x08000;
   if(count>0x08000) pad_size = 0x10000;
   if(count>0x10000) pad_size = 0x20000;
   if(count>0x20000) pad_size = 0x40000;

   }

   if(!(map=AllocateMem(pad_size))) return NULL;

   memset(map,0x00,pad_size);

   tb=0;
   for(ta=0;ta<count;ta++){
      td=0;
      te=0;
      for(tc=0;tc<size;tc++,tb++){
         if(source[tb])
            td=1;
         else
            te=1;
      }
      if((td==0)&&(te==1)) map[ta]=0;		// All pixels are 0: Don't Draw
      if((td==1)&&(te==1)) map[ta]=1;		// Mixed: Draw Trans
      if((td==1)&&(te==0)) map[ta]=2;		// All pixels are !0: Draw Solid
   }

   return map;
}

UBYTE *make_solid_mask_8x8(UBYTE *source, ULONG count)
{
   UBYTE *map;

   map = make_solid_mask(source, count, 8*8, 0);

   tile_list[tile_list_count].count  = count;
   tile_list[tile_list_count].type   = TILE_TYPE_8x8;
   tile_list[tile_list_count].width  = 8;
   tile_list[tile_list_count].height = 8;
   tile_list[tile_list_count].data   = source;
   tile_list[tile_list_count].mask   = map;
   tile_list[tile_list_count].rotate = 0;
   tile_list[tile_list_count].flip   = 0;

   tile_list_count++;

   return map;
}

UBYTE *make_solid_mask_16x8(UBYTE *source, ULONG count)
{
   return make_solid_mask(source, count, 16*8, 0);
}

UBYTE *make_solid_mask_16x16(UBYTE *source, ULONG count)
{
   UBYTE *map;

   map = make_solid_mask(source, count, 16*16, 0);

   tile_list[tile_list_count].count  = count;
   tile_list[tile_list_count].type   = TILE_TYPE_16x16;
   tile_list[tile_list_count].width  = 16;
   tile_list[tile_list_count].height = 16;
   tile_list[tile_list_count].data   = source;
   tile_list[tile_list_count].mask   = map;
   tile_list[tile_list_count].rotate = 0;
   tile_list[tile_list_count].flip   = 0;

   tile_list_count++;

   return map;
}

UBYTE *make_solid_mask_64x64(UBYTE *source, ULONG count)
{
   return make_solid_mask(source, count, 64*64, 0);
}

UBYTE *make_solid_mask_pad_8x8(UBYTE *source, ULONG count, ULONG pad_size)
{
   UBYTE *map;

   map = make_solid_mask(source, count, 8*8, pad_size);

   tile_list[tile_list_count].count  = count;
   tile_list[tile_list_count].type   = TILE_TYPE_8x8;
   tile_list[tile_list_count].width  = 8;
   tile_list[tile_list_count].height = 8;
   tile_list[tile_list_count].data   = source;
   tile_list[tile_list_count].mask   = map;
   tile_list[tile_list_count].rotate = 0;
   tile_list[tile_list_count].flip   = 0;

   tile_list_count++;

   return map;
}

UBYTE *make_solid_mask_pad_16x16(UBYTE *source, ULONG count, ULONG pad_size)
{
   UBYTE *map;

   map = make_solid_mask(source, count, 16*16, pad_size);

   tile_list[tile_list_count].count  = count;
   tile_list[tile_list_count].type   = TILE_TYPE_16x16;
   tile_list[tile_list_count].width  = 16;
   tile_list[tile_list_count].height = 16;
   tile_list[tile_list_count].data   = source;
   tile_list[tile_list_count].mask   = map;
   tile_list[tile_list_count].rotate = 0;
   tile_list[tile_list_count].flip   = 0;

   tile_list_count++;

   return map;
}

UBYTE *make_solid_mask_4bpp(UBYTE *source, ULONG count, ULONG size)
{
   ULONG ta,tb,tc,td,te;
   UBYTE *map;

   ta = 0x1000;
   if(count>0x01000) ta = 0x02000;
   if(count>0x02000) ta = 0x04000;
   if(count>0x04000) ta = 0x08000;
   if(count>0x08000) ta = 0x10000;
   if(count>0x10000) ta = 0x20000;
   if(count>0x20000) ta = 0x40000;

   if(!(map=AllocateMem(ta)))return(NULL);

   memset(map,0x00,ta);

   tb=0;
   for(ta=0;ta<count;ta++){
      td=0;
      te=0;
      for(tc=0;tc<size;tc+=2,tb++){
         if((source[tb]&0xF0)!=0) td=1;
         else te=1;
         if((source[tb]&0x0F)!=0) td=1;
         else te=1;
      }
      if((td==0)&&(te==1)) map[ta]=0;		// All pixels are 0: Don't Draw
      if((td==1)&&(te==1)) map[ta]=1;		// Mixed: Draw Trans
      if((td==1)&&(te==0)) map[ta]=2;		// All pixels are !0: Draw Solid
   }

   return(map);
}

UBYTE *make_solid_mask_8x8_4bpp(UBYTE *source, ULONG count)
{
   return make_solid_mask_4bpp(source,count,8*8);
}

UBYTE *make_solid_mask_16x16_4bpp(UBYTE *source, ULONG count)
{
   return make_solid_mask_4bpp(source,count,16*16);
}

UBYTE *make_solid_mask_64x64_4bpp(UBYTE *source, ULONG count)
{
   return make_solid_mask_4bpp(source,count,64*64);
}


UBYTE *make_colour_count(UBYTE *source, ULONG count, ULONG size)
{
   ULONG ta,tb,tc,td,te;
   UBYTE *map;

   ta = 0x1000;
   if(count>0x01000) ta = 0x02000;
   if(count>0x02000) ta = 0x04000;
   if(count>0x04000) ta = 0x08000;
   if(count>0x08000) ta = 0x10000;
   if(count>0x10000) ta = 0x20000;
   if(count>0x20000) ta = 0x40000;

   if(!(map=AllocateMem(ta))) return(NULL);

   memset(map,0x00,ta);

   tb=0;
   for(ta=0;ta<count;ta++){
      td = 0;
      for(tc=0;tc<size;tc++,tb++){
         if(source[tb] > td) td = source[tb];
      }
      te = 0;
      if(td>0x00) te=0x01;
      if(td>0x01) te=0x02;
      if(td>0x02) te=0x04;
      if(td>0x04) te=0x08;
      if(td>0x08) te=0x10;
      if(td>0x10) te=0x20;
      if(td>0x20) te=0x40;
      if(td>0x40) te=0x80;
      if(td>0x80) te=0xFF;

      map[ta] = te;
   }

   return map;
}

UBYTE *make_colour_count_16x16(UBYTE *source, ULONG count)
{
   return make_colour_count(source,count,16*16);
}

