/******************************************************************************/
/*                                                                            */
/*                               TILE MODIFIERS                               */
/*                                                                            */
/******************************************************************************/

#ifndef _tilemod_h_
#define _tilemod_h_

#include "deftypes.h"

enum tile_type_num
{
   TILE_TYPE_UNKNOWN = 0,
   TILE_TYPE_8x8,
   TILE_TYPE_16x16,
   TILE_TYPE_MAX,
};

extern char *tile_type[TILE_TYPE_MAX];

typedef struct TILE_INFO
{
   ULONG count;
   ULONG type;
   ULONG width;
   ULONG height;
   UBYTE *data;
   UBYTE *mask;
   ULONG rotate;
   ULONG flip;
} TILE_INFO;

struct TILE_INFO tile_list[0x10];

ULONG tile_list_count;

extern UBYTE tile_8x8_map[8 * 8];
extern UBYTE tile_16x16_map[16 * 16];

/**********************************************************/
/* Rotation by 90 degrees                                 */
/**********************************************************/

void Rotate8x8(UBYTE *source, ULONG count);
void Rotate8x8_4bpp(UBYTE *source, ULONG count);
void Rotate8x8_2pl(UBYTE *source, ULONG count);
void Rotate16x16(UBYTE *source, ULONG count);

/**********************************************************/
/* Flip on X Axis                                         */
/**********************************************************/

void Flip8x8_X(UBYTE *source, ULONG count);
void Flip8x8_4bpp_X(UBYTE *source, ULONG count);
void Flip8x8_2pl_X(UBYTE *source, ULONG count);
void Flip16x16_X(UBYTE *source, ULONG count);

/**********************************************************/
/* Flip on Y Axis                                         */
/**********************************************************/

void Flip8x8_Y(UBYTE *source, ULONG count);
void Flip16x16_Y(UBYTE *source, ULONG count);

/**********************************************************/
/* Flip on X and Y Axis                                   */
/**********************************************************/

void Flip8x8_XY(UBYTE *source, ULONG count);
void Flip16x16_XY(UBYTE *source, ULONG count);

/**********************************************************/
/* Count Colour bits (for F3-System 64 colour tile check) */
/**********************************************************/

void CountColours(UBYTE *source, ULONG count);

/**********************************************************/
/* Make Tile Solid/Transparent map                        */
/**********************************************************/

UBYTE *MakeSolidTileMap16x16(UBYTE *source, ULONG count);
UBYTE *MakeSolidTileMap64x64(UBYTE *source, ULONG count);

UBYTE *make_solid_mask_8x8(UBYTE *source, ULONG count);
UBYTE *make_solid_mask_16x8(UBYTE *source, ULONG count);
UBYTE *make_solid_mask_16x16(UBYTE *source, ULONG count);
UBYTE *make_solid_mask_64x64(UBYTE *source, ULONG count);

UBYTE *make_solid_mask_8x8_4bpp(UBYTE *source, ULONG count);
UBYTE *make_solid_mask_16x16_4bpp(UBYTE *source, ULONG count);
UBYTE *make_solid_mask_64x64_4bpp(UBYTE *source, ULONG count);

UBYTE *make_solid_mask_pad_8x8(UBYTE *source, ULONG count, ULONG pad_size);
UBYTE *make_solid_mask_pad_16x16(UBYTE *source, ULONG count, ULONG pad_size);

/**********************************************************/
/* Make Used colour count map                             */
/**********************************************************/

UBYTE *make_colour_count_16x16(UBYTE *source, ULONG count);

UBYTE check_tile_solid(UBYTE *src, ULONG size);

#endif //_tilemod_h_
