/******************************************************************************/
/*                                                                            */
/*                          SAMPLE SUPPORT FOR RAINE                          */
/*                                                                            */
/*                        modified by Hiromitsu Shioya                        */
/*                        change SEAL library 09/20/98                        */
/*                                                                            */
/******************************************************************************/

#include <audio.h>				// SEAL library
#include "raine.h"
#include "sasound.h"

/*
 *  Volume Section
 */

#define VOLUME_MIN	(0)			// Minimum volume
#define VOLUME_MAX	(255)			// Maximum volume

/*
 *  Panning Section
 */

#define PAN_LEFT	(0)			// Left panning
#define PAN_CENTRE	(128)			// Central panning
#define PAN_RIGHT	(255)			// Right panning

/*
 *  Sample Control Section
 */

#define CPU_FPS		(60)			// Frames per second of emulated hardware

/*
 *  Other stuff
 */

#define UPDATE_60_FPS				// Turn off the counting stuff. Emulation
						// speed stays more constant this way.

unsigned char No_FM;				// fm.c control

static int samp_modeb_timer;

SoundRec      *SndMachine = NULL, snd_entry, *nowSndRec;

/* audio related stuff */

HAC    hVoice[NUMVOICES];
LPAUDIOWAVE    lpWave[NUMVOICES];
int    Volumi[NUMVOICES];
int    MasterVolume;
int    playing[NUMVOICES];
int    vchan[NUMVOICES];
int    ventry[NUMVOICES];
int    vbover_err, vbunder_err;
int    vmode[NUMVOICES];
int    *vpan[NUMVOICES];

enum vmode_num
{
  VOICE_NONE = 0,
  VOICE_STREAM,
  VOICE_PLAY
};

AUDIOINFO    audio_info;
AUDIOCAPS    audio_caps;
int    SampleRate;
int    audio_sample_rate;
int    sound_emu_count;

static int reserved_channel = 0;

int sound_stream_mode = SOUND_STREAM_WAIT;

static int stream_buffer_max;

static int pause_sound;

void saCheckPlayStream( void );

/*******************************************************************************************/
/**** osd_??? functions                                                                 ****/
/*******************************************************************************************/
int get_play_channels( int request ){
  return saGetPlayChannels( request );
}

void reset_play_channels(void){
  saResetPlayChannels();
}

void osd_play_streamed_sample(int channel,signed char *data,int len,int freq,int volume){
  if( !sound_stream_mode )  saPlayStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 8, 128 );
  else                      saPlayBufferedStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 8, 128 );
}

void osd_play_streamed_sample_16(int channel,signed short *data,int len,int freq,int volume){
  if( !sound_stream_mode )  saPlayStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 16, 128 );
  else                      saPlayBufferedStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 16, 128 );
}

void osd_play_streamed_sample2(int channel,signed char *data,int len,int freq,int volume, int pan){
  if( !sound_stream_mode )  saPlayStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 8, pan );
  else                      saPlayBufferedStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 8, pan );
}

void osd_play_streamed_sample2_16(int channel,signed short *data,int len,int freq,int volume, int pan){
  if( !sound_stream_mode )  saPlayStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 16, pan );
  else                      saPlayBufferedStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 16, pan );
}

void osd_ym2203_update(void){
}

/*******************************************************************************************/
/*  streams v35x                                                                           */
/*******************************************************************************************/

#include "streams.c"

/*******************************************************************************************/
/*  sa???Sound                                                                             */
/*******************************************************************************************/
/******************************************/
/*    update sound                        */
/******************************************/
void saUpdateSound( int nowclock )
{
   int i;

   if( GameSound == 0 ) return;			// [J3d!]: Extra Check
   if( RaineSoundCard == 0 ) return;		// [J3d!]: Extra Check
   if( audio_sample_rate == 0 ) return;
   if( SndMachine == NULL ) return;

   if( nowclock != 0 ){

      #ifdef UPDATE_60_FPS

      streams_sh_update();
      for( i = 0; i < SndMachine->control_max; i++ )
	 if( SndMachine->update[i] != NULL ) SndMachine->update[i]();

      #else

      SndMachine->sound_timing += nowclock;
      while( SndMachine->sound_timing >= SndMachine->sound_timingmax ){
         SndMachine->sound_timing -= SndMachine->sound_timingmax;
         SndMachine->sound_count ++;
      }

      if( SndMachine->sound_count >= SndMachine->sound_countmax ){
         SndMachine->sound_count -= SndMachine->sound_countmax;
         streams_sh_update();
         for( i = 0; i < SndMachine->control_max; i++ )
	    if( SndMachine->update[i] != NULL ) SndMachine->update[i]();
      }

      #endif

      if( !sound_stream_mode ){
         AUpdateAudio();
      }

   }
   else{

      if( sound_stream_mode ){
         AUpdateAudio();
         saCheckPlayStream();
      }

   }
}

/******************************************/
/*    setup sound                         */
/******************************************/
BOOL saInitSoundCard( int soundcard, int buffer_mode, int sample_rate )
{
  int i, j, k;

  if( AInitialize() == AUDIO_ERROR_NONE ){

    if( soundcard <= 0 ){		// soundcard = 'none'
       audio_sample_rate = 0;
       #ifdef HACKS
       PrintDebug("saInitSoundCard(): OK [No Card]\n");
       #endif
       return TRUE;			// always ok
    }
    else{				// soundcard = <nn seal library>

      // Set card id
      audio_info.nDeviceId = soundcard;

      // Always use 16 bit mixing if possible, better quality and same speed of 8 bit.
      audio_info.wFormat = AUDIO_FORMAT_16BITS | AUDIO_FORMAT_STEREO;

      // I don't know why, but playing at 44100Hz causes some crackling, which
      // disappears at 44099Hz. So here we just subtract one from the target sample rate.
      audio_info.nSampleRate = audio_sample_rate - 1;

      if( AOpenAudio(&audio_info) != AUDIO_ERROR_NONE ){
         #ifdef HACKS
	 PrintDebug("saInitSoundCard(): Audio Initialization Failed!\n");
         #endif
	 audio_sample_rate = 0;
	 return FALSE;
      }

      AGetAudioDevCaps( audio_info.nDeviceId, &audio_caps );

      // Update sample rate (incase seal chose a different one)
      audio_sample_rate = audio_info.nSampleRate;

      // Open and allocate voices, allocate waveforms
      if( AOpenVoices(NUMVOICES) != AUDIO_ERROR_NONE ){
         #ifdef HACKS
	 PrintDebug("saInitSoundCard(): Voice Initialization Failed!\n");
         #endif
	 audio_sample_rate = 0;
	 return FALSE;
      }
	  
      for( i = 0; i < NUMVOICES; i++ ){
	if( ACreateAudioVoice(&hVoice[i]) != AUDIO_ERROR_NONE ){
          #ifdef HACKS
	  PrintDebug("saInitSoundCard(): Voice #%d Creation Failed!\n",i);
          #endif
	  audio_sample_rate = 0;
	  return FALSE;
	}
	ASetVoicePanning( hVoice[i], 128 );
	lpWave[i] = 0;
	playing[i] = 0;
	vmode[i] = VOICE_NONE;
      }

    }

  }

  AUpdateAudio();

   #ifdef HACKS
   PrintDebug("saInitSoundCard(): SEAL initialize OK\n");
   #endif

  /**** stream buffer mode ****/
  sound_stream_mode = buffer_mode;
  if( !sound_stream_mode )  stream_buffer_max = STREAM_BUFFER_MAXA;
  else                      stream_buffer_max = STREAM_BUFFER_MAXB;
  samp_modeb_timer = 0;

  MasterVolume = 100;
  //reserved_channel = 0;

   if(change_sample_rate!=0){		// Sample rate has changed, so all emulators must
      change_sample_rate=0;		// be restarted!
      saStopSoundEmulators();
   }

  if( SndMachine != NULL ){
    if( !SndMachine->first ){
      saInitVolPan();		/* moved. (hiro-shi) */
      SndMachine->first = 1;	/* first flag clear */
      streams_sh_start();	/* streaming system initialize & start */
      pause_sound = 0;		/* pause flag off */
      vbover_err = vbunder_err = 0; /* error initial */
      No_FM = 1;			/* FM DAC mode */
      for( j = 0; j < SndMachine->control_max; j++ ){
	switch( SndMachine->init[j] ){
	case SOUND_YM2203:
	  i = YM2203_sh_start( SndMachine->intf[j] );
	break;
	case SOUND_YM2151S:
	  i = YM2151_sh_start( SndMachine->intf[j] );
	break;
	case SOUND_YM2151J:
	  i = YM2151_ALT_sh_start( SndMachine->intf[j] );
	break;
	case SOUND_YM2610:
	  i = YM2610_sh_start( SndMachine->intf[j] );
	break;
	case SOUND_YM2610B:
	  i = YM2610B_sh_start( SndMachine->intf[j] );
	break;
	case SOUND_MSM5205:
	  i = MSM5205buffer_sh_start( SndMachine->intf[j] );
	break;
	case SOUND_M6295:
	  i = M6295buffer_sh_start( SndMachine->intf[j] );
	break;
	case SOUND_PSG:
	  i = AY8910_sh_start( SndMachine->intf[j] );
	break;
	case SOUND_YM3812:
	  i = YM3812_sh_start( SndMachine->intf[j] );
	break;
	}
	if( i ){
	  audio_sample_rate = 0;
	  SndMachine = NULL;
	  return FALSE;
	}
      }
    }
  }
  return TRUE;
}

/*********************************************************************/
/* stop sound emulators: this is needed for when sample rate changes */
/*********************************************************************/
void saStopSoundEmulators(void)
{
   int i;

   #ifdef HACKS
   PrintDebug("saStopSoundEmulators(): Removing Sound Emulators\n");
   #endif

   saResetPlayChannels();

   if( SndMachine == NULL ){
      SndMachine = &snd_entry;
   }
   else{
      if(SndMachine->first!=0){		// YM3812 fix crashes
         streams_sh_stop();
            for( i = 0; i < SndMachine->control_max; i++ ){
               if( SndMachine->stop[i] != NULL )  SndMachine->stop[i]();
            }
         SndMachine->first           = 0;
         SndMachine->sound_timing    = 0;
         SndMachine->sound_count     = 0;
      }
   }

   #ifdef HACKS
   PrintDebug("saStopSoundEmulators(): OK\n");
   #endif
}

/******************************************/
/*    setup sound                         */
/******************************************/

void saSetupSound( SoundRecEntry *rec )
{
  int i;

  if( rec == NULL ) return;

  saStopSoundEmulators();

  /**** init sound control work ****/
  for( i = 0; i < SND_CONTROL_MAX; i++ ){
    SndMachine->init[i]      = -1;
    SndMachine->update[i]    = NULL;
    SndMachine->stop[i]      = NULL;
    SndMachine->intf[i]      = NULL;
  }
  SndMachine->sound_timing    = 0;
  SndMachine->sound_count     = 0;
  SndMachine->sound_countmax  = rec->count;
  SndMachine->sound_timingmax = rec->sync / SndMachine->sound_countmax;

  SndMachine->init[0]      = rec->init;
  SndMachine->update[0]    = rec->update;
  SndMachine->stop[0]      = rec->stop;
  SndMachine->intf[0]      = rec->intf;

  SndMachine->first = 0;
  SndMachine->control_max = 1;
}

/******************************************/
/*    add sound                           */
/******************************************/

void saAddSound( SoundRecEntry *rec )
{
  int i;
  if( SndMachine->control_max >= SND_CONTROL_MAX )  return;
  i = SndMachine->control_max;
  SndMachine->init[i]      = rec->init;
  SndMachine->update[i]    = rec->update;
  SndMachine->stop[i]      = rec->stop;
  SndMachine->intf[i]      = rec->intf;
  SndMachine->control_max++;
}
/******************************************/
/*    destroy sound                       */
/******************************************/

void saDestroyChannel( int chan )
{
   if( lpWave[chan] ){
      AStopVoice( hVoice[chan] );
      ADestroyAudioData( lpWave[chan] );
      free( lpWave[chan] );
      free( vpan[chan] );
      lpWave[chan] = 0;
      playing[chan] = 0;
      vmode[chan] = VOICE_NONE;
      vpan[chan] = 0;
   }
}

void saDestroySound( int remove_all_resources )
{
   int i;

   #ifdef HACKS
   PrintDebug("saDestroySound: Removing SEAL\n");
   #endif

   pause_raine_ym3812();

   if(remove_all_resources!=0) saStopSoundEmulators();

   for( i = 0; i < NUMVOICES; i++ ){
      saDestroyChannel( i );
      ADestroyAudioVoice( hVoice[i] );
   }
   ACloseVoices();
   ACloseAudio();

   //SndMachine = NULL;
   #ifdef HACKS
   PrintDebug("saDestroySound: OK\n");
   #endif
}

void sa_pause_sound(void)
{
   int i;

   pause_sound      = 1;
   samp_modeb_timer = 0;
   vbover_err       = 0;
   vbunder_err      = 0;

   pause_raine_ym3812();
   for( i = 0; i < NUMVOICES; i++ ){
      saDestroyChannel( i );
   }
}

void sa_unpause_sound(void)
{
   pause_sound      = 0;
}

/******************************************/
/*    buffer check                        */
/******************************************/
void saCheckPlayStream( void )
{
   int i, len, rlen;
   DWORD pos;
   int vp, ve;

   if( pause_sound != 0){

      samp_modeb_timer = (samp_modeb_timer+1) % MODEB_FRAME_SIZE;	// <60> res = [0..59]

      vp = samp_modeb_timer / MODEB_UPDATE_COUNT;			// <3>  res = [0..19]

      for( i = 0; i < NUMVOICES; i++ ){					// ***** SCAN VOICES

         if( playing[i] ){						// ***** PROCESS ENABLED VOICE

	    ve = ventry[i] % MODEB_MASK;				// <60/3=20> res = [0..19]
	
	    if( (vmode[i] == VOICE_STREAM) && (vp == ve )){		// ***** PROCESS ENABLED STREAM VOICE

	       /**** buffer update check ****/
	       AGetVoicePosition( hVoice[i], &pos );
	       len = (lpWave[i]->dwLength / stream_buffer_max) * MODEB_UPDATE_COUNT;	// length of <3> buffers
	       rlen = len * vp;								// buffer point where sample should go

	       if( (pos < rlen) || (pos >= (rlen+len)) ){		// Make sure it is not playing this buffer right now

	          /**** check play buffer/write buffer position ****/
	          ve = (ventry[i] + 1) * MODEB_UPDATE_COUNT;		// 

	          /*if( (vchan[i] - ve) > 0 ){*/
	          /*if( (vchan[i] - ve) > -MODEB_UPDATE_COUNT ){*/

	          if( ((vchan[i] - ve) > (0-MODEB_UPDATE_COUNT)) && (vbunder_err < (MODEB_ERROR_MAX*2))){

	             if( (vchan[i] - ve) < 0 )  vbunder_err++;
	             else{
		        if( vbunder_err >= 0 )  vbunder_err--;
	             }
	             /**** all buffer update (not overrun) ****/
                     for( i = 0; i < NUMVOICES; i++ ){
		        if( playing[i] && vmode[i] == VOICE_STREAM ){
		           vbover_err = 0;
		           len  = (lpWave[i]->dwLength / stream_buffer_max) * MODEB_UPDATE_COUNT;
		           rlen = len * vp;
		           ASetVoicePanning( hVoice[i], (UINT)*(vpan[i] + samp_modeb_timer) );
		           ventry[i]++;
		        }
	             }

	          }
                  else{

                     #ifdef HACKS
	             PrintIngame(120,"Now buf-p[over] %08x %08x", vchan[i], ve );
                     #endif
	             /**** all buffer update (over run) ****/
                     for( i = 0; i < NUMVOICES; i++ ){
		        if( playing[i] && vmode[i] == VOICE_STREAM ){
		           //AStopVoice( hVoice[i] );
		           vbunder_err = 0;
		           vbover_err = 0;
		           len  = (lpWave[i]->dwLength / stream_buffer_max) * MODEB_UPDATE_COUNT;
		           rlen = len * vp;
		           //AWriteAudioData( lpWave[i], rlen, len );
		           //APlayVoice( hVoice[i], lpWave[i] );
		           ASetVoicePanning( hVoice[i], (UINT)*(vpan[i] + samp_modeb_timer) );
		           vchan[i] = (ventry[i] + 1) * MODEB_UPDATE_COUNT;
		           ventry[i]++;
		        }
	             }

	          }
	          return;		// Completed
	       }
               else{			// ERROR: Required buffer was being played

	          vbover_err++;		/* error count */
	          //PrintIngame(120,"ve:%08x, vc:%08x, verr:%08x", ventry[0]*MODEB_UPDATE_COUNT, vchan[0], vbover_err );
	          if( vbover_err >= MODEB_ERROR_MAX ){
	             /**** all buffer restart ****/
	             vbover_err       = 0;
	             vbunder_err      = 0;
	             samp_modeb_timer = 0;

                     for( i = 0; i < NUMVOICES; i++ ){
                        #if 0
		        ventry[i] = 1;
		        vchan[i] = MODEB_UPDATE_COUNT;
                        #else
		        saDestroyChannel( i );
                        #endif
	             }

                     #ifdef HACKS
	             PrintIngame(60,"Buffer stop: Restarting streams.");
                     #endif
	          }
	          return;		// Completed

	       }
	       /**** pos check end ****/
	    }			// ***** END PROCESS ENABLED STREAM VOICE
         }			// ***** END PROCESS ENABLED VOICE
      }				// ***** END SCAN VOICES
   }				// ***** END PLAY SOUND
}

/*******************************************************************************************/
/*******************************************************************************************/
/******************************************/
/*    play samples                        */
/******************************************/
void saPlayBufferedStreamedSampleBase( int channel, signed char *data, int len, int freq, int volume, int bits , int pan ){
  int i;
  if( audio_sample_rate == 0 || channel >= NUMVOICES )  return;
  if( SndMachine == NULL )  return;
  if( !playing[channel] ){
    if( lpWave[channel] ){
      AStopVoice( hVoice[channel] );
      ADestroyAudioData( lpWave[channel] );
      free( lpWave[channel] );
      lpWave[channel] = 0;
    }
    if( (lpWave[channel] = (LPAUDIOWAVE)malloc(sizeof(AUDIOWAVE))) == 0 )  return;
    if( (vpan[channel] = (int *)malloc(sizeof(int)*stream_buffer_max)) == NULL )  return;

    lpWave[channel]->wFormat     = (bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS) | AUDIO_FORMAT_STEREO | AUDIO_FORMAT_LOOP;
    lpWave[channel]->nSampleRate = audio_sample_rate;
    lpWave[channel]->dwLength    = stream_buffer_max*len;
    lpWave[channel]->dwLoopStart = 0;
    lpWave[channel]->dwLoopEnd   = stream_buffer_max*len;
    if( ACreateAudioData(lpWave[channel]) != AUDIO_ERROR_NONE ){
      free( lpWave[channel] );
      free( vpan[channel] );
      lpWave[channel] = 0;
      vpan[channel] = 0;
      return;
    }
    memset( lpWave[channel]->lpData, 0, stream_buffer_max*len );
    APlayVoice( hVoice[channel], lpWave[channel] );
    ASetVoiceFrequency( hVoice[channel], freq );
    ASetVoiceVolume( hVoice[channel], MasterVolume*volume/400 );
    ASetVoicePanning( hVoice[channel], (UINT)pan );
    playing[channel] = 1;	/* use front surface */
    vmode[channel] = VOICE_STREAM;
    /**** make sound temp. buffer ****/
    vchan[channel] = MODEB_UPDATE_COUNT;
    memcpy( &lpWave[channel]->lpData[len * MODEB_UPDATE_COUNT], data, len );
    for( i = 0; i <= MODEB_UPDATE_COUNT; i++ ){
      *(vpan[channel] + i) = pan;
    }
    vchan[channel]++;
    ventry[channel] = 1;
  } else{
    int s_pos = vchan[channel]%stream_buffer_max;
    memcpy( &lpWave[channel]->lpData[len * s_pos], data, len );
    *(vpan[channel] + s_pos) = pan;
    vchan[channel]++;
  }
  Volumi[channel] = volume/4;
}

/******************************************/
/*    play samples                        */
/******************************************/
void saPlayStreamedSampleBase( int channel, signed char *data, int len, int freq, int volume, int bits , int pan ){
  if( audio_sample_rate == 0 || channel >= NUMVOICES )  return;
  if( SndMachine == NULL )  return;
  if( !playing[channel] ){
    if( lpWave[channel] ){
      AStopVoice( hVoice[channel] );
      ADestroyAudioData( lpWave[channel] );
      free( lpWave[channel] );
      lpWave[channel] = 0;
    }
    if( (lpWave[channel] = (LPAUDIOWAVE)malloc(sizeof(AUDIOWAVE))) == 0 )  return;

    lpWave[channel]->wFormat = (bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS) | AUDIO_FORMAT_STEREO | AUDIO_FORMAT_LOOP;
    lpWave[channel]->nSampleRate = audio_sample_rate;
    lpWave[channel]->dwLength = stream_buffer_max*len;
    lpWave[channel]->dwLoopStart = 0;
    lpWave[channel]->dwLoopEnd = stream_buffer_max*len;
    if( ACreateAudioData(lpWave[channel]) != AUDIO_ERROR_NONE ){
      free( lpWave[channel] );
      lpWave[channel] = 0;
      return;
    }
    memset(lpWave[channel]->lpData,0,stream_buffer_max*len);
    memcpy(lpWave[channel]->lpData,data,len);
    /* upload the data to the audio DRAM local memory */
    AWriteAudioData(lpWave[channel],0,stream_buffer_max*len);
    APlayVoice(hVoice[channel],lpWave[channel]);
    ASetVoiceFrequency(hVoice[channel],freq);
    ASetVoiceVolume(hVoice[channel],MasterVolume*volume/400);
    ASetVoicePanning(hVoice[channel], (UINT)pan );
    playing[channel] = 1;	/* use front surface */
    vchan[channel] = 1;
    ventry[channel] = 1;
    vmode[channel] = VOICE_STREAM;

    //PrintDebug("first stream entry. [%d:%d:%d:%d]\n", channel, len, freq, volume );
  } else{
    DWORD pos;
#if 0
    int nownum = vchan[channel]%STREAM_BUFFER_MAXA;
    for(;;){
      AUpdateAudio();
      AGetVoicePosition(hVoice[channel],&pos);
      
      if( !nownum && pos >= len )  break;
      if( nownum == (stream_buffer_max-1) ){
	if( pos < (len*nownum) )  break;
      } else{
	if( (pos < len*(nownum) || pos >= (len*(nownum+1))) )  break;
      }
      /*
      if ((vchan[channel]%STREAM_BUFFER_MAXA) == 0 && pos >= len) break;
      if ((vchan[channel]%STREAM_BUFFER_MAXA) == 1 && (pos < len || pos >= (2*len))) break;
      if ((vchan[channel]%STREAM_BUFFER_MAXA) == 2 && pos < (2*len)) break;
      */
    }
#else
    AGetVoicePosition(hVoice[channel],&pos);
    AUpdateAudio();
#endif

    memcpy( &lpWave[channel]->lpData[len * (vchan[channel]%STREAM_BUFFER_MAXA)], data, len );
    AWriteAudioData( lpWave[channel], len*(vchan[channel]%STREAM_BUFFER_MAXA), len );
    vchan[channel]++;
    //PrintDebug("next stream entry. [%d:%d:%d:%d::%d]\n", channel, len, freq, volume, pos );
  }
  Volumi[channel] = volume/4;
}

/******************************************/
/*    play samples                        */
/******************************************/
void saPlayStreamedSample16( int channel, signed short *data, int len, int freq, int volume )
{
  if( !sound_stream_mode )  saPlayStreamedSampleBase( channel, (signed char *)data, 2*len, freq, volume, 16, 128 );
  else                      saPlayBufferedStreamedSampleBase( channel, (signed char *)data, 2*len, freq, volume, 16, 128 );
}
/******************************************/
/*    play samples                        */
/******************************************/
void saPlayStreamedSample16Pan( int channel, signed short *data, int len, int freq, int volume, int pan )
{
  if( !sound_stream_mode )  saPlayStreamedSampleBase( channel, (signed char *)data, 2*len, freq, volume, 16, pan );
  else                      saPlayBufferedStreamedSampleBase( channel, (signed char *)data, 2*len, freq, volume, 16, pan );
}

/******************************************/
/*    play samples                        */
/******************************************/
void saPlaySampleBase( int channel, signed char *data, int len, int freq, int volume, int loop, int bits )
{
  if( audio_sample_rate == 0 || channel >= NUMVOICES ) return;
  if( SndMachine == NULL )  return;

  if ( lpWave[channel] && lpWave[channel]->dwLength != len ){
    AStopVoice(hVoice[channel]);
    ADestroyAudioData(lpWave[channel]);
    free(lpWave[channel]);
    lpWave[channel] = 0;
  }
  if( lpWave[channel] == 0 ){
    if( (lpWave[channel] = (LPAUDIOWAVE)malloc(sizeof(AUDIOWAVE))) == 0 )  return;

    if( loop )  lpWave[channel]->wFormat = ( bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS) |
		  AUDIO_FORMAT_MONO | AUDIO_FORMAT_LOOP;
    else        lpWave[channel]->wFormat = (bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS) |
		  AUDIO_FORMAT_MONO;
    lpWave[channel]->nSampleRate = audio_sample_rate;
    lpWave[channel]->dwLength = len;
    lpWave[channel]->dwLoopStart = 0;
    lpWave[channel]->dwLoopEnd = len;
    if( ACreateAudioData(lpWave[channel]) != AUDIO_ERROR_NONE ){
      free(lpWave[channel]);
      lpWave[channel] = 0;
      return;
    }
  } else{
    /*
    if( loop )  lpWave[channel]->wFormat = AUDIO_FORMAT_8BITS | AUDIO_FORMAT_MONO | AUDIO_FORMAT_LOOP;
    else        lpWave[channel]->wFormat = AUDIO_FORMAT_8BITS | AUDIO_FORMAT_MONO;
    */
    if( loop )  lpWave[channel]->wFormat = ( bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS) |
		  AUDIO_FORMAT_MONO | AUDIO_FORMAT_LOOP;
    else        lpWave[channel]->wFormat = (bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS) |
		  AUDIO_FORMAT_MONO;
  }
  memcpy(lpWave[channel]->lpData,data,len);
  /* upload the data to the audio DRAM local memory */
  AWriteAudioData(lpWave[channel],0,len);
  APlayVoice(hVoice[channel],lpWave[channel]);
  /* need to cast to double because freq*nominal_sample_rate can exceed the size of an int */
  ASetVoiceFrequency(hVoice[channel],freq );
  ASetVoiceVolume(hVoice[channel],MasterVolume*volume/400);
  Volumi[channel] = volume/4;
}

/******************************************/
/*    play samples                        */
/******************************************/
void saPlaySample8(int channel,signed char *data,int len,int freq,int volume,int loop)
{
   saPlaySampleBase(channel,data,len,freq,volume,loop,8);
}

void saPlaySample16(int channel,signed short *data,int len,int freq,int volume,int loop)
{
   saPlaySampleBase(channel,(signed char *)data,2*len,freq,volume,loop,16);
}

/******************************************/
/*    stop samples                        */
/******************************************/
void saStopSample(int channel)
{
    AStopVoice(hVoice[channel]);
    ADestroyAudioData(lpWave[channel]);
    free(lpWave[channel]);
    lpWave[channel] = 0;
}

/******************************************************************************/
/*                                                                            */
/*                        SOUND CHANNEL ALLOCATION                            */
/*                                                                            */
/******************************************************************************/

int saGetPlayChannels( int request )
{
   int ret_value = reserved_channel;
   reserved_channel += request;
   return ret_value;
}

void saResetPlayChannels( void )
{
   reserved_channel = 0;
}

/******************************************************************************/
/*                                                                            */
/*                          SOUND CARD INFORMATION                            */
/*                                                                            */
/******************************************************************************/

char *sound_card_name( int num )
{
   static char *sound_name_list[] =
   {
      "Silence",
      "Sound Blaster",
      "Sound Blaster AWE32",
      "Pro Audio Spectrum",
      "Ultrasound Max (CS4231 Codec)",
      "Ultrasound",
      "Windows Sound System",
      "Ensoniq Soundscape",
   };

   return sound_name_list[num];
}

/******************************************************************************/
/*                                                                            */
/*                        VOLUME / PANNING CONTROL                            */
/*                                                                            */
/******************************************************************************/

void saInitVolPan( void )
{
   int i;
   for( i = 0; i < VOL_PAN_MAX; i++ ){
     SampleVol[i] = VOLUME_MAX;
     SamplePan[i] = PAN_CENTRE;
   }
}

void saSetVolume( int num, int data )
{
   SampleVol[num] = data <= VOLUME_MAX ? data : VOLUME_MAX;
}

void saSetPan( int num, int data )
{
   SamplePan[num] = data;
}

void saSetVolumeNolimit( int num, int data )
{
   SampleVol[num] = data;
}

void saSetVolumeMuli( int num, int data )
{
   int  i;
   for( i = 0; i < stream_joined_channels[num]; i++ )
      SampleVol[num + i] = data <= VOLUME_MAX ? data : VOLUME_MAX;
}

void saSetPanMulti( int num, int data )
{
   int  i;
   for( i = 0; i < stream_joined_channels[num]; i++ )
      SamplePan[num + i] = data;
}

void saSetVolumeMuliNolimit( int num, int data )
{
   int  i;
   for( i = 0; i < stream_joined_channels[num]; i++ )
      SampleVol[num + i] = data;
}

/******************************************************************************/
/*                                                                            */
/*                     OLD SOUND ROUTINES (REMOVE SOON)                       */
/*                                                                            */
/******************************************************************************/

signed short *SampleDest;
int SampleSpace;

int ad_sample;
int ad_index;

int ad_index_table[8] =
{
   -1, -1, -1, -1, 2, 4, 6, 8
};

int ad_offset_table[49] =
{
     16,  17,  19,  21,  23,  25,  28,  31,  34,  37,  41,  45,  50,  55,  60,  66,
     73,  80,  88,  97, 107, 118, 130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
    337, 371, 408, 449, 494, 544, 598, 658, 724, 796, 876, 963,1060,1166,1282,1411,
   1552
};

void ad_decode(int ad_code)
{
   int ad_base;
   int ad_offset;

   ad_base=ad_offset_table[ad_index];
   ad_offset=ad_base/8;

   if(ad_code&0x01)ad_offset+=ad_base>>2;
   if(ad_code&0x02)ad_offset+=ad_base>>1;
   if(ad_code&0x04)ad_offset+=ad_base;
   if(ad_code&0x08)ad_offset=0-ad_offset;

   ad_sample+=ad_offset;

   if(ad_sample<-2048)ad_sample=-2048;
   if(ad_sample>2047)ad_sample=2047;

   ad_index+=ad_index_table[ad_code&7];

   if(ad_index<0)ad_index=0;
   if(ad_index>48)ad_index=48;
}

// Used in operation wolf and cabal. hopefully it can be removed soon...

void AddSampleADPCM(int start, int end, UBYTE *SOURCE)
{
   int ta,tb;

   ad_sample = 0;
   ad_index = 0;

   tb=0;

   for(ta=start;ta<=end;ta++){
      ad_decode(SOURCE[ta]>>4);
      SampleDest[tb++] = ad_sample<<4;
      ad_decode(SOURCE[ta]&15);
      SampleDest[tb++] = ad_sample<<4;
   }

   SampleList[SampleCount].bits=16;
   SampleList[SampleCount].freq=8000;
   SampleList[SampleCount].priority=128;
   SampleList[SampleCount].len=tb;
   SampleList[SampleCount].loop_start=0;
   SampleList[SampleCount].loop_end=tb;
   SampleList[SampleCount].param=-1;
   SampleList[SampleCount].data=(UBYTE *)&SampleDest[0];

   #ifdef HACKS
      PrintDebug(
	"Sample Added...... $%02x\n"
	"Sample Size....... $%06x\n"
	"ADPCM Area........ $%06x-$%06x\n"
	"Total Sample Sizes $%06x\n"
	,SampleCount,tb,start,end,SampleSpace+tb
      );
   #endif

   SampleSpace+=tb;
   SampleDest=&SampleDest[tb];
   SampleCount++;
}

// Used Nichibutsu games...

void AddSample8Bit(int start, int end, UBYTE *SOURCE)
{
   int ta,tb;

   ad_sample = 0;
   ad_index = 0;

   tb=0;

   for(ta=start;ta<=end;ta++){
      SampleDest[tb++] = (SOURCE[ta]-0x80)<<8;
   }

   SampleList[SampleCount].bits=16;
   SampleList[SampleCount].freq=8000;
   SampleList[SampleCount].priority=128;
   SampleList[SampleCount].len=tb;
   SampleList[SampleCount].loop_start=0;
   SampleList[SampleCount].loop_end=tb;
   SampleList[SampleCount].param=-1;
   SampleList[SampleCount].data=(UBYTE *)&SampleDest[0];

   #ifdef HACKS
      PrintDebug(
	"Sample Added...... $%02x\n"
	"Sample Size....... $%06x\n"
	"8-Bit Area........ $%06x-$%06x\n"
	"Total Sample Sizes $%06x\n"
	,SampleCount,tb,start,end,SampleSpace+tb
      );
   #endif

   SampleSpace+=tb;
   SampleDest=&SampleDest[tb];
   SampleCount++;
}

void SetSampleDest(int start, UBYTE *DEST)
{
  SampleDest = (WORD *)DEST;
  SampleCount= start;
  SampleSpace= 0;
}

/******************************* END OF FILE **********************************/
