/******************************************************************************/
/*                                                                            */
/*                RAINE TILE FLIPPING AND ROTATION SUPPORT                    */
/*                                                                            */
/******************************************************************************/

#include "raine.h"
#include "games.h"
#include "newspr.h"

#include "spz16x16.h"

static int disp_x;
static int disp_y;
static int disp_x_8;
static int disp_y_8;
static int disp_x_16;
static int disp_y_16;
#undef RAINE_DOS
/*

rotation support, using jumptables and precomputing as much as we can
it's a real macro overload, but rotation overhead is now *very small*

*/

#define MAKE_ROT_JUMP( FUNC, XOFS, YOFS )                                      \
                                                                               \
                                                                               \
static void FUNC##_R090(UBYTE *SPR, int x, int y, ARGS)                        \
{                                                                              \
   FUNC##(SPR, YOFS## - y, x,          ARGS_2);                                \
}                                                                              \
                                                                               \
static void FUNC##_R180(UBYTE *SPR, int x, int y, ARGS)                        \
{                                                                              \
   FUNC##(SPR, XOFS## - x, YOFS## - y, ARGS_1);                                \
}                                                                              \
                                                                               \
static void FUNC##_R270(UBYTE *SPR, int x, int y, ARGS)                        \
{                                                                              \
   FUNC##(SPR, y,          XOFS## - x, ARGS_2);                                \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R090_X(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipY(SPR, YOFS## - y, x,          ARGS_2);                          \
}                                                                              \
                                                                               \
static void FUNC##_R180_X(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipX(SPR, XOFS## - x, YOFS## - y, ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R270_X(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipY(SPR, y,          XOFS## - x, ARGS_2);                          \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R090_Y(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipX(SPR, YOFS## - y, x,          ARGS_2);                          \
}                                                                              \
                                                                               \
static void FUNC##_R180_Y(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipY(SPR, XOFS## - x, YOFS## - y, ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R270_Y(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipX(SPR, y,          XOFS## - x, ARGS_2);                          \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R090_XY(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_FlipXY(SPR, YOFS## - y, x,          ARGS_2);                         \
}                                                                              \
                                                                               \
static void FUNC##_R180_XY(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_FlipXY(SPR, XOFS## - x, YOFS## - y, ARGS_1);                         \
}                                                                              \
                                                                               \
static void FUNC##_R270_XY(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_FlipXY(SPR, y,          XOFS## - x, ARGS_2);                         \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R000_FX(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##(SPR, XOFS## - x, y,          ARGS_1);                                \
}                                                                              \
                                                                               \
static void FUNC##_R090_FX(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##(SPR, y,          x,          ARGS_2);                                \
}                                                                              \
                                                                               \
static void FUNC##_R180_FX(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##(SPR, x,          YOFS## - y, ARGS_1);                                \
}                                                                              \
                                                                               \
static void FUNC##_R270_FX(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##(SPR, YOFS## - y, XOFS## - x, ARGS_2);                                \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R000_X_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipX(SPR, XOFS## - x, y,          ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R090_X_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipY(SPR, y,          x,          ARGS_2);                          \
}                                                                              \
                                                                               \
static void FUNC##_R180_X_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipX(SPR, x,          YOFS## - y, ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R270_X_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipY(SPR, YOFS## - y, XOFS## - x, ARGS_2);                          \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R000_Y_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipY(SPR, XOFS## - x, y,          ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R090_Y_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipX(SPR, y,          x,          ARGS_2);                          \
}                                                                              \
                                                                               \
static void FUNC##_R180_Y_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipY(SPR, x,          YOFS## - y, ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R270_Y_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipX(SPR, YOFS## - y, XOFS## - x, ARGS_2);                          \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R000_XY_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_FlipXY(SPR, XOFS## - x, y,          ARGS_1);                         \
}                                                                              \
                                                                               \
static void FUNC##_R090_XY_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_FlipXY(SPR, y,          x,          ARGS_2);                         \
}                                                                              \
                                                                               \
static void FUNC##_R180_XY_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_FlipXY(SPR, x,          YOFS## - y, ARGS_1);                         \
}                                                                              \
                                                                               \
static void FUNC##_R270_XY_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_FlipXY(SPR, YOFS## - y, XOFS## - x, ARGS_2);                         \
}                                                                              \
                                                                               \
                                                                               \
CMAP_FUNC *FUNC##_Flip_Rot[4];                                                 \
                                                                               \
static void FUNC##_init_jump_table(void)                                       \
{                                                                              \
   ULONG i;                                                                    \
                                                                               \
   i  = display_cfg.rotate;                                                    \
   i ^= display_cfg.flip & 2;                                                  \
                                                                               \
   if((display_cfg.flip == 0) || (display_cfg.flip == 3)){                     \
                                                                               \
   switch(i){                                                                  \
   case 0:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##;                                            \
      FUNC##_Flip_Rot[1] = &FUNC##_FlipY;                                      \
      FUNC##_Flip_Rot[2] = &FUNC##_FlipX;                                      \
      FUNC##_Flip_Rot[3] = &FUNC##_FlipXY;                                     \
   break;                                                                      \
   case 1:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R090;                                       \
      FUNC##_Flip_Rot[1] = &FUNC##_R090_Y;                                     \
      FUNC##_Flip_Rot[2] = &FUNC##_R090_X;                                     \
      FUNC##_Flip_Rot[3] = &FUNC##_R090_XY;                                    \
   break;                                                                      \
   case 2:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R180;                                       \
      FUNC##_Flip_Rot[1] = &FUNC##_R180_Y;                                     \
      FUNC##_Flip_Rot[2] = &FUNC##_R180_X;                                     \
      FUNC##_Flip_Rot[3] = &FUNC##_R180_XY;                                    \
   break;                                                                      \
   case 3:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R270;                                       \
      FUNC##_Flip_Rot[1] = &FUNC##_R270_Y;                                     \
      FUNC##_Flip_Rot[2] = &FUNC##_R270_X;                                     \
      FUNC##_Flip_Rot[3] = &FUNC##_R270_XY;                                    \
   break;                                                                      \
   };                                                                          \
                                                                               \
   }                                                                           \
   else{                                                                       \
                                                                               \
   switch(i){                                                                  \
   case 0:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R000_FX;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_R000_Y_FX;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_R000_X_FX;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_R000_XY_FX;                                 \
   break;                                                                      \
   case 1:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R090_FX;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_R090_Y_FX;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_R090_X_FX;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_R090_XY_FX;                                 \
   break;                                                                      \
   case 2:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R180_FX;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_R180_Y_FX;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_R180_X_FX;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_R180_XY_FX;                                 \
   break;                                                                      \
   case 3:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R270_FX;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_R270_Y_FX;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_R270_X_FX;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_R270_XY_FX;                                 \
   break;                                                                      \
   };                                                                          \
                                                                               \
   };                                                                          \
};                                                                             \
                                                                               \



/*

rotation support, using jumptables and precomputing as much as we can
it's a real macro overload, but rotation overhead is now *very small*

*/

#define MAKE_ROT_JUMP_16( FUNC, XOFS, YOFS )                                   \
                                                                               \
                                                                               \
static void FUNC##_R090(UBYTE *SPR, int x, int y, ARGS)                        \
{                                                                              \
   FUNC##(SPR, YOFS## - y, x,          ARGS_2);                                \
}                                                                              \
                                                                               \
static void FUNC##_R180(UBYTE *SPR, int x, int y, ARGS)                        \
{                                                                              \
   FUNC##(SPR, XOFS## - x, YOFS## - y, ARGS_1);                                \
}                                                                              \
                                                                               \
static void FUNC##_R270(UBYTE *SPR, int x, int y, ARGS)                        \
{                                                                              \
   FUNC##(SPR, y,          XOFS## - x, ARGS_2);                                \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R090_X(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipY(SPR, YOFS## - y, x,          ARGS_2);                          \
}                                                                              \
                                                                               \
static void FUNC##_R180_X(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipX(SPR, XOFS## - x, YOFS## - y, ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R270_X(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipY(SPR, y,          XOFS## - x, ARGS_2);                          \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R090_Y(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipX(SPR, YOFS## - y, x,          ARGS_2);                          \
}                                                                              \
                                                                               \
static void FUNC##_R180_Y(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipY(SPR, XOFS## - x, YOFS## - y, ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R270_Y(UBYTE *SPR, int x, int y, ARGS)                      \
{                                                                              \
   FUNC##_FlipX(SPR, y,          XOFS## - x, ARGS_2);                          \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R090_XY(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_FlipXY(SPR, YOFS## - y, x,          ARGS_2);                         \
}                                                                              \
                                                                               \
static void FUNC##_R180_XY(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_FlipXY(SPR, XOFS## - x, YOFS## - y, ARGS_1);                         \
}                                                                              \
                                                                               \
static void FUNC##_R270_XY(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_FlipXY(SPR, y,          XOFS## - x, ARGS_2);                         \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R000_FX(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##(SPR, XOFS## - x, y,          ARGS_1);                                \
}                                                                              \
                                                                               \
static void FUNC##_R090_FX(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##(SPR, y,          x,          ARGS_2);                                \
}                                                                              \
                                                                               \
static void FUNC##_R180_FX(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##(SPR, x,          YOFS## - y, ARGS_1);                                \
}                                                                              \
                                                                               \
static void FUNC##_R270_FX(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##(SPR, YOFS## - y, XOFS## - x, ARGS_2);                                \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R000_X_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipX(SPR, XOFS## - x, y,          ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R090_X_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipY(SPR, y,          x,          ARGS_2);                          \
}                                                                              \
                                                                               \
static void FUNC##_R180_X_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipX(SPR, x,          YOFS## - y, ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R270_X_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipY(SPR, YOFS## - y, XOFS## - x, ARGS_2);                          \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R000_Y_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipY(SPR, XOFS## - x, y,          ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R090_Y_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipX(SPR, y,          x,          ARGS_2);                          \
}                                                                              \
                                                                               \
static void FUNC##_R180_Y_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipY(SPR, x,          YOFS## - y, ARGS_1);                          \
}                                                                              \
                                                                               \
static void FUNC##_R270_Y_FX(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_FlipX(SPR, YOFS## - y, XOFS## - x, ARGS_2);                          \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_R000_XY_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_FlipXY(SPR, XOFS## - x, y,          ARGS_1);                         \
}                                                                              \
                                                                               \
static void FUNC##_R090_XY_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_FlipXY(SPR, y,          x,          ARGS_2);                         \
}                                                                              \
                                                                               \
static void FUNC##_R180_XY_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_FlipXY(SPR, x,          YOFS## - y, ARGS_1);                         \
}                                                                              \
                                                                               \
static void FUNC##_R270_XY_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_FlipXY(SPR, YOFS## - y, XOFS## - x, ARGS_2);                         \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_16_R090(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_16(SPR, YOFS## - y, x,          ARGS_2);                             \
}                                                                              \
                                                                               \
static void FUNC##_16_R180(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_16(SPR, XOFS## - x, YOFS## - y, ARGS_1);                             \
}                                                                              \
                                                                               \
static void FUNC##_16_R270(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_16(SPR, y,          XOFS## - x, ARGS_2);                             \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_16_R090_X(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_16_FlipY(SPR, YOFS## - y, x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_16_R180_X(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_16_FlipX(SPR, XOFS## - x, YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_16_R270_X(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_16_FlipY(SPR, y,          XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_16_R090_Y(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_16_FlipX(SPR, YOFS## - y, x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_16_R180_Y(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_16_FlipY(SPR, XOFS## - x, YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_16_R270_Y(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_16_FlipX(SPR, y,          XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_16_R090_XY(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_16_FlipXY(SPR, YOFS## - y, x,          ARGS_2);                      \
}                                                                              \
                                                                               \
static void FUNC##_16_R180_XY(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_16_FlipXY(SPR, XOFS## - x, YOFS## - y, ARGS_1);                      \
}                                                                              \
                                                                               \
static void FUNC##_16_R270_XY(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_16_FlipXY(SPR, y,          XOFS## - x, ARGS_2);                      \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_16_R000_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_16(SPR, XOFS## - x, y,          ARGS_1);                             \
}                                                                              \
                                                                               \
static void FUNC##_16_R090_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_16(SPR, y,          x,          ARGS_2);                             \
}                                                                              \
                                                                               \
static void FUNC##_16_R180_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_16(SPR, x,          YOFS## - y, ARGS_1);                             \
}                                                                              \
                                                                               \
static void FUNC##_16_R270_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_16(SPR, YOFS## - y, XOFS## - x, ARGS_2);                             \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_16_R000_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_16_FlipX(SPR, XOFS## - x, y,          ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_16_R090_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_16_FlipY(SPR, y,          x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_16_R180_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_16_FlipX(SPR, x,          YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_16_R270_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_16_FlipY(SPR, YOFS## - y, XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_16_R000_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_16_FlipY(SPR, XOFS## - x, y,          ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_16_R090_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_16_FlipX(SPR, y,          x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_16_R180_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_16_FlipY(SPR, x,          YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_16_R270_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_16_FlipX(SPR, YOFS## - y, XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_16_R000_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_16_FlipXY(SPR, XOFS## - x, y,          ARGS_1);                      \
}                                                                              \
                                                                               \
static void FUNC##_16_R090_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_16_FlipXY(SPR, y,          x,          ARGS_2);                      \
}                                                                              \
                                                                               \
static void FUNC##_16_R180_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_16_FlipXY(SPR, x,          YOFS## - y, ARGS_1);                      \
}                                                                              \
                                                                               \
static void FUNC##_16_R270_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_16_FlipXY(SPR, YOFS## - y, XOFS## - x, ARGS_2);                      \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_24_R090(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_24(SPR, YOFS## - y, x,          ARGS_2);                             \
}                                                                              \
                                                                               \
static void FUNC##_24_R180(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_24(SPR, XOFS## - x, YOFS## - y, ARGS_1);                             \
}                                                                              \
                                                                               \
static void FUNC##_24_R270(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_24(SPR, y,          XOFS## - x, ARGS_2);                             \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_24_R090_X(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_24_FlipY(SPR, YOFS## - y, x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_24_R180_X(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_24_FlipX(SPR, XOFS## - x, YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_24_R270_X(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_24_FlipY(SPR, y,          XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_24_R090_Y(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_24_FlipX(SPR, YOFS## - y, x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_24_R180_Y(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_24_FlipY(SPR, XOFS## - x, YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_24_R270_Y(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_24_FlipX(SPR, y,          XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_24_R090_XY(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_24_FlipXY(SPR, YOFS## - y, x,          ARGS_2);                      \
}                                                                              \
                                                                               \
static void FUNC##_24_R180_XY(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_24_FlipXY(SPR, XOFS## - x, YOFS## - y, ARGS_1);                      \
}                                                                              \
                                                                               \
static void FUNC##_24_R270_XY(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_24_FlipXY(SPR, y,          XOFS## - x, ARGS_2);                      \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_24_R000_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_24(SPR, XOFS## - x, y,          ARGS_1);                             \
}                                                                              \
                                                                               \
static void FUNC##_24_R090_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_24(SPR, y,          x,          ARGS_2);                             \
}                                                                              \
                                                                               \
static void FUNC##_24_R180_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_24(SPR, x,          YOFS## - y, ARGS_1);                             \
}                                                                              \
                                                                               \
static void FUNC##_24_R270_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_24(SPR, YOFS## - y, XOFS## - x, ARGS_2);                             \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_24_R000_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_24_FlipX(SPR, XOFS## - x, y,          ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_24_R090_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_24_FlipY(SPR, y,          x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_24_R180_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_24_FlipX(SPR, x,          YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_24_R270_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_24_FlipY(SPR, YOFS## - y, XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_24_R000_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_24_FlipY(SPR, XOFS## - x, y,          ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_24_R090_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_24_FlipX(SPR, y,          x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_24_R180_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_24_FlipY(SPR, x,          YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_24_R270_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_24_FlipX(SPR, YOFS## - y, XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_24_R000_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_24_FlipXY(SPR, XOFS## - x, y,          ARGS_1);                      \
}                                                                              \
                                                                               \
static void FUNC##_24_R090_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_24_FlipXY(SPR, y,          x,          ARGS_2);                      \
}                                                                              \
                                                                               \
static void FUNC##_24_R180_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_24_FlipXY(SPR, x,          YOFS## - y, ARGS_1);                      \
}                                                                              \
                                                                               \
static void FUNC##_24_R270_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_24_FlipXY(SPR, YOFS## - y, XOFS## - x, ARGS_2);                      \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_32_R090(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_32(SPR, YOFS## - y, x,          ARGS_2);                             \
}                                                                              \
                                                                               \
static void FUNC##_32_R180(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_32(SPR, XOFS## - x, YOFS## - y, ARGS_1);                             \
}                                                                              \
                                                                               \
static void FUNC##_32_R270(UBYTE *SPR, int x, int y, ARGS)                     \
{                                                                              \
   FUNC##_32(SPR, y,          XOFS## - x, ARGS_2);                             \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_32_R090_X(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_32_FlipY(SPR, YOFS## - y, x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_32_R180_X(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_32_FlipX(SPR, XOFS## - x, YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_32_R270_X(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_32_FlipY(SPR, y,          XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_32_R090_Y(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_32_FlipX(SPR, YOFS## - y, x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_32_R180_Y(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_32_FlipY(SPR, XOFS## - x, YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_32_R270_Y(UBYTE *SPR, int x, int y, ARGS)                   \
{                                                                              \
   FUNC##_32_FlipX(SPR, y,          XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_32_R090_XY(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_32_FlipXY(SPR, YOFS## - y, x,          ARGS_2);                      \
}                                                                              \
                                                                               \
static void FUNC##_32_R180_XY(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_32_FlipXY(SPR, XOFS## - x, YOFS## - y, ARGS_1);                      \
}                                                                              \
                                                                               \
static void FUNC##_32_R270_XY(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_32_FlipXY(SPR, y,          XOFS## - x, ARGS_2);                      \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_32_R000_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_32(SPR, XOFS## - x, y,          ARGS_1);                             \
}                                                                              \
                                                                               \
static void FUNC##_32_R090_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_32(SPR, y,          x,          ARGS_2);                             \
}                                                                              \
                                                                               \
static void FUNC##_32_R180_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_32(SPR, x,          YOFS## - y, ARGS_1);                             \
}                                                                              \
                                                                               \
static void FUNC##_32_R270_FX(UBYTE *SPR, int x, int y, ARGS)                  \
{                                                                              \
   FUNC##_32(SPR, YOFS## - y, XOFS## - x, ARGS_2);                             \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_32_R000_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_32_FlipX(SPR, XOFS## - x, y,          ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_32_R090_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_32_FlipY(SPR, y,          x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_32_R180_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_32_FlipX(SPR, x,          YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_32_R270_X_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_32_FlipY(SPR, YOFS## - y, XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_32_R000_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_32_FlipY(SPR, XOFS## - x, y,          ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_32_R090_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_32_FlipX(SPR, y,          x,          ARGS_2);                       \
}                                                                              \
                                                                               \
static void FUNC##_32_R180_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_32_FlipY(SPR, x,          YOFS## - y, ARGS_1);                       \
}                                                                              \
                                                                               \
static void FUNC##_32_R270_Y_FX(UBYTE *SPR, int x, int y, ARGS)                \
{                                                                              \
   FUNC##_32_FlipX(SPR, YOFS## - y, XOFS## - x, ARGS_2);                       \
}                                                                              \
                                                                               \
                                                                               \
static void FUNC##_32_R000_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_32_FlipXY(SPR, XOFS## - x, y,          ARGS_1);                      \
}                                                                              \
                                                                               \
static void FUNC##_32_R090_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_32_FlipXY(SPR, y,          x,          ARGS_2);                      \
}                                                                              \
                                                                               \
static void FUNC##_32_R180_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_32_FlipXY(SPR, x,          YOFS## - y, ARGS_1);                      \
}                                                                              \
                                                                               \
static void FUNC##_32_R270_XY_FX(UBYTE *SPR, int x, int y, ARGS)               \
{                                                                              \
   FUNC##_32_FlipXY(SPR, YOFS## - y, XOFS## - x, ARGS_2);                      \
}                                                                              \
                                                                               \
                                                                               \
CMAP_FUNC *FUNC##_Flip_Rot[4];                                                 \
                                                                               \
static void FUNC##_init_jump_table(void)                                       \
{                                                                              \
   ULONG i;                                                                    \
                                                                               \
   i  = display_cfg.rotate;                                                    \
   i ^= display_cfg.flip & 2;                                                  \
                                                                               \
   switch(display_cfg.bpp){                                                    \
   case 8:                                                                     \
                                                                               \
   if((display_cfg.flip == 0) || (display_cfg.flip == 3)){                     \
                                                                               \
   switch(i){                                                                  \
   case 0:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##;                                            \
      FUNC##_Flip_Rot[1] = &FUNC##_FlipY;                                      \
      FUNC##_Flip_Rot[2] = &FUNC##_FlipX;                                      \
      FUNC##_Flip_Rot[3] = &FUNC##_FlipXY;                                     \
   break;                                                                      \
   case 1:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R090;                                       \
      FUNC##_Flip_Rot[1] = &FUNC##_R090_Y;                                     \
      FUNC##_Flip_Rot[2] = &FUNC##_R090_X;                                     \
      FUNC##_Flip_Rot[3] = &FUNC##_R090_XY;                                    \
   break;                                                                      \
   case 2:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R180;                                       \
      FUNC##_Flip_Rot[1] = &FUNC##_R180_Y;                                     \
      FUNC##_Flip_Rot[2] = &FUNC##_R180_X;                                     \
      FUNC##_Flip_Rot[3] = &FUNC##_R180_XY;                                    \
   break;                                                                      \
   case 3:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R270;                                       \
      FUNC##_Flip_Rot[1] = &FUNC##_R270_Y;                                     \
      FUNC##_Flip_Rot[2] = &FUNC##_R270_X;                                     \
      FUNC##_Flip_Rot[3] = &FUNC##_R270_XY;                                    \
   break;                                                                      \
   };                                                                          \
                                                                               \
   }                                                                           \
   else{                                                                       \
                                                                               \
   switch(i){                                                                  \
   case 0:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R000_FX;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_R000_Y_FX;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_R000_X_FX;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_R000_XY_FX;                                 \
   break;                                                                      \
   case 1:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R090_FX;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_R090_Y_FX;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_R090_X_FX;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_R090_XY_FX;                                 \
   break;                                                                      \
   case 2:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R180_FX;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_R180_Y_FX;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_R180_X_FX;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_R180_XY_FX;                                 \
   break;                                                                      \
   case 3:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_R270_FX;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_R270_Y_FX;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_R270_X_FX;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_R270_XY_FX;                                 \
   break;                                                                      \
   };                                                                          \
                                                                               \
   };                                                                          \
                                                                               \
   break;                                                                      \
   case 15:                                                                    \
   case 16:                                                                    \
                                                                               \
   if((display_cfg.flip == 0) || (display_cfg.flip == 3)){                     \
                                                                               \
   switch(i){                                                                  \
   case 0:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_16;                                         \
      FUNC##_Flip_Rot[1] = &FUNC##_16_FlipY;                                   \
      FUNC##_Flip_Rot[2] = &FUNC##_16_FlipX;                                   \
      FUNC##_Flip_Rot[3] = &FUNC##_16_FlipXY;                                  \
   break;                                                                      \
   case 1:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_16_R090;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_16_R090_Y;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_16_R090_X;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_16_R090_XY;                                 \
   break;                                                                      \
   case 2:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_16_R180;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_16_R180_Y;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_16_R180_X;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_16_R180_XY;                                 \
   break;                                                                      \
   case 3:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_16_R270;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_16_R270_Y;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_16_R270_X;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_16_R270_XY;                                 \
   break;                                                                      \
   };                                                                          \
                                                                               \
   }                                                                           \
   else{                                                                       \
                                                                               \
   switch(i){                                                                  \
   case 0:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_16_R000_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_16_R000_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_16_R000_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_16_R000_XY_FX;                              \
   break;                                                                      \
   case 1:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_16_R090_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_16_R090_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_16_R090_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_16_R090_XY_FX;                              \
   break;                                                                      \
   case 2:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_16_R180_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_16_R180_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_16_R180_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_16_R180_XY_FX;                              \
   break;                                                                      \
   case 3:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_16_R270_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_16_R270_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_16_R270_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_16_R270_XY_FX;                              \
   break;                                                                      \
   };                                                                          \
                                                                               \
   };                                                                          \
                                                                               \
   break;                                                                      \
   case 24:                                                                    \
                                                                               \
   if((display_cfg.flip == 0) || (display_cfg.flip == 3)){                     \
                                                                               \
   switch(i){                                                                  \
   case 0:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_24;                                         \
      FUNC##_Flip_Rot[1] = &FUNC##_24_FlipY;                                   \
      FUNC##_Flip_Rot[2] = &FUNC##_24_FlipX;                                   \
      FUNC##_Flip_Rot[3] = &FUNC##_24_FlipXY;                                  \
   break;                                                                      \
   case 1:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_24_R090;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_24_R090_Y;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_24_R090_X;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_24_R090_XY;                                 \
   break;                                                                      \
   case 2:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_24_R180;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_24_R180_Y;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_24_R180_X;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_24_R180_XY;                                 \
   break;                                                                      \
   case 3:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_24_R270;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_24_R270_Y;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_24_R270_X;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_24_R270_XY;                                 \
   break;                                                                      \
   };                                                                          \
                                                                               \
   }                                                                           \
   else{                                                                       \
                                                                               \
   switch(i){                                                                  \
   case 0:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_24_R000_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_24_R000_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_24_R000_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_24_R000_XY_FX;                              \
   break;                                                                      \
   case 1:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_24_R090_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_24_R090_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_24_R090_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_24_R090_XY_FX;                              \
   break;                                                                      \
   case 2:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_24_R180_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_24_R180_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_24_R180_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_24_R180_XY_FX;                              \
   break;                                                                      \
   case 3:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_24_R270_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_24_R270_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_24_R270_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_24_R270_XY_FX;                              \
   break;                                                                      \
   };                                                                          \
                                                                               \
   };                                                                          \
                                                                               \
   break;                                                                      \
   case 32:                                                                    \
                                                                               \
   if((display_cfg.flip == 0) || (display_cfg.flip == 3)){                     \
                                                                               \
   switch(i){                                                                  \
   case 0:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_32;                                         \
      FUNC##_Flip_Rot[1] = &FUNC##_32_FlipY;                                   \
      FUNC##_Flip_Rot[2] = &FUNC##_32_FlipX;                                   \
      FUNC##_Flip_Rot[3] = &FUNC##_32_FlipXY;                                  \
   break;                                                                      \
   case 1:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_32_R090;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_32_R090_Y;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_32_R090_X;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_32_R090_XY;                                 \
   break;                                                                      \
   case 2:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_32_R180;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_32_R180_Y;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_32_R180_X;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_32_R180_XY;                                 \
   break;                                                                      \
   case 3:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_32_R270;                                    \
      FUNC##_Flip_Rot[1] = &FUNC##_32_R270_Y;                                  \
      FUNC##_Flip_Rot[2] = &FUNC##_32_R270_X;                                  \
      FUNC##_Flip_Rot[3] = &FUNC##_32_R270_XY;                                 \
   break;                                                                      \
   };                                                                          \
                                                                               \
   }                                                                           \
   else{                                                                       \
                                                                               \
   switch(i){                                                                  \
   case 0:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_32_R000_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_32_R000_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_32_R000_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_32_R000_XY_FX;                              \
   break;                                                                      \
   case 1:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_32_R090_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_32_R090_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_32_R090_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_32_R090_XY_FX;                              \
   break;                                                                      \
   case 2:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_32_R180_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_32_R180_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_32_R180_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_32_R180_XY_FX;                              \
   break;                                                                      \
   case 3:                                                                     \
      FUNC##_Flip_Rot[0] = &FUNC##_32_R270_FX;                                 \
      FUNC##_Flip_Rot[1] = &FUNC##_32_R270_Y_FX;                               \
      FUNC##_Flip_Rot[2] = &FUNC##_32_R270_X_FX;                               \
      FUNC##_Flip_Rot[3] = &FUNC##_32_R270_XY_FX;                              \
   break;                                                                      \
   };                                                                          \
                                                                               \
   };                                                                          \
                                                                               \
   break;                                                                      \
   };                                                                          \
};                                                                             \
                                                                               \

/*

prepare for mapper direct routines

*/

#undef ARGS
#undef ARGS_1
#undef ARGS_2
#undef CMAP_FUNC

#define ARGS      UBYTE cmap
#define ARGS_1    cmap
#define ARGS_2    cmap
#define CMAP_FUNC draw_func

/*

generate direct mapped routines

*/

MAKE_ROT_JUMP( Draw8x8,       disp_x_8, disp_y_8 )

MAKE_ROT_JUMP( Draw8x8_Trans, disp_x_8, disp_y_8 )


/*

prepare for colour mapped routines

*/

#undef ARGS
#undef ARGS_1
#undef ARGS_2
#undef CMAP_FUNC

#define ARGS      UBYTE *cmap
#define ARGS_1    cmap
#define ARGS_2    cmap
#define CMAP_FUNC draw_mapped_func

/*

generate mapped routines

*/

MAKE_ROT_JUMP_16( Draw8x8_Mapped,         disp_x_8,  disp_y_8  )

MAKE_ROT_JUMP_16( Draw8x8_Trans_Mapped,   disp_x_8,  disp_y_8  )

MAKE_ROT_JUMP_16( Draw16x16_Mapped,       disp_x_16, disp_y_16 )

MAKE_ROT_JUMP_16( Draw16x16_Trans_Mapped, disp_x_16, disp_y_16 )

/*

prepare for colour mapped zoom xy routines

*/

#undef ARGS
#undef ARGS_1
#undef ARGS_2
#undef CMAP_FUNC

#define ARGS      UBYTE *cmap, int zoom_x, int zoom_y
#define ARGS_1    cmap, zoom_x, zoom_y
#define ARGS_2    cmap, zoom_y, zoom_x
#define CMAP_FUNC draw_mapped_zoom_xy_func

/*

generate mapped routines

*/

MAKE_ROT_JUMP( Draw16x16_Mapped_ZoomXY,       (disp_x - zoom_x), (disp_y - zoom_y) )

MAKE_ROT_JUMP( Draw16x16_Trans_Mapped_ZoomXY, (disp_x - zoom_x), (disp_y - zoom_y) )

/******************************************************************************/

/* Normal Display */

draw_mapped_func *Draw16x16_Mapped_JumpList[4] =
{
   Draw16x16_Mapped,
   Draw16x16_Mapped_FlipY,
   Draw16x16_Mapped_FlipX,
   Draw16x16_Mapped_FlipXY,
};

/* X+Y Flipped Display */

draw_mapped_func *Draw16x16_Mapped_JumpListR[4] =
{
   Draw16x16_Mapped_FlipXY,
   Draw16x16_Mapped_FlipX,
   Draw16x16_Mapped_FlipY,
   Draw16x16_Mapped,
};

/* X+Y Flipped Display */

draw_mapped_func *Draw16x16_Trans_Mapped_JumpListR[4] =
{
   Draw16x16_Trans_Mapped_FlipXY,
   Draw16x16_Trans_Mapped_FlipX,
   Draw16x16_Trans_Mapped_FlipY,
   Draw16x16_Trans_Mapped,
};

/******************************************************************************/

/* Only used Tecmo System */

void Draw16x16_Part8x8_Trans_Mapped(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc8+2
	xorl	%%edx,%%edx

	movl	$8,%%ebp	
9:
	movl	(%%esi),%%eax
	movl	4(%%esi),%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
7:
	addl	$16,%%esi		
smc8:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx"
   );
#endif
}

/********************************************************************/

/* Only used Space Gun */

void Draw64x64_Mapped(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc17+2
	xorl	%%edx,%%edx

	movl	$64,%%ebp	
9:
	movl	(%%esi),%%eax
	movl	4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,(%%edi)
	movl	%%ebx,4(%%edi)

	movl	8(%%esi),%%eax
	movl	12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,8(%%edi)
	movl	%%ebx,12(%%edi)

	movl	16(%%esi),%%eax
	movl	16+4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,16(%%edi)
	movl	%%ebx,16+4(%%edi)

	movl	16+8(%%esi),%%eax
	movl	16+12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,16+8(%%edi)
	movl	%%ebx,16+12(%%edi)

	movl	32(%%esi),%%eax
	movl	32+4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,32(%%edi)
	movl	%%ebx,32+4(%%edi)

	movl	32+8(%%esi),%%eax
	movl	32+12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,32+8(%%edi)
	movl	%%ebx,32+12(%%edi)

	movl	48(%%esi),%%eax
	movl	48+4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,48(%%edi)
	movl	%%ebx,48+4(%%edi)

	movl	48+8(%%esi),%%eax
	movl	48+12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,48+8(%%edi)
	movl	%%ebx,48+12(%%edi)

	addl	$64,%%esi		
smc17:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx"
   );
#endif
}

void Draw64x64_Mapped_FlipY(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc18+2
	xorl	%%edx,%%edx

	movl	$64,%%ebp	
9:
	movl	(%%esi),%%eax
	movl	4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,60(%%edi)
	movl	%%ebx,56(%%edi)

	movl	8(%%esi),%%eax
	movl	12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,52(%%edi)
	movl	%%ebx,48(%%edi)

	movl	16(%%esi),%%eax
	movl	16+4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,44(%%edi)
	movl	%%ebx,40(%%edi)

	movl	16+8(%%esi),%%eax
	movl	16+12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,36(%%edi)
	movl	%%ebx,32(%%edi)

	movl	32(%%esi),%%eax
	movl	36(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,28(%%edi)
	movl	%%ebx,24(%%edi)

	movl	40(%%esi),%%eax
	movl	44(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,20(%%edi)
	movl	%%ebx,16(%%edi)

	movl	48(%%esi),%%eax
	movl	52(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,12(%%edi)
	movl	%%ebx,8(%%edi)

	movl	56(%%esi),%%eax
	movl	60(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,4(%%edi)
	movl	%%ebx,(%%edi)

	addl	$64,%%esi		
smc18:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx"
   );
#endif
}

void Draw64x64_Mapped_FlipX(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc19+2
	xorl	%%edx,%%edx

	movl	$64,%%ebp	
9:
	movl	(%%esi),%%eax
	movl	4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,(%%edi)
	movl	%%ebx,4(%%edi)

	movl	8(%%esi),%%eax
	movl	12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,8(%%edi)
	movl	%%ebx,12(%%edi)

	movl	16(%%esi),%%eax
	movl	16+4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,16(%%edi)
	movl	%%ebx,16+4(%%edi)

	movl	16+8(%%esi),%%eax
	movl	16+12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,16+8(%%edi)
	movl	%%ebx,16+12(%%edi)

	movl	32(%%esi),%%eax
	movl	32+4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,32(%%edi)
	movl	%%ebx,32+4(%%edi)

	movl	32+8(%%esi),%%eax
	movl	32+12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,32+8(%%edi)
	movl	%%ebx,32+12(%%edi)

	movl	48(%%esi),%%eax
	movl	48+4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,48(%%edi)
	movl	%%ebx,48+4(%%edi)

	movl	48+8(%%esi),%%eax
	movl	48+12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movl	%%eax,48+8(%%edi)
	movl	%%ebx,48+12(%%edi)

	addl	$64,%%esi		
smc19:	subl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y+63]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx"
   );
#endif
}

void Draw64x64_Mapped_FlipXY(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc20+2
	xorl	%%edx,%%edx

	movl	$64,%%ebp	
9:
	movl	(%%esi),%%eax
	movl	4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,60(%%edi)
	movl	%%ebx,56(%%edi)

	movl	8(%%esi),%%eax
	movl	12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,52(%%edi)
	movl	%%ebx,48(%%edi)

	movl	16(%%esi),%%eax
	movl	16+4(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,44(%%edi)
	movl	%%ebx,40(%%edi)

	movl	16+8(%%esi),%%eax
	movl	16+12(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,36(%%edi)
	movl	%%ebx,32(%%edi)

	movl	32(%%esi),%%eax
	movl	36(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,28(%%edi)
	movl	%%ebx,24(%%edi)

	movl	40(%%esi),%%eax
	movl	44(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,20(%%edi)
	movl	%%ebx,16(%%edi)

	movl	48(%%esi),%%eax
	movl	52(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,12(%%edi)
	movl	%%ebx,8(%%edi)

	movl	56(%%esi),%%eax
	movl	60(%%esi),%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	bswap	%%eax
	bswap	%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%bl
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%ah
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%bh

	movl	%%eax,4(%%edi)
	movl	%%ebx,(%%edi)

	addl	$64,%%esi		
smc20:	subl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y+63]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx"
   );
#endif
}

/********************************************************************/

void Draw64x64_Trans_Mapped(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc21+2
	xorl	%%edx,%%edx

	movl	$64,%%ebp	
9:
	movl	(%%esi),%%eax
	movl	4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
7:
	movl	8(%%esi),%%eax
	movl	12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,12(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,9(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,13(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,10(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,14(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,11(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,15(%%edi)
7:
	movl	16(%%esi),%%eax
	movl	16+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+4(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+1(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+2(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+6(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+3(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+7(%%edi)
7:
	movl	16+8(%%esi),%%eax
	movl	16+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+8(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+12(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+9(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+13(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+10(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+14(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+11(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+15(%%edi)
7:
	movl	32(%%esi),%%eax
	movl	32+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+4(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+1(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+2(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+6(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+3(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+7(%%edi)
7:
	movl	32+8(%%esi),%%eax
	movl	32+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+8(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+12(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+9(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+13(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+10(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+14(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+11(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+15(%%edi)
7:
	movl	48(%%esi),%%eax
	movl	48+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+4(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+1(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+2(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+6(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+3(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+7(%%edi)
7:
	movl	48+8(%%esi),%%eax
	movl	48+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+8(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+12(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+9(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+13(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+10(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+14(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+11(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+15(%%edi)
7:
	addl	$64,%%esi	
smc21:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx"
   );
#endif
}

void Draw64x64_Trans_Mapped_FlipY(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc22+2
	xorl	%%edx,%%edx

	movl	$64,%%ebp	
9:
	movl	(%%esi),%%eax
	movl	4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-4(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-1(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-2(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-6(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-3(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-7(%%edi)
7:
	movl	8(%%esi),%%eax
	movl	12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-8(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-12(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-9(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-13(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-10(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-14(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-11(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-15(%%edi)
7:
	movl	16(%%esi),%%eax
	movl	16+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-16(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-20(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-17(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-21(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-18(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-22(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-19(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-23(%%edi)
7:
	movl	16+8(%%esi),%%eax
	movl	16+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-24(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-28(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-25(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-29(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-26(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-30(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-27(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-31(%%edi)
7:
	movl	32(%%esi),%%eax
	movl	32+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-32(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-36(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-33(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-37(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-34(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-38(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-35(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-39(%%edi)
7:
	movl	32+8(%%esi),%%eax
	movl	32+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-40(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-44(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-41(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-45(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-42(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-46(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-43(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-47(%%edi)
7:
	movl	48(%%esi),%%eax
	movl	48+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-48(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-52(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-49(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-53(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-50(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-54(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-51(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-55(%%edi)
7:
	movl	48+8(%%esi),%%eax
	movl	48+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-56(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-60(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-57(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-61(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-58(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-62(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-59(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:
	addl	$64,%%esi		
smc22:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx"
   );
#endif
}

void Draw64x64_Trans_Mapped_FlipX(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc23+2
	xorl	%%edx,%%edx

	movl	$64,%%ebp		
9:
	movl	(%%esi),%%eax
	movl	4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
7:
	movl	8(%%esi),%%eax
	movl	12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,12(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,9(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,13(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,10(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,14(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,11(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,15(%%edi)
7:
	movl	16(%%esi),%%eax
	movl	16+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+4(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+1(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+2(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+6(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+3(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+7(%%edi)
7:
	movl	16+8(%%esi),%%eax
	movl	16+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+8(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+12(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+9(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+13(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+10(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+14(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+11(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,16+15(%%edi)
7:
	movl	32(%%esi),%%eax
	movl	32+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+4(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+1(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+2(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+6(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+3(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+7(%%edi)
7:
	movl	32+8(%%esi),%%eax
	movl	32+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+8(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+12(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+9(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+13(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+10(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+14(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+11(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,32+15(%%edi)
7:
	movl	48(%%esi),%%eax
	movl	48+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+4(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+1(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+2(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+6(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+3(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+7(%%edi)
7:
	movl	48+8(%%esi),%%eax
	movl	48+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+8(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+12(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+9(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+13(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+10(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+14(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+11(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,48+15(%%edi)
7:
	addl	$64,%%esi		
smc23:	subl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y+63]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx"
   );
#endif
}

void Draw64x64_Trans_Mapped_FlipXY(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc24+2
	xorl	%%edx,%%edx

	movl	$64,%%ebp	
9:
	movl	(%%esi),%%eax
	movl	4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-4(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-1(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-2(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-6(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-3(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-7(%%edi)
7:
	movl	8(%%esi),%%eax
	movl	12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-8(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-12(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-9(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-13(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-10(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-14(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-11(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-15(%%edi)
7:
	movl	16(%%esi),%%eax
	movl	16+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-16(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-20(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-17(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-21(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-18(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-22(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-19(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-23(%%edi)
7:
	movl	16+8(%%esi),%%eax
	movl	16+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-24(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-28(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-25(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-29(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-26(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-30(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-27(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-31(%%edi)
7:
	movl	32(%%esi),%%eax
	movl	32+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-32(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-36(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-33(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-37(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-34(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-38(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-35(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-39(%%edi)
7:
	movl	32+8(%%esi),%%eax
	movl	32+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-40(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-44(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-41(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-45(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-42(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-46(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-43(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-47(%%edi)
7:
	movl	48(%%esi),%%eax
	movl	48+4(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-48(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-52(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-49(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-53(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-50(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-54(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-51(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-55(%%edi)
7:
	movl	48+8(%%esi),%%eax
	movl	48+12(%%esi),%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-56(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-60(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-57(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-61(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	cmpb	%%dh,%%al
	je	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-58(%%edi)
7:	cmpb	%%dh,%%bl
	je	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-62(%%edi)
7:	cmpb	%%dh,%%ah
	je	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,63-59(%%edi)
7:	cmpb	%%dh,%%bh
	je	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:
	addl	$64,%%esi		
smc24:	subl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y+63]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx"
   );
#endif
}

/************************************************************/

void Draw8xH_Trans_Packed_Mapped_Column(UBYTE *SPR, int x, int y, int height, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc33+2
	movl	%%ebx,%%ebp

	xorl	%%edx,%%edx
9:
	movl	(%%esi),%%eax
	testl	%%eax,%%eax		
	jz	8f
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
7:	testb	%%bh,%%bh
	jz	8f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
8:
	addl	$4,%%esi		
smc33:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w) , "b" (height)
	: "%edx"
   );
#endif
}

void Draw8xH_Trans_Packed_Mapped_FlipY_Column(UBYTE *SPR, int x, int y, int height, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smc34+2
	movl	%%ebx,%%ebp

	xorl	%%edx,%%edx
9:
	movl	(%%esi),%%eax
	testl	%%eax,%%eax		
	jz	8f
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:	testb	%%bh,%%bh
	jz	8f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
8:
	addl	$4,%%esi		
smc34:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w) , "b" (height)
	: "%edx"
   );
#endif
}

/****************************************/

void Draw16x16_Trans_Mapped_4bpp(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smp25+2
	xorl	%%edx,%%edx

	movl	$16,%%ebp		
9:
	movl	(%%esi),%%eax
	testl	%%eax,%%eax		
	jz	8f
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
7:	testb	%%bh,%%bh
	jz	8f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
8:
	movl	4(%%esi),%%eax
	testl	%%eax,%%eax		
	jz	8f
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+1(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+3(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+2(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+5(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+4(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+7(%%edi)
7:	testb	%%bh,%%bh
	jz	8f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+6(%%edi)
8:
	addl	$8,%%esi		

smp25:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx", "%ebp", "cc"
   );
#endif
}

void Draw16x16_Trans_Mapped_FlipY_4bpp(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smp26+2
	xorl	%%edx,%%edx

	movl	$16,%%ebp			
9:
	movl	(%%esi),%%eax
	testl	%%eax,%%eax		
	jz	8f
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+6(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+7(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+4(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+2(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+3(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8(%%edi)
7:	testb	%%bh,%%bh
	jz	8f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+1(%%edi)
8:
	movl	4(%%esi),%%eax
	testl	%%eax,%%eax		
	jz	8f
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:	testb	%%bh,%%bh
	jz	8f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
8:
	addl	$8,%%esi		
smp26:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx", "%ebp", "cc"
   );
#endif
}

void Draw16x16_Trans_Mapped_FlipX_4bpp(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smp27+2
	xorl	%%edx,%%edx

	movl	$16,%%ebp			
9:
	movl	(%%esi),%%eax
	testl	%%eax,%%eax		
	jz	8f
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
7:	testb	%%bh,%%bh
	jz	8f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
8:
	movl	4(%%esi),%%eax
	testl	%%eax,%%eax		
	jz	8f
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+1(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+3(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+2(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+5(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+4(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+7(%%edi)
7:	testb	%%bh,%%bh
	jz	8f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+6(%%edi)
8:
	addl	$8,%%esi		
smp27:	subl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y+15]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx", "%ebp", "cc"
   );
#endif
}


void Draw16x16_Trans_Mapped_FlipXY_4bpp(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smp28+2
	xorl	%%edx,%%edx

	movl	$16,%%ebp			
9:
	movl	(%%esi),%%eax
	testl	%%eax,%%eax		
	jz	8f
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+6(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+7(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+4(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+2(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+3(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8(%%edi)
7:	testb	%%bh,%%bh
	jz	8f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+1(%%edi)
8:
	movl	4(%%esi),%%eax
	testl	%%eax,%%eax		
	jz	8f
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
7:	testb	%%bh,%%bh
	jz	7f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
7:
	shr	$16,%%eax
	shr	$16,%%ebx

	testb	%%al,%%al
	jz	7f
	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
7:	testb	%%bl,%%bl
	jz	7f
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
7:	testb	%%ah,%%ah
	jz	7f
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
7:	testb	%%bh,%%bh
	jz	8f
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
8:
	addl	$8,%%esi		
smp28:	subl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y+15]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx", "%ebp", "cc"
   );
#endif
}

void Draw16x16_Mapped_4bpp(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smo25+2
	xorl	%%edx,%%edx

	movl	$16,%%ebp	
9:
	movl	(%%esi),%%eax
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)

	shr	$16,%%eax
	shr	$16,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
8:
	movl	4(%%esi),%%eax
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+1(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+3(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+2(%%edi)

	shr	$16,%%eax
	shr	$16,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+5(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+4(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+7(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+6(%%edi)
8:
	addl	$8,%%esi		

smo25:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx", "%ebp", "cc"
   );
#endif
}


void Draw16x16_Mapped_FlipY_4bpp(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smo26+2
	xorl	%%edx,%%edx

	movl	$16,%%ebp			
9:
	movl	(%%esi),%%eax
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+6(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+7(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+4(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+5(%%edi)

	shr	$16,%%eax
	shr	$16,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+2(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+3(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+1(%%edi)
8:
	movl	4(%%esi),%%eax
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)

	shr	$16,%%eax
	shr	$16,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
8:
	addl	$8,%%esi		
smo26:	addl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx", "%ebp", "cc"
   );
#endif
}

void Draw16x16_Mapped_FlipX_4bpp(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smo27+2
	xorl	%%edx,%%edx

	movl	$16,%%ebp			
9:
	movl	(%%esi),%%eax
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)

	shr	$16,%%eax
	shr	$16,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
8:
	movl	4(%%esi),%%eax
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+1(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+3(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+2(%%edi)

	shr	$16,%%eax
	shr	$16,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+5(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+4(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+7(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+6(%%edi)
8:
	addl	$8,%%esi		
smo27:	subl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y+15]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx", "%ebp", "cc"
   );
#endif
}


void Draw16x16_Mapped_FlipXY_4bpp(UBYTE *SPR, int x, int y, UBYTE *cmap)
{
#ifdef RAINE_DOS
   asm("
	pushl	%%ebp
	movl	%%eax,smo28+2
	xorl	%%edx,%%edx

	movl	$16,%%ebp			
9:
	movl	(%%esi),%%eax
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+6(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+7(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+4(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+5(%%edi)

	shr	$16,%%eax
	shr	$16,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+2(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+3(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,8+1(%%edi)
8:
	movl	4(%%esi),%%eax
	movl	%%eax,%%ebx
	andl	$0x0F0F0F0F,%%eax

	shrl	$4,%%ebx
	andl	$0x0F0F0F0F,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,6(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,7(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,4(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,5(%%edi)

	shr	$16,%%eax
	shr	$16,%%ebx

	movb	%%al,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,2(%%edi)
	movb	%%bl,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,3(%%edi)
	movb	%%ah,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,(%%edi)
	movb	%%bh,%%dl
	movb	(%%ecx,%%edx,1),%%al
	movb	%%al,1(%%edi)
8:
	addl	$8,%%esi		
smo28:	subl	$0x2164334A,%%edi	

	decl	%%ebp
	jne	9b
	popl	%%ebp
	"
	: 
	: "S" (SPR) , "D" (GameBitmap->line[y+15]+x) , "c" (cmap) , "a" (GameBitmap->w)
	: "%ebx", "%edx", "%ebp", "cc"
   );
#endif
}

/******************************************************************************/

text_func *textout_fast;

void init_video_core(void)
{
   VIDEO_INFO *vid_info;

   /* convenient screen data */

   vid_info = current_game->video_info;

   disp_x = vid_info->border_size + vid_info->screen_x + vid_info->border_size;
   disp_y = vid_info->border_size + vid_info->screen_y + vid_info->border_size;

   disp_x_8 = disp_x - 8;
   disp_y_8 = disp_y - 8;

   disp_x_16 = disp_x - 16;
   disp_y_16 = disp_y - 16;

   /* rotation jumptables */

   Draw8x8_init_jump_table();
   Draw8x8_Trans_init_jump_table();

   Draw8x8_Mapped_init_jump_table();
   Draw8x8_Trans_Mapped_init_jump_table();

   Draw16x16_Mapped_init_jump_table();
   Draw16x16_Trans_Mapped_init_jump_table();

   Draw16x16_Mapped_ZoomXY_init_jump_table();
   Draw16x16_Trans_Mapped_ZoomXY_init_jump_table();

   /* 8bpp sprites */

   init_spr8x8asm();
   init_spr8x8asm_16();
   init_spr8x8asm_24();
   init_spr8x8asm_32();

   init_spr16x8asm();

   init_spr16x16asm();
   init_spr16x16asm_16();
   init_spr16x16asm_24();
   init_spr16x16asm_32();

   /* 4bpp sprites */

   init_spp8x8asm();

   /* 1bpp text */

   init_str6x8asm();
   init_str6x8asm_16();
   init_str6x8asm_24();
   init_str6x8asm_32();

   switch(display_cfg.bpp){
      case 8:
         textout_fast = &draw_string_6x8;
      break;
      case 15:
      case 16:
         textout_fast = &draw_string_6x8_16;
      break;
      case 24:
         textout_fast = &draw_string_6x8_24;
      break;
      case 32:
         textout_fast = &draw_string_6x8_32;
      break;
   }
}

/******************************************************************************/
